Package: tidyaudit
Title: Pipeline Audit Trails and Data Diagnostics for 'tidyverse'
        Workflows
Version: 0.2.1
Authors@R: 
    person("Fernando", "Cordeiro", , "fernandolpcordeiro@gmail.com", role = c("aut", "cre", "cph"))
Description: Provides pipeline audit trails and data diagnostics for
    'tidyverse' workflows. The audit trail system captures lightweight
    metadata snapshots at each step of a pipeline, building a structured
    record without storing the data itself. Operation-aware taps enrich
    snapshots with join match rates and filter drop statistics.
    Trails can be serialized to 'JSON' or 'RDS' and exported as
    self-contained 'HTML' visualizations. Also includes diagnostic
    functions for interactive data analysis including frequency tables,
    string quality auditing, and data comparison.
License: LGPL (>= 3)
URL: https://fpcordeiro.github.io/tidyaudit/,
        https://github.com/fpcordeiro/tidyaudit
BugReports: https://github.com/fpcordeiro/tidyaudit/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr (>= 1.2.0), glue, rlang (>= 1.0.0), stats, tools,
        utils
Suggests: covr, jsonlite, knitr, pkgdown, rmarkdown, spelling, stringi,
        testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 11:57:54 UTC; fernando
Author: Fernando Cordeiro [aut, cre, cph]
Maintainer: Fernando Cordeiro <fernandolpcordeiro@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-08 12:30:02 UTC
Built: R 4.4.3; ; 2026-05-08 15:12:18 UTC; unix

Package: tidyaudit
Title: Pipeline Audit Trails and Data Diagnostics for 'tidyverse'
        Workflows
Version: 0.2.1
Authors@R: 
    person("Fernando", "Cordeiro", , "fernandolpcordeiro@gmail.com", role = c("aut", "cre", "cph"))
Description: Provides pipeline audit trails and data diagnostics for
    'tidyverse' workflows. The audit trail system captures lightweight
    metadata snapshots at each step of a pipeline, building a structured
    record without storing the data itself. Operation-aware taps enrich
    snapshots with join match rates and filter drop statistics.
    Trails can be serialized to 'JSON' or 'RDS' and exported as
    self-contained 'HTML' visualizations. Also includes diagnostic
    functions for interactive data analysis including frequency tables,
    string quality auditing, and data comparison.
License: LGPL (>= 3)
URL: https://fpcordeiro.github.io/tidyaudit/,
        https://github.com/fpcordeiro/tidyaudit
BugReports: https://github.com/fpcordeiro/tidyaudit/issues
Depends: R (>= 4.1.0)
Imports: cli, dplyr (>= 1.2.0), glue, rlang (>= 1.0.0), stats, tools,
        utils
Suggests: covr, jsonlite, knitr, pkgdown, rmarkdown, spelling, stringi,
        testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-08 11:57:54 UTC; fernando
Author: Fernando Cordeiro [aut, cre, cph]
Maintainer: Fernando Cordeiro <fernandolpcordeiro@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-08 12:30:02 UTC
Built: R 4.5.2; ; 2026-05-08 12:44:56 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,compare_tbl)
S3method(as.data.frame,tidyaudit_tab)
S3method(filter_drop,data.frame)
S3method(filter_keep,data.frame)
S3method(print,audit_diff)
S3method(print,audit_snap)
S3method(print,audit_trail)
S3method(print,audit_transform)
S3method(print,compare_tbl)
S3method(print,diagnose_na)
S3method(print,diagnose_strings)
S3method(print,tidyaudit_tab)
S3method(print,validate_join)
S3method(print,validate_pk)
S3method(print,validate_var_rel)
S3method(summary,validate_join)
export(anti_join_tap)
export(audit_diff)
export(audit_export)
export(audit_report)
export(audit_tap)
export(audit_trail)
export(audit_transform)
export(compare_tables)
export(diagnose_nas)
export(diagnose_strings)
export(filter_drop)
export(filter_keep)
export(filter_out_tap)
export(filter_tap)
export(full_join_tap)
export(get_summary_table)
export(inner_join_tap)
export(left_join_tap)
export(read_trail)
export(right_join_tap)
export(semi_join_tap)
export(summarize_column)
export(tab)
export(tab_tap)
export(trail_to_df)
export(trail_to_list)
export(validate_join)
export(validate_primary_keys)
export(validate_var_relationship)
export(write_trail)
importFrom(rlang,"%||%")
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,setNames)
importFrom(utils,head)

# tidyaudit

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/tidyaudit)](https://CRAN.R-project.org/package=tidyaudit)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/tidyaudit)](https://cran.r-project.org/package=tidyaudit)
[![R-CMD-check](https://github.com/fpcordeiro/tidyaudit/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/fpcordeiro/tidyaudit/actions/workflows/R-CMD-check.yaml)
[![Codecov test coverage](https://codecov.io/gh/fpcordeiro/tidyaudit/graph/badge.svg)](https://app.codecov.io/gh/fpcordeiro/tidyaudit)
[![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
<!-- badges: end -->

## Pipeline audit trails and data diagnostics for tidyverse workflows

**Audit trails** track what happens at every step of a dplyr pipeline by
recording metadata-only snapshots: row counts, column changes, NA totals,
numeric shifts, and custom functions. Build a trail by dropping **taps**
into your pipe — transparent pass-throughs that record a snapshot and let the
data flow on unchanged. Operation-aware taps, such as `left_join_tap()` and `filter_tap()` go
further, capturing match rates and drop counts. The result is a
structured trail you can print, diff, export as HTML, or serialize to JSON. You
can learn more in `vignette("tidyaudit")`.

tidyaudit also includes a **diagnostic toolkit** for interactive data
exploration — join validation, key checks, table comparison, and more —
described in `vignette("diagnostics")`.

## Quick start

```r
library(tidyaudit)
library(dplyr)
set.seed(123)

orders  <- data.frame(id = 1:100, amount = runif(100, 10, 500), region_id = sample(1:5, 100, TRUE))
regions <- data.frame(region_id = 1:4, name = c("North", "South", "East", "West"))

trail <- audit_trail("order_pipeline")

result <- orders |>
  audit_tap(trail, "raw") |>
  left_join_tap(regions, by = "region_id", .trail = trail, .label = "with_region") |>
  filter_tap(amount > 100, .trail = trail, .label = "high_value", .stat = amount)
#> i filter_tap: amount > 100
#> Dropped 18 of 100 rows (18.0%)
#> Stat amount: dropped 1,062.191 of 25,429.39

print(trail)
#> -- Audit Trail: "order_pipeline" -----------------------------------------------
#> Created: 2026-02-21 14:36:35
#> Snapshots: 3
#>
#>   #  Label        Rows  Cols  NAs  Type
#>   -  -----------  ----  ----  ---  ------------------------------------
#>   1  raw           100     3    0  tap
#>   2  with_region   100     4   23  left_join (many-to-one, 77% matched)
#>   3  high_value     82     4   20  filter (dropped 18 rows, 18%)
#>
#> Changes:
#>   raw -> with_region: = rows, +1 cols, +23 NAs
#>   with_region -> high_value: -18 rows, = cols, -3 NAs

audit_diff(trail, "raw", "high_value")
#> -- Audit Diff: "raw" -> "high_value" --
#>
#>   Metric  Before  After  Delta
#>   ------  ------  -----  -----
#>   Rows       100     82    -18
#>   Cols         3      4     +1
#>   NAs          0     20    +20
#>
#> Columns added: name
#>
#> Numeric shifts (common columns):
#>     Column     Mean before  Mean after   Shift
#>     ---------  -----------  ----------  ------
#>     id               50.50       49.66   -0.84
#>     amount          254.29      297.16  +42.87
#>     region_id         3.08        3.05   -0.03
```

Three taps. Three snapshots. A complete record of what the pipeline did to
your data — and what it cost.

### Export as HTML

Share a trail as a self-contained HTML file — one file you can email, attach to
a report, or drop into a compliance folder:

```r
audit_export(trail, "order_pipeline.html")
```

The output is an interactive flow diagram with clickable nodes and edges,
light/dark theme toggle, and embedded JSON export. No server or internet
required.

![audit_export demo](man/figures/audit_export_readme.gif)

## Features

### Audit trail system

Build a structured timeline of your pipeline's behavior. Drop taps into
any dplyr pipe and get a traceable, diffable, exportable record of every step.

- `audit_trail()` / `audit_tap()` — create a trail and record snapshots
  inside pipes
- `left_join_tap()`, `filter_tap()`, and friends — **operation-aware taps**
  that capture match rates, drop counts, stat impact, and relationship types
- `tab_tap()` — track frequency distributions across pipeline steps
- `audit_diff()` — before/after comparison of any two snapshots
- `audit_report()` — full pipeline report in one call
- `audit_export()` — self-contained HTML visualization
- `write_trail()` / `read_trail()` — serialize to RDS or JSON for CI
  pipelines and dashboards
- Snapshot controls (`.numeric_summary`, `.cols_include`, `.cols_exclude`) —
  fine-tune what each tap captures

### Diagnostic toolkit

Standalone functions for interactive data exploration — the questions you ask
in the console before, during, and after building a pipeline.

- `validate_join()` — analyze a join before performing it (match rates,
  duplicates, unmatched keys)
- `validate_primary_keys()` / `validate_var_relationship()` — key and
  relationship validation
- `compare_tables()` — column, row, numeric, and categorical comparison
  between two data frames
- `filter_keep()` / `filter_drop()` — filter with diagnostic output and
  configurable warning thresholds
- `diagnose_nas()` / `summarize_column()` / `get_summary_table()` — data
  quality diagnostics
- `diagnose_strings()` / `audit_transform()` — string quality auditing and
  type-aware transformation diagnostics
- `tab()` — frequency tables and crosstabulations with sorting, cutoffs, and
  weighting

## Installation

```r
# Install from CRAN
install.packages("tidyaudit")

# Install development version
pak::pak("fpcordeiro/tidyaudit")
```

## Learn more

- [Audit trail walkthrough](https://fpcordeiro.github.io/tidyaudit/articles/tidyaudit.html)
  — build your first trail, use operation-aware taps, export and share results
- [Diagnostic functions guide](https://fpcordeiro.github.io/tidyaudit/articles/diagnostics.html)
  — validate joins, compare tables, diagnose data quality, and more

## Relationship to dtaudit

tidyaudit is the tidyverse-native counterpart to
[dtaudit](https://github.com/fpcordeiro/dtaudit), a data.table-based package
on CRAN. Same design vocabulary, independent implementations — choose the
one that matches your stack.

## License

LGPL (>= 3)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, message=FALSE-----------------------------------------------------
library(tidyaudit)
library(dplyr)

## ----basic-trail--------------------------------------------------------------
# Sample data
orders <- data.frame(
  id       = 1:20,
  customer = rep(c("Alice", "Bob", "Carol", "Dan", "Eve"), 4),
  amount   = c(150, 200, 50, 300, 75, 120, 400, 90, 250, 60,
               180, 210, 45, 320, 85, 130, 380, 95, 270, 55),
  status   = rep(c("complete", "pending", "complete", "cancelled", "complete"), 4)
)

trail <- audit_trail("order_pipeline")

result <- orders |>
  audit_tap(trail, "raw") |>
  filter(status == "complete") |>
  audit_tap(trail, "complete_only") |>
  mutate(tax = amount * 0.1) |>
  audit_tap(trail, "with_tax")

## ----print-trail--------------------------------------------------------------
print(trail)

## ----join-tap-----------------------------------------------------------------
customers <- data.frame(
  customer = c("Alice", "Bob", "Carol", "Dan"),
  region   = c("East", "West", "East", "North")
)

trail2 <- audit_trail("join_pipeline")

result2 <- orders |>
  audit_tap(trail2, "raw") |>
  left_join_tap(customers, by = "customer",
                .trail = trail2, .label = "with_region")

print(trail2)

## ----filter-tap---------------------------------------------------------------
trail3 <- audit_trail("filter_pipeline")

result3 <- orders |>
  audit_tap(trail3, "raw") |>
  filter_tap(status == "complete",
             .trail = trail3, .label = "complete_only") |>
  filter_tap(amount > 100,
             .trail = trail3, .label = "high_value",
             .stat = amount)

print(trail3)

## ----audit-diff---------------------------------------------------------------
audit_diff(trail3, "raw", "high_value")

## ----audit-report-------------------------------------------------------------
audit_report(trail3)

## ----custom-fns---------------------------------------------------------------
trail4 <- audit_trail("custom_example")

result4 <- orders |>
  audit_tap(trail4, "raw", .fns = list(
    n_complete   = ~sum(.$status == "complete"),          # scalar
    amount_stats = ~c(mean = mean(.$amount),              # named vector
                      max  = max(.$amount))
  )) |>
  filter(status == "complete") |>
  audit_tap(trail4, "complete_only", .fns = list(
    n_complete   = ~sum(.$status == "complete"),
    amount_stats = ~c(mean = mean(.$amount),
                      max  = max(.$amount))
  ))

## ----print-custom-------------------------------------------------------------
print(trail4)

## ----show-custom-false--------------------------------------------------------
print(trail4, show_custom = FALSE)

## ----snapshot-controls--------------------------------------------------------
wide_data <- cbind(orders, matrix(rnorm(20 * 50), nrow = 20))

trail_ctrl <- audit_trail("snapshot_controls")

wide_data |>
  audit_tap(trail_ctrl, "full_snapshot") |>
  audit_tap(trail_ctrl, "minimal",
            .numeric_summary = FALSE,
            .cols_include = c("id", "amount", "status"))

print(trail_ctrl)

## ----tab-standalone-----------------------------------------------------------
tab(orders, status)
tab(orders, status, customer)

## ----tab-tap------------------------------------------------------------------
trail_tab <- audit_trail("tab_pipeline")

result_tab <- orders |>
  tab_tap(status, .trail = trail_tab, .label = "status_dist") |>
  filter(status == "complete") |>
  tab_tap(customer, .trail = trail_tab, .label = "customer_dist",
          .sort = "freq_desc")

print(trail_tab)

## ----null-trail---------------------------------------------------------------
# Plain filter -- no diagnostics
orders |> filter_tap(amount > 100) |> nrow()

# Diagnostics without a trail
orders |> filter_tap(amount > 100, .stat = amount) |> invisible()

## ----trail-to-objects---------------------------------------------------------
# As a plain R list (suitable for jsonlite::toJSON())
trail_list <- trail_to_list(trail3)
str(trail_list, max.level = 2)

# As a data.frame (one row per snapshot)
trail_df <- trail_to_df(trail3)
print(trail_df)

## ----trail-rds----------------------------------------------------------------
tmp_rds <- tempfile(fileext = ".rds")
write_trail(trail3, tmp_rds)
restored <- read_trail(tmp_rds)
print(restored)

## ----trail-json---------------------------------------------------------------
tmp_json <- tempfile(fileext = ".json")
write_trail(trail3, tmp_json, format = "json")

## ----trail-html---------------------------------------------------------------
audit_export(trail3, tempfile(fileext = ".html"))

#' Diagnose String Column Quality
#'
#' Audits a character vector for common data quality issues including missing
#' values, empty strings, whitespace problems, non-ASCII characters, and case
#' inconsistencies. Requires the stringi package (in Suggests).
#'
#' @param x Character vector to diagnose.
#' @param name Optional name for the variable (used in output). If `NULL`,
#'   captures the variable name from the call.
#'
#' @returns An S3 object of class `diagnose_strings` containing:
#' \describe{
#'   \item{name}{Name of the variable}
#'   \item{n_total}{Total number of elements}
#'   \item{n_na}{Count of NA values}
#'   \item{n_empty}{Count of empty strings}
#'   \item{n_whitespace_only}{Count of whitespace-only strings}
#'   \item{n_leading_ws}{Count of strings with leading whitespace}
#'   \item{n_trailing_ws}{Count of strings with trailing whitespace}
#'   \item{n_non_ascii}{Count of strings with non-ASCII characters}
#'   \item{n_case_variants}{Number of unique values with case variants}
#'   \item{n_case_variant_groups}{Number of groups of case-insensitive duplicates}
#'   \item{case_variant_examples}{Data.frame with examples of case variants}
#' }
#'
#' @examples
#' firms <- c("Apple", "APPLE", "apple", "  Microsoft ", "Google", NA, "")
#' diagnose_strings(firms)
#'
#' @family data quality
#' @export
diagnose_strings <- function(x, name = NULL) {
  rlang::check_installed("stringi", reason = "for string diagnostics")

  if (is.null(name)) {
    name <- deparse(substitute(x))
  }

  x <- as.character(x)
  n_total <- length(x)
  n_na <- sum(is.na(x))

  non_na <- x[!is.na(x)]

  n_empty <- sum(non_na == "")
  n_whitespace_only <- sum(stringi::stri_detect_regex(non_na, "^\\s+$"))

  # Leading/trailing whitespace (excluding empty and whitespace-only)
  content_strings <- non_na[non_na != "" & !stringi::stri_detect_regex(non_na, "^\\s*$")]
  n_leading_ws <- sum(stringi::stri_detect_regex(content_strings, "^\\s"))
  n_trailing_ws <- sum(stringi::stri_detect_regex(content_strings, "\\s$"))

  n_non_ascii <- sum(stringi::stri_detect_regex(non_na, "[^\\x00-\\x7F]"))

  # Case variants
  if (length(content_strings) > 0L) {
    lower_versions <- stringi::stri_trans_tolower(content_strings)
    case_df <- data.frame(original = content_strings, lower = lower_versions,
                          stringsAsFactors = FALSE)

    # Group by lowercase and find groups with multiple distinct originals
    case_groups <- lapply(split(case_df$original, case_df$lower), function(vals) {
      uvals <- unique(vals)
      if (length(uvals) > 1L) {
        data.frame(
          lower = vals[1], # lowercase version via the split key
          n_variants = length(uvals),
          examples = paste(uvals[seq_len(min(3L, length(uvals)))], collapse = ", "),
          stringsAsFactors = FALSE
        )
      } else {
        NULL
      }
    })
    case_groups <- Filter(Negate(is.null), case_groups)

    if (length(case_groups) > 0L) {
      case_variant_examples <- do.call(rbind, case_groups)
      # Fix: the "lower" column should be the split key, not the original value
      case_variant_examples$lower <- names(case_groups)
      row.names(case_variant_examples) <- NULL
      n_case_variant_groups <- nrow(case_variant_examples)
      n_case_variants <- sum(case_variant_examples$n_variants)
    } else {
      n_case_variant_groups <- 0L
      n_case_variants <- 0L
      case_variant_examples <- data.frame(
        lower = character(0), n_variants = integer(0), examples = character(0),
        stringsAsFactors = FALSE
      )
    }
  } else {
    n_case_variant_groups <- 0L
    n_case_variants <- 0L
    case_variant_examples <- data.frame(
      lower = character(0), n_variants = integer(0), examples = character(0),
      stringsAsFactors = FALSE
    )
  }

  out <- list(
    name = name,
    n_total = n_total,
    n_na = n_na,
    n_empty = n_empty,
    n_whitespace_only = n_whitespace_only,
    n_leading_ws = n_leading_ws,
    n_trailing_ws = n_trailing_ws,
    n_non_ascii = n_non_ascii,
    n_case_variants = n_case_variants,
    n_case_variant_groups = n_case_variant_groups,
    case_variant_examples = case_variant_examples
  )
  structure(out, class = c("diagnose_strings", "list"))
}

#' @rdname diagnose_strings
#' @param ... Additional arguments (currently unused).
#' @export
print.diagnose_strings <- function(x, ...) {
  fmt_int <- function(z) format(z, big.mark = ",", scientific = FALSE, trim = TRUE)
  fmt_pct <- function(n, total) {
    if (total == 0L) return("0.0%")
    sprintf("%.1f%%", 100 * n / total)
  }

  cli::cli_h1("String Column Diagnosis: {x$name}")
  cli::cli_text("Total elements: {fmt_int(x$n_total)}")
  cli::cli_text("")

  cli::cli_text("{.strong Missing & Empty:}")
  cli::cli_bullets(c(
    "*" = "NA values: {fmt_int(x$n_na)} ({fmt_pct(x$n_na, x$n_total)})",
    "*" = "Empty strings: {fmt_int(x$n_empty)} ({fmt_pct(x$n_empty, x$n_total)})",
    "*" = "Whitespace-only: {fmt_int(x$n_whitespace_only)} ({fmt_pct(x$n_whitespace_only, x$n_total)})"
  ))

  cli::cli_text("")
  cli::cli_text("{.strong Whitespace Issues:}")
  cli::cli_bullets(c(
    "*" = "Leading whitespace: {fmt_int(x$n_leading_ws)}",
    "*" = "Trailing whitespace: {fmt_int(x$n_trailing_ws)}"
  ))

  cli::cli_text("")
  cli::cli_text("{.strong Encoding:}")
  cli::cli_bullets(c("*" = "Non-ASCII chars: {fmt_int(x$n_non_ascii)}"))

  cli::cli_text("")
  cli::cli_text("{.strong Case Inconsistencies:}")
  cli::cli_bullets(c(
    "*" = "Variant groups: {fmt_int(x$n_case_variant_groups)}",
    "*" = "Total variants: {fmt_int(x$n_case_variants)}"
  ))

  if (x$n_case_variant_groups > 0L) {
    cli::cli_text("")
    cli::cli_text("Case variant examples (up to 5 groups):")
    print(utils::head(x$case_variant_examples, 5L), row.names = FALSE)
  }

  invisible(x)
}

test_that("diagnose_strings detects basic issues", {
  skip_if_not_installed("stringi")
  x <- c("Apple", "APPLE", "apple", "  Microsoft ", "Google", NA, "")
  result <- diagnose_strings(x)

  expect_s3_class(result, "diagnose_strings")
  expect_equal(result$n_total, 7L)
  expect_equal(result$n_na, 1L)
  expect_equal(result$n_empty, 1L)
})

test_that("diagnose_strings detects whitespace issues", {
  skip_if_not_installed("stringi")
  x <- c("  leading", "trailing  ", "  both  ", "normal", "   ")
  result <- diagnose_strings(x)

  expect_equal(result$n_leading_ws, 2L)   # "  leading", "  both  "
  expect_equal(result$n_trailing_ws, 2L)  # "trailing  ", "  both  "
  expect_equal(result$n_whitespace_only, 1L)  # "   "
})

test_that("diagnose_strings detects case variants", {
  skip_if_not_installed("stringi")
  x <- c("Apple", "APPLE", "apple", "Google", "google")
  result <- diagnose_strings(x)

  expect_equal(result$n_case_variant_groups, 2L)
  expect_true(nrow(result$case_variant_examples) > 0L)
})

test_that("diagnose_strings handles all-NA input", {
  skip_if_not_installed("stringi")
  x <- c(NA_character_, NA_character_)
  result <- diagnose_strings(x)

  expect_equal(result$n_total, 2L)
  expect_equal(result$n_na, 2L)
  expect_equal(result$n_empty, 0L)
  expect_equal(result$n_case_variant_groups, 0L)
})

test_that("diagnose_strings handles clean input", {
  skip_if_not_installed("stringi")
  x <- c("alpha", "beta", "gamma")
  result <- diagnose_strings(x)

  expect_equal(result$n_na, 0L)
  expect_equal(result$n_empty, 0L)
  expect_equal(result$n_whitespace_only, 0L)
  expect_equal(result$n_leading_ws, 0L)
  expect_equal(result$n_trailing_ws, 0L)
  expect_equal(result$n_case_variant_groups, 0L)
})

test_that("diagnose_strings captures variable name", {
  skip_if_not_installed("stringi")
  my_var <- c("a", "b")
  result <- diagnose_strings(my_var)

  expect_equal(result$name, "my_var")
})

test_that("diagnose_strings uses explicit name", {
  skip_if_not_installed("stringi")
  result <- diagnose_strings(c("a", "b"), name = "test_col")

  expect_equal(result$name, "test_col")
})

test_that("print.diagnose_strings produces output", {
  skip_if_not_installed("stringi")
  x <- c("Apple", "APPLE", NA, "")
  result <- diagnose_strings(x)

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("String Column Diagnosis", combined))
})

test_that("diagnose_strings detects non-ASCII", {
  skip_if_not_installed("stringi")
  x <- c("hello", "caf\u00e9", "na\u00efve")
  result <- diagnose_strings(x)

  expect_equal(result$n_non_ascii, 2L)
})

#' Validate Join Operations Between Two Tables
#'
#' Analyzes a potential join between two data.frames or tibbles without
#' performing the full join. Reports relationship type (one-to-one,
#' one-to-many, etc.), match rates, duplicate keys, and unmatched rows.
#' Optionally tracks a numeric statistic column through the join to quantify
#' impact.
#'
#' @param x A data.frame or tibble (left table).
#' @param y A data.frame or tibble (right table).
#' @param by A character vector of column names to join on. Use a named vector
#'   `c("key_x" = "key_y")` when column names differ between tables. Unnamed
#'   elements are used for both tables.
#' @param stat Optional single column name (string) to track in both tables when
#'   the column name is the same. Ignored if `stat_x` or `stat_y` is provided.
#' @param stat_x Optional column name (string) for a numeric statistic in `x`.
#' @param stat_y Optional column name (string) for a numeric statistic in `y`.
#'
#' @returns An S3 object of class `validate_join` containing:
#' \describe{
#'   \item{x_name, y_name}{Names of the input tables from the original call}
#'   \item{by_x, by_y}{Key columns used for the join}
#'   \item{counts}{List with row counts, match rates, and overlap statistics}
#'   \item{stat}{When `stat`, `stat_x`, or `stat_y` is provided, a list with
#'     stat diagnostics per table. `NULL` when no stat is provided.}
#'   \item{duplicates}{List with duplicate key information for each table}
#'   \item{summary_table}{A data.frame summarizing the join diagnostics}
#'   \item{relation}{Character string describing the relationship}
#'   \item{keys_only_in_x}{Unmatched keys from x}
#'   \item{keys_only_in_y}{Unmatched keys from y}
#' }
#'
#' @examples
#' x <- data.frame(id = c(1L, 2L, 3L, 3L), value = c("a", "b", "c", "d"))
#' y <- data.frame(id = c(2L, 3L, 4L), score = c(10, 20, 30))
#' result <- validate_join(x, y, by = "id")
#' print(result)
#'
#' # Track a stat column with different names in each table
#' x2 <- data.frame(id = 1:3, sales = c(100, 200, 300))
#' y2 <- data.frame(id = 2:4, cost = c(10, 20, 30))
#' validate_join(x2, y2, by = "id", stat_x = "sales", stat_y = "cost")
#'
#' @family join validation
#' @export
validate_join <- function(x, y, by = NULL, stat = NULL, stat_x = NULL,
                          stat_y = NULL) {
  if (!is.data.frame(x)) {
    cli::cli_abort("{.arg x} must be a data.frame or tibble.")
  }
  if (!is.data.frame(y)) {
    cli::cli_abort("{.arg y} must be a data.frame or tibble.")
  }

  x_name <- deparse(substitute(x))
  y_name <- deparse(substitute(y))

  # Resolve by into by_x / by_y
  if (is.null(by)) {
    cli::cli_abort("{.arg by} must be provided.")
  }
  if (!is.character(by) || length(by) == 0L) {
    cli::cli_abort("{.arg by} must be a non-empty character vector.")
  }

  by_names <- names(by)
  if (is.null(by_names)) {
    by_x <- by
    by_y <- by
  } else {
    by_x <- ifelse(by_names == "", by, by_names)
    by_y <- unname(by)
  }

  # Column existence checks
  miss_x <- setdiff(by_x, names(x))
  if (length(miss_x) > 0L) {
    cli::cli_abort("{.arg x} is missing key column{?s}: {.field {miss_x}}")
  }
  miss_y <- setdiff(by_y, names(y))
  if (length(miss_y) > 0L) {
    cli::cli_abort("{.arg y} is missing key column{?s}: {.field {miss_y}}")
  }

  # Resolve stat columns: stat_x/stat_y take precedence over stat
  have_stat_x <- FALSE
  have_stat_y <- FALSE
  stat_col_x <- NULL
  stat_col_y <- NULL

  if (!is.null(stat_x) || !is.null(stat_y)) {
    if (!is.null(stat_x)) {
      if (!is.character(stat_x) || length(stat_x) != 1L) {
        cli::cli_abort("{.arg stat_x} must be a single column name string.")
      }
      stat_col_x <- stat_x
      have_stat_x <- TRUE
    }
    if (!is.null(stat_y)) {
      if (!is.character(stat_y) || length(stat_y) != 1L) {
        cli::cli_abort("{.arg stat_y} must be a single column name string.")
      }
      stat_col_y <- stat_y
      have_stat_y <- TRUE
    }
  } else if (!is.null(stat)) {
    if (!is.character(stat) || length(stat) != 1L) {
      cli::cli_abort("{.arg stat} must be a single column name string.")
    }
    stat_col_x <- stat_col_y <- stat
    have_stat_x <- have_stat_y <- TRUE
  }

  if (have_stat_x) {
    if (!stat_col_x %in% names(x)) {
      cli::cli_abort("Column {.field {stat_col_x}} not found in {.arg x}.")
    }
    if (!is.numeric(x[[stat_col_x]])) {
      cli::cli_abort("Column {.field {stat_col_x}} in {.arg x} is not numeric.")
    }
  }
  if (have_stat_y) {
    if (!stat_col_y %in% names(y)) {
      cli::cli_abort("Column {.field {stat_col_y}} not found in {.arg y}.")
    }
    if (!is.numeric(y[[stat_col_y]])) {
      cli::cli_abort("Column {.field {stat_col_y}} in {.arg y} is not numeric.")
    }
  }

  # Row counts
  x_rows <- nrow(x)
  y_rows <- nrow(y)

  # Aggregate to counts per key (no cartesian explosion)
  xc <- dplyr::count(x, dplyr::across(dplyr::all_of(by_x)), name = "N")
  yc <- dplyr::count(y, dplyr::across(dplyr::all_of(by_y)), name = "N")

  x_unique <- nrow(xc)
  y_unique <- nrow(yc)
  x_has_dups <- any(xc[["N"]] > 1L)
  y_has_dups <- any(yc[["N"]] > 1L)

  # Build join_by for merging aggregate counts
  merge_by <- setNames(by_y, by_x)
  comb <- dplyr::full_join(xc, yc, by = merge_by, suffix = c(".x", ".y"))

  Nx <- comb[["N.x"]]
  Ny <- comb[["N.y"]]
  if (is.null(Nx)) Nx <- integer(0)
  if (is.null(Ny)) Ny <- integer(0)
  Nx0 <- ifelse(is.na(Nx), 0L, Nx)
  Ny0 <- ifelse(is.na(Ny), 0L, Ny)

  # Merge metrics
  n_matched <- sum(as.double(Nx0) * as.double(Ny0))
  n_only_x <- sum(Nx0[Ny0 == 0L])
  n_only_y <- sum(Ny0[Nx0 == 0L])
  n_key_overlap <- sum(Nx0 > 0L & Ny0 > 0L)
  match_rate_x <- if (x_rows > 0L) 100 * ((x_rows - n_only_x) / x_rows) else NA_real_
  match_rate_y <- if (y_rows > 0L) 100 * ((y_rows - n_only_y) / y_rows) else NA_real_

  # Stat diagnostics
  stat_info <- NULL
  if (have_stat_x || have_stat_y) {
    stat_info <- list()

    # Identify matched keys in x
    matched_mask <- Nx0 > 0L & Ny0 > 0L
    matched_keys_df <- comb[matched_mask, by_x, drop = FALSE]

    if (have_stat_x) {
      stat_total_x <- sum(x[[stat_col_x]], na.rm = TRUE)
      stat_na_x <- sum(is.na(x[[stat_col_x]]))
      stat_matched_x <- if (nrow(matched_keys_df) > 0L) {
        matched_x <- dplyr::semi_join(x, matched_keys_df, by = setNames(by_x, by_x))
        sum(matched_x[[stat_col_x]], na.rm = TRUE)
      } else {
        0
      }
      stat_only_x <- stat_total_x - stat_matched_x
      stat_rate_x <- if (stat_total_x != 0) 100 * stat_matched_x / stat_total_x else NA_real_
      stat_info$stat_col_x <- stat_col_x
      stat_info$x <- list(
        total = stat_total_x, matched = stat_matched_x,
        only = stat_only_x, rate = stat_rate_x, n_na = stat_na_x
      )
    }

    if (have_stat_y) {
      stat_total_y <- sum(y[[stat_col_y]], na.rm = TRUE)
      stat_na_y <- sum(is.na(y[[stat_col_y]]))
      # Build matched keys for y (may need column renaming)
      if (!identical(by_x, by_y)) {
        matched_keys_y <- matched_keys_df
        names(matched_keys_y) <- by_y
      } else {
        matched_keys_y <- matched_keys_df
      }
      stat_matched_y <- if (nrow(matched_keys_y) > 0L) {
        matched_y <- dplyr::semi_join(y, matched_keys_y, by = setNames(by_y, by_y))
        sum(matched_y[[stat_col_y]], na.rm = TRUE)
      } else {
        0
      }
      stat_only_y <- stat_total_y - stat_matched_y
      stat_rate_y <- if (stat_total_y != 0) 100 * stat_matched_y / stat_total_y else NA_real_
      stat_info$stat_col_y <- stat_col_y
      stat_info$y <- list(
        total = stat_total_y, matched = stat_matched_y,
        only = stat_only_y, rate = stat_rate_y, n_na = stat_na_y
      )
    }
  }

  # Relationship classification
  relation <- if (n_key_overlap == 0L) {
    "no matches"
  } else if (!x_has_dups && !y_has_dups) {
    "one-to-one"
  } else if (!x_has_dups && y_has_dups) {
    "one-to-many"
  } else if (x_has_dups && !y_has_dups) {
    "many-to-one"
  } else {
    "many-to-many"
  }

  # Build summary table
  fmt_int <- function(z) format(z, big.mark = ",", scientific = FALSE, trim = TRUE)
  key_lbl_x <- paste(by_x, collapse = ", ")
  key_lbl_y <- paste(by_y, collapse = ", ")

  summary_table <- data.frame(
    Item = c(
      "Relationship",
      sprintf("Key(s) in %s   [%s]", x_name, key_lbl_x),
      sprintf("Key(s) in %s   [%s]", y_name, key_lbl_y),
      sprintf("Rows in %s", x_name),
      sprintf("Distinct key combos in %s", x_name),
      sprintf("Rows in %s", y_name),
      sprintf("Distinct key combos in %s", y_name),
      "Overlapping distinct key combos",
      "Matched row pairs (cartesian)",
      sprintf("Match rate from %s", x_name),
      sprintf("Match rate from %s", y_name),
      sprintf("Rows only in %s (no match in %s)", x_name, y_name),
      sprintf("Rows only in %s (no match in %s)", y_name, x_name)
    ),
    Value = c(
      relation,
      sprintf("(%d col%s)", length(by_x), ifelse(length(by_x) == 1L, "", "s")),
      sprintf("(%d col%s)", length(by_y), ifelse(length(by_y) == 1L, "", "s")),
      fmt_int(x_rows),
      fmt_int(x_unique),
      fmt_int(y_rows),
      fmt_int(y_unique),
      fmt_int(n_key_overlap),
      fmt_int(n_matched),
      if (is.na(match_rate_x)) "N/A" else sprintf("%.2f%%", match_rate_x),
      if (is.na(match_rate_y)) "N/A" else sprintf("%.2f%%", match_rate_y),
      fmt_int(n_only_x),
      fmt_int(n_only_y)
    ),
    stringsAsFactors = FALSE
  )

  # Duplicate key tables
  x_dupe_keys <- xc[xc$N > 1L, , drop = FALSE]
  y_dupe_keys <- yc[yc$N > 1L, , drop = FALSE]

  # Unmatched keys
  keys_only_in_x <- comb[is.na(Ny), , drop = FALSE]
  keys_only_in_y <- comb[is.na(Nx), , drop = FALSE]

  out <- list(
    x_name = x_name,
    y_name = y_name,
    by_x = by_x,
    by_y = by_y,
    counts = list(
      x_rows = x_rows, y_rows = y_rows,
      x_unique = x_unique, y_unique = y_unique,
      n_key_overlap = n_key_overlap,
      n_matched_pairs = n_matched,
      match_rate_x = match_rate_x,
      match_rate_y = match_rate_y,
      n_only_x = n_only_x,
      n_only_y = n_only_y
    ),
    stat = stat_info,
    duplicates = list(
      x_has_dups = x_has_dups, y_has_dups = y_has_dups,
      x_dupe_keys = x_dupe_keys,
      y_dupe_keys = y_dupe_keys
    ),
    summary_table = summary_table,
    relation = relation,
    keys_only_in_x = keys_only_in_x,
    keys_only_in_y = keys_only_in_y
  )
  structure(out, class = c("validate_join", "list"))
}

#' @rdname validate_join
#' @param ... Additional arguments (currently unused).
#' @export
print.validate_join <- function(x, ...) {
  cli::cli_h1("Join Validation: {x$x_name} \u2194 {x$y_name}")
  cli::cli_text("Keys in {x$x_name}: {.field {x$by_x}}")
  cli::cli_text("Keys in {x$y_name}: {.field {x$by_y}}")
  cli::cli_text("")

  # Summary table via .cli_table
  .cli_table(x$summary_table, right_align = "Value")

  # Stat diagnostics section
  if (!is.null(x$stat)) {
    cli::cli_text("")
    cli::cli_rule("Stat diagnostics")

    if (!is.null(x$stat$x)) {
      sx <- x$stat$x
      scx <- x$stat$stat_col_x
      cli::cli_text("")
      cli::cli_text("{.strong {scx}} in {x$x_name}:")
      cli::cli_bullets(c(
        "*" = "Total: {format(sx$total, big.mark = ',', trim = TRUE)}",
        "*" = "Matched: {format(sx$matched, big.mark = ',', trim = TRUE)}  ({if (is.na(sx$rate)) 'N/A' else sprintf('%.2f%%', sx$rate)})",
        "*" = "Unmatched: {format(sx$only, big.mark = ',', trim = TRUE)}  ({if (is.na(sx$rate)) 'N/A' else sprintf('%.2f%%', 100 - sx$rate)})"
      ))
    }

    if (!is.null(x$stat$y)) {
      sy <- x$stat$y
      scy <- x$stat$stat_col_y
      cli::cli_text("")
      cli::cli_text("{.strong {scy}} in {x$y_name}:")
      cli::cli_bullets(c(
        "*" = "Total: {format(sy$total, big.mark = ',', trim = TRUE)}",
        "*" = "Matched: {format(sy$matched, big.mark = ',', trim = TRUE)}  ({if (is.na(sy$rate)) 'N/A' else sprintf('%.2f%%', sy$rate)})",
        "*" = "Unmatched: {format(sy$only, big.mark = ',', trim = TRUE)}  ({if (is.na(sy$rate)) 'N/A' else sprintf('%.2f%%', 100 - sy$rate)})"
      ))
    }
  }

  # Duplicates footer
  cli::cli_text("")
  du <- x$duplicates
  x_dup <- if (du$x_has_dups) "yes" else "no"
  y_dup <- if (du$y_has_dups) "yes" else "no"
  cli::cli_text("Duplicates: {x$x_name}={x_dup}  {x$y_name}={y_dup}")

  invisible(x)
}

#' @rdname validate_join
#' @param object A `validate_join` object to summarize.
#' @export
summary.validate_join <- function(object, ...) {
  print(object)
  invisible(object$summary_table)
}

test_that("validate_join detects one-to-one relationship", {
  x <- data.frame(id = 1:3, val = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 2:4, score = c(10, 20, 30))
  result <- validate_join(x, y, by = "id")

  expect_s3_class(result, "validate_join")
  expect_equal(result$relation, "one-to-one")
  expect_equal(result$counts$x_rows, 3L)
  expect_equal(result$counts$y_rows, 3L)
  expect_equal(result$counts$n_key_overlap, 2L)
  expect_equal(result$counts$n_only_x, 1L)
  expect_equal(result$counts$n_only_y, 1L)
})

test_that("validate_join detects many-to-one relationship", {
  x <- data.frame(id = c(1L, 2L, 3L, 3L), val = c("a", "b", "c", "d"),
                   stringsAsFactors = FALSE)
  y <- data.frame(id = c(2L, 3L, 4L), score = c(10, 20, 30))
  result <- validate_join(x, y, by = "id")

  expect_equal(result$relation, "many-to-one")
  expect_true(result$duplicates$x_has_dups)
  expect_false(result$duplicates$y_has_dups)
})

test_that("validate_join detects no matches", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 4:6)
  result <- validate_join(x, y, by = "id")

  expect_equal(result$relation, "no matches")
  expect_equal(result$counts$n_key_overlap, 0L)
})

test_that("validate_join works with named by vector", {
  x <- data.frame(key_x = c(1L, 2L, 3L), val = c("a", "b", "c"),
                   stringsAsFactors = FALSE)
  y <- data.frame(key_y = c(2L, 3L, 4L), score = c(10, 20, 30))
  result <- validate_join(x, y, by = c("key_x" = "key_y"))

  expect_equal(result$by_x, "key_x")
  expect_equal(result$by_y, "key_y")
  expect_equal(result$counts$n_key_overlap, 2L)
})

test_that("validate_join tracks stat column (same name)", {
  x <- data.frame(id = 1:4, revenue = c(100, 200, 300, 400))
  y <- data.frame(id = 2:5, revenue = c(10, 20, 30, 40))
  result <- validate_join(x, y, by = "id", stat = "revenue")

  expect_false(is.null(result$stat))
  expect_equal(result$stat$stat_col_x, "revenue")
  expect_equal(result$stat$stat_col_y, "revenue")
  expect_equal(result$stat$x$total, 1000)
  expect_equal(result$stat$y$total, 100)
})

test_that("validate_join tracks stat with stat_x and stat_y", {
  x <- data.frame(id = 1:3, sales = c(100, 200, 300))
  y <- data.frame(id = 2:4, cost = c(10, 20, 30))
  result <- validate_join(x, y, by = "id", stat_x = "sales", stat_y = "cost")

  expect_equal(result$stat$stat_col_x, "sales")
  expect_equal(result$stat$stat_col_y, "cost")
  expect_false(is.null(result$stat$x))
  expect_false(is.null(result$stat$y))
})

test_that("validate_join errors on missing key columns", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 1:3)

  expect_error(validate_join(x, y, by = "nonexistent"), "missing key")
})

test_that("validate_join errors without by argument", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 1:3)

  expect_error(validate_join(x, y), "must be provided")
})

test_that("validate_join errors on non-numeric stat", {
  x <- data.frame(id = 1:3, name = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3)

  expect_error(validate_join(x, y, by = "id", stat = "name"), "not numeric")
})

test_that("validate_join has correct match rates", {
  x <- data.frame(id = 1:10, val = 1:10)
  y <- data.frame(id = 6:15, val = 6:15)
  result <- validate_join(x, y, by = "id")

  expect_equal(result$counts$match_rate_x, 50)
  expect_equal(result$counts$match_rate_y, 50)
})

test_that("validate_join summary_table is a data.frame", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 2:4)
  result <- validate_join(x, y, by = "id")

  expect_true(is.data.frame(result$summary_table))
  expect_true("Item" %in% names(result$summary_table))
  expect_true("Value" %in% names(result$summary_table))
})

test_that("print.validate_join produces output", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 2:4)
  result <- validate_join(x, y, by = "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Join Validation", combined))
})

test_that("summary.validate_join returns summary_table", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 2:4)
  result <- validate_join(x, y, by = "id")

  st <- capture.output(s <- summary(result), type = "message")
  expect_true(is.data.frame(s))
})

test_that("validate_join works with composite keys", {
  x <- data.frame(a = c(1L, 1L, 2L), b = c("x", "y", "x"),
                   stringsAsFactors = FALSE)
  y <- data.frame(a = c(1L, 2L), b = c("x", "x"),
                   stringsAsFactors = FALSE)
  result <- validate_join(x, y, by = c("a", "b"))

  expect_equal(result$counts$n_key_overlap, 2L)
  expect_equal(result$counts$n_only_x, 1L)
})

test_that("validate_join works with tibbles", {
  skip_if_not_installed("dplyr")
  x <- dplyr::tibble(id = 1:3, val = c(10, 20, 30))
  y <- dplyr::tibble(id = 2:4, val = c(40, 50, 60))
  result <- validate_join(x, y, by = "id")

  expect_s3_class(result, "validate_join")
  expect_equal(result$relation, "one-to-one")
})

test_that("validate_join stat_x only tracks x side", {
  x <- data.frame(id = 1:3, sales = c(100, 200, 300))
  y <- data.frame(id = 2:4, cost = c(10, 20, 30))
  result <- validate_join(x, y, by = "id", stat_x = "sales")

  expect_false(is.null(result$stat))
  expect_false(is.null(result$stat$x))
  expect_null(result$stat$y)
  expect_equal(result$stat$stat_col_x, "sales")
})

test_that("validate_join stat_y only tracks y side", {
  x <- data.frame(id = 1:3, sales = c(100, 200, 300))
  y <- data.frame(id = 2:4, cost = c(10, 20, 30))
  result <- validate_join(x, y, by = "id", stat_y = "cost")

  expect_false(is.null(result$stat))
  expect_null(result$stat$x)
  expect_false(is.null(result$stat$y))
  expect_equal(result$stat$stat_col_y, "cost")
})

test_that("validate_join stat_x/stat_y override stat", {
  x <- data.frame(id = 1:3, sales = c(100, 200, 300), revenue = c(1, 2, 3))
  y <- data.frame(id = 2:4, cost = c(10, 20, 30), revenue = c(4, 5, 6))
  # stat_x/stat_y should take precedence over stat
  result <- validate_join(x, y, by = "id", stat = "revenue",
                          stat_x = "sales", stat_y = "cost")

  expect_equal(result$stat$stat_col_x, "sales")
  expect_equal(result$stat$stat_col_y, "cost")
})

#' Create an Audit Trail
#'
#' Creates an audit trail object that captures metadata snapshots at each step
#' of a data pipeline. The trail uses environment-based reference semantics so
#' it can be modified in place inside pipes via [audit_tap()].
#'
#' @param name Optional name for the trail. If `NULL`, a timestamped name is
#'   generated automatically.
#'
#' @returns An `audit_trail` object (S3 class wrapping an environment).
#'
#' @examples
#' trail <- audit_trail("my_analysis")
#' print(trail)
#'
#' @family audit trail
#' @export
audit_trail <- function(name = NULL) {
  trail_env <- new.env(parent = emptyenv())
  trail_env$name <- name %||% format(Sys.time(), "trail_%Y%m%d_%H%M%S")
  trail_env$created_at <- Sys.time()
  trail_env$snapshots <- list()
  trail_env$labels <- character()
  structure(trail_env, class = c("audit_trail", "environment"))
}

#' @rdname audit_trail
#' @param x An object to print.
#' @param show_custom Logical. If `TRUE` (default), inline annotations (one
#'   indented line per custom function) are printed below each snapshot that has
#'   custom diagnostics. Set to `FALSE` to suppress them and display only the
#'   main timeline table.
#' @param ... Additional arguments (currently unused).
#' @export
print.audit_trail <- function(x, show_custom = TRUE, ...) {
  n_snaps <- length(x$snapshots)

  cli::cli_h1("Audit Trail: {.val {x$name}}")
  cli::cli_text("Created: {format(x$created_at, '%Y-%m-%d %H:%M:%S')}")
  cli::cli_text("Snapshots: {n_snaps}")

  if (n_snaps == 0L) {
    cli::cli_text("")
    cli::cli_alert_info("No snapshots recorded yet. Use {.fn audit_tap} in a pipe to add snapshots.")
    return(invisible(x))
  }

  cli::cli_text("")

  # Build timeline table
  tbl <- data.frame(
    `#`    = as.character(seq_len(n_snaps)),
    Label  = vapply(x$snapshots, `[[`, character(1), "label"),
    Rows   = format(vapply(x$snapshots, `[[`, integer(1), "nrow"),
                    big.mark = ",", trim = TRUE),
    Cols   = as.character(vapply(x$snapshots, `[[`, integer(1), "ncol")),
    NAs    = format(vapply(x$snapshots, `[[`, integer(1), "total_nas"),
                    big.mark = ",", trim = TRUE),
    Type   = vapply(x$snapshots, .format_snap_type, character(1)),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  # Build per-row annotation lines for custom diagnostics
  row_annotations <- NULL
  if (show_custom) {
    row_annotations <- vector("list", n_snaps)
    for (i in seq_len(n_snaps)) {
      snap <- x$snapshots[[i]]
      if (!is.null(snap$custom) && length(snap$custom) > 0L) {
        row_annotations[[i]] <- vapply(
          names(snap$custom),
          function(nm) .format_custom_result(nm, snap$custom[[nm]]),
          character(1)
        )
      }
    }
  }

  .cli_table(tbl, right_align = c("Rows", "Cols", "NAs"), row_annotations = row_annotations)

  # Print change summaries as a table
  if (n_snaps > 1L) {
  change_rows <- lapply(seq(2L, n_snaps), function(i) {
    snap <- x$snapshots[[i]]
    if (is.null(snap$changes)) return(NULL)
    ch <- snap$changes
    list(
      From = x$snapshots[[i - 1L]]$label,
      To   = snap$label,
      Rows = .format_delta(ch$row_delta),
      Cols = .format_delta(ch$col_delta),
      NAs  = .format_delta(ch$na_delta)
    )
  })
  change_rows <- Filter(Negate(is.null), change_rows)
  }

  if (n_snaps > 1L && length(change_rows) > 0L) {
    cli::cli_text("")
    cli::cli_text("{.strong Changes:}")
    changes_tbl <- data.frame(
      From = vapply(change_rows, `[[`, character(1), "From"),
      To   = vapply(change_rows, `[[`, character(1), "To"),
      Rows = vapply(change_rows, `[[`, character(1), "Rows"),
      Cols = vapply(change_rows, `[[`, character(1), "Cols"),
      NAs  = vapply(change_rows, `[[`, character(1), "NAs"),
      stringsAsFactors = FALSE
    )
    .cli_table(changes_tbl, right_align = c("Rows", "Cols", "NAs"))
  }

  invisible(x)
}

#' Print a formatted table with header separator via cli_verbatim
#'
#' @param tbl A data.frame of character values (already formatted for display).
#' @param right_align Character vector of column names to right-align. All other
#'   columns are left-aligned.
#' @param indent Number of leading spaces (default 2).
#'
#' @noRd
.cli_table <- function(tbl, right_align = character(), indent = 2L, row_annotations = NULL) {
  nms <- names(tbl)

  # Compute column widths (max of header and data)
  col_widths <- vapply(nms, function(nm) {
    data_width <- if (nrow(tbl) > 0L) max(nchar(tbl[[nm]]), na.rm = TRUE) else 0L
    max(nchar(nm), data_width)
  }, integer(1))

  pad <- strrep(" ", indent)
  gap <- "  "

  # Format one row
  format_row <- function(values) {
    cells <- vapply(seq_along(nms), function(j) {
      w <- col_widths[j]
      if (nms[j] %in% right_align) {
        formatC(values[j], width = w)
      } else {
        formatC(values[j], width = w, flag = "-")
      }
    }, character(1))
    paste0(pad, paste(cells, collapse = gap))
  }

  # Header
  header_line <- format_row(nms)

  # Separator: thin line matching each column width
  sep_cells <- vapply(seq_along(nms), function(j) {
    strrep("\u2500", col_widths[j])
  }, character(1))
  sep_line <- paste0(pad, paste(sep_cells, collapse = gap))

  # Data rows — interleave annotation lines where present
  data_lines <- character(0)
  for (i in seq_len(nrow(tbl))) {
    data_lines <- c(data_lines, format_row(as.character(tbl[i, ])))
    if (!is.null(row_annotations) && !is.null(row_annotations[[i]])) {
      data_lines <- c(data_lines, row_annotations[[i]])
    }
  }

  cli::cli_verbatim(paste(c(header_line, sep_line, data_lines), collapse = "\n"))
}

#' Format snapshot type for the timeline table
#' @noRd
.format_snap_type <- function(snap) {
  type <- snap$type %||% "tap"
  diag <- snap$diagnostics

  if (type == "join" && !is.null(diag)) {
    join_type <- diag$join_type %||% "join"
    relation <- diag$relation
    match_x <- diag$match_rate$x
    if (!is.null(match_x) && !is.na(match_x) && !is.null(relation)) {
      return(glue::glue("{join_type} ({relation}, {round(match_x, 1)}% matched)"))
    }
    if (!is.null(relation)) {
      return(glue::glue("{join_type} ({relation})"))
    }
    return(join_type)
  }

  if (type == "filter" && !is.null(diag)) {
    filter_type <- diag$filter_type %||% "keep"
    n_dropped <- diag$n_dropped %||% 0L
    pct_dropped <- diag$pct_dropped %||% 0
    action <- if (filter_type == "keep") "dropped" else "removed"
    return(glue::glue("filter ({action} {format(n_dropped, big.mark = ',')} rows, {round(pct_dropped, 1)}%)"))
  }

  type
}

#' Format a numeric delta as a signed string for the changes table
#' @noRd
.format_delta <- function(d) {
  if (is.null(d) || is.na(d)) return("=")
  if (d > 0L) paste0("+", format(d, big.mark = ",", trim = TRUE))
  else if (d < 0L) format(d, big.mark = ",", trim = TRUE)
  else "="
}

#' Format a change summary for display
#' @noRd
.format_change_summary <- function(changes) {
  parts <- character()

  # Row delta
  rd <- changes$row_delta
  if (rd > 0) {
    parts <- c(parts, glue::glue("+{format(rd, big.mark = ',')} rows"))
  } else if (rd < 0) {
    parts <- c(parts, glue::glue("{format(rd, big.mark = ',')} rows"))
  } else {
    parts <- c(parts, "= rows")
  }

  # Col delta
  cd <- changes$col_delta
  if (cd > 0) {
    parts <- c(parts, glue::glue("+{cd} cols"))
  } else if (cd < 0) {
    parts <- c(parts, glue::glue("{cd} cols"))
  } else {
    parts <- c(parts, "= cols")
  }

  # NA delta
  nd <- changes$na_delta
  if (nd > 0) {
    parts <- c(parts, glue::glue("+{format(nd, big.mark = ',')} NAs"))
  } else if (nd < 0) {
    parts <- c(parts, glue::glue("{format(nd, big.mark = ',')} NAs"))
  } else {
    parts <- c(parts, "= NAs")
  }

  paste(parts, collapse = ", ")
}

#' Detect if x is a named collection of scalar values (Case 2 rendering)
#'
#' Returns TRUE for named atomic vectors and named lists whose every element is
#' a length-1 atomic. Returns FALSE for unnamed structures, data frames, and
#' anything with complex/nested elements.
#'
#' @param x Any R object.
#' @returns Logical scalar.
#' @noRd
.is_named_scalars <- function(x) {
  if (length(x) == 0L) return(FALSE)
  if (is.null(names(x)) || !all(nzchar(names(x)))) return(FALSE)
  if (is.data.frame(x)) return(FALSE)
  if (is.atomic(x) && !is.list(x)) return(TRUE)   # named atomic vector
  if (is.list(x)) {
    return(all(vapply(x, function(e) is.atomic(e) && length(e) == 1L, logical(1))))
  }
  FALSE
}

#' Format one custom diagnostic result for inline display
#'
#' @param name Character scalar: the function name as it appears in `snap$custom`.
#' @param result The value returned by the custom function.
#' @param indent Integer: number of leading spaces before the arrow (default 5).
#' @returns A single character string, e.g. `"     ↳ fn_name: key=val"`.
#' @noRd
.format_custom_result <- function(name, result, indent = 5L) {
  pad <- strrep(" ", indent)
  prefix <- paste0(pad, "\u21b3 ", name, ": ")

  # Case 1: single scalar
  if (is.atomic(result) && length(result) == 1L && !is.list(result)) {
    return(paste0(prefix, format(result)))
  }

  # Case 2: named atomic vector or named list of length-1 atomics
  if (.is_named_scalars(result)) {
    pairs <- paste(
      names(result),
      vapply(result, function(v) format(v[[1L]]), character(1)),  # [[1L]] works for both atomic elements and length-1 list elements
      sep = "="
    )
    value_str <- paste(pairs, collapse = ", ")
    if (nchar(value_str) > 60L) {
      value_str <- paste0(substr(value_str, 1L, 60L), "...")
    }
    return(paste0(prefix, value_str))
  }

  # Case 3: complex object
  paste0(prefix, "[complex \u2014 see audit_report()]")
}

test_that("tab_tap stores tabulation in trail and returns data unchanged", {
  trail <- audit_trail("test")
  result <- mtcars |>
    tab_tap(cyl, .trail = trail, .label = "by_cyl")

  expect_identical(result, mtcars)
  expect_equal(length(trail$snapshots), 1L)

  snap <- trail$snapshots[[1L]]
  expect_equal(snap$label, "by_cyl")
  expect_true("tab(cyl)" %in% names(snap$custom))

  tab_result <- snap$custom[["tab(cyl)"]]
  expect_s3_class(tab_result, "tidyaudit_tab")
  expect_equal(tab_result$var_name, "cyl")
})

test_that("tab_tap works with multiple taps in a pipeline", {
  trail <- audit_trail("pipeline")
  result <- mtcars |>
    tab_tap(cyl, .trail = trail, .label = "raw") |>
    dplyr::filter(mpg > 20) |>
    tab_tap(cyl, .trail = trail, .label = "filtered")

  expect_equal(length(trail$snapshots), 2L)
  expect_equal(nrow(result), sum(mtcars$mpg > 20))
})

test_that("tab_tap passes .sort and .cutoff through", {
  trail <- audit_trail("test")
  mtcars |>
    tab_tap(cyl, .trail = trail, .label = "s1",
            .sort = "freq_desc", .cutoff = 2)

  tab_result <- trail$snapshots[[1L]]$custom[["tab(cyl)"]]
  expect_equal(tab_result$sort, "freq_desc")
  expect_equal(tab_result$n_other, 1L)
})

test_that("tab_tap passes .wt through", {
  trail <- audit_trail("test")
  mtcars |>
    tab_tap(cyl, .trail = trail, .label = "s1", .wt = mpg)

  tab_result <- trail$snapshots[[1L]]$custom[["tab(cyl)"]]
  expect_true(tab_result$weighted)
  expect_equal(tab_result$wt_name, "mpg")
})

test_that("tab_tap errors without audit_trail", {
  expect_error(
    tab_tap(mtcars, cyl, .trail = "not a trail", .label = "x"),
    "audit_trail"
  )
})

test_that("tab_tap with two variables names the diagnostic correctly", {
  trail <- audit_trail("test")
  mtcars |>
    tab_tap(cyl, gear, .trail = trail, .label = "cross")

  snap <- trail$snapshots[[1L]]
  expect_true("tab(cyl, gear)" %in% names(snap$custom))

  tab_result <- snap$custom[["tab(cyl, gear)"]]
  expect_equal(tab_result$type, "twoway")
})

test_that("audit_report works with console format", {
  trail <- audit_trail("report_test")
  mtcars |> audit_tap(trail, "raw")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")
  dplyr::mutate(dplyr::filter(mtcars, mpg > 20), mpg2 = mpg * 2) |>
    audit_tap(trail, "final")

  output <- capture.output(audit_report(trail), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Audit Report", combined))
})

test_that("audit_report returns trail invisibly", {
  trail <- audit_trail("invisible_test")
  mtcars |> audit_tap(trail, "raw")

  capture.output(result <- audit_report(trail), type = "message")
  expect_identical(result, trail)
})

test_that("audit_report errors on non-trail", {
  expect_error(audit_report("not_a_trail"), "audit_trail")
})

test_that("audit_report rejects unknown format", {
  trail <- audit_trail("format_test")
  expect_error(audit_report(trail, format = "rmd"))
})

test_that("audit_report works with empty trail", {
  trail <- audit_trail("empty_report")
  output <- capture.output(audit_report(trail), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("No snapshots", combined))
})

test_that("audit_report shows custom diagnostics", {
  trail <- audit_trail("custom_report")
  mtcars |> audit_tap(trail, "raw", .fns = list(
    n_cyl_types = ~length(unique(.x$cyl))
  ))

  output <- capture.output(audit_report(trail), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Custom Diagnostics", combined))
  expect_true(grepl("n_cyl_types", combined))
})

test_that("audit_report shows unnamed .fns with auto-generated names", {
  trail <- audit_trail("unnamed_fns_report")
  mtcars |> audit_tap(trail, "raw", .fns = list(nrow, ~ncol(.x)))

  output <- capture.output(audit_report(trail), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Custom Diagnostics", combined))
  expect_true(grepl("fn_1", combined))
  expect_true(grepl("fn_2", combined))
})

test_that("audit_report shows snapshot controls when non-default", {
  trail <- audit_trail("controls_report")
  mtcars |> audit_tap(trail, "raw")
  mtcars |> audit_tap(trail, "filtered_cols",
                       .cols_include = c("mpg", "cyl"),
                       .numeric_summary = FALSE)

  output <- capture.output(audit_report(trail), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Snapshot Controls", combined))
  expect_true(grepl("filtered_cols", combined))
  expect_true(grepl("cols_include", combined))
  expect_true(grepl("numeric_summary", combined))
})

test_that("audit_trail() creates a valid trail object", {
  trail <- audit_trail("test_trail")

  expect_s3_class(trail, "audit_trail")
  expect_s3_class(trail, "environment")
  expect_equal(trail$name, "test_trail")
  expect_s3_class(trail$created_at, "POSIXct")
  expect_equal(trail$snapshots, list())
  expect_equal(trail$labels, character())
})

test_that("audit_trail() auto-generates name when NULL", {
  trail <- audit_trail()

  expect_true(grepl("^trail_", trail$name))
})

test_that("audit_trail uses reference semantics", {
  trail <- audit_trail("ref_test")
  trail_ref <- trail

  # Modify via one reference
  trail$snapshots[[1]] <- list(label = "test")

  # Should be visible via the other reference
  expect_equal(trail_ref$snapshots[[1]]$label, "test")
})

test_that("print.audit_trail works with no snapshots", {
  trail <- audit_trail("empty")

  output <- capture.output(print(trail), type = "message")
  combined <- paste(output, collapse = "\n")

  expect_true(grepl("Audit Trail", combined))
  expect_true(grepl("Snapshots: 0", combined))
  expect_true(grepl("No snapshots recorded", combined))
})

test_that("print.audit_trail works with snapshots", {
  trail <- audit_trail("print_test")
  mtcars |> audit_tap(trail, "raw")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")

  output <- capture.output(print(trail), type = "message")
  combined <- paste(output, collapse = "\n")

  expect_true(grepl("print_test", combined))
  expect_true(grepl("raw", combined))
  expect_true(grepl("filtered", combined))
})

test_that(".is_named_scalars() detects Case 2 correctly", {
  # TRUE cases
  expect_true(tidyaudit:::.is_named_scalars(c(a = 1, b = 2)))
  expect_true(tidyaudit:::.is_named_scalars(c(x = "foo", y = "bar")))
  expect_true(tidyaudit:::.is_named_scalars(list(n = 5L, pct = 98.2)))

  # FALSE cases
  expect_false(tidyaudit:::.is_named_scalars(42))
  expect_false(tidyaudit:::.is_named_scalars(c(1, 2, 3)))
  expect_false(tidyaudit:::.is_named_scalars(list(a = 1:3)))
  expect_false(tidyaudit:::.is_named_scalars(data.frame(a = 1, b = 2)))
  expect_false(tidyaudit:::.is_named_scalars(list(a = list(x = 1))))
  expect_false(tidyaudit:::.is_named_scalars(setNames(character(0), character(0))))  # zero-length
})

test_that(".format_custom_result() renders Case 1 (scalar)", {
  result <- tidyaudit:::.format_custom_result("n_rows", 42L)
  expect_true(grepl("\u21b3 n_rows: 42", result))
})

test_that(".format_custom_result() renders Case 2 (named vector)", {
  result <- tidyaudit:::.format_custom_result("check", c(a = 1, b = 2))
  expect_true(grepl("\u21b3 check:", result))
  expect_true(grepl("a=1", result))
  expect_true(grepl("b=2", result))
})

test_that(".format_custom_result() renders Case 2 (named list of scalars)", {
  result <- tidyaudit:::.format_custom_result("quality", list(n_nulls = 5L, pct = 98.2))
  expect_true(grepl("\u21b3 quality:", result))
  expect_true(grepl("n_nulls=5", result))
  expect_true(grepl("pct=98.2", result))
})

test_that(".format_custom_result() truncates Case 2 output beyond 60 chars", {
  long_result <- setNames(as.double(1:20), paste0("p", 1:20))
  result <- tidyaudit:::.format_custom_result("deciles", long_result)
  value_portion <- sub(".*\u21b3 deciles: ", "", result)
  expect_true(endsWith(value_portion, "..."))
  expect_equal(nchar(value_portion), 63L)  # exactly 60 chars + "..."
})

test_that(".format_custom_result() renders Case 3 (complex)", {
  result <- tidyaudit:::.format_custom_result("report", data.frame(x = 1))
  expect_true(grepl("\u21b3 report:", result))
  expect_true(grepl("complex", result))
  expect_true(grepl("audit_report", result))

  result2 <- tidyaudit:::.format_custom_result("vec", 1:5)  # unnamed length-5 vector
  expect_true(grepl("complex", result2))
})

test_that("print.audit_trail shows inline annotations for custom diagnostics", {
  trail <- audit_trail("custom_test")

  mtcars |> audit_tap(trail, "raw", .fns = list(
    n_rows  = ~nrow(.),
    quality = ~c(n_na = sum(is.na(.)), n_cols = ncol(.))
  ))

  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered", .fns = list(
    n_rows  = ~nrow(.),
    quality = ~c(n_na = sum(is.na(.)), n_cols = ncol(.))
  ))

  output <- capture.output(print(trail), type = "message")
  combined <- paste(output, collapse = "\n")

  expect_true(grepl("\u21b3 n_rows", combined))
  expect_true(grepl("\u21b3 quality", combined))

  expect_equal(sum(grepl("\u21b3 n_rows", output)), 2L)
})

test_that("print.audit_trail shows no annotation lines when .fns is not used", {
  trail <- audit_trail("no_custom")
  mtcars |> audit_tap(trail, "raw")

  output <- capture.output(print(trail), type = "message")
  combined <- paste(output, collapse = "\n")

  expect_false(grepl("\u21b3", combined))
})

test_that("print.audit_trail shows annotation only for snapshots with custom diagnostics", {
  trail <- audit_trail("partial_custom")
  mtcars |> audit_tap(trail, "raw")
  mtcars |> audit_tap(trail, "checked", .fns = list(n = ~nrow(.)))

  output <- capture.output(print(trail), type = "message")

  expect_equal(sum(grepl("\u21b3", output)), 1L)
})

test_that("print.audit_trail respects show_custom = FALSE", {
  trail <- audit_trail("show_custom_test")
  mtcars |> audit_tap(trail, "raw", .fns = list(n = ~nrow(.)))

  output <- capture.output(print(trail, show_custom = FALSE), type = "message")
  combined <- paste(output, collapse = "\n")

  expect_false(grepl("\u21b3", combined))
  expect_true(grepl("raw", combined))  # row still present
})

test_that("audit_tap returns data unchanged (pipe transparency)", {
  trail <- audit_trail("transparency")
  input <- mtcars

  output <- audit_tap(input, trail, "step1")

  expect_identical(output, input)
})

test_that("audit_tap appends snapshot to trail", {
  trail <- audit_trail("append_test")
  mtcars |> audit_tap(trail, "raw")

  expect_equal(length(trail$snapshots), 1L)
  expect_equal(trail$labels, "raw")
  expect_s3_class(trail$snapshots[[1]], "audit_snap")
  expect_equal(trail$snapshots[[1]]$label, "raw")
  expect_equal(trail$snapshots[[1]]$nrow, nrow(mtcars))
  expect_equal(trail$snapshots[[1]]$ncol, ncol(mtcars))
})

test_that("audit_tap auto-generates labels", {
  trail <- audit_trail("auto_label")
  mtcars |> audit_tap(trail)
  mtcars |> audit_tap(trail)

  expect_equal(trail$labels, c("step_1", "step_2"))
})

test_that("audit_tap rejects duplicate labels", {
  trail <- audit_trail("dup_test")
  mtcars |> audit_tap(trail, "same")

  expect_error(
    audit_tap(mtcars, trail, "same"),
    "already exists"
  )
})

test_that("audit_tap rejects non-trail object", {
  expect_error(
    audit_tap(mtcars, "not_a_trail", "step"),
    "audit_trail"
  )
})

test_that("audit_tap detects row changes", {
  trail <- audit_trail("row_changes")
  mtcars |> audit_tap(trail, "raw")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")

  snap2 <- trail$snapshots[[2]]
  expect_true(!is.null(snap2$changes))
  expect_true(snap2$changes$row_delta < 0L)
})

test_that("audit_tap detects column changes", {
  trail <- audit_trail("col_changes")
  mtcars |> audit_tap(trail, "raw")
  dplyr::mutate(mtcars, new_col = mpg * 2) |> audit_tap(trail, "added")

  snap2 <- trail$snapshots[[2]]
  expect_equal(snap2$changes$col_delta, 1L)
  expect_equal(snap2$changes$cols_added, "new_col")
})

test_that("audit_tap detects NA changes", {
  trail <- audit_trail("na_changes")
  df <- data.frame(x = 1:5, y = c(1, NA, 3, NA, 5))
  df |> audit_tap(trail, "with_na")
  df[complete.cases(df), ] |> audit_tap(trail, "no_na")

  snap2 <- trail$snapshots[[2]]
  expect_true(snap2$changes$na_delta < 0L)
})

test_that("audit_tap works with custom .fns", {
  trail <- audit_trail("custom_fns")
  mtcars |> audit_tap(trail, "raw", .fns = list(
    n_rows = nrow,
    mean_mpg = ~mean(.x$mpg)
  ))

  snap <- trail$snapshots[[1]]
  expect_equal(snap$custom$n_rows, 32L)
  expect_equal(snap$custom$mean_mpg, mean(mtcars$mpg))
})

test_that("audit_tap works on zero-row data", {
  trail <- audit_trail("empty_data")
  empty_df <- mtcars[0, ]
  result <- audit_tap(empty_df, trail, "empty")

  expect_identical(result, empty_df)
  expect_equal(trail$snapshots[[1]]$nrow, 0L)
  expect_equal(trail$snapshots[[1]]$total_nas, 0L)
})

test_that("audit_tap modifies trail by reference inside pipe", {
  trail <- audit_trail("ref_pipe")
  result <- mtcars |>
    audit_tap(trail, "step1") |>
    dplyr::filter(mpg > 20) |>
    audit_tap(trail, "step2") |>
    dplyr::mutate(mpg2 = mpg * 2) |>
    audit_tap(trail, "step3")

  expect_equal(length(trail$snapshots), 3L)
  expect_equal(trail$labels, c("step1", "step2", "step3"))
})

test_that("audit_tap works with grouped tibbles", {
  trail <- audit_trail("grouped")
  grouped_df <- dplyr::group_by(mtcars, cyl)
  result <- audit_tap(grouped_df, trail, "grouped")

  # Data passes through with groups intact
  expect_true(dplyr::is_grouped_df(result))
  expect_equal(dplyr::group_vars(result), "cyl")

  # Snapshot diagnostics are computed on ungrouped data
  snap <- trail$snapshots[[1]]
  expect_equal(snap$nrow, nrow(mtcars))
})

test_that("audit_tap rejects non-data.frame input", {
  trail <- audit_trail("bad_data")
  expect_error(audit_tap(1:10, trail, "vec"), "data.frame or tibble")
  expect_error(audit_tap(matrix(1:4, 2), trail, "mat"), "data.frame or tibble")
})

test_that("audit_tap rejects invalid label types", {
  trail <- audit_trail("bad_label")
  expect_error(audit_tap(mtcars, trail, .label = 42), "single character string")
  expect_error(audit_tap(mtcars, trail, .label = TRUE), "single character string")
  expect_error(audit_tap(mtcars, trail, .label = c("a", "b")), "single character string")
  expect_error(audit_tap(mtcars, trail, .label = NA_character_), "single character string")
})

test_that("audit_tap rejects non-list .fns", {
  trail <- audit_trail("bad_fns")
  expect_error(audit_tap(mtcars, trail, "s", .fns = mean), "named list")
})

test_that("audit_tap auto-names unnamed .fns", {
  trail <- audit_trail("autoname_fns")
  mtcars |> audit_tap(trail, "raw", .fns = list(nrow, ~ncol(.x)))

  snap <- trail$snapshots[[1]]
  expect_equal(names(snap$custom), c("fn_1", "fn_2"))
  expect_equal(snap$custom$fn_1, 32L)
  expect_equal(snap$custom$fn_2, 11L)
})

test_that("audit_tap auto-names partially named .fns", {
  trail <- audit_trail("partname_fns")
  mtcars |> audit_tap(trail, "raw", .fns = list(rows = nrow, ~ncol(.x)))

  snap <- trail$snapshots[[1]]
  expect_equal(names(snap$custom), c("rows", "fn_2"))
})

test_that("audit_tap preserves tibble class", {
  trail <- audit_trail("tibble_class")
  tbl <- dplyr::as_tibble(mtcars)
  result <- audit_tap(tbl, trail, "tbl_step")

  expect_s3_class(result, "tbl_df")
  expect_identical(result, tbl)
})

test_that("audit_tap validates .trail before forcing .data", {
  # If .trail is validated first, an expensive .data should NOT be evaluated
  call_count <- 0L
  expensive_df <- {
    call_count <<- call_count + 1L
    mtcars
  }
  # With a non-trail object, the error should fire before forcing .data
  # We can't perfectly test evaluation order, but we verify the error message
  expect_error(audit_tap(mtcars, "not_a_trail", "s"), "audit_trail")
})

# ---------------------------------------------------------------------------
# .fns with NA names (Fix 3)
# ---------------------------------------------------------------------------

test_that("audit_tap handles .fns list with NA name without crashing", {
  trail <- audit_trail("fns_na_name")
  fns <- list(nrow)
  names(fns) <- NA_character_

  # Must not crash; the NA name should be auto-renamed to fn_1
  mtcars |> audit_tap(trail, "raw", .fns = fns)
  snap <- trail$snapshots[[1]]
  expect_equal(names(snap$custom), "fn_1")
  expect_equal(snap$custom$fn_1, 32L)
})

test_that("audit_tap handles mixed NA and empty names in .fns", {
  trail <- audit_trail("fns_mixed_names")
  fns <- list(nrow, ncol)
  names(fns) <- c("keep", NA_character_)

  mtcars |> audit_tap(trail, "raw", .fns = fns)
  snap <- trail$snapshots[[1]]
  expect_equal(names(snap$custom), c("keep", "fn_2"))
})

# ---------------------------------------------------------------------------
# invisible return (Fix 5)
# ---------------------------------------------------------------------------

test_that("audit_tap returns invisibly", {
  trail <- audit_trail("invisible_test")
  result <- withVisible(audit_tap(mtcars, trail, "step1"))
  expect_false(result$visible)
  expect_identical(result$value, mtcars)
})

# ---------------------------------------------------------------------------
# Snapshot controls
# ---------------------------------------------------------------------------

test_that("audit_tap passes .numeric_summary = FALSE to snapshot", {
  trail <- audit_trail("ns_false")
  mtcars |> audit_tap(trail, "raw", .numeric_summary = FALSE)
  snap <- trail$snapshots[[1]]
  expect_null(snap$numeric_summary)
  expect_equal(snap$nrow, 32L)
})

test_that("audit_tap passes .cols_include to snapshot", {
  trail <- audit_trail("cols_inc")
  mtcars |> audit_tap(trail, "raw", .cols_include = c("mpg", "cyl"))
  snap <- trail$snapshots[[1]]
  expect_equal(snap$schema$column, c("mpg", "cyl"))
  expect_equal(snap$ncol, 11L)
})

test_that("audit_tap passes .cols_exclude to snapshot", {
  trail <- audit_trail("cols_exc")
  df <- data.frame(a = 1:3, b = 4:6, c = 7:9)
  df |> audit_tap(trail, "raw", .cols_exclude = "b")
  snap <- trail$snapshots[[1]]
  expect_equal(snap$schema$column, c("a", "c"))
  expect_equal(snap$ncol, 3L)
})

test_that(".detect_changes works with filtered schemas", {
  trail <- audit_trail("changes_filtered")
  df1 <- data.frame(a = 1:5, b = 6:10, c = 11:15)
  df2 <- data.frame(a = 1:3, b = 4:6, c = 7:9)
  df1 |> audit_tap(trail, "s1", .cols_include = c("a", "b"))
  df2 |> audit_tap(trail, "s2", .cols_include = c("a", "b"))
  snap2 <- trail$snapshots[[2]]
  expect_equal(snap2$changes$row_delta, -2L)
  expect_equal(length(snap2$changes$cols_added), 0L)
})

test_that(".detect_changes uses all_columns, not filtered schema", {
  trail <- audit_trail("changes_diff_filters")
  df <- data.frame(a = 1:5, b = 6:10, c = 11:15)
  # Different column filters on same data — no real column changes

  df |> audit_tap(trail, "s1", .cols_include = c("a", "b"))
  df |> audit_tap(trail, "s2", .cols_include = c("a", "c"))
  snap2 <- trail$snapshots[[2]]
  # all_columns are identical, so no phantom additions/removals
  expect_equal(length(snap2$changes$cols_added), 0L)
  expect_equal(length(snap2$changes$cols_removed), 0L)
})

test_that(".detect_changes reports real column addition via all_columns", {
  trail <- audit_trail("changes_real_add")
  df1 <- data.frame(a = 1:3, b = 4:6)
  df2 <- data.frame(a = 1:3, b = 4:6, c = 7:9)
  df1 |> audit_tap(trail, "s1", .cols_include = "a")
  df2 |> audit_tap(trail, "s2", .cols_include = "a")
  snap2 <- trail$snapshots[[2]]
  expect_equal(snap2$changes$cols_added, "c")
  expect_equal(snap2$changes$col_delta, 1L)
})

test_that("audit_diff computes correct deltas", {
  trail <- audit_trail("diff_test")
  mtcars |> audit_tap(trail, "raw")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")

  diff_obj <- audit_diff(trail, "raw", "filtered")

  expect_s3_class(diff_obj, "audit_diff")
  expect_equal(diff_obj$from_label, "raw")
  expect_equal(diff_obj$to_label, "filtered")
  expect_true(diff_obj$row_delta < 0)
  expect_equal(diff_obj$col_delta, 0L)
  expect_equal(diff_obj$from_nrow, nrow(mtcars))
  expect_equal(diff_obj$to_nrow, sum(mtcars$mpg > 20))
})

test_that("audit_diff works with integer indices", {
  trail <- audit_trail("idx_test")
  mtcars |> audit_tap(trail, "raw")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")

  diff_obj <- audit_diff(trail, 1L, 2L)

  expect_equal(diff_obj$from_label, "raw")
  expect_equal(diff_obj$to_label, "filtered")
})

test_that("audit_diff detects columns added/removed", {
  trail <- audit_trail("col_diff")
  mtcars |> audit_tap(trail, "raw")
  dplyr::select(mtcars, -cyl, -disp) |>
    dplyr::mutate(new_col = 1) |>
    audit_tap(trail, "modified")

  diff_obj <- audit_diff(trail, "raw", "modified")

  expect_true("new_col" %in% diff_obj$cols_added)
  expect_true("cyl" %in% diff_obj$cols_removed)
  expect_true("disp" %in% diff_obj$cols_removed)
})

test_that("audit_diff detects type changes", {
  trail <- audit_trail("type_diff")
  df <- data.frame(x = 1:5, y = letters[1:5], stringsAsFactors = FALSE)
  df |> audit_tap(trail, "before")
  dplyr::mutate(df, x = as.character(x)) |> audit_tap(trail, "after")

  diff_obj <- audit_diff(trail, "before", "after")

  expect_false(is.null(diff_obj$type_changes))
  expect_true("x" %in% diff_obj$type_changes$column)
  expect_equal(diff_obj$type_changes$from[diff_obj$type_changes$column == "x"], "integer")
  expect_equal(diff_obj$type_changes$to[diff_obj$type_changes$column == "x"], "character")
})

test_that("audit_diff detects NA changes per column", {
  trail <- audit_trail("na_diff")
  df <- data.frame(x = 1:10, y = c(1:5, rep(NA, 5)))
  df |> audit_tap(trail, "before")
  df[1:5, ] |> audit_tap(trail, "after")

  diff_obj <- audit_diff(trail, "before", "after")

  expect_false(is.null(diff_obj$na_changes))
  # y had 5 NAs, now has 0
  y_change <- diff_obj$na_changes[diff_obj$na_changes$column == "y", ]
  expect_equal(y_change$from_na, 5L)
  expect_equal(y_change$to_na, 0L)
  expect_equal(y_change$delta, -5L)
})

test_that("audit_diff computes numeric shifts", {
  trail <- audit_trail("num_diff")
  df <- data.frame(x = 1:100, y = rnorm(100, mean = 10))
  df |> audit_tap(trail, "before")
  dplyr::mutate(df, x = x * 2) |> audit_tap(trail, "after")

  diff_obj <- audit_diff(trail, "before", "after")

  expect_false(is.null(diff_obj$numeric_shifts))
  x_shift <- diff_obj$numeric_shifts[diff_obj$numeric_shifts$column == "x", ]
  expect_true(x_shift$mean_shift > 0)
})

test_that("audit_diff errors on invalid label", {
  trail <- audit_trail("err_test")
  mtcars |> audit_tap(trail, "raw")

  expect_error(audit_diff(trail, "raw", "nonexistent"), "not found")
})

test_that("audit_diff errors on out-of-range index", {
  trail <- audit_trail("err_test2")
  mtcars |> audit_tap(trail, "raw")

  expect_error(audit_diff(trail, 1L, 5L), "out of range")
})

test_that("print.audit_diff produces output", {
  trail <- audit_trail("print_diff")
  mtcars |> audit_tap(trail, "raw")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")

  diff_obj <- audit_diff(trail, "raw", "filtered")
  output <- capture.output(print(diff_obj), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Audit Diff", combined))
})

test_that("audit_diff rejects non-integerish numeric refs", {
  trail <- audit_trail("integerish_test")
  mtcars |> audit_tap(trail, "raw")

  expect_error(audit_diff(trail, 1.5, 1L), "whole number")
  expect_error(audit_diff(trail, 1L, 1.9), "whole number")
  expect_error(audit_diff(trail, NA_real_, 1L), "whole number")
  expect_error(audit_diff(trail, 1L, Inf), "whole number")
})

test_that("audit_diff comparing same snapshot to itself gives zero deltas", {
  trail <- audit_trail("self_diff")
  mtcars |> audit_tap(trail, "raw")

  diff_obj <- audit_diff(trail, "raw", "raw")
  expect_equal(diff_obj$row_delta, 0L)
  expect_equal(diff_obj$col_delta, 0L)
  expect_equal(diff_obj$na_delta, 0L)
  expect_length(diff_obj$cols_added, 0L)
  expect_length(diff_obj$cols_removed, 0L)
  expect_null(diff_obj$type_changes)
  expect_null(diff_obj$na_changes)
})

test_that("audit_diff compares non-consecutive snapshots", {
  trail <- audit_trail("nonconsec_test")
  mtcars |> audit_tap(trail, "s1")
  dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "s2")
  dplyr::select(dplyr::filter(mtcars, mpg > 20), -cyl) |> audit_tap(trail, "s3")

  # Compare s1 to s3 directly, skipping s2
  diff_obj <- audit_diff(trail, "s1", "s3")
  expect_true(diff_obj$row_delta < 0L)
  expect_equal(diff_obj$col_delta, -1L)
  expect_true("cyl" %in% diff_obj$cols_removed)
})

#' Build a Metadata Snapshot
#'
#' Internal function that captures a metadata-only snapshot of the data at one
#' point in the pipeline. Does NOT store the data itself.
#'
#' @param .data A data.frame or tibble.
#' @param label Character label for this snapshot.
#' @param index Integer position in the trail.
#' @param .numeric_summary Logical. If `FALSE`, skip numeric summary
#'   computation.
#' @param .cols_include Character vector of column names to include in the
#'   schema, or `NULL` (the default) to include all columns.
#' @param .cols_exclude Character vector of column names to exclude from the
#'   schema, or `NULL` (the default) to exclude none.
#'
#' @returns An `audit_snap` object (S3 list).
#'
#' @noRd
.build_snapshot <- function(.data, label, index,
                            .numeric_summary = TRUE,
                            .cols_include = NULL,
                            .cols_exclude = NULL) {
  # --- Validate snapshot controls ---
  if (!is.logical(.numeric_summary) || length(.numeric_summary) != 1L ||
      is.na(.numeric_summary)) {
    cli::cli_abort("{.arg .numeric_summary} must be {.val TRUE} or {.val FALSE}.")
  }
  if (!is.null(.cols_include) && !is.character(.cols_include)) {
    cli::cli_abort("{.arg .cols_include} must be a character vector or {.val NULL}.")
  }
  if (!is.null(.cols_exclude) && !is.character(.cols_exclude)) {
    cli::cli_abort("{.arg .cols_exclude} must be a character vector or {.val NULL}.")
  }
  if (!is.null(.cols_include) && !is.null(.cols_exclude)) {
    cli::cli_abort(
      "Supply {.arg .cols_include} or {.arg .cols_exclude}, not both."
    )
  }

  # Diagnostic aggregates on ungrouped data for deterministic results
  ungrouped <- dplyr::ungroup(.data)

  # Full schema — always computed for invariants (nrow, ncol, total_nas)
  full_schema <- data.frame(
    column = names(ungrouped),
    type   = vapply(ungrouped, function(x) class(x)[[1L]], character(1)),
    n_na   = vapply(ungrouped, function(x) sum(is.na(x)), integer(1)),
    stringsAsFactors = FALSE
  )
  total_nas <- as.integer(sum(full_schema$n_na))

  # Apply column filter for the stored schema
  if (!is.null(.cols_include)) {
    if (length(.cols_include) == 0L) {
      cli::cli_warn("{.arg .cols_include} is empty; snapshot schema will have no columns.")
    }
    unmatched <- setdiff(.cols_include, names(ungrouped))
    if (length(unmatched) > 0L) {
      cli::cli_warn(
        "{.arg .cols_include} column{?s} not found in data: {.val {unmatched}}."
      )
    }
    schema <- full_schema[full_schema$column %in% .cols_include, , drop = FALSE]
  } else if (!is.null(.cols_exclude)) {
    schema <- full_schema[!full_schema$column %in% .cols_exclude, , drop = FALSE]
  } else {
    schema <- full_schema
  }
  row.names(schema) <- NULL

  # Numeric summaries (gated on .numeric_summary; scoped to filtered schema)
  numeric_summary <- NULL
  if (.numeric_summary) {
    numeric_cols <- schema$column[schema$type %in% c("numeric", "integer")]
    if (length(numeric_cols) > 0L) {
      summaries <- lapply(numeric_cols, function(col) {
        vals <- ungrouped[[col]]
        vals_clean <- vals[!is.na(vals)]
        if (length(vals_clean) == 0L) {
          data.frame(
            column = col, min = NA_real_, q25 = NA_real_, median = NA_real_,
            mean = NA_real_, q75 = NA_real_, max = NA_real_,
            stringsAsFactors = FALSE
          )
        } else {
          qs <- quantile(vals_clean, probs = c(0.25, 0.75), names = FALSE)
          data.frame(
            column = col,
            min    = min(vals_clean),
            q25    = qs[1L],
            median = median(vals_clean),
            mean   = mean(vals_clean),
            q75    = qs[2L],
            max    = max(vals_clean),
            stringsAsFactors = FALSE
          )
        }
      })
      numeric_summary <- do.call(rbind, summaries)
    }
  }

  # Store controls metadata when non-default
  has_controls <- !isTRUE(.numeric_summary) || !is.null(.cols_include) ||
    !is.null(.cols_exclude)
  controls <- if (has_controls) {
    list(
      numeric_summary = .numeric_summary,
      cols_include    = .cols_include,
      cols_exclude    = .cols_exclude
    )
  }

  snap <- list(
    label           = label,
    index           = index,
    timestamp       = Sys.time(),
    type            = "tap",
    nrow            = as.integer(nrow(ungrouped)),
    ncol            = as.integer(ncol(ungrouped)),
    all_columns     = names(ungrouped),
    schema          = schema,
    total_nas       = total_nas,
    numeric_summary = numeric_summary,
    diagnostics     = NULL,
    pipeline        = NULL,
    changes         = NULL,
    custom          = NULL,
    controls        = controls
  )
  structure(snap, class = c("audit_snap", "list"))
}

#' @rdname audit_trail
#' @export
print.audit_snap <- function(x, ...) {
  cli::cli_h2("Snapshot: {.val {x$label}} ({format(x$nrow, big.mark = ',')} rows x {x$ncol} cols)")
  cli::cli_text("Index: {x$index} | Type: {x$type} | Time: {format(x$timestamp, '%H:%M:%S')}")

  # Pipeline
  if (!is.null(x$pipeline) && length(x$pipeline) > 0L) {
    cli::cli_text("")
    cli::cli_text("{.strong Pipeline:}")
    n <- length(x$pipeline)
    lines <- character(n)
    for (i in seq_along(x$pipeline)) {
      step <- x$pipeline[[i]]
      if (i == 1L) {
        lines[i] <- paste0("  ", step, if (n > 1L) " |>" else "")
      } else if (i == n) {
        lines[i] <- paste0("    ", step)
      } else {
        lines[i] <- paste0("    ", step, " |>")
      }
    }
    cli::cli_verbatim(paste(lines, collapse = "\n"))
  }

  # Schema — grouped by type
  cli::cli_text("")
  n_shown <- nrow(x$schema)
  n_total <- x$ncol
  if (n_shown < n_total) {
    cli::cli_text("{.strong Columns ({n_shown} of {n_total}):}")
  } else {
    cli::cli_text("{.strong Columns ({n_total}):}")
  }
  schema <- x$schema
  types <- unique(schema$type)
  for (tp in types) {
    cols_of_type <- schema$column[schema$type == tp]
    na_of_type <- schema$n_na[schema$type == tp]
    n_type <- length(cols_of_type)

    # Build column list with NA annotations
    col_labels <- vapply(seq_along(cols_of_type), function(j) {
      if (na_of_type[j] > 0L) {
        paste0(cols_of_type[j], " [", na_of_type[j], " NA]")
      } else {
        cols_of_type[j]
      }
    }, character(1))

    col_str <- paste(col_labels, collapse = ", ")
    cli::cli_text("  {.emph {tp}} ({n_type}): {col_str}")
  }

  # Total NAs
  if (x$total_nas > 0L) {
    cli::cli_text("")
    cli::cli_alert_warning("Total NAs: {format(x$total_nas, big.mark = ',')}")
  }

  # Changes from previous
  if (!is.null(x$changes)) {
    cli::cli_text("")
    cli::cli_text("{.strong Changes from previous:}")
    summary <- .format_change_summary(x$changes)
    cli::cli_text("  {summary}")
  }

  invisible(x)
}

#' Capture Pipeline Call Chain
#'
#' Reverse-engineers the pipe chain from an unevaluated expression captured via
#' `substitute(.data)`. Only works with the base R `|>` pipe.
#'
#' @param data_expr An unevaluated expression from `substitute(.data)`.
#' @param .max_steps Maximum number of steps to keep (default 5).
#'
#' @returns A character vector of pipeline steps, or `NULL` on failure.
#'
#' @noRd
.capture_pipeline <- function(data_expr, .max_steps = 5L) {
  steps <- .decompose_pipe(data_expr)
  if (length(steps) > .max_steps + 1L) {
    steps <- c("...", steps[(length(steps) - .max_steps + 1L):length(steps)])
  }
  steps
}

#' Decompose a Nested Pipe Expression
#'
#' Recursively walks nested call expressions to extract individual pipe steps.
#'
#' @param expr An unevaluated R expression.
#'
#' @returns A character vector of deparsed steps.
#'
#' @noRd
.decompose_pipe <- function(expr) {
  if (is.call(expr) && length(expr) >= 2L) {
    inner <- .decompose_pipe(expr[[2L]])
    step <- expr
    step[[2L]] <- quote(.)
    c(inner, deparse(step, width.cutoff = 500L))
  } else {
    deparse(expr)
  }
}

test_that(".build_snapshot captures correct dimensions", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L)

  expect_s3_class(snap, "audit_snap")
  expect_equal(snap$label, "test")
  expect_equal(snap$index, 1L)
  expect_equal(snap$nrow, 32L)
  expect_equal(snap$ncol, 11L)
  expect_equal(snap$type, "tap")
  expect_equal(snap$all_columns, names(mtcars))
})

test_that(".build_snapshot captures schema", {
  df <- data.frame(x = 1:5, y = c("a", "b", "c", NA, "e"),
                   z = c(1.1, 2.2, NA, 4.4, 5.5),
                   stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L)

  expect_equal(nrow(snap$schema), 3L)
  expect_equal(snap$schema$column, c("x", "y", "z"))
  expect_equal(snap$schema$n_na, c(0L, 1L, 1L))
  expect_equal(snap$total_nas, 2L)
})

test_that(".build_snapshot computes numeric summaries", {
  df <- data.frame(x = 1:10, y = letters[1:10], stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L)

  expect_false(is.null(snap$numeric_summary))
  expect_equal(snap$numeric_summary$column, "x")
  expect_equal(snap$numeric_summary$mean, 5.5)
  expect_equal(snap$numeric_summary$min, 1)
  expect_equal(snap$numeric_summary$max, 10)
})

test_that(".build_snapshot handles all-NA numeric column", {
  df <- data.frame(x = rep(NA_real_, 5))
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L)

  expect_false(is.null(snap$numeric_summary))
  expect_true(is.na(snap$numeric_summary$mean))
})

test_that(".build_snapshot handles no numeric columns", {
  df <- data.frame(x = letters[1:5], y = letters[6:10], stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L)

  expect_null(snap$numeric_summary)
})

test_that(".build_snapshot computes on ungrouped data", {
  grouped <- dplyr::group_by(mtcars, cyl)
  snap <- tidyaudit:::.build_snapshot(grouped, label = "test", index = 1L)

  expect_equal(snap$nrow, 32L)
})

test_that("print.audit_snap produces output", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L)
  output <- capture.output(print(snap), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Snapshot", combined))
})

# ── Snapshot controls ─────────────────────────────────────────────────────────

test_that(".numeric_summary = FALSE skips numeric summary", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L,
                                       .numeric_summary = FALSE)
  expect_null(snap$numeric_summary)
  # Schema is still populated
  expect_equal(nrow(snap$schema), 11L)
  # Core invariants still computed
  expect_equal(snap$nrow, 32L)
  expect_equal(snap$ncol, 11L)
})

test_that(".cols_include filters schema to specified columns", {
  df <- data.frame(a = 1:5, b = letters[1:5], c = 1.1:5.1,
                   stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L,
                                       .cols_include = c("a", "c"))
  expect_equal(snap$schema$column, c("a", "c"))
  expect_equal(nrow(snap$schema), 2L)
  # nrow/ncol from full data
  expect_equal(snap$ncol, 3L)
  # all_columns always has full set
  expect_equal(snap$all_columns, c("a", "b", "c"))
})

test_that(".cols_exclude removes specified columns from schema", {
  df <- data.frame(a = 1:5, b = letters[1:5], c = 1.1:5.1,
                   stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L,
                                       .cols_exclude = "b")
  expect_equal(snap$schema$column, c("a", "c"))
  expect_equal(snap$ncol, 3L)
})

test_that("total_nas always computed from all columns", {
  df <- data.frame(a = c(1, NA, 3), b = c(NA, NA, "x"),
                   stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L,
                                       .cols_include = "a")
  # total_nas counts NAs across ALL columns (a has 1, b has 2 = 3 total)
  expect_equal(snap$total_nas, 3L)
  # But schema only has column "a"
  expect_equal(snap$schema$column, "a")
  expect_equal(snap$schema$n_na, 1L)
})

test_that(".cols_include limits numeric summary scope", {
  df <- data.frame(x = 1:10, y = 11:20, z = letters[1:10],
                   stringsAsFactors = FALSE)
  snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L,
                                       .cols_include = c("x", "z"))
  # Only x is numeric in the filtered schema
  expect_equal(snap$numeric_summary$column, "x")
})

test_that("non-existent column names warn and are excluded", {
  df <- data.frame(a = 1:3, b = 4:6)
  expect_warning(
    snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L,
                                         .cols_include = c("a", "nonexistent")),
    "not found in data"
  )
  expect_equal(snap$schema$column, "a")
})

test_that("empty .cols_include warns", {
  df <- data.frame(a = 1:3, b = 4:6)
  expect_warning(
    snap <- tidyaudit:::.build_snapshot(df, label = "test", index = 1L,
                                         .cols_include = character(0)),
    "empty"
  )
  expect_equal(nrow(snap$schema), 0L)
  # Invariants still computed from full data
  expect_equal(snap$ncol, 2L)
  expect_equal(snap$nrow, 3L)
})

test_that("both .cols_include and .cols_exclude errors", {
  expect_error(
    tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L,
                                 .cols_include = "mpg", .cols_exclude = "cyl"),
    "cols_include.*cols_exclude"
  )
})

test_that(".numeric_summary must be logical", {
  expect_error(
    tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L,
                                 .numeric_summary = "yes"),
    "numeric_summary"
  )
})

test_that("controls field stored when non-default", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L,
                                       .numeric_summary = FALSE)
  expect_false(is.null(snap$controls))
  expect_false(snap$controls$numeric_summary)

  snap2 <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L,
                                        .cols_include = c("mpg", "cyl"))
  expect_false(is.null(snap2$controls))
  expect_equal(snap2$controls$cols_include, c("mpg", "cyl"))
})

test_that("controls field is NULL when all defaults", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L)
  expect_null(snap$controls)
})

test_that("print.audit_snap shows 'N of M' when schema is filtered", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L,
                                       .cols_include = c("mpg", "cyl"))
  output <- capture.output(print(snap), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("2 of 11", combined))
})

test_that("print.audit_snap shows plain count when schema is unfiltered", {
  snap <- tidyaudit:::.build_snapshot(mtcars, label = "test", index = 1L)
  output <- capture.output(print(snap), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Columns \\(11\\)", combined))
  expect_false(grepl("of", combined))
})

test_that(".decompose_pipe handles simple symbol", {
  result <- tidyaudit:::.decompose_pipe(quote(df))
  expect_equal(result, "df")
})

test_that(".decompose_pipe handles nested calls", {
  # Simulates: df |> filter(x > 0) |> mutate(y = 1)
  # Which R parses as: mutate(filter(df, x > 0), y = 1)
  expr <- quote(mutate(filter(df, x > 0), y = 1))
  result <- tidyaudit:::.decompose_pipe(expr)

  expect_equal(length(result), 3L)
  expect_equal(result[1], "df")
  expect_equal(result[2], "filter(., x > 0)")
  expect_equal(result[3], "mutate(., y = 1)")
})

test_that(".capture_pipeline truncates long chains", {
  # Build a deeply nested expression (8 steps)
  expr <- quote(f8(f7(f6(f5(f4(f3(f2(f1(df)))))))))
  result <- tidyaudit:::.capture_pipeline(expr, .max_steps = 5L)

  # Should have "..." prefix + 5 steps
  expect_equal(result[1], "...")
  expect_equal(length(result), 6L)
})

test_that(".capture_pipeline returns short chains intact", {
  expr <- quote(filter(df, x > 0))
  result <- tidyaudit:::.capture_pipeline(expr, .max_steps = 5L)

  expect_equal(length(result), 2L)
  expect_equal(result[1], "df")
  expect_equal(result[2], "filter(., x > 0)")
})

#' Audit a Vector Transformation
#'
#' Applies a transformation function to a vector and reports what changed.
#' Works with any vector type: character, numeric, Date/POSIXct, factor, or
#' logical. Diagnostics are adapted to the detected input type.
#'
#' @param x Vector to transform. Accepted types: character, numeric,
#'   Date, POSIXct, factor, or logical.
#' @param clean_fn A function applied to `x` that returns a vector of the
#'   same length, **or** a pre-computed vector of the same length (used
#'   directly as the transformation result).
#' @param name Optional name for the variable (used in output). If `NULL`,
#'   captures the variable name from the call.
#' @param .tolerance Numeric tolerance used for the "changed beyond tolerance"
#'   diagnostic (numeric type only). Defaults to `sqrt(.Machine$double.eps)`.
#'
#' @returns An S3 object of class `audit_transform` containing:
#' \describe{
#'   \item{name}{Name of the variable}
#'   \item{clean_fn_name}{Name of the transformation function, or
#'     `"<pre-computed>"` when a vector was supplied directly}
#'   \item{type_class}{Detected type: `"character"`, `"numeric"`,
#'     `"Date"`, `"POSIXct"`, `"factor"`, or `"logical"`}
#'   \item{n_total}{Total number of elements}
#'   \item{n_changed}{Count of values that changed (including NA status changes)}
#'   \item{n_unchanged}{Count of values that stayed the same}
#'   \item{n_na_before}{Count of NA values before transformation}
#'   \item{n_na_after}{Count of NA values after transformation}
#'   \item{pct_changed}{Percentage of total elements that changed}
#'   \item{change_examples}{Data frame with before/after pairs (up to 10)}
#'   \item{diagnostics}{Type-specific diagnostic list, or `NULL` for character}
#'   \item{cleaned}{The transformed vector, retaining its type}
#' }
#'
#' @examples
#' # Character
#' x <- c("  hello ", "WORLD", "  foo  ", NA)
#' result <- audit_transform(x, trimws)
#' result$cleaned
#'
#' # Numeric
#' prices <- c(10.5, 20.0, NA, 30.0)
#' audit_transform(prices, function(v) round(v))
#'
#' # Pre-computed result
#' audit_transform(prices, round(prices))
#'
#' @family data quality
#' @seealso [diagnose_strings()]
#' @export
audit_transform <- function(x, clean_fn, name = NULL,
                            .tolerance = sqrt(.Machine$double.eps)) {
  if (is.null(name)) {
    name <- deparse(substitute(x))
  }

  if (is.function(clean_fn)) {
    clean_fn_name <- deparse(substitute(clean_fn))
    cleaned <- clean_fn(x)
  } else {
    clean_fn_name <- "<pre-computed>"
    cleaned <- clean_fn
  }

  if (length(cleaned) != length(x)) {
    input_len <- length(x)
    output_len <- length(cleaned)
    cli::cli_abort(
      "{.arg clean_fn} must return a vector of the same length as the input ({input_len}), but returned length {output_len}."
    )
  }

  type_class <- if (inherits(x, c("Date", "POSIXct"))) {
    class(x)[1]
  } else if (is.numeric(x)) {
    "numeric"
  } else if (is.factor(x)) {
    "factor"
  } else if (is.logical(x)) {
    "logical"
  } else {
    "character"
  }

  n_total      <- length(x)
  is_na_before <- is.na(x)
  is_na_after  <- is.na(cleaned)
  n_na_before  <- sum(is_na_before)
  n_na_after   <- sum(is_na_after)

  # Changed: NA status changed, OR both non-NA and values differ.
  # For factors with different level sets, compare via character to avoid error.
  if (is.factor(x) || is.factor(cleaned)) {
    values_differ <- as.character(x) != as.character(cleaned)
  } else {
    values_differ <- x != cleaned
  }
  changed <- (is_na_before != is_na_after) |
    (!is_na_before & !is_na_after & values_differ)

  n_changed   <- sum(changed)
  n_unchanged <- n_total - n_changed
  pct_changed <- if (n_total > 0L) 100 * n_changed / n_total else 0

  if (n_changed > 0L) {
    changed_idx <- which(changed)
    sample_idx  <- changed_idx[seq_len(min(10L, length(changed_idx)))]
    change_examples <- data.frame(
      before = .for_examples(x[sample_idx]),
      after  = .for_examples(cleaned[sample_idx]),
      stringsAsFactors = FALSE
    )
  } else {
    change_examples <- data.frame(
      before = character(0),
      after  = character(0),
      stringsAsFactors = FALSE
    )
  }

  diagnostics <- .audit_transform_diagnostics(
    x, cleaned, type_class, .tolerance, changed
  )

  out <- list(
    name          = name,
    clean_fn_name = clean_fn_name,
    type_class    = type_class,
    n_total       = n_total,
    n_changed     = n_changed,
    n_unchanged   = n_unchanged,
    n_na_before   = n_na_before,
    n_na_after    = n_na_after,
    pct_changed   = pct_changed,
    change_examples = change_examples,
    diagnostics   = diagnostics,
    cleaned       = cleaned
  )
  structure(out, class = c("audit_transform", "list"))
}

# Convert a vector to character for display in change_examples.
# Date/POSIXct uses format() for ISO representation; everything else as.character().
#' @noRd
.for_examples <- function(v) {
  if (inherits(v, c("Date", "POSIXct"))) format(v) else as.character(v)
}

# Compute type-specific diagnostic list
#' @noRd
.audit_transform_diagnostics <- function(x, cleaned, type_class, .tolerance,
                                         changed) {
  switch(type_class,
    "numeric" = {
      delta <- abs(cleaned - x)  # NA where either operand is NA
      n_non_na_delta <- sum(!is.na(delta))
      n_beyond_tol   <- sum(delta > .tolerance, na.rm = TRUE)

      list(
        mean_before    = mean(x,       na.rm = TRUE),
        mean_after     = mean(cleaned, na.rm = TRUE),
        sd_before      = stats::sd(x,       na.rm = TRUE),
        sd_after       = stats::sd(cleaned, na.rm = TRUE),
        min_before     = .safe_stat(x,       min),
        min_after      = .safe_stat(cleaned, min),
        median_before  = .safe_stat(x,       stats::median),
        median_after   = .safe_stat(cleaned, stats::median),
        max_before     = .safe_stat(x,       max),
        max_after      = .safe_stat(cleaned, max),
        n_nan_before   = sum(is.nan(x)),
        n_nan_after    = sum(is.nan(cleaned)),
        n_inf_before   = sum(is.infinite(x)),
        n_inf_after    = sum(is.infinite(cleaned)),
        # na.rm = TRUE: NA-status changes (value→NA, NA→value) produce NA delta;
        # silently excluded so mean reflects only finite-to-finite changes.
        mean_abs_delta = if (any(changed)) mean(delta[changed], na.rm = TRUE) else 0,
        n_beyond_tol   = n_beyond_tol,
        pct_beyond_tol = if (n_non_na_delta > 0L) 100 * n_beyond_tol / n_non_na_delta else 0
      )
    },
    "Date" = ,
    "POSIXct" = {
      list(
        min_before        = .safe_stat(x,       min),
        min_after         = .safe_stat(cleaned, min),
        max_before        = .safe_stat(x,       max),
        max_after         = .safe_stat(cleaned, max),
        range_days_before = .date_range_days(x),
        range_days_after  = .date_range_days(cleaned)
      )
    },
    "factor" = {
      lvl_before <- levels(x)
      lvl_after  <- if (is.factor(cleaned)) {
        levels(cleaned)
      } else {
        sort(unique(as.character(cleaned[!is.na(cleaned)])))
      }
      list(
        levels_before       = lvl_before,
        levels_after        = lvl_after,
        levels_added        = setdiff(lvl_after,  lvl_before),
        levels_removed      = setdiff(lvl_before, lvl_after),
        level_counts_before = .factor_counts(x),
        level_counts_after  = .factor_counts(cleaned)
      )
    },
    "logical" = {
      list(
        n_true_before  = sum( x,       na.rm = TRUE),
        n_true_after   = sum( cleaned, na.rm = TRUE),
        n_false_before = sum(!x,       na.rm = TRUE),
        n_false_after  = sum(!cleaned, na.rm = TRUE)
      )
    },
    # character — no extra diagnostics needed
    NULL
  )
}

# min/max that returns NA instead of Inf/-Inf on all-NA input
#' @noRd
.safe_stat <- function(x, fn) {
  x_valid <- x[!is.na(x)]
  if (length(x_valid) == 0L) NA else fn(x_valid)
}

# Frequency table for a factor/vector as a data.frame with standard columns.
# Built manually rather than via as.data.frame(table(...)) — under R-devel
# (r89994+), as.data.frame.table() routes the count vector through
# as.data.frame.integer(), which now rejects NA in names. With useNA = "always"
# the table's NA bucket carries an NA name and trips that check.
#' @noRd
.factor_counts <- function(x) {
  tbl <- table(x, useNA = "always")
  data.frame(
    Level = names(tbl),
    N     = unname(as.integer(tbl)),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

# Date/POSIXct range in days; NA if fewer than two non-NA values
#' @noRd
.date_range_days <- function(x) {
  x_valid <- x[!is.na(x)]
  if (length(x_valid) < 2L) {
    return(NA_real_)
  }
  as.numeric(difftime(max(x_valid), min(x_valid), units = "days"))
}

#' @rdname audit_transform
#' @param ... Additional arguments (currently unused).
#' @export
print.audit_transform <- function(x, ...) {
  fmt_int <- function(z) format(z, big.mark = ",", scientific = FALSE, trim = TRUE)
  fmt_pct <- function(z) sprintf("%.1f%%", z)

  type_label <- x$type_class
  x_name     <- x$name
  cli::cli_h1("Transformation Audit [{type_label}]: {x_name}")

  fn_name <- x$clean_fn_name
  cli::cli_text("Function: {fn_name}")
  cli::cli_text("")

  # Common metrics table
  tbl <- data.frame(
    Metric = c("Total elements", "NA (before)", "NA (after)", "Changed",
               "Unchanged"),
    Value  = c(
      fmt_int(x$n_total),
      fmt_int(x$n_na_before),
      fmt_int(x$n_na_after),
      paste0(fmt_int(x$n_changed),
             " (", fmt_pct(x$pct_changed), " of total)"),
      fmt_int(x$n_unchanged)
    ),
    stringsAsFactors = FALSE
  )
  .cli_table(tbl, right_align = "Value")

  # Type-specific summary
  .print_transform_diagnostics(x, fmt_int, fmt_pct)

  # Change examples
  if (x$n_changed > 0L) {
    n_ex <- nrow(x$change_examples)
    cli::cli_text("")
    cli::cli_text(
      "Examples of changes (showing {n_ex} of {fmt_int(x$n_changed)}):"
    )
    print(x$change_examples, row.names = FALSE)
  }

  cli::cli_text("")
  cli::cli_text("Access cleaned vector with: {.code result$cleaned}")

  invisible(x)
}

# Dispatch type-specific summary section
#' @noRd
.print_transform_diagnostics <- function(x, fmt_int, fmt_pct) {
  d <- x$diagnostics
  if (is.null(d)) return(invisible(NULL))

  fmt_num <- function(z) {
    if (is.na(z)) "NA" else sprintf("%.4g", z)
  }

  tc <- x$type_class

  if (tc == "numeric") {
    cli::cli_text("")
    cli::cli_text("Numeric summary:")
    stats_tbl <- data.frame(
      Metric = c("Mean", "SD", "Min", "Median", "Max", "NaN", "Inf"),
      Before = c(
        fmt_num(d$mean_before), fmt_num(d$sd_before),
        fmt_num(d$min_before),  fmt_num(d$median_before),
        fmt_num(d$max_before),
        fmt_int(d$n_nan_before), fmt_int(d$n_inf_before)
      ),
      After  = c(
        fmt_num(d$mean_after), fmt_num(d$sd_after),
        fmt_num(d$min_after),  fmt_num(d$median_after),
        fmt_num(d$max_after),
        fmt_int(d$n_nan_after), fmt_int(d$n_inf_after)
      ),
      stringsAsFactors = FALSE
    )
    .cli_table(stats_tbl, right_align = c("Before", "After"))
    cli::cli_text("")
    mean_delta <- fmt_num(d$mean_abs_delta)
    cli::cli_text("Mean absolute delta: {mean_delta}")
    n_bt  <- fmt_int(d$n_beyond_tol)
    p_bt  <- fmt_pct(d$pct_beyond_tol)
    cli::cli_text("Changed beyond tolerance: {n_bt} ({p_bt})")

  } else if (tc %in% c("Date", "POSIXct")) {
    cli::cli_text("")
    cli::cli_text("Date range:")
    fmt_date <- function(z) if (is.na(z)) "NA" else format(z)
    fmt_days <- function(z) if (is.na(z)) "NA" else sprintf("%.1f", z)
    range_tbl <- data.frame(
      Metric = c("Min", "Max", "Span (days)"),
      Before = c(fmt_date(d$min_before), fmt_date(d$max_before),
                 fmt_days(d$range_days_before)),
      After  = c(fmt_date(d$min_after),  fmt_date(d$max_after),
                 fmt_days(d$range_days_after)),
      stringsAsFactors = FALSE
    )
    .cli_table(range_tbl, right_align = c("Before", "After"))

  } else if (tc == "factor") {
    cli::cli_text("")
    cli::cli_text("Factor levels:")
    if (length(d$levels_added) > 0L) {
      added <- paste(d$levels_added, collapse = ", ")
      cli::cli_bullets(c("*" = "Levels added:   {added}"))
    }
    if (length(d$levels_removed) > 0L) {
      removed <- paste(d$levels_removed, collapse = ", ")
      cli::cli_bullets(c("*" = "Levels removed: {removed}"))
    }
    if (length(d$levels_added) == 0L && length(d$levels_removed) == 0L) {
      cli::cli_text("No level changes.")
    }
    cli::cli_text("")
    cli::cli_text("Level counts (before | after):")
    cnt_b <- d$level_counts_before
    cnt_a <- d$level_counts_after
    cnt_b$Level[is.na(cnt_b$Level)] <- "<NA>"
    cnt_a$Level[is.na(cnt_a$Level)] <- "<NA>"
    names(cnt_b) <- c("Level", "Before")
    names(cnt_a) <- c("Level", "After")
    cnt <- merge(cnt_b, cnt_a, by = "Level", all = TRUE)
    cnt$Before[is.na(cnt$Before)] <- 0L
    cnt$After[is.na(cnt$After)]   <- 0L
    cnt$Before <- fmt_int(cnt$Before)
    cnt$After  <- fmt_int(cnt$After)
    .cli_table(cnt, right_align = c("Before", "After"))

  } else if (tc == "logical") {
    cli::cli_text("")
    cli::cli_text("Logical balance:")
    na_b <- x$n_na_before
    na_a <- x$n_na_after
    bal_tbl <- data.frame(
      Value  = c("TRUE", "FALSE", "NA"),
      Before = fmt_int(c(d$n_true_before, d$n_false_before, na_b)),
      After  = fmt_int(c(d$n_true_after,  d$n_false_after,  na_a)),
      stringsAsFactors = FALSE
    )
    .cli_table(bal_tbl, right_align = c("Before", "After"))
  }

  invisible(NULL)
}

test_that("audit_transform [character] reports changes from trimws", {
  x <- c("  hello ", "world", "  foo  ", NA)
  result <- audit_transform(x, trimws)

  expect_s3_class(result, "audit_transform")
  expect_equal(result$type_class, "character")
  expect_equal(result$n_total, 4L)
  expect_equal(result$n_na_before, 1L)
  expect_equal(result$n_changed, 2L)  # "  hello " and "  foo  " change
  expect_equal(result$n_unchanged, 2L) # "world" and NA unchanged
  expect_equal(length(result$cleaned), 4L)
  expect_equal(result$cleaned[1], "hello")
  expect_equal(result$cleaned[3], "foo")
  expect_true(is.na(result$cleaned[4]))
  expect_null(result$diagnostics)
})

test_that("audit_transform [character] captures function name", {
  x <- c("a", "b")
  result <- audit_transform(x, toupper)

  expect_equal(result$clean_fn_name, "toupper")
  expect_equal(result$cleaned, c("A", "B"))
  expect_equal(result$n_changed, 2L)
})

test_that("audit_transform [character] handles no changes", {
  x <- c("HELLO", "WORLD")
  result <- audit_transform(x, toupper)

  expect_equal(result$n_changed, 0L)
  expect_equal(nrow(result$change_examples), 0L)
  expect_equal(result$pct_changed, 0)
})

test_that("audit_transform [character] provides change examples", {
  x <- c("  a  ", "b", "  c  ", "d", "  e  ")
  result <- audit_transform(x, trimws)

  expect_true(nrow(result$change_examples) > 0L)
  expect_true(all(c("before", "after") %in% names(result$change_examples)))
})

test_that("audit_transform [character] handles all-NA input", {
  x <- c(NA_character_, NA_character_)
  result <- audit_transform(x, toupper)

  expect_equal(result$n_changed, 0L)
  expect_equal(result$n_na_before, 2L)
  expect_equal(result$pct_changed, 0)
})

test_that("audit_transform [character] captures custom name", {
  result <- audit_transform(c("a", "b"), toupper, name = "test_col")
  expect_equal(result$name, "test_col")
})

test_that("print.audit_transform [character] produces output", {
  x <- c("  hello ", "world")
  result <- audit_transform(x, trimws)

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Transformation Audit", combined))
  expect_true(grepl("character", combined))
  expect_true(grepl("trimws", combined))
})

test_that("audit_transform [character] pct_changed uses n_total as denominator", {
  # 4 of 5 elements change (NA stays NA); pct = 4/5 * 100 = 80
  x <- c("a", "b", "c", "d", NA)
  result <- audit_transform(x, toupper)

  expect_equal(result$pct_changed, 80)
})

test_that("audit_transform [character] change examples limited to 10", {
  x <- paste0("item_", 1:20)
  result <- audit_transform(x, toupper)

  expect_true(nrow(result$change_examples) <= 10L)
})

test_that("audit_transform errors when clean_fn returns wrong length", {
  expect_error(
    audit_transform(c("a", "b", "c"), function(x) x[1]),
    "same length"
  )
})

test_that("audit_transform errors when clean_fn returns longer vector", {
  expect_error(
    audit_transform(c("a", "b"), function(x) rep(x, 2)),
    "same length"
  )
})

# ---------------------------------------------------------------------------
# Numeric
# ---------------------------------------------------------------------------

test_that("audit_transform [numeric] detects type and computes diagnostics", {
  x <- c(1, 2, NA, 4)
  result <- audit_transform(x, function(v) v * 2)

  expect_s3_class(result, "audit_transform")
  expect_equal(result$type_class, "numeric")
  expect_equal(result$n_total, 4L)
  expect_equal(result$n_changed, 3L)   # 1→2, 2→4, 4→8 (NA stays NA)
  expect_equal(result$n_unchanged, 1L) # NA→NA counts as unchanged
  expect_equal(result$n_na_before, 1L)
  expect_equal(result$n_na_after, 1L)
  expect_equal(result$cleaned, c(2, 4, NA, 8))
  expect_false(is.null(result$diagnostics))
})

test_that("audit_transform [numeric] diagnostics fields are correct", {
  x <- c(1, 2, 3, 4)
  result <- audit_transform(x, function(v) v * 2)
  d <- result$diagnostics

  expect_equal(d$mean_before, 2.5)
  expect_equal(d$mean_after, 5.0)
  expect_equal(d$min_before, 1)
  expect_equal(d$max_before, 4)
  expect_equal(d$min_after, 2)
  expect_equal(d$max_after, 8)
  expect_equal(d$n_nan_before, 0L)
  expect_equal(d$n_inf_before, 0L)
  expect_true(!is.null(d$mean_abs_delta))
  expect_equal(d$mean_abs_delta, 2.5)  # avg delta = mean(1,2,3,4) = 2.5
})

test_that("audit_transform [numeric] n_beyond_tol uses tolerance", {
  x <- c(1.0, 2.0, 3.0)
  # Add a tiny delta below default tolerance to one element
  tiny <- sqrt(.Machine$double.eps) / 10
  result <- audit_transform(x, function(v) v + c(0, tiny, 1))

  d <- result$diagnostics
  expect_equal(d$n_beyond_tol, 1L)  # only the +1 exceeds tolerance
})

test_that("audit_transform [numeric] handles all-NA", {
  x <- c(NA_real_, NA_real_)
  result <- audit_transform(x, function(v) v + 1)

  expect_equal(result$n_changed, 0L)
  expect_equal(result$n_na_before, 2L)
  expect_true(is.na(result$diagnostics$mean_before))
})

test_that("audit_transform [numeric] zero changes", {
  x <- c(1.0, 2.0, 3.0)
  result <- audit_transform(x, function(v) v)

  expect_equal(result$n_changed, 0L)
  expect_equal(result$diagnostics$mean_abs_delta, 0)
})

test_that("print.audit_transform [numeric] mentions type and summary", {
  x <- c(1, 2, 3)
  result <- audit_transform(x, function(v) v * 2)
  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("numeric", combined))
  expect_true(grepl("Mean", combined))
})

# ---------------------------------------------------------------------------
# Date
# ---------------------------------------------------------------------------

test_that("audit_transform [Date] detects type and computes range diagnostics", {
  x <- as.Date(c("2020-01-01", "2020-06-15", "2020-12-31", NA))
  result <- audit_transform(x, function(d) d + 1L)

  expect_equal(result$type_class, "Date")
  expect_equal(result$n_changed, 3L)
  expect_equal(result$n_na_before, 1L)
  d <- result$diagnostics
  expect_false(is.null(d))
  expect_equal(d$min_before, as.Date("2020-01-01"))
  expect_equal(d$max_before, as.Date("2020-12-31"))
  expect_equal(d$range_days_before, 365)
})

test_that("print.audit_transform [Date] mentions type and date range", {
  x <- as.Date(c("2020-01-01", "2020-12-31"))
  result <- audit_transform(x, function(d) d + 10L)
  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Date", combined))
  expect_true(grepl("Span", combined))
})

# ---------------------------------------------------------------------------
# Factor
# ---------------------------------------------------------------------------

test_that("audit_transform [factor] detects level changes", {
  f <- factor(c("a", "b", "b", "c", "c"))
  # Rename level "c" → "d": removes "c", adds "d", same length
  result <- audit_transform(f, function(v) {
    levels(v)[levels(v) == "c"] <- "d"
    v
  })

  expect_equal(result$type_class, "factor")
  d <- result$diagnostics
  expect_true("c" %in% d$levels_removed)
  expect_true("d" %in% d$levels_added)
})

test_that("audit_transform [factor] level_counts have correct structure", {
  f <- factor(c("a", "b", "a"))
  result <- audit_transform(f, function(v) factor(toupper(as.character(v))))
  d <- result$diagnostics

  expect_true(is.data.frame(d$level_counts_before))
  expect_true(is.data.frame(d$level_counts_after))
  expect_true("Level" %in% names(d$level_counts_before))
})

test_that("audit_transform [factor] handles NA values in input", {
  # Regression: under R-devel (r89994+), as.data.frame.table() rejects NA in
  # names, so the internal level_counts builder must avoid that path.
  f <- factor(c("a", "b", NA, "a", NA))
  result <- audit_transform(f, function(v) factor(toupper(as.character(v))))

  expect_equal(result$type_class, "factor")
  expect_equal(result$n_na_before, 2L)

  cnt <- result$diagnostics$level_counts_before
  expect_true(is.data.frame(cnt))
  expect_true(any(is.na(cnt$Level)))
  expect_equal(cnt$N[is.na(cnt$Level)], 2L)

  # print() exercises the merge path on level counts with the NA bucket
  expect_no_error(capture.output(print(result), type = "message"))
})

test_that("print.audit_transform [factor] mentions type", {
  f <- factor(c("a", "b", "a"))
  result <- audit_transform(f, function(v) factor(toupper(as.character(v))))
  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("factor", combined))
})

# ---------------------------------------------------------------------------
# Logical
# ---------------------------------------------------------------------------

test_that("audit_transform [logical] flips TRUE/FALSE correctly", {
  x <- c(TRUE, FALSE, TRUE, NA, FALSE)
  result <- audit_transform(x, `!`)

  expect_equal(result$type_class, "logical")
  expect_equal(result$n_changed, 4L)  # 4 non-NA values all flip
  expect_equal(result$n_na_before, 1L)
  d <- result$diagnostics
  expect_equal(d$n_true_before,  2L)
  expect_equal(d$n_false_before, 2L)
  expect_equal(d$n_true_after,   2L)
  expect_equal(d$n_false_after,  2L)
})

test_that("print.audit_transform [logical] shows balance table", {
  x <- c(TRUE, FALSE, TRUE)
  result <- audit_transform(x, `!`)
  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("logical", combined))
  expect_true(grepl("TRUE", combined))
  expect_true(grepl("FALSE", combined))
})

# ---------------------------------------------------------------------------
# Pre-computed vector
# ---------------------------------------------------------------------------

test_that("audit_transform accepts pre-computed vector", {
  x <- c(1, 2, NA, 4)
  cleaned <- c(2, 4, NA, 8)
  result <- audit_transform(x, cleaned)

  expect_equal(result$clean_fn_name, "<pre-computed>")
  expect_equal(result$cleaned, cleaned)
  expect_equal(result$n_changed, 3L)
})

test_that("audit_transform errors on pre-computed vector of wrong length", {
  expect_error(
    audit_transform(c(1, 2, 3), c(1, 2)),
    "same length"
  )
})

# ---------------------------------------------------------------------------
# Edge cases — empty vector
# ---------------------------------------------------------------------------

test_that("audit_transform handles empty character vector", {
  result <- audit_transform(character(0), toupper)

  expect_equal(result$n_total, 0L)
  expect_equal(result$n_changed, 0L)
  expect_equal(result$pct_changed, 0)
  expect_equal(nrow(result$change_examples), 0L)
})

test_that("audit_transform handles empty numeric vector", {
  result <- audit_transform(numeric(0), function(v) v * 2)

  expect_equal(result$type_class, "numeric")
  expect_equal(result$n_total, 0L)
  expect_equal(result$pct_changed, 0)
})

# ---------------------------------------------------------------------------
# Integer input → classified as "numeric"
# ---------------------------------------------------------------------------

test_that("audit_transform classifies integer as numeric", {
  x <- 1L:5L
  result <- audit_transform(x, rev)

  expect_equal(result$type_class, "numeric")
  expect_false(is.null(result$diagnostics))
  expect_equal(result$diagnostics$mean_before, 3)
})

# ---------------------------------------------------------------------------
# POSIXct
# ---------------------------------------------------------------------------

test_that("audit_transform [POSIXct] detects type and range diagnostics", {
  x <- as.POSIXct(c("2020-01-01 00:00:00", "2020-06-15 12:00:00",
                     "2020-12-31 23:59:59"), tz = "UTC")
  result <- audit_transform(x, function(v) v + 3600)  # add 1 hour

  expect_equal(result$type_class, "POSIXct")
  expect_equal(result$n_changed, 3L)
  d <- result$diagnostics
  expect_false(is.null(d))
  expect_true(!is.na(d$range_days_before))
  expect_true(!is.na(d$range_days_after))
})

# ---------------------------------------------------------------------------
# Numeric: NaN and Inf
# ---------------------------------------------------------------------------

test_that("audit_transform [numeric] counts NaN and Inf correctly", {
  x <- c(1, NaN, Inf, -Inf, NA)
  result <- audit_transform(x, function(v) v * 2)
  d <- result$diagnostics

  expect_equal(d$n_nan_before, 1L)
  expect_equal(d$n_inf_before, 2L)  # Inf and -Inf
  expect_equal(d$n_nan_after, 1L)
  expect_equal(d$n_inf_after, 2L)
})

# ---------------------------------------------------------------------------
# Factor: clean_fn returns character (not factor)
# ---------------------------------------------------------------------------

test_that("audit_transform [factor] handles clean_fn returning character", {
  f <- factor(c("a", "b", "c"))
  result <- audit_transform(f, function(v) as.character(v))

  expect_equal(result$type_class, "factor")
  d <- result$diagnostics
  # levels_after derived from unique values of character output
  expect_true(is.character(d$levels_after))
  expect_equal(sort(d$levels_after), c("a", "b", "c"))
})

# ---------------------------------------------------------------------------
# Transform introduces new NAs (value → NA)
# ---------------------------------------------------------------------------

test_that("audit_transform detects value-to-NA transitions as changed", {
  x <- c(1, 2, 3)
  result <- audit_transform(x, function(v) ifelse(v == 2, NA_real_, v))

  expect_equal(result$n_changed, 1L)
  expect_equal(result$n_na_before, 0L)
  expect_equal(result$n_na_after, 1L)
  # pct_changed = 1/3 * 100
  expect_equal(result$pct_changed, 100 / 3)
})

test_that("audit_transform detects NA-to-value transitions as changed", {
  x <- c(NA_real_, 2, 3)
  result <- audit_transform(x, function(v) ifelse(is.na(v), 99, v))

  expect_equal(result$n_changed, 1L)
  expect_equal(result$n_na_before, 1L)
  expect_equal(result$n_na_after, 0L)
})

# ---------------------------------------------------------------------------
# Custom .tolerance
# ---------------------------------------------------------------------------

test_that("audit_transform [numeric] respects custom .tolerance", {
  x <- c(1.0, 2.0, 3.0)
  # Delta of 0.5 exceeds tight tolerance of 0.1 but not loose tolerance of 1.0
  result_tight <- audit_transform(x, function(v) v + 0.5, .tolerance = 0.1)
  result_loose <- audit_transform(x, function(v) v + 0.5, .tolerance = 1.0)

  expect_equal(result_tight$diagnostics$n_beyond_tol, 3L)
  expect_equal(result_loose$diagnostics$n_beyond_tol, 0L)
})

#' Compare Two Audit Trail Snapshots
#'
#' Computes detailed differences between any two snapshots in an audit trail,
#' including row/column/NA deltas, columns added/removed, type changes,
#' per-column NA changes, and numeric distribution shifts.
#'
#' @param .trail An [audit_trail()] object.
#' @param from Label (character) or index (integer) of the first snapshot.
#' @param to Label (character) or index (integer) of the second snapshot.
#'
#' @returns An `audit_diff` object (S3 list).
#'
#' @examples
#' trail <- audit_trail("example")
#' mtcars |>
#'   audit_tap(trail, "raw") |>
#'   dplyr::filter(mpg > 20) |>
#'   audit_tap(trail, "filtered")
#' audit_diff(trail, "raw", "filtered")
#'
#' @family audit trail
#' @export
audit_diff <- function(.trail, from, to) {
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  snap_from <- .resolve_snapshot(.trail, from)
  snap_to   <- .resolve_snapshot(.trail, to)

  diff_obj <- list(
    from_label     = snap_from$label,
    to_label       = snap_to$label,
    from_nrow      = snap_from$nrow,
    to_nrow        = snap_to$nrow,
    from_ncol      = snap_from$ncol,
    to_ncol        = snap_to$ncol,
    from_nas       = snap_from$total_nas,
    to_nas         = snap_to$total_nas,
    row_delta      = snap_to$nrow - snap_from$nrow,
    col_delta      = snap_to$ncol - snap_from$ncol,
    na_delta       = snap_to$total_nas - snap_from$total_nas,
    cols_added     = setdiff(snap_to$all_columns, snap_from$all_columns),
    cols_removed   = setdiff(snap_from$all_columns, snap_to$all_columns),
    type_changes   = .detect_type_changes(snap_from$schema, snap_to$schema),
    na_changes     = .compare_na_counts(snap_from$schema, snap_to$schema),
    numeric_shifts = .compare_numeric_summaries(snap_from, snap_to)
  )
  structure(diff_obj, class = c("audit_diff", "list"))
}

#' Resolve a Snapshot by Label or Index
#'
#' @param .trail An audit trail.
#' @param ref Character label or integer index.
#'
#' @returns The matching `audit_snap` object.
#'
#' @noRd
.resolve_snapshot <- function(.trail, ref) {
  if (is.character(ref)) {
    idx <- match(ref, .trail$labels)
    if (is.na(idx)) {
      trail_name <- .trail$name
      cli::cli_abort("Label {.val {ref}} not found in trail {.val {trail_name}}.")
    }
    return(.trail$snapshots[[idx]])
  }

  if (is.numeric(ref)) {
    if (length(ref) != 1L || is.na(ref) || !is.finite(ref) || ref != trunc(ref)) {
      cli::cli_abort("{.arg from}/{.arg to} must be a whole number or character label, not {.val {ref}}.")
    }
    ref <- as.integer(ref)
    if (ref < 1L || ref > length(.trail$snapshots)) {
      n <- length(.trail$snapshots)
      cli::cli_abort("Index {.val {ref}} out of range (trail has {n} snapshots).")
    }
    return(.trail$snapshots[[ref]])
  }

  cli::cli_abort("{.arg from}/{.arg to} must be a character label or integer index.")
}

#' Compare NA Counts Between Two Snapshots
#'
#' @param from_schema Schema data.frame from the source snapshot.
#' @param to_schema Schema data.frame from the target snapshot.
#'
#' @returns A data.frame of per-column NA changes for common columns, or `NULL`.
#'
#' @noRd
.compare_na_counts <- function(from_schema, to_schema) {
  common <- intersect(from_schema$column, to_schema$column)
  if (length(common) == 0L) return(NULL)

  from_nas <- setNames(from_schema$n_na, from_schema$column)[common]
  to_nas   <- setNames(to_schema$n_na, to_schema$column)[common]
  delta    <- to_nas - from_nas

  changed <- common[delta != 0L]
  if (length(changed) == 0L) return(NULL)

  data.frame(
    column  = changed,
    from_na = from_nas[changed],
    to_na   = to_nas[changed],
    delta   = delta[changed],
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' Compare Numeric Summaries Between Two Snapshots
#'
#' @param snap_from Source snapshot.
#' @param snap_to Target snapshot.
#'
#' @returns A data.frame of numeric distribution shifts, or `NULL`.
#'
#' @noRd
.compare_numeric_summaries <- function(snap_from, snap_to) {
  ns_from <- snap_from$numeric_summary
  ns_to   <- snap_to$numeric_summary
  if (is.null(ns_from) || is.null(ns_to)) return(NULL)

  common <- intersect(ns_from$column, ns_to$column)
  if (length(common) == 0L) return(NULL)

  from_sub <- ns_from[ns_from$column %in% common, , drop = FALSE]
  to_sub   <- ns_to[ns_to$column %in% common, , drop = FALSE]

  # Align by column order
  from_sub <- from_sub[match(common, from_sub$column), , drop = FALSE]
  to_sub   <- to_sub[match(common, to_sub$column), , drop = FALSE]

  data.frame(
    column      = common,
    mean_before = from_sub$mean,
    mean_after  = to_sub$mean,
    mean_shift  = to_sub$mean - from_sub$mean,
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' @rdname audit_diff
#' @param x An `audit_diff` object to print.
#' @param ... Additional arguments (currently unused).
#' @export
print.audit_diff <- function(x, ...) {
  cli::cli_h2("Audit Diff: {.val {x$from_label}} \u2192 {.val {x$to_label}}")

  # Dimension changes as aligned table
  delta_tbl <- data.frame(
    Metric = c("Rows", "Cols", "NAs"),
    Before = c(format(x$from_nrow, big.mark = ",", trim = TRUE),
               as.character(x$from_ncol),
               format(x$from_nas, big.mark = ",", trim = TRUE)),
    After  = c(format(x$to_nrow, big.mark = ",", trim = TRUE),
               as.character(x$to_ncol),
               format(x$to_nas, big.mark = ",", trim = TRUE)),
    Delta  = c(.format_delta(x$row_delta), .format_delta(x$col_delta), .format_delta(x$na_delta)),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )
  .cli_table(delta_tbl, right_align = c("Before", "After", "Delta"))

  # Columns added/removed
  cli::cli_text("")
  if (length(x$cols_added) > 0L) {
    cli::cli_alert_success("Columns added: {.field {x$cols_added}}")
  }
  if (length(x$cols_removed) > 0L) {
    cli::cli_alert_danger("Columns removed: {.field {x$cols_removed}}")
  }
  if (length(x$cols_added) == 0L && length(x$cols_removed) == 0L) {
    cli::cli_alert_info("No columns added or removed")
  }

  # Type changes
  if (!is.null(x$type_changes)) {
    cli::cli_text("")
    cli::cli_text("{.strong Type changes:}")
    bullets <- setNames(
      vapply(seq_len(nrow(x$type_changes)), function(i) {
        tc <- x$type_changes[i, ]
        glue::glue("{tc$column}: {tc$from} \u2192 {tc$to}")
      }, character(1)),
      rep("*", nrow(x$type_changes))
    )
    cli::cli_bullets(bullets)
  }

  # NA changes
  if (!is.null(x$na_changes)) {
    cli::cli_text("")
    cli::cli_text("{.strong NA changes:}")
    bullets <- setNames(
      vapply(seq_len(nrow(x$na_changes)), function(i) {
        nc <- x$na_changes[i, ]
        delta_str <- .format_delta(nc$delta)
        paste0(nc$column, ": ", nc$from_na, " \u2192 ", nc$to_na, " (", delta_str, ")")
      }, character(1)),
      rep("*", nrow(x$na_changes))
    )
    cli::cli_bullets(bullets)
  }

  # Numeric shifts
  if (!is.null(x$numeric_shifts)) {
    cli::cli_text("")
    cli::cli_text("{.strong Numeric shifts (common columns):}")

    ns <- x$numeric_shifts
    shift_tbl <- data.frame(
      Column        = ns$column,
      `Mean before` = format(round(ns$mean_before, 2), big.mark = ",", trim = TRUE),
      `Mean after`  = format(round(ns$mean_after, 2), big.mark = ",", trim = TRUE),
      Shift         = .format_delta_numeric(ns$mean_shift),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
    .cli_table(shift_tbl, right_align = c("Mean before", "Mean after", "Shift"), indent = 4L)
  }

  invisible(x)
}

#' Format a numeric (double) delta with sign prefix
#' @noRd
.format_delta_numeric <- function(shifts) {
  vapply(shifts, function(d) {
    if (is.na(d)) return("NA")
    formatted <- format(round(d, 2), big.mark = ",", trim = TRUE)
    if (d > 0) paste0("+", formatted) else formatted
  }, character(1))
}

#' Generate an Audit Report
#'
#' Prints a full audit report for a trail, including the trail summary, all
#' diffs between consecutive snapshots, custom diagnostic results, and a final
#' data profile.
#'
#' @param .trail An [audit_trail()] object.
#' @param format Report format. Currently only `"console"` is supported.
#'
#' @returns `.trail`, invisibly.
#'
#' @examples
#' trail <- audit_trail("example")
#' mtcars |>
#'   audit_tap(trail, "raw") |>
#'   dplyr::filter(mpg > 20) |>
#'   audit_tap(trail, "filtered")
#' audit_report(trail)
#'
#' @family audit trail
#' @export
audit_report <- function(.trail, format = "console") {
  format <- match.arg(format)

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  .print_full_report(.trail)
  invisible(.trail)
}

#' Print Full Console Report
#'
#' @param .trail An audit trail.
#'
#' @noRd
.print_full_report <- function(.trail) {
  n_snaps <- length(.trail$snapshots)

  trail_name <- .trail$name
  cli::cli_rule(left = "Audit Report: {.val {trail_name}}")
  cli::cli_text("Created: {format(.trail$created_at, '%Y-%m-%d %H:%M:%S')}")
  cli::cli_text("Total snapshots: {n_snaps}")

  if (n_snaps == 0L) {
    cli::cli_text("")
    cli::cli_alert_info("No snapshots recorded.")
    cli::cli_rule()
    return(invisible(NULL))
  }

  # Trail summary (reuse print method)
  cli::cli_text("")
  print(.trail)

  # Detailed diffs between consecutive snapshots
  if (n_snaps > 1L) {
    cli::cli_text("")
    cli::cli_rule(left = "Detailed Diffs")

    for (i in seq(2L, n_snaps)) {
      diff_obj <- audit_diff(.trail, i - 1L, i)
      cli::cli_text("")
      print(diff_obj)
    }
  }

  # Snapshot controls (when non-default)
  has_controls <- any(vapply(.trail$snapshots, function(s) !is.null(s$controls), logical(1)))
  if (has_controls) {
    cli::cli_text("")
    cli::cli_rule(left = "Snapshot Controls")

    for (snap in .trail$snapshots) {
      if (!is.null(snap$controls)) {
        cli::cli_text("")
        cli::cli_text("{.strong {snap$label}:}")
        ctrl <- snap$controls
        if (!isTRUE(ctrl$numeric_summary)) {
          cli::cli_text("  numeric_summary: {.val {ctrl$numeric_summary}}")
        }
        if (!is.null(ctrl$cols_include)) {
          cli::cli_text("  cols_include: {.val {ctrl$cols_include}}")
        }
        if (!is.null(ctrl$cols_exclude)) {
          cli::cli_text("  cols_exclude: {.val {ctrl$cols_exclude}}")
        }
      }
    }
  }

  # Custom diagnostics
  has_custom <- any(vapply(.trail$snapshots, function(s) !is.null(s$custom), logical(1)))
  if (has_custom) {
    cli::cli_text("")
    cli::cli_rule(left = "Custom Diagnostics")

    for (snap in .trail$snapshots) {
      if (!is.null(snap$custom)) {
        cli::cli_text("")
        cli::cli_text("{.strong {snap$label}:}")
        for (nm in names(snap$custom)) {
          val <- snap$custom[[nm]]
          cli::cli_text("  {nm}: {format(val)}")
        }
      }
    }
  }

  # Final snapshot profile
  cli::cli_text("")
  cli::cli_rule(left = "Final Snapshot Profile")
  last_snap <- .trail$snapshots[[n_snaps]]
  cli::cli_text("")
  cli::cli_text("{.strong {last_snap$label}} ({format(last_snap$nrow, big.mark = ',')} rows x {last_snap$ncol} cols)")

  # Column types summary
  schema <- last_snap$schema
  type_counts <- table(schema$type)
  type_str <- paste(
    vapply(names(type_counts), function(t) glue::glue("{type_counts[t]} {t}"), character(1)),
    collapse = ", "
  )
  cli::cli_text("Column types: {type_str}")

  # NA summary
  if (last_snap$total_nas > 0L) {
    na_cols <- schema[schema$n_na > 0L, , drop = FALSE]
    pct <- round(100 * na_cols$n_na / last_snap$nrow, 1)
    na_tbl <- data.frame(
      Column = na_cols$column,
      NAs    = format(na_cols$n_na, big.mark = ",", trim = TRUE),
      `%`    = paste0(format(pct, nsmall = 1, trim = TRUE), "%"),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
    cli::cli_text("{.strong Columns with NAs ({nrow(na_cols)}):}")
    .cli_table(na_tbl, right_align = c("NAs", "%"), indent = 4L)
  } else {
    cli::cli_alert_success("No missing values")
  }

  # Numeric summary
  if (!is.null(last_snap$numeric_summary)) {
    cli::cli_text("")
    cli::cli_text("{.strong Numeric summary:}")
    ns <- last_snap$numeric_summary
    num_tbl <- data.frame(
      Column = ns$column,
      Min    = format(round(ns$min, 2), big.mark = ",", trim = TRUE),
      Mean   = format(round(ns$mean, 2), big.mark = ",", trim = TRUE),
      Median = format(round(ns$median, 2), big.mark = ",", trim = TRUE),
      Max    = format(round(ns$max, 2), big.mark = ",", trim = TRUE),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
    .cli_table(num_tbl, right_align = c("Min", "Mean", "Median", "Max"), indent = 4L)
  }

  cli::cli_text("")
  cli::cli_rule()
}

# ── Helpers: serialisation ──────────────────────────────────────────────────

#' @noRd
.fmt_posixct <- function(x) {
  if (is.null(x) || (length(x) == 1L && is.na(x))) return(NA_character_)
  format(x, "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
}

#' @noRd
.parse_posixct <- function(x) {
  if (is.null(x) || (length(x) == 1L && is.na(x))) {
    return(as.POSIXct(NA_real_, origin = "1970-01-01", tz = "UTC"))
  }
  as.POSIXct(x, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
}

#' Convert a data.frame to a list of named rows
#' @noRd
.df_to_rows <- function(df) {
  if (is.null(df)) return(NULL)
  lapply(seq_len(nrow(df)), function(i) as.list(df[i, , drop = FALSE]))
}

#' Reconstruct a data.frame from a list of named rows
#'
#' @param rows   List of named lists (one per row). NULL or length-0 returns NULL.
#' @param cols   Character vector of expected column names.
#' @param int_cols Character vector of columns to coerce to integer.
#' @param dbl_cols Character vector of columns to coerce to double.
#' @noRd
.rows_to_df <- function(rows, cols, int_cols = character(), dbl_cols = character()) {
  if (is.null(rows) || length(rows) == 0L) return(NULL)
  out <- data.frame(
    lapply(cols, function(col) {
      vals <- lapply(rows, function(r) {
        v <- r[[col]]
        if (is.null(v)) NA else v
      })
      vals <- unlist(vals)
      if (col %in% int_cols) as.integer(vals)
      else if (col %in% dbl_cols) as.double(vals)
      else vals
    }),
    stringsAsFactors = FALSE
  )
  names(out) <- cols
  out
}

#' @noRd
.changes_to_list <- function(ch) {
  if (is.null(ch)) return(NULL)
  ch$type_changes <- .df_to_rows(ch$type_changes)
  ch
}

#' @noRd
.list_to_changes <- function(ch) {
  if (is.null(ch)) return(NULL)
  # Coerce integer fields that jsonlite inflates to double
  ch$row_delta <- as.integer(ch$row_delta)
  ch$col_delta <- as.integer(ch$col_delta)
  ch$na_delta  <- as.integer(ch$na_delta)
  # Coerce empty lists back to character(0)
  ch$cols_added   <- unlist(ch$cols_added)   %||% character(0)
  ch$cols_removed <- unlist(ch$cols_removed) %||% character(0)
  # Reconstruct type_changes as data.frame
  if (!is.null(ch$type_changes) && is.list(ch$type_changes) &&
      !is.data.frame(ch$type_changes)) {
    ch$type_changes <- .rows_to_df(
      ch$type_changes,
      cols = c("column", "from", "to")
    )
  }
  ch
}


# ── Helpers: RDS form ────────────────────────────────────────────────────────

#' @noRd
.trail_to_rds_form <- function(.trail) {
  list(
    name      = .trail$name,
    created_at = .trail$created_at,
    labels    = .trail$labels,
    snapshots = lapply(.trail$snapshots, unclass)
  )
}

#' @noRd
.rds_form_to_trail <- function(lst) {
  trail <- audit_trail(lst$name)
  trail$created_at <- lst$created_at
  trail$labels     <- lst$labels
  trail$snapshots  <- lapply(lst$snapshots, function(snap) {
    structure(snap, class = c("audit_snap", "list"))
  })
  trail
}


# ── Helpers: JSON form ────────────────────────────────────────────────────────

#' @noRd
.detect_format <- function(file) {
  ext <- tolower(tools::file_ext(file))
  if (ext == "rds")  return("rds")
  if (ext == "json") return("json")
  cli::cli_abort(
    "Cannot infer file format from extension {.val {ext}}. Supply {.arg format} explicitly."
  )
}

#' @noRd
.list_to_trail <- function(lst) {
  trail <- audit_trail(lst$name)
  trail$created_at <- .parse_posixct(lst$created_at)

  snap_labels <- names(lst$snapshots) %||% character(0)
  trail$labels <- snap_labels

  trail$snapshots <- lapply(seq_along(lst$snapshots), function(i) {
    s <- lst$snapshots[[i]]

    schema <- .rows_to_df(
      s$schema,
      cols     = c("column", "type", "n_na"),
      int_cols = "n_na"
    ) %||% data.frame(column = character(), type = character(),
                      n_na = integer(), stringsAsFactors = FALSE)

    numeric_summary <- .rows_to_df(
      s$numeric_summary,
      cols     = c("column", "min", "q25", "median", "mean", "q75", "max"),
      dbl_cols = c("min", "q25", "median", "mean", "q75", "max")
    )

    snap <- list(
      label           = s$label,
      index           = as.integer(s$index),
      timestamp       = .parse_posixct(s$timestamp),
      type            = s$type,
      nrow            = as.integer(s$nrow),
      ncol            = as.integer(s$ncol),
      all_columns     = unlist(s$all_columns) %||% character(0),
      schema          = schema,
      total_nas       = as.integer(s$total_nas),
      numeric_summary = numeric_summary,
      diagnostics     = s$diagnostics,
      pipeline        = unlist(s$pipeline),
      changes         = .list_to_changes(s$changes),
      custom          = s$custom,
      controls        = s$controls
    )
    structure(snap, class = c("audit_snap", "list"))
  })

  trail
}


# ── Exported: trail_to_list ──────────────────────────────────────────────────

#' Convert an Audit Trail to a Plain List
#'
#' Converts an [audit_trail()] object to a plain R list suitable for
#' serialisation with [jsonlite::toJSON()]. All POSIXct timestamps are
#' converted to ISO 8601 character strings and data.frames are converted
#' to lists of named rows for JSON compatibility.
#'
#' @param .trail An [audit_trail()] object.
#'
#' @returns A named list with elements `name`, `created_at` (ISO 8601 string),
#'   `n_snapshots`, and `snapshots` (a named list keyed by snapshot label).
#'
#' @examples
#' trail <- audit_trail("example")
#' mtcars |> audit_tap(trail, "raw")
#' lst <- trail_to_list(trail)
#' str(lst, max.level = 2)
#'
#' @family audit export
#' @export
trail_to_list <- function(.trail) {
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  snap_list <- lapply(.trail$snapshots, function(snap) {
    list(
      label           = snap$label,
      index           = snap$index,
      timestamp       = .fmt_posixct(snap$timestamp),
      type            = snap$type,
      nrow            = snap$nrow,
      ncol            = snap$ncol,
      all_columns     = snap$all_columns,
      total_nas       = snap$total_nas,
      schema          = .df_to_rows(snap$schema),
      numeric_summary = .df_to_rows(snap$numeric_summary),
      diagnostics     = snap$diagnostics,
      pipeline        = snap$pipeline,
      changes         = .changes_to_list(snap$changes),
      custom          = snap$custom,
      controls        = snap$controls
    )
  })

  if (length(snap_list) > 0L) {
    names(snap_list) <- .trail$labels
  }

  list(
    name        = .trail$name,
    created_at  = .fmt_posixct(.trail$created_at),
    n_snapshots = length(.trail$snapshots),
    snapshots   = snap_list
  )
}


# ── Exported: trail_to_df ─────────────────────────────────────────────────────

#' Convert an Audit Trail to a Data Frame
#'
#' Returns a plain `data.frame` with one row per snapshot. Nested fields
#' (`all_columns`, `schema`, `numeric_summary`, `changes`, `diagnostics`,
#' `custom`, `pipeline`, `controls`) become list-columns. Trail metadata is
#' stored as attributes on the result.
#'
#' @param .trail An [audit_trail()] object.
#'
#' @returns A `data.frame` with columns `index`, `label`, `type`, `timestamp`,
#'   `nrow`, `ncol`, `total_nas`, `all_columns`, `schema`, `numeric_summary`,
#'   `changes`, `diagnostics`, `custom`, `pipeline`, and `controls`. Trail
#'   `name` and `created_at` are stored as attributes `"trail_name"` and
#'   `"created_at"`.
#'
#' @examples
#' trail <- audit_trail("example")
#' mtcars |> audit_tap(trail, "raw")
#' dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")
#' df <- trail_to_df(trail)
#' print(df)
#' attr(df, "trail_name")
#'
#' @family audit export
#' @export
trail_to_df <- function(.trail) {
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  n <- length(.trail$snapshots)

  if (n == 0L) {
    result <- data.frame(
      index           = integer(),
      label           = character(),
      type            = character(),
      timestamp       = as.POSIXct(character()),
      nrow            = integer(),
      ncol            = integer(),
      total_nas       = integer(),
      all_columns     = I(list()),
      schema          = I(list()),
      numeric_summary = I(list()),
      changes         = I(list()),
      diagnostics     = I(list()),
      custom          = I(list()),
      pipeline        = I(list()),
      controls        = I(list()),
      stringsAsFactors = FALSE
    )
  } else {
    ts_vec <- do.call(c, lapply(.trail$snapshots, `[[`, "timestamp"))

    result <- data.frame(
      index           = vapply(.trail$snapshots, `[[`, integer(1L), "index"),
      label           = vapply(.trail$snapshots, `[[`, character(1L), "label"),
      type            = vapply(.trail$snapshots, `[[`, character(1L), "type"),
      timestamp       = ts_vec,
      nrow            = vapply(.trail$snapshots, `[[`, integer(1L), "nrow"),
      ncol            = vapply(.trail$snapshots, `[[`, integer(1L), "ncol"),
      total_nas       = vapply(.trail$snapshots, `[[`, integer(1L), "total_nas"),
      all_columns     = I(lapply(.trail$snapshots, `[[`, "all_columns")),
      schema          = I(lapply(.trail$snapshots, `[[`, "schema")),
      numeric_summary = I(lapply(.trail$snapshots, `[[`, "numeric_summary")),
      changes         = I(lapply(.trail$snapshots, `[[`, "changes")),
      diagnostics     = I(lapply(.trail$snapshots, `[[`, "diagnostics")),
      custom          = I(lapply(.trail$snapshots, `[[`, "custom")),
      pipeline        = I(lapply(.trail$snapshots, `[[`, "pipeline")),
      controls        = I(lapply(.trail$snapshots, `[[`, "controls")),
      stringsAsFactors = FALSE
    )
  }

  attr(result, "trail_name") <- .trail$name
  attr(result, "created_at") <- .trail$created_at
  result
}


# ── Exported: write_trail ─────────────────────────────────────────────────────

#' Write an Audit Trail to a File
#'
#' Saves an [audit_trail()] to disk as either an RDS file (default) or a
#' JSON file. The RDS format preserves all R types and can be restored
#' perfectly with [read_trail()]. The JSON format produces a human-readable
#' representation suitable for archiving or interoperability with other tools.
#'
#' @param .trail An [audit_trail()] object.
#' @param file   Path to the output file. A `.rds` extension is conventional
#'   for `format = "rds"`; `.json` for `format = "json"`.
#' @param format One of `"rds"` (default) or `"json"`. The JSON format
#'   requires the \pkg{jsonlite} package to be installed.
#'
#' @returns `.trail`, invisibly.
#'
#' @note Custom diagnostic results (the `custom` field, populated via `.fns`
#'   in [audit_tap()]) are serialised on a best-effort basis for JSON output.
#'   Complex R objects such as environments or functions cannot be represented
#'   in JSON and will cause an error.
#'
#' @examples
#' trail <- audit_trail("example")
#' mtcars |> audit_tap(trail, "raw")
#' tmp <- tempfile(fileext = ".rds")
#' write_trail(trail, tmp)
#' restored <- read_trail(tmp)
#'
#' @family audit export
#' @seealso [read_trail()], [trail_to_list()]
#' @export
write_trail <- function(.trail, file, format = c("rds", "json")) {
  format <- match.arg(format)

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }
  if (!is.character(file) || length(file) != 1L || is.na(file)) {
    cli::cli_abort("{.arg file} must be a single non-missing character string.")
  }

  if (format == "rds") {
    saveRDS(.trail_to_rds_form(.trail), file = file)
  } else {
    pkg <- "jsonlite"
    if (!requireNamespace(pkg, quietly = TRUE)) {
      cli::cli_abort(
        c(
          "Package {.pkg {pkg}} is required for JSON export.",
          "i" = "Install it with {.code install.packages(\"{pkg}\")}."
        )
      )
    }
    lst      <- trail_to_list(.trail)
    json_txt <- jsonlite::toJSON(lst, auto_unbox = TRUE, pretty = TRUE, null = "null")
    writeLines(as.character(json_txt), con = file)
  }

  invisible(.trail)
}


# ── Exported: read_trail ─────────────────────────────────────────────────────

#' Read an Audit Trail from a File
#'
#' Restores an [audit_trail()] previously saved with [write_trail()]. The
#' file format is detected automatically from the file extension (`.rds` for
#' RDS, `.json` for JSON), or can be specified explicitly via `format`.
#'
#' @param file   Path to an RDS or JSON file created by [write_trail()].
#' @param format One of `"rds"`, `"json"`, or `NULL` (default). When `NULL`,
#'   the format is inferred from the file extension.
#'
#' @returns A reconstructed [audit_trail()] object with all S3 classes
#'   restored.
#'
#' @examples
#' trail <- audit_trail("example")
#' mtcars |> audit_tap(trail, "raw")
#' tmp <- tempfile(fileext = ".rds")
#' write_trail(trail, tmp)
#' restored <- read_trail(tmp)
#' print(restored)
#'
#' @family audit export
#' @seealso [write_trail()]
#' @export
read_trail <- function(file, format = NULL) {
  if (!is.character(file) || length(file) != 1L || is.na(file)) {
    cli::cli_abort("{.arg file} must be a single non-missing character string.")
  }
  if (!file.exists(file)) {
    cli::cli_abort("File not found: {.path {file}}")
  }

  if (is.null(format)) {
    format <- .detect_format(file)
  } else {
    format <- match.arg(format, c("rds", "json"))
  }

  if (format == "rds") {
    lst <- readRDS(file)
    .rds_form_to_trail(lst)
  } else {
    pkg <- "jsonlite"
    if (!requireNamespace(pkg, quietly = TRUE)) {
      cli::cli_abort(
        c(
          "Package {.pkg {pkg}} is required for JSON import.",
          "i" = "Install it with {.code install.packages(\"{pkg}\")}."
        )
      )
    }
    lst <- jsonlite::fromJSON(file, simplifyVector = FALSE)
    .list_to_trail(lst)
  }
}


# ── Exported: audit_export ──────────────────────────────────────────────────

#' Export an Audit Trail as a Self-Contained HTML File
#'
#' Produces a standalone HTML file that visualises the audit trail as an
#' interactive pipeline flow diagram. The file is completely self-contained
#' — no server, internet connection, or R installation is required to view
#' it. Open it in any browser.
#'
#' The trail is serialised via [trail_to_list()] and embedded as JSON inside
#' an HTML template with inline CSS and vanilla JavaScript. The visualisation
#' features:
#' \itemize{
#'   \item Horizontal pipeline flow diagram with colour-coded nodes per
#'         operation type (snapshot, join, filter).
#'   \item Edges annotated with key deltas (match rate, drop \%, columns
#'         added).
#'   \item Clickable nodes expanding to show column schema, operation
#'         diagnostics, and custom \code{.fns} results.
#'   \item Clickable edges showing the full diff between adjacent snapshots.
#'   \item Light / dark theme toggle.
#'   \item Collapsible JSON export panel.
#' }
#'
#' @param .trail An [audit_trail()] object.
#' @param file   Path to the output \code{.html} file. If \code{NULL}
#'   (the default), writes to a temporary file and opens it in the default
#'   browser via [utils::browseURL()].
#'
#' @returns The file path (character), invisibly.
#'
#' @examples
#' \donttest{
#' trail <- audit_trail("demo")
#' mtcars |> audit_tap(trail, "raw")
#' dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")
#' audit_export(trail, tempfile(fileext = ".html"))
#' }
#'
#' @family audit export
#' @seealso [trail_to_list()], [write_trail()]
#' @export
audit_export <- function(.trail, file = NULL) {
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  pkg <- "jsonlite"
  if (!requireNamespace(pkg, quietly = TRUE)) {
    cli::cli_abort(
      c(
        "Package {.pkg {pkg}} is required for HTML export.",
        "i" = "Install it with {.code install.packages(\"{pkg}\")}."
      )
    )
  }

  if (!is.null(file)) {
    if (!is.character(file) || length(file) != 1L || is.na(file)) {
      cli::cli_abort("{.arg file} must be a single non-missing character string.")
    }
  }

  # Serialise trail to JSON
  lst      <- trail_to_list(.trail)
  json_txt <- jsonlite::toJSON(lst, auto_unbox = TRUE, pretty = FALSE,
                               null = "null")
  json_txt <- as.character(json_txt)

  # Prevent inline JSON from breaking the <script> tag
  json_txt <- gsub("</script>", "<\\/script>", json_txt, fixed = TRUE)

  # Read template
  template_path <- system.file("templates", "audit_trail.html",
                               package = "tidyaudit")
  if (template_path == "") {
    cli::cli_abort("HTML template not found. Is the package installed correctly?")
  }
  template <- paste(readLines(template_path, warn = FALSE), collapse = "\n")

  # Inject JSON into template
  html <- sub("/*__TRAIL_DATA__*/null", json_txt, template, fixed = TRUE)

  # Write output
  if (is.null(file)) {
    file <- tempfile(fileext = ".html")
    writeLines(html, con = file)
    if (interactive()) utils::browseURL(file)
  } else {
    writeLines(html, con = file)
  }

  invisible(file)
}

# Shared fixture: 3-snapshot trail with join, filter, and custom diagnostics
make_export_trail <- function() {
  trail   <- audit_trail("export_test")
  orders  <- data.frame(id = 1:4, amount = c(100, 200, 300, 400))
  lookup  <- data.frame(id = c(2L, 3L, 5L), name = c("A", "B", "C"),
                        stringsAsFactors = FALSE)
  orders |>
    audit_tap(trail, "raw") |>
    left_join_tap(lookup, by = "id", .trail = trail, .label = "joined") |>
    dplyr::filter(amount > 100) |>
    audit_tap(trail, "filtered", .fns = list(n = nrow))
  trail
}


# ── trail_to_list ─────────────────────────────────────────────────────────────

test_that("trail_to_list returns a plain list with correct top-level keys", {
  lst <- trail_to_list(make_export_trail())
  expect_type(lst, "list")
  expect_named(lst, c("name", "created_at", "n_snapshots", "snapshots"))
  expect_equal(lst$name, "export_test")
  expect_equal(lst$n_snapshots, 3L)
  expect_type(lst$created_at, "character")
  expect_true(grepl("T", lst$created_at))
})

test_that("trail_to_list snapshots are named by label", {
  lst <- trail_to_list(make_export_trail())
  expect_named(lst$snapshots, c("raw", "joined", "filtered"))
})

test_that("trail_to_list converts schema data.frame to list of rows", {
  lst <- trail_to_list(make_export_trail())
  schema <- lst$snapshots$raw$schema
  expect_type(schema, "list")
  expect_true(is.list(schema[[1]]))
  expect_true("column" %in% names(schema[[1]]))
  expect_true("type" %in% names(schema[[1]]))
  expect_true("n_na" %in% names(schema[[1]]))
})

test_that("trail_to_list converts POSIXct timestamps to ISO 8601 strings", {
  lst <- trail_to_list(make_export_trail())
  expect_type(lst$snapshots$raw$timestamp, "character")
  expect_true(grepl("T", lst$snapshots$raw$timestamp))
})

test_that("trail_to_list handles NULL numeric_summary (character-only data)", {
  trail <- audit_trail("no_num")
  data.frame(x = letters[1:3], stringsAsFactors = FALSE) |>
    audit_tap(trail, "text_only")
  lst <- trail_to_list(trail)
  expect_null(lst$snapshots$text_only$numeric_summary)
})

test_that("trail_to_list has NULL changes on the first snapshot", {
  lst <- trail_to_list(make_export_trail())
  expect_null(lst$snapshots$raw$changes)
})

test_that("trail_to_list converts type_changes in changes to list of rows", {
  trail <- audit_trail("type_change")
  data.frame(x = 1:3) |> audit_tap(trail, "before")
  data.frame(x = as.character(1:3), stringsAsFactors = FALSE) |>
    audit_tap(trail, "after")
  lst <- trail_to_list(trail)
  tc <- lst$snapshots$after$changes$type_changes
  expect_type(tc, "list")
  expect_true(is.list(tc[[1]]))
  expect_true("column" %in% names(tc[[1]]))
})

test_that("trail_to_list output is JSON-serialisable", {
  skip_if_not_installed("jsonlite")
  lst <- trail_to_list(make_export_trail())
  expect_no_error(jsonlite::toJSON(lst, auto_unbox = TRUE))
})

test_that("trail_to_list rejects non-trail input", {
  expect_error(trail_to_list(list()), "audit_trail")
  expect_error(trail_to_list("not a trail"), "audit_trail")
})

test_that("trail_to_list on empty trail returns empty snapshots list", {
  trail <- audit_trail("empty")
  lst   <- trail_to_list(trail)
  expect_equal(lst$n_snapshots, 0L)
  expect_equal(length(lst$snapshots), 0L)
})


# ── trail_to_df ───────────────────────────────────────────────────────────────

test_that("trail_to_df returns a data.frame", {
  df <- trail_to_df(make_export_trail())
  expect_s3_class(df, "data.frame")
  expect_false(inherits(df, "tbl_df"))
})

test_that("trail_to_df has one row per snapshot", {
  df <- trail_to_df(make_export_trail())
  expect_equal(nrow(df), 3L)
})

test_that("trail_to_df has the expected column names", {
  df <- trail_to_df(make_export_trail())
  expected <- c("index", "label", "type", "timestamp", "nrow", "ncol",
                "total_nas", "all_columns", "schema", "numeric_summary",
                "changes", "diagnostics", "custom", "pipeline", "controls")
  expect_named(df, expected)
})

test_that("trail_to_df scalar columns have correct types", {
  df <- trail_to_df(make_export_trail())
  expect_type(df$index, "integer")
  expect_type(df$label, "character")
  expect_type(df$type, "character")
  expect_s3_class(df$timestamp, "POSIXct")
  expect_type(df$nrow, "integer")
  expect_type(df$ncol, "integer")
  expect_type(df$total_nas, "integer")
})

test_that("trail_to_df schema is a list-column of data.frames", {
  df <- trail_to_df(make_export_trail())
  expect_type(df$schema, "list")
  expect_s3_class(df$schema[[1]], "data.frame")
})

test_that("trail_to_df changes is NULL for the first row", {
  df <- trail_to_df(make_export_trail())
  expect_null(df$changes[[1]])
})

test_that("trail_to_df attaches trail metadata as attributes", {
  df <- trail_to_df(make_export_trail())
  expect_equal(attr(df, "trail_name"), "export_test")
  expect_s3_class(attr(df, "created_at"), "POSIXct")
})

test_that("trail_to_df on empty trail returns zero-row data.frame with correct columns", {
  df <- trail_to_df(audit_trail("empty"))
  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 0L)
  expect_equal(ncol(df), 15L)
})

test_that("trail_to_df diagnostics is NULL for plain taps", {
  trail <- audit_trail("plain")
  mtcars |> audit_tap(trail, "step1")
  df <- trail_to_df(trail)
  expect_null(df$diagnostics[[1]])
})

test_that("trail_to_df rejects non-trail input", {
  expect_error(trail_to_df(list()), "audit_trail")
})


# ── write_trail / read_trail — RDS ───────────────────────────────────────────

test_that("write_trail (rds) creates a file", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp)
  expect_true(file.exists(tmp))
})

test_that("write_trail returns .trail invisibly", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  trail  <- make_export_trail()
  result <- withVisible(write_trail(trail, tmp))
  expect_false(result$visible)
  expect_identical(result$value, trail)
})

test_that("read_trail (rds) restores audit_trail class and structure", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp)
  restored <- read_trail(tmp)

  expect_s3_class(restored, "audit_trail")
  expect_s3_class(restored, "environment")
  expect_equal(restored$name, "export_test")
  expect_s3_class(restored$created_at, "POSIXct")
  expect_equal(length(restored$snapshots), 3L)
  expect_equal(restored$labels, c("raw", "joined", "filtered"))
})

test_that("read_trail (rds) restores audit_snap S3 class on every snapshot", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp)
  restored <- read_trail(tmp)
  for (snap in restored$snapshots) {
    expect_s3_class(snap, "audit_snap")
  }
})

test_that("read_trail (rds) preserves snapshot scalar fields", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp)
  restored <- read_trail(tmp)

  orig_raw <- original$snapshots[[1]]
  rest_raw <- restored$snapshots[[1]]
  expect_equal(rest_raw$label,     orig_raw$label)
  expect_equal(rest_raw$nrow,      orig_raw$nrow)
  expect_equal(rest_raw$ncol,      orig_raw$ncol)
  expect_equal(rest_raw$total_nas, orig_raw$total_nas)
  expect_equal(rest_raw$type,      orig_raw$type)
})

test_that("read_trail (rds) preserves schema data.frame", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp)
  restored <- read_trail(tmp)
  expect_equal(restored$snapshots[[1]]$schema, original$snapshots[[1]]$schema)
})

test_that("read_trail (rds) preserves changes list", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp)
  restored <- read_trail(tmp)

  expect_null(restored$snapshots[[1]]$changes)
  expect_equal(
    restored$snapshots[[2]]$changes$row_delta,
    original$snapshots[[2]]$changes$row_delta
  )
})

test_that("read_trail (rds) preserves custom diagnostics", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp)
  restored <- read_trail(tmp)
  # 3rd snapshot has .fns = list(n = nrow)
  expect_equal(restored$snapshots[[3]]$custom$n,
               original$snapshots[[3]]$custom$n)
})

test_that("read_trail auto-detects rds format from extension", {
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp)
  expect_s3_class(read_trail(tmp), "audit_trail")
})

test_that("read_trail with explicit format = 'rds' works for non-.rds extension", {
  tmp <- tempfile(fileext = ".dat")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "rds")
  restored <- read_trail(tmp, format = "rds")
  expect_s3_class(restored, "audit_trail")
  expect_equal(length(restored$snapshots), 3L)
})


# ── write_trail / read_trail — JSON ──────────────────────────────────────────

test_that("write_trail (json) creates a valid JSON file", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  expect_true(file.exists(tmp))
  expect_no_error(jsonlite::fromJSON(tmp))
})

test_that("read_trail (json) restores audit_trail class and snapshot count", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  restored <- read_trail(tmp)

  expect_s3_class(restored, "audit_trail")
  expect_equal(restored$name, "export_test")
  expect_equal(length(restored$snapshots), 3L)
  expect_equal(restored$labels, c("raw", "joined", "filtered"))
})

test_that("read_trail (json) restores audit_snap S3 class on every snapshot", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  restored <- read_trail(tmp)
  for (snap in restored$snapshots) {
    expect_s3_class(snap, "audit_snap")
  }
})

test_that("read_trail (json) restores scalar fields with correct types", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  restored <- read_trail(tmp)

  snap <- restored$snapshots[[1]]
  expect_type(snap$index, "integer")
  expect_type(snap$nrow, "integer")
  expect_type(snap$ncol, "integer")
  expect_type(snap$total_nas, "integer")
  expect_s3_class(snap$timestamp, "POSIXct")
})

test_that("read_trail (json) restores schema as data.frame with correct column names", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  restored <- read_trail(tmp)

  schema <- restored$snapshots[[1]]$schema
  expect_s3_class(schema, "data.frame")
  expect_named(schema, c("column", "type", "n_na"))
  expect_type(schema$n_na, "integer")
})

test_that("read_trail (json) restores type_changes as data.frame", {
  skip_if_not_installed("jsonlite")
  trail <- audit_trail("type_change_json")
  data.frame(x = 1:3) |> audit_tap(trail, "before")
  data.frame(x = as.character(1:3), stringsAsFactors = FALSE) |>
    audit_tap(trail, "after")

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(trail, tmp, format = "json")
  restored <- read_trail(tmp)

  tc <- restored$snapshots[[2]]$changes$type_changes
  expect_s3_class(tc, "data.frame")
  expect_named(tc, c("column", "from", "to"))
})

test_that("read_trail (json) restores NULL changes on first snapshot", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  restored <- read_trail(tmp)
  expect_null(restored$snapshots[[1]]$changes)
})

test_that("read_trail (json) restores changes integer deltas correctly", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp, format = "json")
  restored <- read_trail(tmp)

  orig_delta <- original$snapshots[[2]]$changes$row_delta
  rest_delta <- restored$snapshots[[2]]$changes$row_delta
  expect_equal(rest_delta, orig_delta)
  expect_type(rest_delta, "integer")
})

test_that("read_trail auto-detects json format from extension", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(make_export_trail(), tmp, format = "json")
  expect_s3_class(read_trail(tmp), "audit_trail")
})

test_that("read_trail (json) restores numeric_summary values as double", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp, format = "json")
  restored <- read_trail(tmp)

  orig_ns <- original$snapshots[[1]]$numeric_summary
  rest_ns <- restored$snapshots[[1]]$numeric_summary
  expect_s3_class(rest_ns, "data.frame")
  expect_equal(rest_ns$min,    orig_ns$min)
  expect_equal(rest_ns$median, orig_ns$median)
  expect_type(rest_ns$min,    "double")
  expect_type(rest_ns$median, "double")
  expect_type(rest_ns$max,    "double")
})

test_that("read_trail (json) restores custom diagnostics", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  original <- make_export_trail()
  write_trail(original, tmp, format = "json")
  restored <- read_trail(tmp)
  # 3rd snapshot has .fns = list(n = nrow)
  expect_equal(restored$snapshots[[3]]$custom$n,
               original$snapshots[[3]]$custom$n)
})

test_that(".parse_posixct returns a POSIXct-classed NA for NULL input", {
  result <- tidyaudit:::.parse_posixct(NULL)
  expect_s3_class(result, "POSIXct")
  expect_true(is.na(result))
})

test_that("read_trail (json) restores schema as 0-row data.frame for zero-column input", {
  skip_if_not_installed("jsonlite")
  trail <- audit_trail("zero_cols")
  data.frame(row.names = 1:3) |> audit_tap(trail, "empty")

  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(trail, tmp, format = "json")
  restored <- read_trail(tmp)

  schema <- restored$snapshots[[1]]$schema
  expect_s3_class(schema, "data.frame")
  expect_named(schema, c("column", "type", "n_na"))
  expect_equal(nrow(schema), 0L)
})


# ── Error handling ────────────────────────────────────────────────────────────

test_that("write_trail errors on non-trail input", {
  expect_error(write_trail(list(), tempfile()), "audit_trail")
})

test_that("write_trail errors on invalid file argument", {
  trail <- make_export_trail()
  expect_error(write_trail(trail, NA_character_), "file")
  expect_error(write_trail(trail, c("a", "b")),   "file")
})

test_that("read_trail errors when file does not exist", {
  expect_error(read_trail("/nonexistent/path/trail.rds"), "not found")
})

test_that("read_trail errors on unknown extension without explicit format", {
  tmp <- tempfile(fileext = ".csv")
  file.create(tmp)
  on.exit(unlink(tmp))
  expect_error(read_trail(tmp), "infer")
})

test_that("read_trail errors on invalid file argument", {
  expect_error(read_trail(NA_character_), "file")
})


# ── audit_export ─────────────────────────────────────────────────────────────

test_that("audit_export creates an HTML file", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- make_export_trail()
  audit_export(trail, file = tmp)
  expect_true(file.exists(tmp))
})

test_that("audit_export output is valid HTML containing the trail JSON", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- make_export_trail()
  audit_export(trail, file = tmp)
  content <- paste(readLines(tmp), collapse = "\n")
  expect_true(grepl("<!DOCTYPE html>", content, fixed = TRUE))
  expect_true(grepl("TRAIL_DATA", content, fixed = TRUE))
  expect_true(grepl("export_test", content, fixed = TRUE))
})

test_that("audit_export returns the file path invisibly", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- make_export_trail()
  result <- withVisible(audit_export(trail, file = tmp))
  expect_false(result$visible)
  expect_equal(result$value, tmp)
})

test_that("audit_export embeds all snapshot labels in the output", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- make_export_trail()
  audit_export(trail, file = tmp)
  content <- paste(readLines(tmp), collapse = "\n")
  expect_true(grepl('"raw"', content, fixed = TRUE))
  expect_true(grepl('"joined"', content, fixed = TRUE))
  expect_true(grepl('"filtered"', content, fixed = TRUE))
})

test_that("audit_export errors on non-trail input", {
  expect_error(audit_export(list()), "audit_trail")
})

test_that("audit_export errors on invalid file argument", {
  skip_if_not_installed("jsonlite")
  trail <- make_export_trail()
  expect_error(audit_export(trail, file = NA_character_), "file")
  expect_error(audit_export(trail, file = c("a.html", "b.html")), "file")
})

test_that("audit_export works with empty trail", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- audit_trail("empty")
  audit_export(trail, file = tmp)
  expect_true(file.exists(tmp))
  content <- paste(readLines(tmp), collapse = "\n")
  expect_true(grepl("TRAIL_DATA", content, fixed = TRUE))
})

test_that("audit_export embeds join diagnostics", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- make_export_trail()
  audit_export(trail, file = tmp)
  content <- paste(readLines(tmp), collapse = "\n")
  expect_true(grepl("match_rate", content, fixed = TRUE))
})

test_that("audit_export embeds filter diagnostics when present", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- audit_trail("filter_test")
  data.frame(id = 1:10, amount = 1:10 * 100) |>
    audit_tap(trail, "raw") |>
    filter_tap(amount > 300, .trail = trail, .label = "filtered")
  audit_export(trail, file = tmp)
  content <- paste(readLines(tmp), collapse = "\n")
  expect_true(grepl("pct_dropped", content, fixed = TRUE))
})

test_that("audit_export with file=NULL creates a temp file", {
  skip_if_not_installed("jsonlite")
  trail <- make_export_trail()
  # Stub browseURL to prevent opening a browser in tests
  local_mocked_bindings(browseURL = function(...) invisible(NULL), .package = "utils")
  result <- audit_export(trail)
  on.exit(unlink(result))
  expect_true(file.exists(result))
  expect_true(grepl("\\.html$", result))
})

test_that("audit_export escapes </script> in trail content", {
  skip_if_not_installed("jsonlite")
  tmp <- tempfile(fileext = ".html")
  on.exit(unlink(tmp))
  trail <- audit_trail("<script>alert(1)</script>")
  mtcars |> audit_tap(trail, "<b>bad</b>")
  audit_export(trail, file = tmp)
  content <- paste(readLines(tmp), collapse = "\n")
  # A raw </script> inside the JSON blob would break the HTML <script> tag.
  # The R function escapes it to <\/script>.
  # Count </script> occurrences — there should be exactly 1 (the real closing tag).
  expect_equal(length(gregexpr("</script>", content, fixed = TRUE)[[1L]]), 1L)
})

# ---------------------------------------------------------------------------
# Snapshot controls serialization
# ---------------------------------------------------------------------------

test_that("JSON round-trip preserves controls", {
  skip_if_not_installed("jsonlite")
  trail <- audit_trail("controls_json")
  mtcars |> audit_tap(trail, "raw", .numeric_summary = FALSE,
                      .cols_include = c("mpg", "cyl"))
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(trail, tmp, format = "json")
  restored <- read_trail(tmp, format = "json")
  snap <- restored$snapshots[[1]]
  expect_null(snap$numeric_summary)
  expect_equal(snap$schema$column, c("mpg", "cyl"))
  expect_false(is.null(snap$controls))
  # jsonlite may deserialize character vectors as lists; compare via unlist
  expect_equal(unlist(snap$controls$cols_include), c("mpg", "cyl"))
  expect_false(snap$controls$numeric_summary)
})

test_that("RDS round-trip preserves controls", {
  trail <- audit_trail("controls_rds")
  mtcars |> audit_tap(trail, "raw", .cols_exclude = c("disp", "hp"))
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  write_trail(trail, tmp, format = "rds")
  restored <- read_trail(tmp, format = "rds")
  snap <- restored$snapshots[[1]]
  expect_false("disp" %in% snap$schema$column)
  expect_false("hp" %in% snap$schema$column)
  expect_equal(snap$controls$cols_exclude, c("disp", "hp"))
})

test_that("trail_to_df includes controls column", {
  trail <- audit_trail("controls_df")
  mtcars |> audit_tap(trail, "raw", .numeric_summary = FALSE)
  df <- trail_to_df(trail)
  expect_true("controls" %in% names(df))
  expect_false(is.null(df$controls[[1]]))
})

# ---------------------------------------------------------------------------
# all_columns serialization
# ---------------------------------------------------------------------------

test_that("JSON round-trip preserves all_columns", {
  skip_if_not_installed("jsonlite")
  trail <- audit_trail("allcols_json")
  mtcars |> audit_tap(trail, "raw", .cols_include = c("mpg", "cyl"))
  tmp <- tempfile(fileext = ".json")
  on.exit(unlink(tmp))
  write_trail(trail, tmp, format = "json")
  restored <- read_trail(tmp, format = "json")
  snap <- restored$snapshots[[1]]
  expect_equal(snap$all_columns, names(mtcars))
  expect_equal(snap$schema$column, c("mpg", "cyl"))
})

test_that("RDS round-trip preserves all_columns", {
  trail <- audit_trail("allcols_rds")
  mtcars |> audit_tap(trail, "raw", .cols_exclude = "disp")
  tmp <- tempfile(fileext = ".rds")
  on.exit(unlink(tmp))
  write_trail(trail, tmp, format = "rds")
  restored <- read_trail(tmp, format = "rds")
  snap <- restored$snapshots[[1]]
  expect_equal(snap$all_columns, names(mtcars))
  expect_false("disp" %in% snap$schema$column)
})

test_that("trail_to_df includes all_columns list-column", {
  trail <- audit_trail("allcols_df")
  mtcars |> audit_tap(trail, "raw")
  df <- trail_to_df(trail)
  expect_true("all_columns" %in% names(df))
  expect_equal(df$all_columns[[1]], names(mtcars))
})

#' Summarize a Single Column
#'
#' Computes summary statistics for a vector. Handles numeric, character,
#' factor, logical, Date, and other types with appropriate statistics for each.
#'
#' @param x A vector to summarize.
#'
#' @returns A named character vector with summary statistics including:
#'   type, unique count, missing count, missing share (proportion from 0 to 1),
#'   most frequent value (for non-numeric), mean, sd, min, quartiles (q25, q50,
#'   q75), max, and three example values.
#'
#' @examples
#' summarize_column(c(1, 2, 3, NA, 5))
#' summarize_column(c("a", "b", "a", "c"))
#'
#' @family data quality
#' @export
summarize_column <- function(x) {
  x_NAs <- is.na(x)
  x_no_NAs <- x[!x_NAs]
  n_valid <- length(x_no_NAs)
  n_uniq <- length(unique(x_no_NAs))

  if (is.numeric(x)) {
    c(
      type = "numeric",
      n_unique = as.character(n_uniq),
      missing = as.character(sum(x_NAs)),
      missing_share = as.character(mean(x_NAs)),
      most_frequent = NA_character_,
      mean = if (n_valid) as.character(mean(x_no_NAs)) else NA_character_,
      sd = if (n_valid > 1L) as.character(stats::sd(x_no_NAs)) else NA_character_,
      min = if (n_valid) as.character(min(x_no_NAs)) else NA_character_,
      q25 = if (n_valid) as.character(quantile(x_no_NAs, 0.25, names = FALSE)) else NA_character_,
      q50 = if (n_valid) as.character(quantile(x_no_NAs, 0.5, names = FALSE)) else NA_character_,
      q75 = if (n_valid) as.character(quantile(x_no_NAs, 0.75, names = FALSE)) else NA_character_,
      max = if (n_valid) as.character(max(x_no_NAs)) else NA_character_,
      example1 = if (n_valid >= 1L) as.character(x_no_NAs[1]) else NA_character_,
      example2 = if (n_valid >= 2L) as.character(x_no_NAs[2]) else NA_character_,
      example3 = if (n_valid >= 3L) as.character(x_no_NAs[3]) else NA_character_
    )
  } else if (is.factor(x)) {
    x_chr <- as.character(x_no_NAs)
    c(
      type = "factor",
      n_unique = as.character(n_uniq),
      missing = as.character(sum(x_NAs)),
      missing_share = as.character(mean(x_NAs)),
      most_frequent = .most_frequent(x),
      mean = NA_character_,
      sd = NA_character_,
      min = if (n_valid) min(x_chr) else NA_character_,
      q25 = NA_character_,
      q50 = NA_character_,
      q75 = NA_character_,
      max = if (n_valid) max(x_chr) else NA_character_,
      example1 = if (n_valid >= 1L) as.character(x_no_NAs[1]) else NA_character_,
      example2 = if (n_valid >= 2L) as.character(x_no_NAs[2]) else NA_character_,
      example3 = if (n_valid >= 3L) as.character(x_no_NAs[3]) else NA_character_
    )
  } else if (is.character(x)) {
    c(
      type = "character",
      n_unique = as.character(n_uniq),
      missing = as.character(sum(x_NAs)),
      missing_share = as.character(mean(x_NAs)),
      most_frequent = .most_frequent(x),
      mean = NA_character_,
      sd = NA_character_,
      min = if (n_valid) min(x_no_NAs) else NA_character_,
      q25 = NA_character_,
      q50 = NA_character_,
      q75 = NA_character_,
      max = if (n_valid) max(x_no_NAs) else NA_character_,
      example1 = if (n_valid >= 1L) x_no_NAs[1] else NA_character_,
      example2 = if (n_valid >= 2L) x_no_NAs[2] else NA_character_,
      example3 = if (n_valid >= 3L) x_no_NAs[3] else NA_character_
    )
  } else if (is.logical(x)) {
    c(
      type = "logical",
      n_unique = as.character(n_uniq),
      missing = as.character(sum(x_NAs)),
      missing_share = as.character(mean(x_NAs)),
      most_frequent = .most_frequent(x),
      mean = if (n_valid) as.character(mean(x_no_NAs)) else NA_character_,
      sd = if (n_valid > 1L) as.character(stats::sd(x_no_NAs)) else NA_character_,
      min = NA_character_,
      q25 = NA_character_,
      q50 = NA_character_,
      q75 = NA_character_,
      max = NA_character_,
      example1 = if (n_valid >= 1L) as.character(x_no_NAs[1]) else NA_character_,
      example2 = if (n_valid >= 2L) as.character(x_no_NAs[2]) else NA_character_,
      example3 = if (n_valid >= 3L) as.character(x_no_NAs[3]) else NA_character_
    )
  } else if ("Date" %in% class(x)) {
    c(
      type = "Date",
      n_unique = as.character(n_uniq),
      missing = as.character(sum(x_NAs)),
      missing_share = as.character(mean(x_NAs)),
      most_frequent = .most_frequent(x),
      mean = NA_character_,
      sd = NA_character_,
      min = if (n_valid) as.character(min(x_no_NAs)) else NA_character_,
      q25 = NA_character_,
      q50 = NA_character_,
      q75 = NA_character_,
      max = if (n_valid) as.character(max(x_no_NAs)) else NA_character_,
      example1 = if (n_valid >= 1L) as.character(x_no_NAs[1]) else NA_character_,
      example2 = if (n_valid >= 2L) as.character(x_no_NAs[2]) else NA_character_,
      example3 = if (n_valid >= 3L) as.character(x_no_NAs[3]) else NA_character_
    )
  } else {
    c(
      type = class(x)[1L],
      n_unique = as.character(n_uniq),
      missing = as.character(sum(x_NAs)),
      missing_share = as.character(mean(x_NAs)),
      most_frequent = .most_frequent(x),
      mean = NA_character_,
      sd = NA_character_,
      min = NA_character_,
      q25 = NA_character_,
      q50 = NA_character_,
      q75 = NA_character_,
      max = NA_character_,
      example1 = if (n_valid >= 1L) as.character(x_no_NAs[1]) else NA_character_,
      example2 = if (n_valid >= 2L) as.character(x_no_NAs[2]) else NA_character_,
      example3 = if (n_valid >= 3L) as.character(x_no_NAs[3]) else NA_character_
    )
  }
}

#' Find Most Frequent Value
#' @noRd
.most_frequent <- function(x) {
  x_no_na <- x[!is.na(x)]
  if (length(x_no_na) == 0L) return(NA_character_)
  tbl <- table(x_no_na)
  as.character(names(tbl)[which.max(tbl)])
}

# Statistic names returned by summarize_column()
.summary_stat_names <- c(
  "type", "n_unique", "missing", "missing_share", "most_frequent",
  "mean", "sd", "min", "q25", "q50", "q75", "max",
  "example1", "example2", "example3"
)

#' Generate Summary Table for a Data Frame
#'
#' Creates a comprehensive summary of all columns in a data.frame, including
#' type, missing values, descriptive statistics, and example values.
#'
#' @param .data A data.frame or tibble to summarize.
#' @param cols Optional character vector of column names to summarize. If
#'   `NULL` (the default), all columns are summarized.
#'
#' @returns A data.frame with one row per column containing summary statistics.
#'
#' @examples
#' df <- data.frame(
#'   id = 1:100,
#'   value = rnorm(100),
#'   category = sample(letters[1:5], 100, replace = TRUE)
#' )
#' get_summary_table(df)
#'
#' @family data quality
#' @export
get_summary_table <- function(.data, cols = NULL) {
  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble.")
  }

  if (!is.null(cols)) {
    miss <- setdiff(cols, names(.data))
    if (length(miss) > 0L) {
      cli::cli_abort("Column{?s} not found in {.arg .data}: {.field {miss}}")
    }
    .data <- .data[cols]
  }

  summaries <- lapply(.data, summarize_column)
  result <- data.frame(
    variable = names(summaries),
    do.call(rbind, summaries),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
  result
}

#' Diagnose Missing Values
#'
#' Reports NA counts and percentages for each column in a data.frame, sorted
#' by missing percentage in descending order.
#'
#' @param .data A data.frame or tibble to diagnose.
#'
#' @returns An S3 object of class `diagnose_na` containing:
#' \describe{
#'   \item{table}{A data.frame with columns `variable`, `n_na`, `pct_na`, and
#'     `n_valid`, sorted by `pct_na` descending.}
#'   \item{n_cols}{Total number of columns in the input.}
#'   \item{n_with_na}{Number of columns that have at least one NA.}
#' }
#'
#' @examples
#' df <- data.frame(
#'   a = c(1, NA, 3),
#'   b = c(NA, NA, "x"),
#'   c = c(TRUE, FALSE, TRUE)
#' )
#' diagnose_nas(df)
#'
#' @family data quality
#' @export
diagnose_nas <- function(.data) {
  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble.")
  }

  n_rows <- nrow(.data)
  na_counts <- vapply(.data, function(col) sum(is.na(col)), integer(1L))

  tbl <- data.frame(
    variable = names(na_counts),
    n_na     = as.integer(na_counts),
    pct_na   = if (n_rows > 0L) round(100 * na_counts / n_rows, 1) else rep(0, length(na_counts)),
    n_valid  = as.integer(n_rows - na_counts),
    stringsAsFactors = FALSE,
    row.names = NULL
  )
  # Sort by pct_na descending
  tbl <- tbl[order(-tbl$pct_na), , drop = FALSE]
  row.names(tbl) <- NULL

  structure(
    list(
      table = tbl,
      n_cols = ncol(.data),
      n_with_na = sum(tbl$n_na > 0L)
    ),
    class = c("diagnose_na", "list")
  )
}

#' @rdname diagnose_nas
#' @param x An object to print.
#' @param ... Additional arguments (currently unused).
#' @export
print.diagnose_na <- function(x, ...) {
  cli::cli_h1("Missing Value Diagnosis")
  cli::cli_text("{x$n_with_na} of {x$n_cols} columns have missing values")

  if (x$n_with_na > 0L) {
    cols_with_na <- x$table[x$table$n_na > 0L, , drop = FALSE]
    display_tbl <- data.frame(
      Variable = cols_with_na$variable,
      `N NA`   = format(cols_with_na$n_na, big.mark = ",", trim = TRUE),
      `Pct NA` = sprintf("%.1f%%", cols_with_na$pct_na),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
    cli::cli_text("")
    .cli_table(display_tbl, right_align = c("N NA", "Pct NA"))
  }

  invisible(x)
}

test_that("summarize_column works for numeric", {
  result <- summarize_column(c(1, 2, 3, NA, 5))

  expect_equal(result[["type"]], "numeric")
  expect_equal(result[["missing"]], "1")
  expect_equal(result[["n_unique"]], "4")
  expect_false(is.na(result[["mean"]]))
  expect_false(is.na(result[["min"]]))
  expect_false(is.na(result[["max"]]))
})

test_that("summarize_column works for character", {
  result <- summarize_column(c("a", "b", "a", "c"))

  expect_equal(result[["type"]], "character")
  expect_equal(result[["n_unique"]], "3")
  expect_equal(result[["missing"]], "0")
  expect_equal(result[["most_frequent"]], "a")
  expect_true(is.na(result[["mean"]]))
})

test_that("summarize_column works for logical", {
  result <- summarize_column(c(TRUE, FALSE, TRUE, NA))

  expect_equal(result[["type"]], "logical")
  expect_equal(result[["missing"]], "1")
  expect_false(is.na(result[["mean"]]))
})

test_that("summarize_column works for factor", {
  result <- summarize_column(factor(c("x", "y", "x")))

  expect_equal(result[["type"]], "factor")
  expect_equal(result[["n_unique"]], "2")
  expect_equal(result[["most_frequent"]], "x")
})

test_that("summarize_column works for Date", {
  result <- summarize_column(as.Date(c("2023-01-01", "2023-06-15", NA)))

  expect_equal(result[["type"]], "Date")
  expect_equal(result[["missing"]], "1")
  expect_equal(result[["min"]], "2023-01-01")
  expect_equal(result[["max"]], "2023-06-15")
})

test_that("summarize_column handles all-NA", {
  result <- summarize_column(c(NA_real_, NA_real_))

  expect_equal(result[["type"]], "numeric")
  expect_equal(result[["missing"]], "2")
  expect_equal(result[["n_unique"]], "0")
  expect_true(is.na(result[["mean"]]))
})

test_that("summarize_column returns named character vector", {
  result <- summarize_column(1:5)
  expect_true(is.character(result))
  expect_equal(length(result), 15L)
  expect_true(all(c("type", "n_unique", "missing", "mean") %in% names(result)))
})

test_that("get_summary_table returns data.frame with all columns", {
  df <- data.frame(
    id = 1:5,
    value = c(10, 20, 30, 40, 50),
    name = c("a", "b", "c", "d", "e"),
    stringsAsFactors = FALSE
  )
  result <- get_summary_table(df)

  expect_true(is.data.frame(result))
  expect_equal(nrow(result), 3L)
  expect_true("variable" %in% names(result))
  expect_true("type" %in% names(result))
})

test_that("get_summary_table filters by cols", {
  df <- data.frame(id = 1:5, value = 1:5, name = letters[1:5],
                   stringsAsFactors = FALSE)
  result <- get_summary_table(df, cols = c("id", "value"))

  expect_equal(nrow(result), 2L)
  expect_true(all(result$variable %in% c("id", "value")))
})

test_that("get_summary_table errors on missing cols", {
  df <- data.frame(id = 1:3)
  expect_error(get_summary_table(df, cols = "nonexistent"), "not found")
})

test_that("diagnose_nas returns correct structure", {
  df <- data.frame(
    a = c(1, NA, 3),
    b = c(NA, NA, "x"),
    c = c(TRUE, FALSE, TRUE),
    stringsAsFactors = FALSE
  )
  result <- diagnose_nas(df)

  expect_s3_class(result, "diagnose_na")
  expect_equal(result$n_cols, 3L)
  expect_equal(result$n_with_na, 2L)

  tbl <- result$table
  expect_true(is.data.frame(tbl))
  expect_equal(nrow(tbl), 3L)
  expect_true(all(c("variable", "n_na", "pct_na", "n_valid") %in% names(tbl)))
})

test_that("diagnose_nas sorted by pct_na descending", {
  df <- data.frame(
    a = c(1, NA, 3),       # 33.3% NA
    b = c(NA, NA, "x"),    # 66.7% NA
    c = c(TRUE, FALSE, TRUE), # 0% NA
    stringsAsFactors = FALSE
  )
  result <- diagnose_nas(df)

  expect_equal(result$table$variable[1], "b")  # highest NA percentage first
  expect_equal(result$table$variable[3], "c")  # lowest NA last
})

test_that("diagnose_nas handles no NAs", {
  df <- data.frame(a = 1:3, b = 4:6)
  result <- diagnose_nas(df)

  expect_equal(result$n_with_na, 0L)
  expect_true(all(result$table$n_na == 0L))
})

test_that("diagnose_nas handles all NAs", {
  df <- data.frame(a = c(NA, NA), b = c(NA, NA))
  result <- diagnose_nas(df)

  expect_equal(result$n_with_na, 2L)
  expect_true(all(result$table$pct_na == 100))
})

test_that("print.diagnose_na produces output", {
  df <- data.frame(a = c(1, NA, 3), b = c(NA, NA, "x"),
                   stringsAsFactors = FALSE)
  result <- diagnose_nas(df)

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Missing Value", combined))
  expect_true(grepl("2 of 2", combined))
})

test_that("diagnose_nas works with tibbles", {
  skip_if_not_installed("dplyr")
  df <- dplyr::tibble(a = c(1, NA, 3), b = c(NA, NA, "x"))
  result <- diagnose_nas(df)

  expect_s3_class(result, "diagnose_na")
  expect_equal(result$n_with_na, 2L)
})

test_that("diagnose_nas handles zero-row data.frame without NaN", {
  df <- data.frame(a = integer(0), b = character(0), stringsAsFactors = FALSE)
  result <- diagnose_nas(df)

  expect_s3_class(result, "diagnose_na")
  expect_equal(result$n_cols, 2L)
  expect_equal(result$n_with_na, 0L)
  expect_true(all(result$table$pct_na == 0))
  expect_false(any(is.nan(result$table$pct_na)))
})

#' Tabulate Variables
#'
#' Produces one-way frequency tables or two-way crosstabulations. One variable
#' gives counts, percentages, and cumulative percentages; two variables give a
#' crosstabulation matrix with row/column totals.
#'
#' @param .data A data.frame or tibble.
#' @param ... One or two unquoted column names to tabulate. One variable
#'   produces a one-way frequency table; two produce a two-way crosstabulation.
#' @param .wt Optional unquoted column to use as frequency weights.
#'   When supplied, frequencies are weighted sums instead of row counts.
#' @param .sort How to order the rows (and columns in two-way tables).
#'   `"value_asc"` (default) sorts alphabetically (or by factor levels),
#'   `"value_desc"` sorts in reverse, `"freq_desc"` sorts by frequency
#'   descending, `"freq_asc"` sorts by frequency ascending.
#' @param .cutoff Controls how many values to display.
#'   An integer >= 1 keeps the top-N values by frequency.
#'   A number in (0, 1) keeps values that cumulatively account for that
#'   proportion of the total. Remaining values are grouped under `"(Other)"`.
#'   For two-way tables, the cutoff applies to the row variable only.
#' @param .na How to handle `NA` values. `"include"` (default) treats `NA` as a
#'   category, `"exclude"` drops `NA` rows before tabulation, `"only"` shows
#'   only `NA` rows.
#' @param .display Cell contents for two-way crosstabulations. One of
#'   `"count"` (default), `"row_pct"`, `"col_pct"`, or `"total_pct"`.
#'   Ignored for one-way tables.
#'
#' @returns An S3 object of class `tidyaudit_tab`. Use [as.data.frame()] to
#'   extract the underlying table.
#'
#' @examples
#' tab(mtcars, cyl)
#' tab(mtcars, cyl, .sort = "freq_desc")
#' tab(mtcars, cyl, gear)
#' tab(mtcars, cyl, gear, .display = "row_pct")
#' tab(mtcars, cyl, .wt = mpg)
#' tab(mtcars, cyl, .cutoff = 2)
#'
#' @family data quality
#' @export
tab <- function(.data, ..., .wt = NULL,
                .sort = c("value_asc", "value_desc", "freq_desc", "freq_asc"),
                .cutoff = NULL,
                .na = c("include", "exclude", "only"),
                .display = c("count", "row_pct", "col_pct", "total_pct")) {

  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble.")
  }

  .sort <- match.arg(.sort)
  .na <- match.arg(.na)
  .display <- match.arg(.display)

  # Validate .cutoff
  if (!is.null(.cutoff)) {
    if (!is.numeric(.cutoff) || length(.cutoff) != 1L || is.na(.cutoff) ||
        .cutoff <= 0) {
      cli::cli_abort(
        "{.arg .cutoff} must be a positive number (integer >= 1 for top-N, or a proportion in (0, 1))."
      )
    }
  }

  # Capture variable names via NSE
  dots <- rlang::enquos(...)
  n_vars <- length(dots)

  if (n_vars == 0L) {
    cli::cli_abort("At least one variable must be supplied to {.fn tab}.")
  }
  if (n_vars > 2L) {
    cli::cli_abort("{.fn tab} supports at most 2 variables, not {n_vars}.")
  }

  var_names <- unname(vapply(dots, rlang::as_label, character(1)))

  # Validate columns exist
  missing_cols <- setdiff(var_names, names(.data))
  if (length(missing_cols) > 0L) {
    cli::cli_abort("Column{?s} not found in {.arg .data}: {.field {missing_cols}}")
  }

  # Capture weight
  wt_quo <- rlang::enquo(.wt)
  have_wt <- !rlang::quo_is_null(wt_quo)
  wt_name <- if (have_wt) rlang::as_label(wt_quo) else NULL

  if (have_wt && !wt_name %in% names(.data)) {
    cli::cli_abort("Weight column {.field {wt_name}} not found in {.arg .data}.")
  }

  # Extract vectors
  n_obs <- nrow(.data)
  var1 <- .data[[var_names[1L]]]
  var2 <- if (n_vars == 2L) .data[[var_names[2L]]] else NULL
  wt_vec <- if (have_wt) .data[[wt_name]] else NULL

  # Count NAs: number of rows where any tabulation variable is NA
  has_any_na <- is.na(var1)
  if (!is.null(var2)) has_any_na <- has_any_na | is.na(var2)
  n_missing <- sum(has_any_na)

  # Handle NAs
  if (.na == "exclude") {
    keep <- !has_any_na
    var1 <- var1[keep]
    if (!is.null(var2)) var2 <- var2[keep]
    if (!is.null(wt_vec)) wt_vec <- wt_vec[keep]
  } else if (.na == "only") {
    keep <- has_any_na
    var1 <- var1[keep]
    if (!is.null(var2)) var2 <- var2[keep]
    if (!is.null(wt_vec)) wt_vec <- wt_vec[keep]
  }

  if (n_vars == 1L) {
    .tab_oneway(var1, wt_vec, var_names[1L], n_obs, n_missing,
                .na, .sort, .cutoff, have_wt, wt_name)
  } else {
    .tab_twoway(var1, var2, wt_vec, var_names, n_obs, n_missing,
                .na, .sort, .cutoff, .display, have_wt, wt_name)
  }
}

# -- Internal helpers ---------------------------------------------------------

#' One-way tabulation
#' @noRd
.tab_oneway <- function(var, wt, var_name, n_obs, n_missing,
                         na_mode, sort, cutoff, weighted, wt_name) {
  # Count NAs before removing them for table()
  n_na <- sum(is.na(var))
  include_na <- (na_mode %in% c("include", "only")) && n_na > 0L

  # Compute frequencies on non-NA values (preserving original type for sorting)
  var_nona <- var[!is.na(var)]
  wt_nona <- if (!is.null(wt)) wt[!is.na(var)] else NULL

  if (is.null(wt_nona)) {
    tbl_raw <- table(var_nona, useNA = "no")
    if (length(tbl_raw) == 0L) {
      freq <- data.frame(Value = character(), Freq = numeric(),
                         stringsAsFactors = FALSE)
    } else {
      freq <- as.data.frame(tbl_raw, stringsAsFactors = FALSE)
      names(freq) <- c("Value", "Freq")
      freq$Freq <- as.numeric(freq$Freq)
    }
  } else {
    # Weighted: exclude NA weights silently
    ok <- !is.na(wt_nona)
    freq_vals <- tapply(wt_nona[ok], var_nona[ok], sum)
    freq <- data.frame(
      Value = names(freq_vals),
      Freq = as.numeric(freq_vals),
      stringsAsFactors = FALSE
    )
    # Drop NaN from tapply on empty groups
    freq <- freq[!is.na(freq$Freq), , drop = FALSE]
  }

  # Append <NA> row if including NAs
  if (include_na) {
    na_freq <- if (is.null(wt)) {
      n_na
    } else {
      sum(wt[is.na(var)], na.rm = TRUE)
    }
    freq <- rbind(freq, data.frame(Value = "<NA>", Freq = na_freq,
                                    stringsAsFactors = FALSE))
  }

  # High-cardinality guard
  if (nrow(freq) > 50L && is.null(cutoff)) {
    cli::cli_warn(
      c("{.field {var_name}} has {nrow(freq)} unique values.",
        "i" = "Consider using {.arg .cutoff} to limit the display.")
    )
  }

  # Apply cutoff
  n_other <- 0L
  if (!is.null(cutoff)) {
    result <- .apply_cutoff(freq, cutoff)
    freq <- result$freq
    n_other <- result$n_other
  }

  # Sort
  freq <- .sort_freq(freq, sort, var)

  # Compute percentages
  total <- sum(freq$Freq)
  if (nrow(freq) > 0L) {
    freq$Percent <- if (total > 0) round(100 * freq$Freq / total, 1) else rep(0, nrow(freq))
    freq$Cum_Percent <- pmin(cumsum(freq$Percent), 100.0)
  } else {
    freq$Percent <- numeric()
    freq$Cum_Percent <- numeric()
  }

  # Add total row
  total_row <- data.frame(
    Value = "Total",
    Freq = total,
    Percent = 100.0,
    Cum_Percent = NA_real_,
    stringsAsFactors = FALSE
  )
  tbl <- rbind(freq, total_row)
  row.names(tbl) <- NULL

  structure(
    list(
      type        = "oneway",
      var_name    = var_name,
      n_obs       = n_obs,
      n_missing   = n_missing,
      na_mode     = na_mode,
      sort        = sort,
      cutoff      = cutoff,
      n_other     = n_other,
      weighted    = weighted,
      wt_name     = wt_name,
      table       = tbl
    ),
    class = c("tidyaudit_tab", "list")
  )
}

#' Two-way crosstabulation
#' @noRd
.tab_twoway <- function(var1, var2, wt, var_names, n_obs, n_missing,
                         na_mode, sort, cutoff, display, weighted, wt_name) {
  # Determine which rows have NAs in either variable
  na1 <- is.na(var1)
  na2 <- is.na(var2)
  include_na <- na_mode %in% c("include", "only")

  # Work with non-NA rows for the main table (preserves original types)
  complete <- !na1 & !na2
  v1_complete <- var1[complete]
  v2_complete <- var2[complete]
  wt_complete <- if (!is.null(wt)) wt[complete] else NULL

  # High-cardinality guard
  n_unique1 <- length(unique(v1_complete))
  n_unique2 <- length(unique(v2_complete))
  if ((n_unique1 > 50L || n_unique2 > 50L) && is.null(cutoff)) {
    cli::cli_warn(
      c("Variables have many unique values ({n_unique1} x {n_unique2}).",
        "i" = "Consider using {.arg .cutoff} to limit the display.")
    )
  }

  # Compute marginal frequencies for cutoff (applied to row variable only)
  n_other <- 0L
  if (!is.null(cutoff)) {
    if (is.null(wt_complete)) {
      marginal <- as.data.frame(table(v1_complete, useNA = "no"), stringsAsFactors = FALSE)
      names(marginal) <- c("Value", "Freq")
      marginal$Freq <- as.numeric(marginal$Freq)
    } else {
      ok <- !is.na(wt_complete)
      mfreq <- tapply(wt_complete[ok], v1_complete[ok], sum)
      marginal <- data.frame(
        Value = names(mfreq), Freq = as.numeric(mfreq),
        stringsAsFactors = FALSE
      )
      marginal <- marginal[!is.na(marginal$Freq), , drop = FALSE]
    }
    result <- .apply_cutoff(marginal, cutoff)
    kept_values <- result$freq$Value
    n_other <- result$n_other

    if (n_other > 0L) {
      v1_complete <- ifelse(
        as.character(v1_complete) %in% kept_values,
        as.character(v1_complete), "(Other)"
      )
    }
  }

  # Build cross-table from complete rows
  if (is.null(wt_complete)) {
    ct <- table(v1_complete, v2_complete, useNA = "no")
  } else {
    ok <- !is.na(wt_complete)
    ct <- tapply(wt_complete[ok], list(v1_complete[ok], v2_complete[ok]), sum)
    ct[is.na(ct)] <- 0
  }

  ct_mat <- as.matrix(ct)

  # Append <NA> row/column if including NAs and there are any
  if (include_na) {
    ct_mat <- .append_na_margins(ct_mat, var1, var2, wt, na1, na2)
  }

  # Sort rows
  row_sums <- rowSums(ct_mat)
  ct_mat <- .sort_matrix_rows(ct_mat, row_sums, sort, var1)

  # Sort columns
  col_sums <- colSums(ct_mat)
  ct_mat <- .sort_matrix_cols(ct_mat, col_sums, sort, var2)

  # Build output data.frame with totals
  row_totals <- rowSums(ct_mat)
  col_totals <- colSums(ct_mat)
  grand_total <- sum(ct_mat)

  # Format cells according to .display
  ct_display <- .format_crosstab(ct_mat, row_totals, col_totals,
                                  grand_total, display)

  # Assemble matrix data.frame
  mat_df <- data.frame(
    V1 = c(rownames(ct_mat), "Total"),
    stringsAsFactors = FALSE
  )
  names(mat_df)[1L] <- var_names[1L]

  for (j in seq_len(ncol(ct_display))) {
    col_total_cell <- .format_crosstab_total(col_totals[j])
    mat_df[[colnames(ct_display)[j]]] <- c(ct_display[, j], col_total_cell)
  }

  # Total column
  total_col <- c(
    vapply(seq_len(nrow(ct_mat)), function(i) {
      .format_crosstab_total(row_totals[i])
    }, character(1)),
    .format_number(grand_total)
  )
  mat_df[["Total"]] <- total_col

  structure(
    list(
      type        = "twoway",
      var1_name   = var_names[1L],
      var2_name   = var_names[2L],
      n_obs       = n_obs,
      n_missing   = n_missing,
      na_mode     = na_mode,
      sort        = sort,
      cutoff      = cutoff,
      n_other     = n_other,
      weighted    = weighted,
      wt_name     = wt_name,
      display     = display,
      matrix      = mat_df,
      counts      = ct_mat,
      row_totals  = row_totals,
      col_totals  = col_totals,
      grand_total = grand_total
    ),
    class = c("tidyaudit_tab", "list")
  )
}

#' Apply cutoff to a frequency table
#' @noRd
.apply_cutoff <- function(freq, cutoff) {
  # Sort by frequency descending for cutoff
  freq <- freq[order(-freq$Freq), , drop = FALSE]

  if (cutoff >= 1) {
    # Top-N
    n_keep <- min(as.integer(cutoff), nrow(freq))
  } else {
    # Cumulative proportion
    total <- sum(freq$Freq)
    cum_prop <- cumsum(freq$Freq) / total
    n_keep <- which(cum_prop >= cutoff)[1L]
    if (is.na(n_keep)) n_keep <- nrow(freq)
  }

  if (n_keep >= nrow(freq)) {
    return(list(freq = freq, n_other = 0L))
  }

  kept <- freq[seq_len(n_keep), , drop = FALSE]
  collapsed <- freq[seq(n_keep + 1L, nrow(freq)), , drop = FALSE]
  n_other <- nrow(collapsed)

  other_row <- data.frame(
    Value = "(Other)",
    Freq = sum(collapsed$Freq),
    stringsAsFactors = FALSE
  )
  freq_out <- rbind(kept, other_row)
  row.names(freq_out) <- NULL

  list(freq = freq_out, n_other = n_other)
}

#' Sort a frequency data.frame
#' @noRd
.sort_freq <- function(freq, sort, original_var) {
  # Keep (Other) and <NA> always last
  is_special <- freq$Value %in% c("(Other)", "<NA>")
  special_rows <- freq[is_special, , drop = FALSE]
  main <- freq[!is_special, , drop = FALSE]

  if (sort == "freq_desc") {
    main <- main[order(-main$Freq), , drop = FALSE]
  } else if (sort == "freq_asc") {
    main <- main[order(main$Freq), , drop = FALSE]
  } else if (sort == "value_asc") {
    main <- main[order(.sort_key(main$Value, original_var)), , drop = FALSE]
  } else if (sort == "value_desc") {
    main <- main[order(.sort_key(main$Value, original_var),
                        decreasing = TRUE), , drop = FALSE]
  }

  result <- rbind(main, special_rows)
  row.names(result) <- NULL
  result
}

#' Compute a sort key that respects the original variable type
#'
#' For factors, returns factor-level position. For numeric-origin columns,
#' returns numeric values so that "2" sorts before "10". Otherwise returns
#' the character values for alphabetical sorting.
#' @noRd
.sort_key <- function(labels, original_var) {
  if (is.factor(original_var)) {
    return(match(labels, levels(original_var)))
  }
  if (is.numeric(original_var)) {
    return(suppressWarnings(as.numeric(labels)))
  }
  labels
}

#' Sort matrix rows by criteria
#' @noRd
.sort_matrix_rows <- function(mat, row_sums, sort, original_var) {
  rn <- rownames(mat)
  key <- .sort_key(rn, original_var)
  idx <- switch(sort,
    "value_asc"  = order(key),
    "value_desc" = order(key, decreasing = TRUE),
    "freq_desc"  = order(-row_sums),
    "freq_asc"   = order(row_sums)
  )
  # Keep (Other) and <NA> last
  is_special <- rn %in% c("(Other)", "<NA>")
  if (any(is_special)) {
    idx <- c(idx[!is_special[idx]], which(is_special))
  }
  mat[idx, , drop = FALSE]
}

#' Sort matrix columns by criteria
#' @noRd
.sort_matrix_cols <- function(mat, col_sums, sort, original_var) {
  cn <- colnames(mat)
  key <- .sort_key(cn, original_var)
  idx <- switch(sort,
    "value_asc"  = order(key),
    "value_desc" = order(key, decreasing = TRUE),
    "freq_desc"  = order(-col_sums),
    "freq_asc"   = order(col_sums)
  )
  # Keep (Other) and <NA> last
  is_special <- cn %in% c("(Other)", "<NA>")
  if (any(is_special)) {
    idx <- c(idx[!is_special[idx]], which(is_special))
  }
  mat[, idx, drop = FALSE]
}

#' Append <NA> row and/or column to a crosstab matrix
#' @noRd
.append_na_margins <- function(ct_mat, var1, var2, wt, na1, na2) {
  has_na1 <- any(na1)
  has_na2 <- any(na2)

  if (!has_na1 && !has_na2) return(ct_mat)

  rn <- rownames(ct_mat)
  cn <- colnames(ct_mat)

  # <NA> column: for each non-NA var1 value, count rows where var2 is NA
  if (has_na2) {
    na_col <- vapply(rn, function(rv) {
      mask <- !na1 & as.character(var1) == rv & na2
      if (is.null(wt)) sum(mask) else sum(wt[mask], na.rm = TRUE)
    }, numeric(1))
    ct_mat <- cbind(ct_mat, `<NA>` = na_col)
  }

  # <NA> row: for each non-NA var2 value, count rows where var1 is NA
  if (has_na1) {
    na_row <- vapply(cn, function(cv) {
      mask <- na1 & !na2 & as.character(var2) == cv
      if (is.null(wt)) sum(mask) else sum(wt[mask], na.rm = TRUE)
    }, numeric(1))
    # Both NA corner cell
    if (has_na2) {
      both_na <- if (is.null(wt)) sum(na1 & na2) else sum(wt[na1 & na2], na.rm = TRUE)
      na_row <- c(na_row, both_na)
    }
    ct_mat <- rbind(ct_mat, `<NA>` = na_row)
  }

  ct_mat
}

#' Format crosstab cells according to display mode
#' @noRd
.format_crosstab <- function(ct_mat, row_totals, col_totals,
                              grand_total, display) {
  out <- matrix("", nrow = nrow(ct_mat), ncol = ncol(ct_mat))
  rownames(out) <- rownames(ct_mat)
  colnames(out) <- colnames(ct_mat)

  for (i in seq_len(nrow(ct_mat))) {
    for (j in seq_len(ncol(ct_mat))) {
      val <- ct_mat[i, j]
      out[i, j] <- switch(display,
        "count"     = .format_number(val),
        "row_pct"   = .format_pct_val(val, row_totals[i]),
        "col_pct"   = .format_pct_val(val, col_totals[j]),
        "total_pct" = .format_pct_val(val, grand_total)
      )
    }
  }
  out
}

#' Format a total cell in crosstab (always shows count regardless of display)
#' @noRd
.format_crosstab_total <- function(value) {
  .format_number(value)
}

#' Format a number for display
#' @noRd
.format_number <- function(x) {
  format(x, big.mark = ",", scientific = FALSE, trim = TRUE)
}

#' Format a percentage value
#' @noRd
.format_pct_val <- function(num, denom) {
  if (denom == 0) return("0.0%")
  sprintf("%.1f%%", 100 * num / denom)
}

# -- Print method -------------------------------------------------------------

#' @rdname tab
#' @param x A `tidyaudit_tab` object.
#' @param ... Additional arguments (currently unused).
#' @export
print.tidyaudit_tab <- function(x, ...) {
  if (x$type == "oneway") {
    .print_tab_oneway(x)
  } else {
    .print_tab_twoway(x)
  }
  invisible(x)
}

#' Print one-way tabulation
#' @noRd
.print_tab_oneway <- function(x) {
  header <- if (x$weighted) {
    glue::glue("Tabulation: {x$var_name} (weighted by {x$wt_name})")
  } else {
    glue::glue("Tabulation: {x$var_name}")
  }
  cli::cli_h1(header)

  # Observation line
  na_note <- switch(x$na_mode,
    "include" = if (x$n_missing > 0L) {
      glue::glue(" ({format(x$n_missing, big.mark = ',')} missing \u2014 included as category)")
    } else {
      ""
    },
    "exclude" = if (x$n_missing > 0L) {
      glue::glue(" ({format(x$n_missing, big.mark = ',')} missing \u2014 excluded)")
    } else {
      ""
    },
    "only" = " (showing only missing rows)"
  )
  cli::cli_text("{format(x$n_obs, big.mark = ',')} observations{na_note}")
  cli::cli_text("")

  # Build display table (without the Total row for .cli_table, add separator)
  tbl <- x$table
  n_data <- nrow(tbl) - 1L  # exclude Total row

  if (n_data == 0L) {
    cli::cli_alert_info("No values to tabulate.")
    return(invisible(NULL))
  }

  data_rows <- tbl[seq_len(n_data), , drop = FALSE]
  total_row <- tbl[nrow(tbl), , drop = FALSE]

  # Format for display
  display_tbl <- data.frame(
    Value   = c(data_rows$Value, total_row$Value),
    Freq    = c(
      .format_number(data_rows$Freq),
      .format_number(total_row$Freq)
    ),
    Percent = c(
      sprintf("%.1f%%", data_rows$Percent),
      sprintf("%.1f%%", total_row$Percent)
    ),
    Cum.    = c(
      sprintf("%.1f%%", data_rows$Cum_Percent),
      ""
    ),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  .cli_table_with_footer(display_tbl, n_data,
                          right_align = c("Freq", "Percent", "Cum."))

  # Cutoff footnote
  if (x$n_other > 0L) {
    cli::cli_text("")
    cli::cli_text(
      "  {.emph (Other) collapses {x$n_other} value{?s} below cutoff}"
    )
  }
}

#' Print two-way crosstabulation
#' @noRd
.print_tab_twoway <- function(x) {
  header <- if (x$weighted) {
    glue::glue("Crosstabulation: {x$var1_name} \u00d7 {x$var2_name} (weighted by {x$wt_name})")
  } else {
    glue::glue("Crosstabulation: {x$var1_name} \u00d7 {x$var2_name}")
  }
  cli::cli_h1(header)

  display_label <- switch(x$display,
    "count"     = "count",
    "row_pct"   = "row %",
    "col_pct"   = "column %",
    "total_pct" = "total %"
  )

  na_note <- switch(x$na_mode,
    "include" = if (x$n_missing > 0L) {
      glue::glue(" ({format(x$n_missing, big.mark = ',')} missing \u2014 included)")
    } else {
      ""
    },
    "exclude" = if (x$n_missing > 0L) {
      glue::glue(" ({format(x$n_missing, big.mark = ',')} missing \u2014 excluded)")
    } else {
      ""
    },
    "only" = " (showing only missing rows)"
  )

  cli::cli_text(
    "{format(x$n_obs, big.mark = ',')} observations{na_note} | Cell contents: {display_label}"
  )
  cli::cli_text("")

  mat <- x$matrix
  n_data <- nrow(mat) - 1L  # exclude Total row

  if (n_data == 0L) {
    cli::cli_alert_info("No values to tabulate.")
    return(invisible(NULL))
  }

  # All columns except the first (var1 name) are numeric-display
  right_cols <- names(mat)[-1L]
  .cli_table_with_footer(mat, n_data, right_align = right_cols)

  if (x$n_other > 0L) {
    cli::cli_text("")
    cli::cli_text(
      "  {.emph (Other) collapses {x$n_other} value{?s} below cutoff}"
    )
  }
}

#' Render a table with a separator before the last (Total) row
#'
#' @param tbl data.frame to display (last row is the footer/Total row).
#' @param n_data Number of data rows (everything except Total).
#' @param right_align Character vector of column names to right-align.
#' @noRd
.cli_table_with_footer <- function(tbl, n_data, right_align = character()) {
  nms <- names(tbl)

  # All values as character
  for (j in seq_along(tbl)) {
    tbl[[j]] <- as.character(tbl[[j]])
  }

  # Column widths
  col_widths <- vapply(nms, function(nm) {
    data_width <- if (nrow(tbl) > 0L) max(nchar(tbl[[nm]]), na.rm = TRUE) else 0L
    max(nchar(nm), data_width)
  }, integer(1))

  pad <- "  "
  gap <- "  "

  format_row <- function(values) {
    cells <- vapply(seq_along(nms), function(j) {
      w <- col_widths[j]
      if (nms[j] %in% right_align) {
        formatC(values[j], width = w)
      } else {
        formatC(values[j], width = w, flag = "-")
      }
    }, character(1))
    paste0(pad, paste(cells, collapse = gap))
  }

  header_line <- format_row(nms)
  sep_cells <- vapply(seq_along(nms), function(j) {
    strrep("\u2500", col_widths[j])
  }, character(1))
  sep_line <- paste0(pad, paste(sep_cells, collapse = gap))

  data_lines <- vapply(seq_len(n_data), function(i) {
    format_row(as.character(tbl[i, ]))
  }, character(1))

  total_line <- format_row(as.character(tbl[nrow(tbl), ]))

  all_lines <- c(header_line, sep_line, data_lines, sep_line, total_line)
  cli::cli_verbatim(paste(all_lines, collapse = "\n"))
}

# -- as.data.frame method ----------------------------------------------------

#' @rdname tab
#' @param row.names Passed to [as.data.frame()]. Default `NULL`.
#' @param optional Passed to [as.data.frame()]. Default `FALSE`.
#' @export
as.data.frame.tidyaudit_tab <- function(x, row.names = NULL,
                                         optional = FALSE, ...) {
  if (x$type == "oneway") {
    # Return data rows only (no Total row)
    tbl <- x$table
    tbl[tbl$Value != "Total", , drop = FALSE]
  } else {
    # Return numeric counts matrix (not the formatted display strings)
    ct <- x$counts
    df <- data.frame(
      V1 = rownames(ct),
      stringsAsFactors = FALSE
    )
    names(df)[1L] <- x$var1_name
    for (j in seq_len(ncol(ct))) {
      df[[colnames(ct)[j]]] <- ct[, j]
    }
    row.names(df) <- NULL
    df
  }
}

# -- One-way tabulation -------------------------------------------------------

test_that("one-way tab returns correct structure", {
  result <- tab(mtcars, cyl)

  expect_s3_class(result, "tidyaudit_tab")
  expect_equal(result$type, "oneway")
  expect_equal(result$var_name, "cyl")
  expect_equal(result$n_obs, 32L)
  expect_equal(result$n_missing, 0L)
  expect_false(result$weighted)
  expect_null(result$wt_name)

  tbl <- result$table
  expect_true("Total" %in% tbl$Value)
  expect_equal(tbl$Freq[tbl$Value == "Total"], 32)
  expect_equal(tbl$Percent[tbl$Value == "Total"], 100.0)
})

test_that("one-way tab frequencies are correct", {
  result <- tab(mtcars, cyl)
  tbl <- result$table[result$table$Value != "Total", ]

  expect_equal(tbl$Freq[tbl$Value == "4"], 11)
  expect_equal(tbl$Freq[tbl$Value == "6"], 7)
  expect_equal(tbl$Freq[tbl$Value == "8"], 14)
})

test_that("one-way tab cumulative percentage does not exceed 100", {
  result <- tab(mtcars, cyl)
  tbl <- result$table[result$table$Value != "Total", ]
  expect_true(all(tbl$Cum_Percent <= 100.0))
})

test_that("one-way tab prints without error", {
  result <- tab(mtcars, cyl)
  output <- capture.output(print(result), type = "message")
  expect_true(any(grepl("Tabulation: cyl", output)))
  expect_true(any(grepl("Total", output)))
})

# -- Sorting ------------------------------------------------------------------

test_that(".sort = 'freq_desc' sorts by descending frequency", {
  result <- tab(mtcars, cyl, .sort = "freq_desc")
  tbl <- result$table[result$table$Value != "Total", ]
  expect_equal(tbl$Value[1], "8")
  expect_equal(tbl$Value[2], "4")
  expect_equal(tbl$Value[3], "6")
})

test_that(".sort = 'freq_asc' sorts by ascending frequency", {
  result <- tab(mtcars, cyl, .sort = "freq_asc")
  tbl <- result$table[result$table$Value != "Total", ]
  expect_equal(tbl$Value[1], "6")
})

test_that(".sort = 'value_desc' reverses alphabetical order", {
  result <- tab(mtcars, cyl, .sort = "value_desc")
  tbl <- result$table[result$table$Value != "Total", ]
  expect_equal(tbl$Value[1], "8")
  expect_equal(tbl$Value[3], "4")
})

# -- Cutoff -------------------------------------------------------------------

test_that(".cutoff top-N groups remaining into (Other)", {
  result <- tab(mtcars, cyl, .sort = "freq_desc", .cutoff = 2)
  tbl <- result$table

  expect_true("(Other)" %in% tbl$Value)
  expect_equal(result$n_other, 1L)
  data_rows <- tbl[!tbl$Value %in% c("Total", "(Other)"), ]
  expect_equal(nrow(data_rows), 2L)
})

test_that(".cutoff proportional threshold works", {
  result <- tab(mtcars, cyl, .cutoff = 0.5)
  tbl <- result$table
  # cyl=8 has 43.8% — need at least 50%, so top 2 by freq (8=14, 4=11) kept
  expect_true("(Other)" %in% tbl$Value)
  data_rows <- tbl[!tbl$Value %in% c("Total", "(Other)"), ]
  expect_equal(nrow(data_rows), 2L)
  # The kept values should cumulatively reach >= 50%
  expect_gte(sum(data_rows$Freq) / 32, 0.5)
})

test_that(".cutoff = 0 errors", {
  expect_error(tab(mtcars, cyl, .cutoff = 0), "positive")
})

test_that(".cutoff = -1 errors", {
  expect_error(tab(mtcars, cyl, .cutoff = -1), "positive")
})

# -- Weighted -----------------------------------------------------------------

test_that("weighted tabulation uses .wt sums", {
  result <- tab(mtcars, cyl, .wt = mpg)

  expect_true(result$weighted)
  expect_equal(result$wt_name, "mpg")

  tbl <- result$table
  total <- tbl$Freq[tbl$Value == "Total"]
  expect_equal(total, sum(mtcars$mpg), tolerance = 0.1)
})

test_that("weighted tabulation prints weight name in header", {
  result <- tab(mtcars, cyl, .wt = mpg)
  output <- capture.output(print(result), type = "message")
  expect_true(any(grepl("weighted by mpg", output)))
})

# -- NA handling --------------------------------------------------------------

test_that(".na = 'include' shows NA as category", {
  df <- data.frame(x = c("a", "b", NA, "a"), stringsAsFactors = FALSE)
  result <- tab(df, x, .na = "include")

  expect_equal(result$n_missing, 1L)
  expect_true("<NA>" %in% result$table$Value)
})

test_that(".na = 'exclude' drops NAs", {
  df <- data.frame(x = c("a", "b", NA, "a"), stringsAsFactors = FALSE)
  result <- tab(df, x, .na = "exclude")

  expect_false("<NA>" %in% result$table$Value)
  total <- result$table$Freq[result$table$Value == "Total"]
  expect_equal(total, 3)
})

test_that(".na = 'only' shows only NA rows", {
  df <- data.frame(x = c("a", "b", NA, "a"), stringsAsFactors = FALSE)
  result <- tab(df, x, .na = "only")
  tbl <- result$table[result$table$Value != "Total", ]
  expect_equal(nrow(tbl), 1L)
  expect_equal(tbl$Value, "<NA>")
})

# -- Two-way crosstabulation --------------------------------------------------

test_that("two-way tab returns correct structure", {
  result <- tab(mtcars, cyl, gear)

  expect_s3_class(result, "tidyaudit_tab")
  expect_equal(result$type, "twoway")
  expect_equal(result$var1_name, "cyl")
  expect_equal(result$var2_name, "gear")
  expect_equal(result$display, "count")
  expect_equal(result$grand_total, 32)
})

test_that("two-way tab matrix has correct dimensions", {
  result <- tab(mtcars, cyl, gear)
  mat <- result$matrix

  # 3 cyl values + Total row = 4 rows
  expect_equal(nrow(mat), 4L)
  # cyl + 3 gear values + Total = 5 columns
  expect_equal(ncol(mat), 5L)
  expect_true("Total" %in% names(mat))
  expect_true("Total" %in% mat[[1L]])
})

test_that("two-way tab prints without error", {
  result <- tab(mtcars, cyl, gear)
  output <- capture.output(print(result), type = "message")
  expect_true(any(grepl("Crosstabulation", output)))
})

test_that(".display = 'row_pct' shows percentages", {
  result <- tab(mtcars, cyl, gear, .display = "row_pct")
  mat <- result$matrix
  # First data row, second column (first gear value) should contain %
  cell <- mat[1, 2]
  expect_true(grepl("%", cell))
})

test_that(".display = 'col_pct' works", {
  result <- tab(mtcars, cyl, gear, .display = "col_pct")
  mat <- result$matrix
  cell <- mat[1, 2]
  expect_true(grepl("%", cell))
})

test_that(".display = 'total_pct' works", {
  result <- tab(mtcars, cyl, gear, .display = "total_pct")
  mat <- result$matrix
  cell <- mat[1, 2]
  expect_true(grepl("%", cell))
})

test_that("two-way weighted works", {
  result <- tab(mtcars, cyl, gear, .wt = mpg)
  expect_true(result$weighted)
  expect_equal(result$grand_total, sum(mtcars$mpg), tolerance = 0.1)
})

# -- as.data.frame ------------------------------------------------------------

test_that("as.data.frame one-way returns data without Total", {
  result <- tab(mtcars, cyl)
  df <- as.data.frame(result)

  expect_false("Total" %in% df$Value)
  expect_equal(nrow(df), 3L)
  expect_true("Freq" %in% names(df))
})

test_that("as.data.frame two-way returns numeric counts without Total", {
  result <- tab(mtcars, cyl, gear)
  df <- as.data.frame(result)

  expect_false("Total" %in% df[[1L]])
  expect_false("Total" %in% names(df))
  expect_equal(nrow(df), 3L)
  # Values should be numeric, not formatted strings
  expect_true(is.numeric(df[[2L]]))
})

# -- Edge cases ---------------------------------------------------------------

test_that("tab errors on 0 variables", {
  expect_error(tab(mtcars), "At least one variable")
})

test_that("tab errors on 3+ variables", {
  expect_error(tab(mtcars, cyl, gear, am), "at most 2")
})

test_that("tab errors on non-existent column", {
  expect_error(tab(mtcars, nonexistent), "not found")
})

test_that("tab errors on non-data.frame", {
  expect_error(tab(1:10, x), "data.frame")
})

test_that("tab works with factor variables", {
  df <- data.frame(
    x = factor(c("b", "a", "c", "a", "b"), levels = c("c", "b", "a"))
  )
  result <- tab(df, x, .sort = "value_asc")
  tbl <- result$table[result$table$Value != "Total", ]
  # Factor level order: c, b, a
  expect_equal(tbl$Value[1], "c")
  expect_equal(tbl$Value[2], "b")
  expect_equal(tbl$Value[3], "a")
})

test_that("tab with all NAs and .na = 'exclude' produces empty table", {
  df <- data.frame(x = c(NA, NA, NA))
  result <- tab(df, x, .na = "exclude")
  tbl <- result$table[result$table$Value != "Total", ]
  expect_equal(nrow(tbl), 0L)
})

test_that("high-cardinality warning fires", {
  df <- data.frame(x = as.character(seq_len(60)), stringsAsFactors = FALSE)
  expect_warning(tab(df, x), "unique values")
})

test_that("n_missing counts rows where any variable is NA", {
  df <- data.frame(
    x = c("a", NA,  "b", "a", NA),
    y = c(NA,  "p", "q", "p", NA),
    stringsAsFactors = FALSE
  )
  # Rows with any NA: 1 (y=NA), 2 (x=NA), 5 (both NA) = 3 rows
  result <- tab(df, x, y)
  expect_equal(result$n_missing, 3L)
})

test_that("numeric values sort correctly with .na = 'include'", {
  df <- data.frame(x = c(2, 10, 100, 2, NA, 10))
  result <- tab(df, x, .sort = "value_asc", .na = "include")
  tbl <- result$table[result$table$Value != "Total", ]
  # <NA> should be last; numeric values should sort as 2, 10, 100 (not "10", "100", "2")
  non_na <- tbl$Value[tbl$Value != "<NA>"]
  expect_equal(non_na, c("2", "10", "100"))
})

test_that("piping works: df |> tab(var)", {
  result <- mtcars |> tab(cyl)
  expect_s3_class(result, "tidyaudit_tab")
  expect_equal(result$table$Freq[result$table$Value == "Total"], 32)
})

#' Record a Pipeline Snapshot
#'
#' Transparent pipe pass-through that captures a metadata snapshot and appends
#' it to an audit trail. Returns `.data` unchanged — the function's only purpose
#' is its side effect on `.trail`.
#'
#' @param .data A data.frame or tibble flowing through the pipe.
#' @param .trail An [audit_trail()] object.
#' @param .label Optional character label for this snapshot. If `NULL`, an
#'   auto-generated label like `"step_1"` is used.
#' @param .fns Optional named list of diagnostic functions (or formula lambdas)
#'   to run on `.data`. Results are stored in the snapshot.
#' @param .numeric_summary Logical. If `FALSE`, skip numeric summary
#'   computation in the snapshot (default `TRUE`).
#' @param .cols_include Character vector of column names to include in the
#'   snapshot schema, or `NULL` (the default) to include all columns. Mutually
#'   exclusive with `.cols_exclude`.
#' @param .cols_exclude Character vector of column names to exclude from the
#'   snapshot schema, or `NULL` (the default). Mutually exclusive with
#'   `.cols_include`.
#'
#' @returns `.data`, unchanged, returned invisibly. The function is a
#'   transparent pass-through; its only effect is the side effect on `.trail`.
#'
#' @examples
#' trail <- audit_trail("example")
#' result <- mtcars |>
#'   audit_tap(trail, "raw") |>
#'   dplyr::filter(mpg > 20) |>
#'   audit_tap(trail, "filtered")
#' print(trail)
#'
#' @family audit trail
#' @export
audit_tap <- function(.data, .trail, .label = NULL, .fns = NULL,
                      .numeric_summary = TRUE,
                      .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)

  # Validate arguments that don't require .data evaluation
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }
  if (!is.null(.label)) {
    if (!is.character(.label) || length(.label) != 1L || is.na(.label)) {
      cli::cli_abort("{.arg .label} must be a single character string or NULL.")
    }
  }
  if (!is.null(.fns) && !is.list(.fns)) {
    cli::cli_abort("{.arg .fns} must be a named list of functions or formulas.")
  }

  # Force .data evaluation — with nested |> pipes, inner audit_tap()
  # calls execute as a side effect of forcing .data. We must force before
  # computing index or checking labels so that the trail state reflects
  # all inner taps.
  force(.data)

  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble, not {.cls {class(.data)}}.")
  }

  # Auto-generate label if not provided
  if (is.null(.label)) {
    .label <- paste0("step_", length(.trail$snapshots) + 1L)
  }

  # Check label uniqueness
  if (.label %in% .trail$labels) {
    trail_name <- .trail$name
    label_val <- .label
    cli::cli_abort("Label {.val {label_val}} already exists in trail {.val {trail_name}}.")
  }

  # Build snapshot (index computed after forcing .data)
  index <- length(.trail$snapshots) + 1L
  snap <- .build_snapshot(.data, label = .label, index = index,
                          .numeric_summary = .numeric_summary,
                          .cols_include = .cols_include,
                          .cols_exclude = .cols_exclude)

  # Capture pipeline (best-effort)
  snap$pipeline <- tryCatch(.capture_pipeline(data_expr), error = function(e) NULL)

  # Run custom diagnostic functions if provided
  if (!is.null(.fns)) {
    # Auto-name unnamed entries so they appear in report output
    if (is.null(names(.fns))) {
      names(.fns) <- paste0("fn_", seq_along(.fns))
    } else {
      empty <- is.na(names(.fns)) | names(.fns) == ""
      if (any(empty)) {
        names(.fns)[empty] <- paste0("fn_", which(empty))
      }
    }
    snap$custom <- lapply(.fns, function(fn) {
      fn <- rlang::as_function(fn)
      fn(.data)
    })
  }

  # Smart change detection (compare to previous snapshot)
  if (index > 1L) {
    prev <- .trail$snapshots[[index - 1L]]
    snap$changes <- .detect_changes(prev, snap)
  }

  # Mutate trail (reference semantics — side effect)
  .trail$snapshots[[index]] <- snap
  .trail$labels <- c(.trail$labels, .label)

  # Return data unchanged, invisibly (side-effect-only function)
  invisible(.data)
}

#' Detect Changes Between Two Consecutive Snapshots
#'
#' Computes row, column, and NA deltas between two snapshots.
#'
#' @param prev Previous `audit_snap`.
#' @param curr Current `audit_snap`.
#'
#' @returns A list of change metrics.
#'
#' @noRd
.detect_changes <- function(prev, curr) {
  list(
    row_delta    = curr$nrow - prev$nrow,
    col_delta    = curr$ncol - prev$ncol,
    na_delta     = curr$total_nas - prev$total_nas,
    cols_added   = setdiff(curr$all_columns, prev$all_columns),
    cols_removed = setdiff(prev$all_columns, curr$all_columns),
    type_changes = .detect_type_changes(prev$schema, curr$schema)
  )
}

#' Detect Column Type Changes Between Snapshots
#'
#' @param prev_schema Schema data.frame from previous snapshot.
#' @param curr_schema Schema data.frame from current snapshot.
#'
#' @returns A data.frame of type changes, or `NULL` if none.
#'
#' @noRd
.detect_type_changes <- function(prev_schema, curr_schema) {
  common <- intersect(prev_schema$column, curr_schema$column)
  if (length(common) == 0L) return(NULL)

  prev_types <- setNames(prev_schema$type, prev_schema$column)[common]
  curr_types <- setNames(curr_schema$type, curr_schema$column)[common]

  changed <- common[prev_types != curr_types]
  if (length(changed) == 0L) return(NULL)

  data.frame(
    column = changed,
    from   = prev_types[changed],
    to     = curr_types[changed],
    stringsAsFactors = FALSE,
    row.names = NULL
  )
}

#' Operation-Aware Join Taps
#'
#' Performs a dplyr join AND records enriched diagnostics in an audit trail.
#' These functions replace the pattern of wrapping a join with two
#' [audit_tap()] calls, capturing information that plain taps cannot:
#' match rates, relationship type, duplicate keys, and unmatched row counts.
#'
#' @param .data A data.frame or tibble (left table in the join).
#' @param y A data.frame or tibble (right table in the join).
#' @param ... Arguments passed to the corresponding `dplyr::*_join()` function,
#'   including `by`, `suffix`, `keep`, `multiple`, `unmatched`, etc. The `by`
#'   argument should be passed by name for enriched diagnostics.
#' @param .trail An [audit_trail()] object, or `NULL` (the default). When
#'   `NULL`, behavior depends on `.stat`: if `.stat` is also `NULL`, a plain
#'   dplyr join is performed; if `.stat` is provided, [validate_join()]
#'   diagnostics are printed before the join.
#' @param .label Optional character label for this snapshot. If `NULL`,
#'   auto-generated as `"left_join_1"` etc.
#' @param .stat An unquoted column name for stat tracking, e.g., `amount`.
#'   Passed to [validate_join()].
#' @param .numeric_summary Logical. If `FALSE`, skip numeric summary
#'   computation in the snapshot (default `TRUE`).
#' @param .cols_include Character vector of column names to include in the
#'   snapshot schema, or `NULL` (the default) to include all columns. Mutually
#'   exclusive with `.cols_exclude`.
#' @param .cols_exclude Character vector of column names to exclude from the
#'   snapshot schema, or `NULL` (the default). Mutually exclusive with
#'   `.cols_include`.
#'
#' @details
#' Enriched diagnostics (match rates, relationship type, duplicate keys) require
#' equality joins — `by` as a character vector, named character vector, or
#' simple equality `join_by()` expression (e.g., `join_by(id)`,
#' `join_by(a == b)`). For non-equi `join_by()` expressions, the tap records
#' a basic snapshot without match-rate diagnostics.
#'
#' All dplyr join features (`join_by`, `multiple`, `unmatched`, `suffix`, etc.)
#' work unchanged via `...`.
#'
#' When `.trail` is `NULL`:
#' \itemize{
#'   \item `.stat` also `NULL`: plain dplyr join
#'   \item `.stat` provided: prints [validate_join()] diagnostics, then joins
#'   \item `.label` provided: warns that label is ignored
#' }
#'
#' @returns The joined data.frame or tibble (same as the corresponding
#'   `dplyr::*_join()`).
#'
#' @examples
#' orders <- data.frame(id = 1:4, amount = c(100, 200, 300, 400))
#' customers <- data.frame(id = c(2, 3, 5), name = c("A", "B", "C"))
#'
#' # With trail
#' trail <- audit_trail("join_example")
#' result <- orders |>
#'   audit_tap(trail, "raw") |>
#'   left_join_tap(customers, by = "id", .trail = trail, .label = "joined")
#' print(trail)
#'
#' # Without trail (plain join)
#' result2 <- left_join_tap(orders, customers, by = "id")
#'
#' @family operation taps
#' @name join_tap
NULL

#' @rdname join_tap
#' @export
left_join_tap <- function(.data, y, ..., .trail = NULL, .label = NULL,
                           .stat = NULL, .numeric_summary = TRUE,
                           .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  y_name <- deparse(substitute(y))
  stat_quo <- rlang::enquo(.stat)
  stat_str <- if (rlang::quo_is_null(stat_quo)) {
    NULL
  } else {
    expr <- rlang::quo_get_expr(stat_quo)
    if (is.character(expr)) expr else rlang::as_label(stat_quo)
  }

  if (is.null(.trail)) {
    return(.join_tap_no_trail(.data, y, ..., .label = .label, .stat = stat_str,
                               .join_fn = dplyr::left_join))
  }

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  force(.data)

  .join_tap_impl(.data, y, ..., .trail = .trail, .label = .label,
                  .stat = stat_str, .join_fn = dplyr::left_join,
                  .join_type = "left_join", .data_expr = data_expr,
                  .y_name = y_name,
                  .numeric_summary = .numeric_summary,
                  .cols_include = .cols_include,
                  .cols_exclude = .cols_exclude)
}

#' @rdname join_tap
#' @export
right_join_tap <- function(.data, y, ..., .trail = NULL, .label = NULL,
                            .stat = NULL, .numeric_summary = TRUE,
                            .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  y_name <- deparse(substitute(y))
  stat_quo <- rlang::enquo(.stat)
  stat_str <- if (rlang::quo_is_null(stat_quo)) {
    NULL
  } else {
    expr <- rlang::quo_get_expr(stat_quo)
    if (is.character(expr)) expr else rlang::as_label(stat_quo)
  }

  if (is.null(.trail)) {
    return(.join_tap_no_trail(.data, y, ..., .label = .label, .stat = stat_str,
                               .join_fn = dplyr::right_join))
  }

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  force(.data)

  .join_tap_impl(.data, y, ..., .trail = .trail, .label = .label,
                  .stat = stat_str, .join_fn = dplyr::right_join,
                  .join_type = "right_join", .data_expr = data_expr,
                  .y_name = y_name,
                  .numeric_summary = .numeric_summary,
                  .cols_include = .cols_include,
                  .cols_exclude = .cols_exclude)
}

#' @rdname join_tap
#' @export
inner_join_tap <- function(.data, y, ..., .trail = NULL, .label = NULL,
                            .stat = NULL, .numeric_summary = TRUE,
                            .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  y_name <- deparse(substitute(y))
  stat_quo <- rlang::enquo(.stat)
  stat_str <- if (rlang::quo_is_null(stat_quo)) {
    NULL
  } else {
    expr <- rlang::quo_get_expr(stat_quo)
    if (is.character(expr)) expr else rlang::as_label(stat_quo)
  }

  if (is.null(.trail)) {
    return(.join_tap_no_trail(.data, y, ..., .label = .label, .stat = stat_str,
                               .join_fn = dplyr::inner_join))
  }

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  force(.data)

  .join_tap_impl(.data, y, ..., .trail = .trail, .label = .label,
                  .stat = stat_str, .join_fn = dplyr::inner_join,
                  .join_type = "inner_join", .data_expr = data_expr,
                  .y_name = y_name,
                  .numeric_summary = .numeric_summary,
                  .cols_include = .cols_include,
                  .cols_exclude = .cols_exclude)
}

#' @rdname join_tap
#' @export
full_join_tap <- function(.data, y, ..., .trail = NULL, .label = NULL,
                           .stat = NULL, .numeric_summary = TRUE,
                           .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  y_name <- deparse(substitute(y))
  stat_quo <- rlang::enquo(.stat)
  stat_str <- if (rlang::quo_is_null(stat_quo)) {
    NULL
  } else {
    expr <- rlang::quo_get_expr(stat_quo)
    if (is.character(expr)) expr else rlang::as_label(stat_quo)
  }

  if (is.null(.trail)) {
    return(.join_tap_no_trail(.data, y, ..., .label = .label, .stat = stat_str,
                               .join_fn = dplyr::full_join))
  }

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  force(.data)

  .join_tap_impl(.data, y, ..., .trail = .trail, .label = .label,
                  .stat = stat_str, .join_fn = dplyr::full_join,
                  .join_type = "full_join", .data_expr = data_expr,
                  .y_name = y_name,
                  .numeric_summary = .numeric_summary,
                  .cols_include = .cols_include,
                  .cols_exclude = .cols_exclude)
}

#' @rdname join_tap
#' @export
anti_join_tap <- function(.data, y, ..., .trail = NULL, .label = NULL,
                           .stat = NULL, .numeric_summary = TRUE,
                           .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  y_name <- deparse(substitute(y))
  stat_quo <- rlang::enquo(.stat)
  stat_str <- if (rlang::quo_is_null(stat_quo)) {
    NULL
  } else {
    expr <- rlang::quo_get_expr(stat_quo)
    if (is.character(expr)) expr else rlang::as_label(stat_quo)
  }

  if (is.null(.trail)) {
    return(.join_tap_no_trail(.data, y, ..., .label = .label, .stat = stat_str,
                               .join_fn = dplyr::anti_join))
  }

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  force(.data)

  .join_tap_impl(.data, y, ..., .trail = .trail, .label = .label,
                  .stat = stat_str, .join_fn = dplyr::anti_join,
                  .join_type = "anti_join", .data_expr = data_expr,
                  .y_name = y_name,
                  .numeric_summary = .numeric_summary,
                  .cols_include = .cols_include,
                  .cols_exclude = .cols_exclude)
}

#' @rdname join_tap
#' @export
semi_join_tap <- function(.data, y, ..., .trail = NULL, .label = NULL,
                           .stat = NULL, .numeric_summary = TRUE,
                           .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  y_name <- deparse(substitute(y))
  stat_quo <- rlang::enquo(.stat)
  stat_str <- if (rlang::quo_is_null(stat_quo)) {
    NULL
  } else {
    expr <- rlang::quo_get_expr(stat_quo)
    if (is.character(expr)) expr else rlang::as_label(stat_quo)
  }

  if (is.null(.trail)) {
    return(.join_tap_no_trail(.data, y, ..., .label = .label, .stat = stat_str,
                               .join_fn = dplyr::semi_join))
  }

  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  force(.data)

  .join_tap_impl(.data, y, ..., .trail = .trail, .label = .label,
                  .stat = stat_str, .join_fn = dplyr::semi_join,
                  .join_type = "semi_join", .data_expr = data_expr,
                  .y_name = y_name,
                  .numeric_summary = .numeric_summary,
                  .cols_include = .cols_include,
                  .cols_exclude = .cols_exclude)
}

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------

#' Normalize `by` Into a Named Character Vector
#'
#' Converts `by` into a named character vector suitable for [validate_join()].
#' Returns `NULL` for non-equi join expressions or unrecognised types.
#'
#' @param by A character vector, named character vector, `dplyr_join_by`
#'   object, or `NULL`.
#'
#' @returns A named character vector, or `NULL`.
#'
#' @noRd
.normalize_by <- function(by) {
  if (is.null(by)) return(NULL)

  if (is.character(by)) return(by)

  if (inherits(by, "dplyr_join_by")) {
    return(tryCatch({
      if (all(by$condition == "==")) {
        setNames(by$y, by$x)
      } else {
        NULL
      }
    }, error = function(e) NULL))
  }

  NULL
}

#' Handle NULL-Trail Join Case
#'
#' When `.trail` is NULL, perform a plain join with optional validate_join
#' diagnostics.
#'
#' @noRd
.join_tap_no_trail <- function(.data, y, ..., .label, .stat, .join_fn) {
  if (!is.null(.label)) {
    cli::cli_warn("{.arg .label} is ignored when {.arg .trail} is NULL.")
  }
  if (!is.null(.stat)) {
    by_norm <- .normalize_by(list(...)$by)
    if (!is.null(by_norm)) {
      vj <- tryCatch(
        validate_join(.data, y, by = by_norm, stat = .stat),
        error = function(e) {
          err_msg <- conditionMessage(e)
          cli::cli_warn("validate_join() failed: {err_msg}")
          NULL
        }
      )
      if (!is.null(vj)) print(vj)
    }
  }
  .join_fn(.data, y, ...)
}

#' Shared Implementation for Join Taps
#'
#' Internal helper used by all six `*_join_tap()` functions. Handles
#' diagnostics, snapshot creation, and trail recording.
#'
#' @param .data Data.frame (already forced by the caller).
#' @param y Right table.
#' @param ... Forwarded dplyr join arguments.
#' @param .trail An audit_trail (validated by the caller).
#' @param .label Label or NULL.
#' @param .stat Stat column name or NULL.
#' @param .join_fn The dplyr join function.
#' @param .join_type Character name of the join type.
#' @param .data_expr Unevaluated expression for pipeline capture.
#' @param .y_name Deparsed name of `y`.
#'
#' @returns The join result.
#'
#' @noRd
.join_tap_impl <- function(.data, y, ..., .trail, .label, .stat,
                            .join_fn, .join_type, .data_expr, .y_name,
                            .numeric_summary = TRUE,
                            .cols_include = NULL, .cols_exclude = NULL) {
  # Validate .data
  if (!is.data.frame(.data)) {
    cli::cli_abort(
      "{.arg .data} must be a data.frame or tibble, not {.cls {class(.data)}}."
    )
  }

  # Validate .label
  if (!is.null(.label)) {
    if (!is.character(.label) || length(.label) != 1L || is.na(.label)) {
      cli::cli_abort("{.arg .label} must be a single character string or NULL.")
    }
  }

  # Extract and normalize `by` for enriched diagnostics
  by_norm <- .normalize_by(list(...)$by)

  # Run validate_join if possible (best-effort)
  # Direct call preserves deparse(substitute()) inside validate_join
  vj <- NULL
  if (!is.null(by_norm)) {
    vj <- if (!is.null(.stat)) {
      tryCatch(validate_join(.data, y, by = by_norm, stat = .stat),
               error = function(e) NULL)
    } else {
      tryCatch(validate_join(.data, y, by = by_norm),
               error = function(e) NULL)
    }
  }

  # Perform the actual join
  result <- .join_fn(.data, y, ...)

  # Auto-generate label if not provided
  index <- length(.trail$snapshots) + 1L
  if (is.null(.label)) {
    .label <- paste0(.join_type, "_", index)
  }

  # Check label uniqueness
  if (.label %in% .trail$labels) {
    trail_name <- .trail$name
    label_val <- .label
    cli::cli_abort(
      "Label {.val {label_val}} already exists in trail {.val {trail_name}}."
    )
  }

  # Build snapshot of the join result
  snap <- .build_snapshot(result, label = .label, index = index,
                          .numeric_summary = .numeric_summary,
                          .cols_include = .cols_include,
                          .cols_exclude = .cols_exclude)

  # Capture pipeline (best-effort)
  snap$pipeline <- tryCatch(.capture_pipeline(.data_expr), error = function(e) NULL)

  # Enrich the snapshot
  snap$type <- "join"

  if (!is.null(vj)) {
    snap$diagnostics <- list(
      join_type  = .join_type,
      relation   = vj$relation,
      match_rate = list(x = vj$counts$match_rate_x, y = vj$counts$match_rate_y),
      n_only_x   = vj$counts$n_only_x,
      n_only_y   = vj$counts$n_only_y,
      x_has_dups = vj$duplicates$x_has_dups,
      y_has_dups = vj$duplicates$y_has_dups,
      stat       = vj$stat,
      y_name     = .y_name
    )
  } else {
    snap$diagnostics <- list(
      join_type = .join_type,
      y_name    = .y_name
    )
  }

  # Compute changes vs previous snapshot
  if (index > 1L) {
    prev <- .trail$snapshots[[index - 1L]]
    snap$changes <- .detect_changes(prev, snap)
  }

  # Append to trail (reference semantics)
  .trail$snapshots[[index]] <- snap
  .trail$labels <- c(.trail$labels, .label)

  result
}

# ---------------------------------------------------------------------------
# Test data
# ---------------------------------------------------------------------------

orders <- data.frame(
  id     = c(1L, 2L, 3L, 4L),
  amount = c(100, 200, 300, 400),
  stringsAsFactors = FALSE
)

customers <- data.frame(
  id   = c(2L, 3L, 5L),
  name = c("Alice", "Bob", "Charlie"),
  stringsAsFactors = FALSE
)

# ---------------------------------------------------------------------------
# left_join_tap — basic with trail
# ---------------------------------------------------------------------------

test_that("left_join_tap produces correct join result", {
  trail <- audit_trail("lj_basic")
  result <- left_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "joined")

  expected <- dplyr::left_join(orders, customers, by = "id")
  expect_equal(result, expected)
})

test_that("left_join_tap records enriched snapshot in trail", {
  trail <- audit_trail("lj_enriched")
  result <- left_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "joined")

  expect_equal(length(trail$snapshots), 1L)
  expect_equal(trail$labels, "joined")

  snap <- trail$snapshots[[1]]
  expect_s3_class(snap, "audit_snap")
  expect_equal(snap$type, "join")
  expect_equal(snap$label, "joined")

  diag <- snap$diagnostics
  expect_equal(diag$join_type, "left_join")
  expect_equal(diag$relation, "one-to-one")
  expect_true(!is.null(diag$match_rate$x))
  expect_true(!is.null(diag$match_rate$y))
})

test_that("left_join_tap with stat tracks stat column", {
  # stat requires the column to exist in both tables
  x_stat <- data.frame(id = 1:4, amount = c(100, 200, 300, 400))
  y_stat <- data.frame(id = c(2L, 3L, 5L), amount = c(50, 75, 100))
  trail <- audit_trail("lj_stat")
  result <- left_join_tap(x_stat, y_stat, by = "id",
                           .trail = trail, .label = "joined",
                           .stat = amount)

  snap <- trail$snapshots[[1]]
  expect_true(!is.null(snap$diagnostics$stat))
})

test_that("left_join_tap accepts .stat as quoted string (tolerant API)", {
  x_stat <- data.frame(id = 1:4, amount = c(100, 200, 300, 400))
  y_stat <- data.frame(id = c(2L, 3L, 5L), amount = c(50, 75, 100))
  trail <- audit_trail("lj_stat_str")
  result <- left_join_tap(x_stat, y_stat, by = "id",
                           .trail = trail, .label = "joined",
                           .stat = "amount")

  snap <- trail$snapshots[[1]]
  expect_true(!is.null(snap$diagnostics$stat))
})

# ---------------------------------------------------------------------------
# All six join types
# ---------------------------------------------------------------------------

test_that("right_join_tap produces correct result and enriched snapshot", {
  trail <- audit_trail("rj")
  result <- right_join_tap(orders, customers, by = "id",
                            .trail = trail, .label = "rj")
  expected <- dplyr::right_join(orders, customers, by = "id")
  expect_equal(result, expected)
  expect_equal(trail$snapshots[[1]]$diagnostics$join_type, "right_join")
})

test_that("inner_join_tap produces correct result and enriched snapshot", {
  trail <- audit_trail("ij")
  result <- inner_join_tap(orders, customers, by = "id",
                            .trail = trail, .label = "ij")
  expected <- dplyr::inner_join(orders, customers, by = "id")
  expect_equal(result, expected)
  expect_equal(trail$snapshots[[1]]$diagnostics$join_type, "inner_join")
})

test_that("full_join_tap produces correct result and enriched snapshot", {
  trail <- audit_trail("fj")
  result <- full_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "fj")
  expected <- dplyr::full_join(orders, customers, by = "id")
  expect_equal(result, expected)
  expect_equal(trail$snapshots[[1]]$diagnostics$join_type, "full_join")
})

test_that("anti_join_tap produces correct result and enriched snapshot", {
  trail <- audit_trail("aj")
  result <- anti_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "aj")
  expected <- dplyr::anti_join(orders, customers, by = "id")
  expect_equal(result, expected)
  expect_equal(trail$snapshots[[1]]$diagnostics$join_type, "anti_join")
})

test_that("semi_join_tap produces correct result and enriched snapshot", {
  trail <- audit_trail("sj")
  result <- semi_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "sj")
  expected <- dplyr::semi_join(orders, customers, by = "id")
  expect_equal(result, expected)
  expect_equal(trail$snapshots[[1]]$diagnostics$join_type, "semi_join")
})

# ---------------------------------------------------------------------------
# inner_join_tap row loss
# ---------------------------------------------------------------------------

test_that("inner_join_tap captures partial overlap diagnostics", {
  trail <- audit_trail("ij_partial")
  result <- inner_join_tap(orders, customers, by = "id",
                            .trail = trail, .label = "inner")
  diag <- trail$snapshots[[1]]$diagnostics
  expect_true(diag$n_only_x > 0)
  expect_true(diag$n_only_y > 0)
})

# ---------------------------------------------------------------------------
# Passthrough of dplyr join args via ...
# ---------------------------------------------------------------------------

test_that("join_tap passes suffix and other args through", {
  trail <- audit_trail("suffix_test")
  x <- data.frame(id = 1:3, value = 10:12)
  y <- data.frame(id = 2:4, value = 20:22)
  result <- left_join_tap(x, y, by = "id", suffix = c("_left", "_right"),
                           .trail = trail, .label = "sfx")
  expect_true("value_left" %in% names(result))
  expect_true("value_right" %in% names(result))
})

# ---------------------------------------------------------------------------
# Join tap without trail
# ---------------------------------------------------------------------------

test_that("join_tap without trail performs plain dplyr join", {
  result <- left_join_tap(orders, customers, by = "id")
  expected <- dplyr::left_join(orders, customers, by = "id")
  expect_equal(result, expected)
})

test_that("join_tap without trail + .stat prints diagnostics", {
  # stat requires the column in both tables
  x_stat <- data.frame(id = 1:4, amount = c(100, 200, 300, 400))
  y_stat <- data.frame(id = c(2L, 3L, 5L), amount = c(50, 75, 100))
  output <- capture.output(
    result <- left_join_tap(x_stat, y_stat, by = "id", .stat = amount),
    type = "message"
  )
  expected <- dplyr::left_join(x_stat, y_stat, by = "id")
  expect_equal(result, expected)
  # validate_join output should appear
  expect_true(any(grepl("Join Validation|Relationship|Match rate", output)))
})

test_that("join_tap without trail + .label warns", {
  expect_warning(
    left_join_tap(orders, customers, by = "id", .label = "foo"),
    "ignored"
  )
})

# ---------------------------------------------------------------------------
# join_by() equality — enriched diagnostics
# ---------------------------------------------------------------------------

test_that("join_tap with join_by() equality produces enriched diagnostics", {
  trail <- audit_trail("jb_eq")
  result <- left_join_tap(orders, customers,
                           by = dplyr::join_by(id),
                           .trail = trail, .label = "jb")
  diag <- trail$snapshots[[1]]$diagnostics
  expect_equal(diag$join_type, "left_join")
  expect_true(!is.null(diag$relation))
  expect_true(!is.null(diag$match_rate$x))
})

test_that("join_tap with join_by(a == b) produces enriched diagnostics", {
  x <- data.frame(key_a = 1:3, val = 10:12)
  y <- data.frame(key_b = 2:4, score = 20:22)
  trail <- audit_trail("jb_eq2")
  result <- left_join_tap(x, y,
                           by = dplyr::join_by(key_a == key_b),
                           .trail = trail, .label = "jb2")
  diag <- trail$snapshots[[1]]$diagnostics
  expect_true(!is.null(diag$relation))
})

# ---------------------------------------------------------------------------
# join_by() non-equi — basic snapshot only
# ---------------------------------------------------------------------------

test_that("join_tap with non-equi join_by produces basic snapshot", {
  x <- data.frame(id = 1:3, start = c(1, 5, 10))
  y <- data.frame(id = 1:3, threshold = c(2, 4, 8))
  trail <- audit_trail("jb_neq")
  result <- left_join_tap(x, y,
                           by = dplyr::join_by(id, start >= threshold),
                           .trail = trail, .label = "neq")
  snap <- trail$snapshots[[1]]
  expect_equal(snap$type, "join")
  diag <- snap$diagnostics
  expect_equal(diag$join_type, "left_join")
  # No relation or match_rate for non-equi
  expect_null(diag$relation)
  expect_null(diag$match_rate)
})

# ---------------------------------------------------------------------------
# Label management
# ---------------------------------------------------------------------------

test_that("join_tap auto-generates labels", {
  trail <- audit_trail("auto_lbl")
  left_join_tap(orders, customers, by = "id", .trail = trail)
  expect_equal(trail$labels, "left_join_1")
})

test_that("join_tap rejects duplicate labels", {
  trail <- audit_trail("dup_lbl")
  left_join_tap(orders, customers, by = "id", .trail = trail, .label = "j")
  expect_error(
    left_join_tap(orders, customers, by = "id", .trail = trail, .label = "j"),
    "already exists"
  )
})

test_that("join_tap rejects invalid trail", {
  expect_error(
    left_join_tap(orders, customers, by = "id", .trail = "not_a_trail"),
    "audit_trail"
  )
})

# ---------------------------------------------------------------------------
# Change detection
# ---------------------------------------------------------------------------

test_that("join_tap computes changes from previous snapshot", {
  trail <- audit_trail("changes")
  orders |> audit_tap(trail, "raw")
  left_join_tap(orders, customers, by = "id", .trail = trail, .label = "joined")

  snap2 <- trail$snapshots[[2]]
  expect_true(!is.null(snap2$changes))
  expect_true(snap2$changes$col_delta > 0L)  # join adds columns
})

# ---------------------------------------------------------------------------
# Trail timeline with mixed types
# ---------------------------------------------------------------------------

test_that("mixed pipeline with audit_tap + join_tap + filter_tap renders correctly", {
  trail <- audit_trail("mixed")
  result <- orders |>
    audit_tap(trail, "raw") |>
    left_join_tap(customers, by = "id", .trail = trail, .label = "joined") |>
    dplyr::filter(amount > 100) |>
    audit_tap(trail, "filtered")

  expect_equal(length(trail$snapshots), 3L)
  expect_equal(trail$snapshots[[1]]$type, "tap")
  expect_equal(trail$snapshots[[2]]$type, "join")
  expect_equal(trail$snapshots[[3]]$type, "tap")

  # Print should work without error
  output <- capture.output(print(trail), type = "message")
  expect_true(any(grepl("left_join", output)))
})

# ---------------------------------------------------------------------------
# audit_diff across typed snapshots
# ---------------------------------------------------------------------------

test_that("audit_diff works across join and tap snapshots", {
  trail <- audit_trail("diff_typed")
  orders |> audit_tap(trail, "raw")
  left_join_tap(orders, customers, by = "id", .trail = trail, .label = "joined")

  diff_obj <- audit_diff(trail, "raw", "joined")
  expect_s3_class(diff_obj, "audit_diff")
  expect_true(diff_obj$col_delta > 0L)
})

# ---------------------------------------------------------------------------
# Pipeline capture in join_tap
# ---------------------------------------------------------------------------

test_that("join_tap captures pipeline", {
  trail <- audit_trail("pipeline")
  result <- orders |>
    dplyr::mutate(x = 1) |>
    left_join_tap(customers, by = "id", .trail = trail, .label = "j")

  snap <- trail$snapshots[[1]]
  expect_true(!is.null(snap$pipeline))
  expect_true(length(snap$pipeline) > 0L)
})

# ---------------------------------------------------------------------------
# Grouped tibble handling
# ---------------------------------------------------------------------------

test_that("join_tap preserves grouping on left table", {
  grouped_orders <- dplyr::group_by(orders, id)
  trail <- audit_trail("grouped")
  result <- left_join_tap(grouped_orders, customers, by = "id",
                           .trail = trail, .label = "gj")

  # Result should be a grouped tibble (dplyr preserves groups on left)
  expect_true(dplyr::is_grouped_df(result))
})

# ---------------------------------------------------------------------------
# Named by vector (different column names)
# ---------------------------------------------------------------------------

test_that("join_tap works with named by vector", {
  x <- data.frame(key_x = 1:3, val = 10:12)
  y <- data.frame(key_y = 2:4, score = 20:22)
  trail <- audit_trail("named_by")
  result <- left_join_tap(x, y, by = c("key_x" = "key_y"),
                           .trail = trail, .label = "nb")
  expected <- dplyr::left_join(x, y, by = c("key_x" = "key_y"))
  expect_equal(result, expected)

  diag <- trail$snapshots[[1]]$diagnostics
  expect_true(!is.null(diag$relation))
})

# ---------------------------------------------------------------------------
# Duplicate key detection
# ---------------------------------------------------------------------------

test_that("join_tap detects duplicate keys", {
  x <- data.frame(id = c(1L, 2L, 2L, 3L), val = 1:4)
  y <- data.frame(id = c(2L, 3L), score = c(10, 20))
  trail <- audit_trail("dups")
  result <- left_join_tap(x, y, by = "id", .trail = trail, .label = "d")

  diag <- trail$snapshots[[1]]$diagnostics
  expect_true(diag$x_has_dups)
  expect_false(diag$y_has_dups)
  expect_equal(diag$relation, "many-to-one")
})

# ---------------------------------------------------------------------------
# .normalize_by internal helper
# ---------------------------------------------------------------------------

test_that(".normalize_by handles character vectors", {
  expect_equal(tidyaudit:::.normalize_by("id"), "id")
  expect_equal(tidyaudit:::.normalize_by(c("a" = "b")), c("a" = "b"))
  expect_equal(tidyaudit:::.normalize_by(c("x", "y")), c("x", "y"))
})

test_that(".normalize_by handles NULL", {
  expect_null(tidyaudit:::.normalize_by(NULL))
})

test_that(".normalize_by handles equality join_by", {
  jb <- dplyr::join_by(id)
  result <- tidyaudit:::.normalize_by(jb)
  expect_equal(result, c("id" = "id"))

  jb2 <- dplyr::join_by(a == b)
  result2 <- tidyaudit:::.normalize_by(jb2)
  expect_equal(result2, c("a" = "b"))
})

test_that(".normalize_by returns NULL for non-equi join_by", {
  jb <- dplyr::join_by(x >= y)
  expect_null(tidyaudit:::.normalize_by(jb))
})

test_that(".normalize_by returns NULL for unknown types", {
  expect_null(tidyaudit:::.normalize_by(42))
  expect_null(tidyaudit:::.normalize_by(list(a = 1)))
})

# ---------------------------------------------------------------------------
# Multiple snapshots via pipe chain
# ---------------------------------------------------------------------------

test_that("multiple join_tap calls in a pipe chain work correctly", {
  lookup_a <- data.frame(id = 1:3, a_val = c("x", "y", "z"))
  lookup_b <- data.frame(id = 2:4, b_val = c(10, 20, 30))

  trail <- audit_trail("chain")
  result <- orders |>
    audit_tap(trail, "raw") |>
    left_join_tap(lookup_a, by = "id", .trail = trail, .label = "join_a") |>
    left_join_tap(lookup_b, by = "id", .trail = trail, .label = "join_b")

  expect_equal(length(trail$snapshots), 3L)
  expect_equal(trail$labels, c("raw", "join_a", "join_b"))

  # Verify correct ordering
  expect_equal(trail$snapshots[[1]]$index, 1L)
  expect_equal(trail$snapshots[[2]]$index, 2L)
  expect_equal(trail$snapshots[[3]]$index, 3L)

  # Changes should be computed
  expect_true(!is.null(trail$snapshots[[2]]$changes))
  expect_true(!is.null(trail$snapshots[[3]]$changes))
})

# ---------------------------------------------------------------------------
# Zero-row tables
# ---------------------------------------------------------------------------

test_that("join_tap handles zero-row tables", {
  empty <- orders[0, ]
  trail <- audit_trail("empty_join")
  result <- left_join_tap(empty, customers, by = "id",
                           .trail = trail, .label = "ej")
  expect_equal(nrow(result), 0L)
  expect_equal(trail$snapshots[[1]]$nrow, 0L)
})

# ---------------------------------------------------------------------------
# y_name capture
# ---------------------------------------------------------------------------

test_that("join_tap captures y_name in diagnostics", {
  trail <- audit_trail("yname")
  left_join_tap(orders, customers, by = "id", .trail = trail, .label = "yn")
  expect_equal(trail$snapshots[[1]]$diagnostics$y_name, "customers")
})

# ---------------------------------------------------------------------------
# .data validation
# ---------------------------------------------------------------------------

test_that("join_tap rejects non-data.frame input", {
  trail <- audit_trail("bad_data")
  expect_error(
    left_join_tap(list(a = 1), customers, by = "id", .trail = trail),
    "data.frame"
  )
  expect_error(
    left_join_tap(42, customers, by = "id", .trail = trail),
    "data.frame"
  )
})

# ---------------------------------------------------------------------------
# Label validation
# ---------------------------------------------------------------------------

test_that("join_tap rejects NA label", {
  trail <- audit_trail("na_label")
  expect_error(
    left_join_tap(orders, customers, by = "id",
                   .trail = trail, .label = NA_character_),
    "single character string"
  )
})

test_that("join_tap rejects multi-element label", {
  trail <- audit_trail("multi_label")
  expect_error(
    left_join_tap(orders, customers, by = "id",
                   .trail = trail, .label = c("a", "b")),
    "single character string"
  )
})

# ---------------------------------------------------------------------------
# Natural join (by = NULL)
# ---------------------------------------------------------------------------

test_that("join_tap with by = NULL performs natural join with basic snapshot", {
  trail <- audit_trail("natural_join")
  result <- left_join_tap(orders, customers, .trail = trail, .label = "nj")

  expected <- suppressMessages(dplyr::left_join(orders, customers, by = "id"))
  expect_equal(result, expected)

  # Natural join => by is NULL => no enriched diagnostics
  snap <- trail$snapshots[[1]]
  expect_equal(snap$type, "join")
  diag <- snap$diagnostics
  expect_equal(diag$join_type, "left_join")
  expect_null(diag$relation)
  expect_null(diag$match_rate)
})

# ---------------------------------------------------------------------------
# validate_join error handling (no-trail mode)
# ---------------------------------------------------------------------------

test_that("join_tap no-trail warns when validate_join fails", {
  # .stat referencing a column that doesn't exist in one table
  expect_warning(
    result <- left_join_tap(orders, customers, by = "id",
                             .stat = nonexistent_col),
    "validate_join.*failed"
  )
  # Join itself should still succeed
  expected <- dplyr::left_join(orders, customers, by = "id")
  expect_equal(result, expected)
})

# ---------------------------------------------------------------------------
# validate_join error handling (with trail)
# ---------------------------------------------------------------------------

test_that("join_tap with trail degrades gracefully when validate_join fails", {
  trail <- audit_trail("vj_fail")
  # .stat referencing a column that doesn't exist
  result <- left_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "graceful",
                           .stat = nonexistent_col)

  # Join result should be correct
  expected <- dplyr::left_join(orders, customers, by = "id")
  expect_equal(result, expected)

  # Snapshot should have basic diagnostics (no enriched)
  snap <- trail$snapshots[[1]]
  expect_equal(snap$type, "join")
  expect_equal(snap$diagnostics$join_type, "left_join")
  expect_null(snap$diagnostics$relation)
})

# ---------------------------------------------------------------------------
# Snapshot controls
# ---------------------------------------------------------------------------

test_that("left_join_tap passes snapshot controls through", {
  trail <- audit_trail("join_controls")
  result <- left_join_tap(orders, customers, by = "id",
                           .trail = trail, .label = "joined",
                           .numeric_summary = FALSE,
                           .cols_include = c("id", "amount"))
  snap <- trail$snapshots[[1]]
  expect_null(snap$numeric_summary)
  expect_equal(snap$schema$column, c("id", "amount"))
  # ncol reflects the actual join result (id, amount, name)
  expect_equal(snap$ncol, 3L)
})

#' Operation-Aware Filter Taps
#'
#' Performs a diagnostic filter AND records filter diagnostics in an audit trail.
#' `filter_tap()` keeps matching rows (like [dplyr::filter()]),
#' `filter_out_tap()` drops matching rows (the inverse).
#'
#' @param .data A data.frame or tibble.
#' @param ... Filter conditions, evaluated in the context of `.data` using
#'   tidy evaluation (same as [dplyr::filter()]).
#' @param .trail An [audit_trail()] object, or `NULL` (the default). When
#'   `NULL`, behavior depends on diagnostic arguments: if none are provided,
#'   a plain `dplyr::filter()` is performed; if `.stat`, `.warn_threshold`, or
#'   `.quiet = TRUE` is provided, delegates to [filter_keep()] or
#'   [filter_drop()].
#' @param .label Optional character label for this snapshot. If `NULL`,
#'   auto-generated as `"filter_1"` etc.
#' @param .stat An unquoted column or expression to total, e.g., `amount`,
#'   `price * qty`. Reports the stat amount dropped and its share of the total.
#' @param .quiet Logical. If `TRUE`, suppress printing diagnostics (default
#'   `FALSE`).
#' @param .warn_threshold Numeric between 0 and 1. If set and the proportion of
#'   dropped rows exceeds this threshold, a warning is issued.
#' @param .numeric_summary Logical. If `FALSE`, skip numeric summary
#'   computation in the snapshot (default `TRUE`).
#' @param .cols_include Character vector of column names to include in the
#'   snapshot schema, or `NULL` (the default) to include all columns. Mutually
#'   exclusive with `.cols_exclude`.
#' @param .cols_exclude Character vector of column names to exclude from the
#'   snapshot schema, or `NULL` (the default). Mutually exclusive with
#'   `.cols_include`.
#'
#' @details
#' When `.trail` is `NULL`:
#' \itemize{
#'   \item No diagnostic args: plain `dplyr::filter()` / `dplyr::filter_out()`
#'   \item Diagnostic args provided: delegates to [filter_keep()] /
#'     [filter_drop()] (prints diagnostics but no trail recording)
#'   \item `.label` provided: warns that label is ignored
#' }
#'
#' @returns The filtered data.frame or tibble.
#'
#' @examples
#' df <- data.frame(id = 1:10, amount = 1:10 * 100, flag = rep(c(TRUE, FALSE), 5))
#'
#' # With trail
#' trail <- audit_trail("filter_example")
#' result <- df |>
#'   audit_tap(trail, "raw") |>
#'   filter_tap(amount > 300, .trail = trail, .label = "big_only")
#' print(trail)
#'
#' # Inverse: drop matching rows
#' trail2 <- audit_trail("filter_out_example")
#' result2 <- df |>
#'   audit_tap(trail2, "raw") |>
#'   filter_out_tap(flag == FALSE, .trail = trail2, .label = "flagged_only")
#' print(trail2)
#'
#' # Without trail (plain filter)
#' result3 <- filter_tap(df, amount > 300)
#'
#' @family operation taps
#' @name filter_tap
NULL

#' @rdname filter_tap
#' @export
filter_tap <- function(.data, ..., .trail = NULL, .label = NULL, .stat = NULL,
                        .quiet = FALSE, .warn_threshold = NULL,
                        .numeric_summary = TRUE,
                        .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  .validate_warn_threshold(.warn_threshold)
  stat_quo <- rlang::enquo(.stat)
  have_stat <- !rlang::quo_is_null(stat_quo)

  # NULL trail handling
  if (is.null(.trail)) {
    if (!is.null(.label)) {
      cli::cli_warn("{.arg .label} is ignored when {.arg .trail} is NULL.")
    }
    has_diag_args <- have_stat || !is.null(.warn_threshold) || .quiet
    if (has_diag_args) {
      return(filter_keep(.data, ..., .stat = !!stat_quo, .quiet = .quiet,
                          .warn_threshold = .warn_threshold))
    }
    return(dplyr::filter(.data, ...))
  }

  # Validate trail
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  # Force .data before reading trail state
  force(.data)

  .filter_tap_impl(.data, rlang::enquos(...), .trail = .trail, .label = .label,
                    .stat_quo = stat_quo, .have_stat = have_stat,
                    .quiet = .quiet, .warn_threshold = .warn_threshold,
                    .filter_fn = dplyr::filter, .filter_type = "keep",
                    .label_prefix = "filter_", .tap_name = "filter_tap",
                    .action = "Dropped", .data_expr = data_expr,
                    .numeric_summary = .numeric_summary,
                    .cols_include = .cols_include,
                    .cols_exclude = .cols_exclude)
}

#' @rdname filter_tap
#' @export
filter_out_tap <- function(.data, ..., .trail = NULL, .label = NULL,
                            .stat = NULL, .quiet = FALSE,
                            .warn_threshold = NULL,
                            .numeric_summary = TRUE,
                            .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)
  .validate_warn_threshold(.warn_threshold)
  stat_quo <- rlang::enquo(.stat)
  have_stat <- !rlang::quo_is_null(stat_quo)

  # NULL trail handling
  if (is.null(.trail)) {
    if (!is.null(.label)) {
      cli::cli_warn("{.arg .label} is ignored when {.arg .trail} is NULL.")
    }
    has_diag_args <- have_stat || !is.null(.warn_threshold) || .quiet
    if (has_diag_args) {
      return(filter_drop(.data, ..., .stat = !!stat_quo, .quiet = .quiet,
                          .warn_threshold = .warn_threshold))
    }
    return(dplyr::filter_out(.data, ...))
  }

  # Validate trail
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }

  # Force .data before reading trail state
  force(.data)

  .filter_tap_impl(.data, rlang::enquos(...), .trail = .trail, .label = .label,
                    .stat_quo = stat_quo, .have_stat = have_stat,
                    .quiet = .quiet, .warn_threshold = .warn_threshold,
                    .filter_fn = dplyr::filter_out, .filter_type = "drop",
                    .label_prefix = "filter_out_", .tap_name = "filter_out_tap",
                    .action = "Removed", .data_expr = data_expr,
                    .numeric_summary = .numeric_summary,
                    .cols_include = .cols_include,
                    .cols_exclude = .cols_exclude)
}

# ---------------------------------------------------------------------------
# Internal helper
# ---------------------------------------------------------------------------

#' Shared Implementation for Filter Taps
#'
#' Internal helper used by both `filter_tap()` and `filter_out_tap()`.
#' Handles diagnostics, snapshot creation, and trail recording.
#'
#' @param .data Data.frame (already forced by the caller).
#' @param dots Quosure list of filter conditions (from `rlang::enquos()`).
#' @param .trail An audit_trail (validated by the caller).
#' @param .label Label or NULL.
#' @param .stat_quo Quosure for the stat column.
#' @param .have_stat Logical indicating whether stat was provided.
#' @param .quiet Logical for suppressing output.
#' @param .warn_threshold Numeric threshold or NULL.
#' @param .filter_fn The dplyr filter function (`dplyr::filter` or
#'   `dplyr::filter_out`).
#' @param .filter_type Character: `"keep"` or `"drop"`.
#' @param .label_prefix Character prefix for auto-generated labels.
#' @param .tap_name Character name for diagnostic messages.
#' @param .action Character verb for diagnostic messages (`"Dropped"` or
#'   `"Removed"`).
#' @param .data_expr Unevaluated expression for pipeline capture.
#'
#' @returns The filtered result.
#'
#' @noRd
.filter_tap_impl <- function(.data, dots, .trail, .label, .stat_quo, .have_stat,
                              .quiet, .warn_threshold, .filter_fn, .filter_type,
                              .label_prefix, .tap_name, .action, .data_expr,
                              .numeric_summary = TRUE,
                              .cols_include = NULL, .cols_exclude = NULL) {
  # Validate .data
  if (!is.data.frame(.data)) {
    cli::cli_abort(
      "{.arg .data} must be a data.frame or tibble, not {.cls {class(.data)}}."
    )
  }

  # Validate .label
  if (!is.null(.label)) {
    if (!is.character(.label) || length(.label) != 1L || is.na(.label)) {
      cli::cli_abort("{.arg .label} must be a single character string or NULL.")
    }
  }

  # Format expression label
  expr_lab <- paste(vapply(dots, rlang::as_label, character(1)), collapse = " & ")

  # Apply filter (guard against zero conditions)
  if (length(dots) == 0L) {
    result <- .data
  } else {
    result <- .filter_fn(.data, !!!dots)
  }

  # Compute diagnostics
  n_total <- nrow(.data)
  n_result <- nrow(result)
  n_dropped <- n_total - n_result
  pct_dropped <- if (n_total > 0L) 100 * n_dropped / n_total else 0

  # Count NAs in the filter condition (best-effort)
  n_na_in_filter <- tryCatch({
    if (length(dots) == 0L) {
      0L
    } else {
      mask_results <- lapply(dots, function(q) rlang::eval_tidy(q, data = .data))
      combined <- Reduce(`&`, mask_results)
      sum(is.na(combined))
    }
  }, error = function(e) NA_integer_)

  # Stat diagnostics
  stat_dropped <- NULL
  stat_total <- NULL
  stat_col <- NULL
  if (.have_stat) {
    stat_total <- rlang::eval_tidy(
      rlang::quo(sum(!!.stat_quo, na.rm = TRUE)), data = .data
    )
    stat_result <- rlang::eval_tidy(
      rlang::quo(sum(!!.stat_quo, na.rm = TRUE)), data = result
    )
    stat_dropped <- stat_total - stat_result
    stat_col <- rlang::as_label(.stat_quo)
  }

  # Build snapshot
  index <- length(.trail$snapshots) + 1L
  if (is.null(.label)) {
    .label <- paste0(.label_prefix, index)
  }
  if (.label %in% .trail$labels) {
    trail_name <- .trail$name
    label_val <- .label
    cli::cli_abort(
      "Label {.val {label_val}} already exists in trail {.val {trail_name}}."
    )
  }

  snap <- .build_snapshot(result, label = .label, index = index,
                          .numeric_summary = .numeric_summary,
                          .cols_include = .cols_include,
                          .cols_exclude = .cols_exclude)
  snap$pipeline <- tryCatch(.capture_pipeline(.data_expr), error = function(e) NULL)

  # Enrich snapshot
  snap$type <- "filter"
  snap$diagnostics <- list(
    filter_type    = .filter_type,
    expr_label     = expr_lab,
    n_dropped      = n_dropped,
    n_total        = n_total,
    pct_dropped    = pct_dropped,
    stat_dropped   = stat_dropped,
    stat_total     = stat_total,
    stat_col       = stat_col,
    n_na_in_filter = n_na_in_filter
  )

  # Compute changes vs previous
  if (index > 1L) {
    prev <- .trail$snapshots[[index - 1L]]
    snap$changes <- .detect_changes(prev, snap)
  }

  # Print diagnostics (unless quiet)
  if (!.quiet) {
    tap_name_local <- .tap_name
    action_local <- .action
    cli::cli_alert_info(
      "{tap_name_local}: {expr_lab}"
    )
    cli::cli_text(
      "  {action_local} {format(n_dropped, big.mark = ',')} of {format(n_total, big.mark = ',')} rows ({sprintf('%.1f%%', pct_dropped)})"
    )
    if (.have_stat) {
      stat_col_local <- stat_col
      cli::cli_text(
        "  Stat {stat_col_local}: {tolower(action_local)} {format(stat_dropped, big.mark = ',', scientific = FALSE)} of {format(stat_total, big.mark = ',', scientific = FALSE)}"
      )
    }
  }

  # Warn if threshold exceeded
  if (!is.null(.warn_threshold)) {
    share_drop <- if (n_total > 0L) n_dropped / n_total else NA_real_
    if (!is.na(share_drop) && share_drop > .warn_threshold) {
      action_local <- .action
      cli::cli_warn(
        "{action_local} {sprintf('%.1f%%', 100 * share_drop)} of rows exceeds threshold ({sprintf('%.1f%%', 100 * .warn_threshold)})."
      )
    }
  }

  # Append to trail
  .trail$snapshots[[index]] <- snap
  .trail$labels <- c(.trail$labels, .label)

  result
}

# ---------------------------------------------------------------------------
# Test data
# ---------------------------------------------------------------------------

df <- data.frame(
  id     = 1:10,
  amount = c(100, 200, 300, 400, 500, 600, 700, 800, 900, 1000),
  flag   = rep(c(TRUE, FALSE), 5),
  stringsAsFactors = FALSE
)

# ---------------------------------------------------------------------------
# filter_tap — basic with trail
# ---------------------------------------------------------------------------

test_that("filter_tap produces correct filter result", {
  trail <- audit_trail("ft_basic")
  result <- filter_tap(df, amount > 300, .trail = trail, .label = "big")

  expected <- dplyr::filter(df, amount > 300)
  expect_equal(result, expected)
})

test_that("filter_tap records enriched snapshot in trail", {
  trail <- audit_trail("ft_enriched")
  suppressMessages(
    result <- filter_tap(df, amount > 300, .trail = trail, .label = "big")
  )

  expect_equal(length(trail$snapshots), 1L)
  expect_equal(trail$labels, "big")

  snap <- trail$snapshots[[1]]
  expect_s3_class(snap, "audit_snap")
  expect_equal(snap$type, "filter")
  expect_equal(snap$label, "big")

  diag <- snap$diagnostics
  expect_equal(diag$filter_type, "keep")
  expect_equal(diag$n_dropped, 3L)
  expect_equal(diag$n_total, 10L)
  expect_equal(diag$pct_dropped, 30)
  expect_equal(diag$expr_label, "amount > 300")
})

test_that("filter_tap with stat tracks stat column", {
  trail <- audit_trail("ft_stat")
  suppressMessages(
    result <- filter_tap(df, amount > 500, .trail = trail, .label = "big",
                          .stat = amount)
  )

  diag <- trail$snapshots[[1]]$diagnostics
  expect_equal(diag$stat_col, "amount")
  expect_true(!is.null(diag$stat_dropped))
  expect_true(!is.null(diag$stat_total))
  expect_equal(diag$stat_total, sum(df$amount))
  expect_equal(diag$stat_dropped, sum(df$amount[df$amount <= 500]))
})

# ---------------------------------------------------------------------------
# filter_out_tap — inverse logic
# ---------------------------------------------------------------------------

test_that("filter_out_tap drops matching rows", {
  trail <- audit_trail("fot_basic")
  suppressMessages(
    result <- filter_out_tap(df, flag == FALSE,
                              .trail = trail, .label = "flagged")
  )

  expected <- dplyr::filter_out(df, flag == FALSE)
  expect_equal(result, expected)
})

test_that("filter_out_tap records enriched snapshot", {
  trail <- audit_trail("fot_enriched")
  suppressMessages(
    result <- filter_out_tap(df, amount > 500,
                              .trail = trail, .label = "small_only")
  )

  snap <- trail$snapshots[[1]]
  expect_equal(snap$type, "filter")

  diag <- snap$diagnostics
  expect_equal(diag$filter_type, "drop")
  expect_equal(diag$n_dropped, 5L)  # 600-1000 dropped
  expect_equal(diag$n_total, 10L)
  expect_equal(diag$pct_dropped, 50)
})

test_that("filter_out_tap with stat tracks dropped stat", {
  trail <- audit_trail("fot_stat")
  suppressMessages(
    result <- filter_out_tap(df, amount > 500,
                              .trail = trail, .label = "low",
                              .stat = amount)
  )

  diag <- trail$snapshots[[1]]$diagnostics
  expect_equal(diag$stat_total, sum(df$amount))
  expect_equal(diag$stat_dropped, sum(df$amount[df$amount > 500]))
})

# ---------------------------------------------------------------------------
# NULL trail — plain filter
# ---------------------------------------------------------------------------

test_that("filter_tap without trail performs plain dplyr::filter", {
  result <- filter_tap(df, amount > 300)
  expected <- dplyr::filter(df, amount > 300)
  expect_equal(result, expected)
})

test_that("filter_out_tap without trail performs plain dplyr::filter_out", {
  result <- filter_out_tap(df, amount > 300)
  expected <- dplyr::filter_out(df, amount > 300)
  expect_equal(result, expected)
})

# ---------------------------------------------------------------------------
# NULL trail + diagnostic args → delegates to filter_keep/filter_drop
# ---------------------------------------------------------------------------

test_that("filter_tap without trail + .stat delegates to filter_keep", {
  output <- capture.output(
    result <- filter_tap(df, amount > 300, .stat = amount),
    type = "message"
  )
  expected <- dplyr::filter(df, amount > 300)
  expect_equal(result, expected)
  expect_true(any(grepl("Dropped|filter_keep", output)))
})

test_that("filter_out_tap without trail + .stat delegates to filter_drop", {
  output <- capture.output(
    result <- filter_out_tap(df, amount > 300, .stat = amount),
    type = "message"
  )
  expected <- dplyr::filter_out(df, amount > 300)
  expect_equal(result, expected)
  expect_true(any(grepl("Dropped|filter_drop", output)))
})

# ---------------------------------------------------------------------------
# NULL trail + .label → warns
# ---------------------------------------------------------------------------

test_that("filter_tap without trail + .label warns", {
  expect_warning(
    filter_tap(df, amount > 300, .label = "foo"),
    "ignored"
  )
})

test_that("filter_out_tap without trail + .label warns", {
  expect_warning(
    filter_out_tap(df, amount > 300, .label = "bar"),
    "ignored"
  )
})

# ---------------------------------------------------------------------------
# .warn_threshold
# ---------------------------------------------------------------------------

test_that("filter_tap warns when threshold exceeded", {
  trail <- audit_trail("ft_warn")
  expect_warning(
    suppressMessages(
      filter_tap(df, amount > 200, .trail = trail, .label = "w",
                  .warn_threshold = 0.1)
    ),
    "exceeds threshold"
  )
})

test_that("filter_tap does not warn when threshold not exceeded", {
  trail <- audit_trail("ft_nowarn")
  expect_no_warning(
    suppressMessages(
      filter_tap(df, amount > 900, .trail = trail, .label = "nw",
                  .warn_threshold = 0.95)
    )
  )
})

# ---------------------------------------------------------------------------
# .quiet
# ---------------------------------------------------------------------------

test_that("filter_tap .quiet suppresses diagnostic output", {
  trail <- audit_trail("ft_quiet")
  output <- capture.output(
    result <- filter_tap(df, amount > 300, .trail = trail, .label = "q",
                          .quiet = TRUE),
    type = "message"
  )
  # Only trail-level output should be absent
  expect_false(any(grepl("filter_tap:", output)))
})

# ---------------------------------------------------------------------------
# Label management
# ---------------------------------------------------------------------------

test_that("filter_tap auto-generates labels", {
  trail <- audit_trail("ft_autolbl")
  suppressMessages(
    filter_tap(df, amount > 300, .trail = trail)
  )
  expect_equal(trail$labels, "filter_1")
})

test_that("filter_out_tap auto-generates labels", {
  trail <- audit_trail("fot_autolbl")
  suppressMessages(
    filter_out_tap(df, amount > 300, .trail = trail)
  )
  expect_equal(trail$labels, "filter_out_1")
})

test_that("filter_tap rejects duplicate labels", {
  trail <- audit_trail("ft_dup")
  suppressMessages(
    filter_tap(df, amount > 300, .trail = trail, .label = "f")
  )
  expect_error(
    filter_tap(df, amount > 500, .trail = trail, .label = "f"),
    "already exists"
  )
})

test_that("filter_tap rejects invalid trail", {
  expect_error(
    filter_tap(df, amount > 300, .trail = "not_a_trail"),
    "audit_trail"
  )
})

# ---------------------------------------------------------------------------
# Change detection
# ---------------------------------------------------------------------------

test_that("filter_tap computes changes from previous snapshot", {
  trail <- audit_trail("ft_changes")
  df |> audit_tap(trail, "raw")
  suppressMessages(
    filter_tap(df, amount > 300, .trail = trail, .label = "filtered")
  )

  snap2 <- trail$snapshots[[2]]
  expect_true(!is.null(snap2$changes))
  expect_true(snap2$changes$row_delta < 0L)
})

# ---------------------------------------------------------------------------
# Pipeline capture
# ---------------------------------------------------------------------------

test_that("filter_tap captures pipeline", {
  trail <- audit_trail("ft_pipeline")
  suppressMessages(
    result <- df |>
      dplyr::mutate(x = 1) |>
      filter_tap(amount > 300, .trail = trail, .label = "f")
  )

  snap <- trail$snapshots[[1]]
  expect_true(!is.null(snap$pipeline))
  expect_true(length(snap$pipeline) > 0L)
})

# ---------------------------------------------------------------------------
# Mixed pipeline
# ---------------------------------------------------------------------------

test_that("mixed pipeline with audit_tap + join_tap + filter_tap works", {
  customers <- data.frame(id = c(1L, 5L, 7L), name = c("A", "B", "C"))

  trail <- audit_trail("mixed")
  suppressMessages(
    result <- df |>
      audit_tap(trail, "raw") |>
      left_join_tap(customers, by = "id", .trail = trail, .label = "joined") |>
      filter_tap(amount > 300, .trail = trail, .label = "big_only")
  )

  expect_equal(length(trail$snapshots), 3L)
  expect_equal(trail$snapshots[[1]]$type, "tap")
  expect_equal(trail$snapshots[[2]]$type, "join")
  expect_equal(trail$snapshots[[3]]$type, "filter")

  # Print should work
  output <- capture.output(print(trail), type = "message")
  expect_true(any(grepl("left_join", output)))
  expect_true(any(grepl("filter", output)))
})

# ---------------------------------------------------------------------------
# audit_diff across filter snapshots
# ---------------------------------------------------------------------------

test_that("audit_diff works between tap and filter snapshots", {
  trail <- audit_trail("diff_filter")
  df |> audit_tap(trail, "raw")
  suppressMessages(
    filter_tap(df, amount > 500, .trail = trail, .label = "big")
  )

  diff_obj <- audit_diff(trail, "raw", "big")
  expect_s3_class(diff_obj, "audit_diff")
  expect_true(diff_obj$row_delta < 0L)
})

# ---------------------------------------------------------------------------
# Multiple conditions
# ---------------------------------------------------------------------------

test_that("filter_tap handles multiple conditions", {
  trail <- audit_trail("ft_multi")
  suppressMessages(
    result <- filter_tap(df, amount > 300, flag == TRUE,
                          .trail = trail, .label = "multi")
  )

  expected <- dplyr::filter(df, amount > 300, flag == TRUE)
  expect_equal(result, expected)

  diag <- trail$snapshots[[1]]$diagnostics
  expect_true(grepl("&", diag$expr_label))
})

# ---------------------------------------------------------------------------
# n_na_in_filter
# ---------------------------------------------------------------------------

test_that("filter_tap counts NAs in filter condition", {
  df_na <- data.frame(
    id = 1:6,
    val = c(10, NA, 30, NA, 50, 60)
  )
  trail <- audit_trail("ft_na")
  suppressMessages(
    result <- filter_tap(df_na, val > 20, .trail = trail, .label = "naf")
  )

  diag <- trail$snapshots[[1]]$diagnostics
  expect_equal(diag$n_na_in_filter, 2L)
})

# ---------------------------------------------------------------------------
# Zero-row data
# ---------------------------------------------------------------------------

test_that("filter_tap handles zero-row data", {
  empty <- df[0, ]
  trail <- audit_trail("ft_empty")
  suppressMessages(
    result <- filter_tap(empty, amount > 300, .trail = trail, .label = "e")
  )
  expect_equal(nrow(result), 0L)
  expect_equal(trail$snapshots[[1]]$diagnostics$n_dropped, 0L)
  expect_equal(trail$snapshots[[1]]$diagnostics$pct_dropped, 0)
})

# ---------------------------------------------------------------------------
# Filter that drops all rows
# ---------------------------------------------------------------------------

test_that("filter_tap handles condition that drops all rows", {
  trail <- audit_trail("ft_all")
  suppressMessages(
    result <- filter_tap(df, amount > 99999, .trail = trail, .label = "none")
  )
  expect_equal(nrow(result), 0L)
  expect_equal(trail$snapshots[[1]]$diagnostics$n_dropped, 10L)
  expect_equal(trail$snapshots[[1]]$diagnostics$pct_dropped, 100)
})

# ---------------------------------------------------------------------------
# Grouped tibble
# ---------------------------------------------------------------------------

test_that("filter_tap preserves grouping", {
  grouped <- dplyr::group_by(df, flag)
  trail <- audit_trail("ft_grouped")
  suppressMessages(
    result <- filter_tap(grouped, amount > 300, .trail = trail, .label = "g")
  )
  expect_true(dplyr::is_grouped_df(result))
  expect_equal(dplyr::group_vars(result), "flag")
})

# ---------------------------------------------------------------------------
# Nested pipe chain
# ---------------------------------------------------------------------------

test_that("filter_tap and filter_out_tap work in nested pipes", {
  trail <- audit_trail("ft_nested")
  suppressMessages(
    result <- df |>
      audit_tap(trail, "raw") |>
      filter_tap(amount > 200, .trail = trail, .label = "f1") |>
      filter_out_tap(flag == FALSE, .trail = trail, .label = "f2")
  )

  expect_equal(length(trail$snapshots), 3L)
  expect_equal(trail$labels, c("raw", "f1", "f2"))
  # Correct ordering
  expect_equal(trail$snapshots[[1]]$index, 1L)
  expect_equal(trail$snapshots[[2]]$index, 2L)
  expect_equal(trail$snapshots[[3]]$index, 3L)
})

# ---------------------------------------------------------------------------
# print.audit_trail Type column rendering
# ---------------------------------------------------------------------------

test_that("print.audit_trail renders Type column for filter snapshots", {
  trail <- audit_trail("type_col")
  df |> audit_tap(trail, "raw")
  suppressMessages(
    filter_tap(df, amount > 300, .trail = trail, .label = "f")
  )
  output <- capture.output(print(trail), type = "message")
  # Should show filter info in the Type column
  expect_true(any(grepl("filter.*dropped.*rows", output)))
})

# ---------------------------------------------------------------------------
# .data validation
# ---------------------------------------------------------------------------

test_that("filter_tap rejects non-data.frame input", {
  trail <- audit_trail("bad_data")
  expect_error(
    filter_tap(list(a = 1), a > 1, .trail = trail),
    "data.frame"
  )
})

test_that("filter_out_tap rejects non-data.frame input", {
  trail <- audit_trail("bad_data2")
  expect_error(
    filter_out_tap(42, a > 1, .trail = trail),
    "data.frame"
  )
})

# ---------------------------------------------------------------------------
# Label validation
# ---------------------------------------------------------------------------

test_that("filter_tap rejects NA label", {
  trail <- audit_trail("na_label")
  expect_error(
    filter_tap(df, amount > 300, .trail = trail, .label = NA_character_),
    "single character string"
  )
})

test_that("filter_tap rejects multi-element label", {
  trail <- audit_trail("multi_label")
  expect_error(
    filter_tap(df, amount > 300, .trail = trail, .label = c("a", "b")),
    "single character string"
  )
})

test_that("filter_out_tap rejects NA label", {
  trail <- audit_trail("na_label2")
  expect_error(
    filter_out_tap(df, amount > 300, .trail = trail, .label = NA_character_),
    "single character string"
  )
})

# ---------------------------------------------------------------------------
# filter_out_tap with .warn_threshold
# ---------------------------------------------------------------------------

test_that("filter_out_tap warns when threshold exceeded", {
  trail <- audit_trail("fot_warn")
  expect_warning(
    suppressMessages(
      filter_out_tap(df, amount > 200, .trail = trail, .label = "w",
                      .warn_threshold = 0.1)
    ),
    "exceeds threshold"
  )
})

test_that("filter_out_tap does not warn when threshold not exceeded", {
  trail <- audit_trail("fot_nowarn")
  expect_no_warning(
    suppressMessages(
      filter_out_tap(df, amount > 900, .trail = trail, .label = "nw",
                      .warn_threshold = 0.95)
    )
  )
})

# ---------------------------------------------------------------------------
# filter_out_tap with .quiet
# ---------------------------------------------------------------------------

test_that("filter_out_tap .quiet suppresses diagnostic output", {
  trail <- audit_trail("fot_quiet")
  output <- capture.output(
    result <- filter_out_tap(df, amount > 300, .trail = trail, .label = "q",
                              .quiet = TRUE),
    type = "message"
  )
  expect_false(any(grepl("filter_out_tap:", output)))
})

# ---------------------------------------------------------------------------
# Zero conditions with trail
# ---------------------------------------------------------------------------

test_that("filter_tap with zero conditions and trail records no-op snapshot", {
  trail <- audit_trail("ft_zero")
  df |> audit_tap(trail, "raw")
  suppressMessages(
    result <- filter_tap(df, .trail = trail, .label = "noop")
  )
  expect_equal(result, df)
  expect_equal(length(trail$snapshots), 2L)
  diag <- trail$snapshots[[2]]$diagnostics
  expect_equal(diag$n_dropped, 0L)
  expect_equal(diag$pct_dropped, 0)
})

test_that("filter_out_tap with zero conditions and trail records no-op snapshot", {
  trail <- audit_trail("fot_zero")
  df |> audit_tap(trail, "raw")
  suppressMessages(
    result <- filter_out_tap(df, .trail = trail, .label = "noop")
  )
  expect_equal(result, df)
  expect_equal(length(trail$snapshots), 2L)
  diag <- trail$snapshots[[2]]$diagnostics
  expect_equal(diag$n_dropped, 0L)
})

# ---------------------------------------------------------------------------
# audit_report with mixed snapshot types
# ---------------------------------------------------------------------------

# ---------------------------------------------------------------------------
# .warn_threshold input validation in filter_tap / filter_out_tap (Fix 1)
# ---------------------------------------------------------------------------

test_that("filter_tap rejects NA_real_ .warn_threshold", {
  trail <- audit_trail("wt_na1")
  expect_error(
    filter_tap(df, amount > 300, .trail = trail, .warn_threshold = NA_real_),
    "warn_threshold"
  )
})

test_that("filter_tap rejects length-2 .warn_threshold", {
  trail <- audit_trail("wt_len")
  expect_error(
    filter_tap(df, amount > 300, .trail = trail, .warn_threshold = c(0.1, 0.2)),
    "warn_threshold"
  )
})

test_that("filter_tap rejects character .warn_threshold", {
  trail <- audit_trail("wt_char")
  expect_error(
    filter_tap(df, amount > 300, .trail = trail, .warn_threshold = "0.5"),
    "warn_threshold"
  )
})

test_that("filter_out_tap rejects NA_real_ .warn_threshold", {
  trail <- audit_trail("wt_na2")
  expect_error(
    filter_out_tap(df, amount > 300, .trail = trail, .warn_threshold = NA_real_),
    "warn_threshold"
  )
})

test_that("filter_tap rejects invalid .warn_threshold without trail too", {
  # NULL trail delegation path also validates
  expect_error(
    filter_tap(df, amount > 300, .warn_threshold = NA_real_),
    "warn_threshold"
  )
})

test_that("audit_report works with mixed snapshot types (tap + join + filter)", {
  customers <- data.frame(id = c(1L, 5L, 7L), name = c("A", "B", "C"))

  trail <- audit_trail("report_mixed")
  suppressMessages({
    df |>
      audit_tap(trail, "raw") |>
      left_join_tap(customers, by = "id", .trail = trail, .label = "joined") |>
      filter_tap(amount > 300, .trail = trail, .label = "big")
  })

  expect_equal(length(trail$snapshots), 3L)

  # audit_report should work without error
  output <- capture.output(audit_report(trail), type = "message")
  expect_true(any(grepl("Audit Report", output)))
  expect_true(any(grepl("left_join", output)))
  expect_true(any(grepl("filter", output)))
  expect_true(any(grepl("Detailed Diffs", output)))
})

# ---------------------------------------------------------------------------
# Snapshot controls
# ---------------------------------------------------------------------------

test_that("filter_tap passes snapshot controls through", {
  trail <- audit_trail("filter_controls")
  suppressMessages({
    df |>
      audit_tap(trail, "raw") |>
      filter_tap(amount > 300, .trail = trail, .label = "big",
                 .numeric_summary = FALSE, .cols_exclude = "flag")
  })
  snap <- trail$snapshots[[2]]
  expect_null(snap$numeric_summary)
  expect_false("flag" %in% snap$schema$column)
  # ncol is still full data column count
  expect_equal(snap$ncol, 3L)
})

#' Filter Data with Diagnostic Statistics (Keep)
#'
#' Filters a data.frame or tibble while reporting statistics about dropped rows
#' and optionally the sum of a statistic column that was dropped. Keeps rows
#' where the conditions are TRUE (same as [dplyr::filter()]).
#'
#' @param .data A data.frame, tibble, or other object.
#' @param ... Arguments passed to methods.
#'
#' @returns The filtered data.frame or tibble.
#'
#' @family filter diagnostics
#' @export
filter_keep <- function(.data, ...) UseMethod("filter_keep")

#' @describeIn filter_keep Method for data.frame objects
#'
#' @param ... Filter conditions, evaluated in the context of `.data` using
#'   tidy evaluation (same as [dplyr::filter()]).
#' @param .stat An unquoted column or expression to total, e.g., `amount`,
#'   `price * qty`. Reports the amount dropped and its share of the total.
#' @param .quiet Logical. If `TRUE`, suppress printing diagnostics.
#' @param .warn_threshold Numeric between 0 and 1. If set and the proportion of
#'   dropped rows exceeds this threshold, a warning is issued.
#'
#' @examples
#' df <- data.frame(
#'   id = 1:6,
#'   keep = c(TRUE, FALSE, TRUE, NA, TRUE, FALSE),
#'   sales = c(100, 50, 200, 25, NA, 75)
#' )
#' filter_keep(df, keep == TRUE)
#' filter_keep(df, keep == TRUE, .stat = sales)
#'
#' @export
filter_keep.data.frame <- function(.data, ..., .stat = NULL,
                                    .quiet = FALSE, .warn_threshold = NULL) {
  data_name <- deparse(substitute(.data))
  .validate_warn_threshold(.warn_threshold)
  dots <- rlang::enquos(...)
  stat_quo <- rlang::enquo(.stat)
  have_stat <- !rlang::quo_is_null(stat_quo)

  expr_lab <- paste(vapply(dots, rlang::as_label, character(1)), collapse = ", ")

  # Evaluate filter: keep matching rows
  filtered <- dplyr::filter(.data, !!!dots)

  .print_filter_diagnostic(
    .data, filtered, data_name, expr_lab, stat_quo, have_stat,
    mode = "keep", quiet = .quiet, warn_threshold = .warn_threshold
  )

  filtered
}

#' Filter Data with Diagnostic Statistics (Drop)
#'
#' Filters a data.frame or tibble by DROPPING rows where the conditions are
#' TRUE, while reporting statistics about dropped rows and optionally the sum
#' of a statistic column that was dropped.
#'
#' @param .data A data.frame, tibble, or other object.
#' @param ... Arguments passed to methods.
#'
#' @returns The filtered data.frame or tibble.
#'
#' @family filter diagnostics
#' @export
filter_drop <- function(.data, ...) UseMethod("filter_drop")

#' @describeIn filter_drop Method for data.frame objects
#'
#' @param ... Filter conditions specifying rows to DROP, evaluated in the
#'   context of `.data` using tidy evaluation.
#' @param .stat An unquoted column or expression to total, e.g., `amount`,
#'   `price * qty`. Reports the amount dropped and its share of the total.
#' @param .quiet Logical. If `TRUE`, suppress printing diagnostics.
#' @param .warn_threshold Numeric between 0 and 1. If set and the proportion of
#'   dropped rows exceeds this threshold, a warning is issued.
#'
#' @examples
#' df <- data.frame(
#'   id = 1:5,
#'   bad = c(FALSE, TRUE, FALSE, TRUE, FALSE),
#'   sales = 10:14
#' )
#' filter_drop(df, bad == TRUE)
#' filter_drop(df, bad == TRUE, .stat = sales)
#'
#' @export
filter_drop.data.frame <- function(.data, ..., .stat = NULL,
                                    .quiet = FALSE, .warn_threshold = NULL) {
  data_name <- deparse(substitute(.data))
  .validate_warn_threshold(.warn_threshold)
  dots <- rlang::enquos(...)
  stat_quo <- rlang::enquo(.stat)
  have_stat <- !rlang::quo_is_null(stat_quo)

  expr_lab <- paste(vapply(dots, rlang::as_label, character(1)), collapse = ", ")

  # Evaluate filter: drop matching rows (keep the inverse)
  if (length(dots) == 0L) {
    filtered <- .data
  } else {
    filtered <- dplyr::filter_out(.data, !!!dots)
  }

  .print_filter_diagnostic(
    .data, filtered, data_name, expr_lab, stat_quo, have_stat,
    mode = "drop", quiet = .quiet, warn_threshold = .warn_threshold
  )

  filtered
}

#' Internal helper for filter diagnostic printing
#' @noRd
.print_filter_diagnostic <- function(.data, filtered, data_name, expr_lab,
                                      stat_quo, have_stat, mode, quiet,
                                      warn_threshold) {
  n_total <- nrow(.data)
  n_filtered <- nrow(filtered)
  n_drop <- n_total - n_filtered
  share_drop <- if (n_total > 0L) n_drop / n_total else NA_real_

  func_name <- if (mode == "keep") "filter_keep" else "filter_drop"

  if (have_stat) {
    total_val <- rlang::eval_tidy(
      rlang::quo(sum(!!stat_quo, na.rm = TRUE)), data = .data
    )
    filtered_val <- rlang::eval_tidy(
      rlang::quo(sum(!!stat_quo, na.rm = TRUE)), data = filtered
    )
    drop_val <- total_val - filtered_val
    share_val <- if (isTRUE(all.equal(total_val, 0))) NA_real_ else drop_val / total_val
    stat_lab <- rlang::as_label(stat_quo)
  }

  if (!quiet) {
    pct_lab <- if (is.na(share_drop)) "N/A" else sprintf("%.2f%%", 100 * share_drop)
    cli::cli_text("{func_name}({data_name}, {expr_lab})")
    cli::cli_text(
      "  Dropped {format(n_drop, big.mark = ',')} of {format(n_total, big.mark = ',')} rows ({pct_lab})."
    )
    if (have_stat) {
      pct_val_lab <- if (is.na(share_val)) "N/A" else sprintf("%.2f%%", 100 * share_val)
      cli::cli_text(
        "  Dropped {format(drop_val, big.mark = ',', scientific = FALSE)} of {format(total_val, big.mark = ',', scientific = FALSE)} for {stat_lab} ({pct_val_lab})."
      )
    }
  }

  if (!is.null(warn_threshold) && !is.na(share_drop) && share_drop > warn_threshold) {
    cli::cli_warn(
      "Dropped {sprintf('%.1f%%', 100 * share_drop)} of rows exceeds threshold ({sprintf('%.1f%%', 100 * warn_threshold)})."
    )
  }
}

#' Validate .warn_threshold argument
#'
#' Checks that .warn_threshold is NULL or a single finite numeric in \[0, 1\].
#' Emits a cli_abort() if the value is invalid.
#'
#' @param x The value to validate.
#'
#' @noRd
.validate_warn_threshold <- function(x) {
  if (is.null(x)) return(invisible(NULL))
  if (!is.numeric(x) || length(x) != 1L || is.na(x) || x < 0 || x > 1) {
    cli::cli_abort(
      "{.arg .warn_threshold} must be a single number between 0 and 1, or NULL."
    )
  }
  invisible(NULL)
}

test_that("filter_keep keeps matching rows", {
  df <- data.frame(id = 1:6, value = c(10, 20, 30, 40, 50, 60))
  result <- filter_keep(df, value > 30, .quiet = TRUE)

  expect_equal(nrow(result), 3L)
  expect_true(all(result$value > 30))
})

test_that("filter_keep prints diagnostics", {
  df <- data.frame(id = 1:10, value = 1:10)

  output <- capture.output(
    result <- filter_keep(df, value > 5),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("filter_keep", combined))
  expect_true(grepl("Dropped", combined))
  expect_true(grepl("5", combined))  # 5 rows dropped
})

test_that("filter_keep suppresses output with .quiet", {
  df <- data.frame(id = 1:10, value = 1:10)

  output <- capture.output(
    result <- filter_keep(df, value > 5, .quiet = TRUE),
    type = "message"
  )
  expect_equal(length(output), 0L)
})

test_that("filter_keep tracks .stat", {
  df <- data.frame(id = 1:4, sales = c(100, 200, 300, 400), keep = c(TRUE, FALSE, TRUE, FALSE))

  output <- capture.output(
    result <- filter_keep(df, keep == TRUE, .stat = sales),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("sales", combined))
  expect_equal(nrow(result), 2L)
})

test_that("filter_keep warns at threshold", {
  df <- data.frame(id = 1:100, value = 1:100)

  expect_warning(
    filter_keep(df, value > 90, .quiet = TRUE, .warn_threshold = 0.5),
    "exceeds threshold"
  )
})

test_that("filter_keep does not warn below threshold", {
  df <- data.frame(id = 1:100, value = 1:100)

  expect_no_warning(
    filter_keep(df, value > 10, .quiet = TRUE, .warn_threshold = 0.95)
  )
})

test_that("filter_keep works with multiple conditions", {
  df <- data.frame(id = 1:10, x = 1:10, y = 10:1)
  result <- filter_keep(df, x > 3, y > 3, .quiet = TRUE)

  expect_true(all(result$x > 3 & result$y > 3))
})

test_that("filter_drop drops matching rows", {
  df <- data.frame(id = 1:6, bad = c(FALSE, TRUE, FALSE, TRUE, FALSE, TRUE))
  result <- filter_drop(df, bad == TRUE, .quiet = TRUE)

  expect_equal(nrow(result), 3L)
  expect_true(all(!result$bad))
})

test_that("filter_drop prints diagnostics", {
  df <- data.frame(id = 1:10, value = 1:10)

  output <- capture.output(
    result <- filter_drop(df, value > 8),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("filter_drop", combined))
  expect_true(grepl("Dropped", combined))
})

test_that("filter_drop tracks .stat", {
  df <- data.frame(id = 1:4, sales = c(100, 200, 300, 400),
                   bad = c(TRUE, FALSE, TRUE, FALSE))

  output <- capture.output(
    result <- filter_drop(df, bad == TRUE, .stat = sales),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("sales", combined))
  # Dropped rows 1 and 3 (sales = 100 + 300 = 400)
  expect_equal(nrow(result), 2L)
})

test_that("filter_drop warns at threshold", {
  df <- data.frame(id = 1:10, value = 1:10)

  expect_warning(
    filter_drop(df, value < 9, .quiet = TRUE, .warn_threshold = 0.5),
    "exceeds threshold"
  )
})

test_that("filter_keep preserves tibble class", {
  skip_if_not_installed("dplyr")
  df <- dplyr::tibble(id = 1:5, value = 1:5)
  result <- filter_keep(df, value > 2, .quiet = TRUE)

  expect_s3_class(result, "tbl_df")
})

test_that("filter_drop preserves tibble class", {
  skip_if_not_installed("dplyr")
  df <- dplyr::tibble(id = 1:5, value = 1:5)
  result <- filter_drop(df, value < 3, .quiet = TRUE)

  expect_s3_class(result, "tbl_df")
})

test_that("filter_keep handles zero-row result", {
  df <- data.frame(id = 1:5, value = 1:5)
  result <- filter_keep(df, value > 100, .quiet = TRUE)

  expect_equal(nrow(result), 0L)
  expect_equal(ncol(result), ncol(df))
})

test_that("filter_drop with all rows matching drops everything", {
  df <- data.frame(id = 1:5, value = 1:5)
  result <- filter_drop(df, value > 0, .quiet = TRUE)

  expect_equal(nrow(result), 0L)
})

test_that("filter_drop works with multiple predicates", {
  df <- data.frame(a = 1:6, b = 6:1)
  # Drop rows where a > 2 AND b < 5 (rows 3,4,5,6)
  result <- filter_drop(df, a > 2, b < 5, .quiet = TRUE)

  expect_equal(nrow(result), 2L)
  # Kept rows should NOT satisfy both conditions
  expect_true(all(!(result$a > 2 & result$b < 5)))
})

test_that("filter_drop with no predicates returns all rows", {
  df <- data.frame(id = 1:5, value = 1:5)
  result <- filter_drop(df, .quiet = TRUE)

  expect_equal(nrow(result), 5L)
})

test_that("filter_drop with multiple predicates and .stat", {
  df <- data.frame(a = 1:6, b = 6:1, sales = c(10, 20, 30, 40, 50, 60))
  output <- capture.output(
    result <- filter_drop(df, a > 3, b < 4, .stat = sales),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("sales", combined))
  # Rows matching a > 3 & b < 4 are rows 4,5,6
  expect_equal(nrow(result), 3L)
})

# ---------------------------------------------------------------------------
# .warn_threshold input validation (Fix 1)
# ---------------------------------------------------------------------------

test_that("filter_keep rejects NA_real_ .warn_threshold", {
  df <- data.frame(x = 1:5)
  expect_error(
    filter_keep(df, x > 2, .warn_threshold = NA_real_),
    "warn_threshold"
  )
})

test_that("filter_keep rejects length-2 .warn_threshold", {
  df <- data.frame(x = 1:5)
  expect_error(
    filter_keep(df, x > 2, .warn_threshold = c(0.1, 0.2)),
    "warn_threshold"
  )
})

test_that("filter_keep rejects character .warn_threshold", {
  df <- data.frame(x = 1:5)
  expect_error(
    filter_keep(df, x > 2, .warn_threshold = "0.5"),
    "warn_threshold"
  )
})

test_that("filter_drop rejects NA_real_ .warn_threshold", {
  df <- data.frame(x = 1:5)
  expect_error(
    filter_drop(df, x > 2, .warn_threshold = NA_real_),
    "warn_threshold"
  )
})

test_that("filter_drop rejects out-of-range .warn_threshold", {
  df <- data.frame(x = 1:5)
  expect_error(
    filter_drop(df, x > 2, .warn_threshold = 1.5),
    "warn_threshold"
  )
  expect_error(
    filter_drop(df, x > 2, .warn_threshold = -0.1),
    "warn_threshold"
  )
})

# ---------------------------------------------------------------------------
# NA% display for zero-row inputs (Fix 4)
# ---------------------------------------------------------------------------

test_that("filter_keep shows N/A instead of NA% for empty input", {
  empty <- data.frame(x = integer(0))
  output <- capture.output(
    filter_keep(empty, x > 2),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_false(grepl("NA%", combined))
  expect_true(grepl("N/A", combined))
})

test_that("filter_drop shows N/A instead of NA% for empty input", {
  empty <- data.frame(x = integer(0))
  output <- capture.output(
    filter_drop(empty, x > 2),
    type = "message"
  )
  combined <- paste(output, collapse = "\n")
  expect_false(grepl("NA%", combined))
  expect_true(grepl("N/A", combined))
})

#' Record a Tabulation Snapshot in a Pipeline
#'
#' Transparent pipe pass-through that runs [tab()] on the data and stores the
#' result as a custom diagnostic annotation in the audit trail snapshot.
#' Returns `.data` unchanged.
#'
#' @inheritParams tab
#' @param .trail An [audit_trail()] object.
#' @param .label Character label for this snapshot.
#' @param .numeric_summary Logical. If `FALSE`, skip numeric summary
#'   computation in the snapshot (default `TRUE`).
#' @param .cols_include Character vector of column names to include in the
#'   snapshot schema, or `NULL` (the default) to include all columns. Mutually
#'   exclusive with `.cols_exclude`.
#' @param .cols_exclude Character vector of column names to exclude from the
#'   snapshot schema, or `NULL` (the default). Mutually exclusive with
#'   `.cols_include`.
#'
#' @returns `.data`, unchanged, returned invisibly.
#'
#' @examples
#' trail <- audit_trail("example")
#' result <- mtcars |>
#'   tab_tap(cyl, .trail = trail, .label = "by_cyl") |>
#'   dplyr::filter(mpg > 20) |>
#'   tab_tap(cyl, .trail = trail, .label = "by_cyl_filtered")
#' print(trail)
#'
#' @family audit trail
#' @export
tab_tap <- function(.data, ..., .trail, .label = NULL, .wt = NULL,
                    .sort = c("value_asc", "value_desc", "freq_desc", "freq_asc"),
                    .cutoff = NULL, .na = c("include", "exclude", "only"),
                    .display = c("count", "row_pct", "col_pct", "total_pct"),
                    .numeric_summary = TRUE,
                    .cols_include = NULL, .cols_exclude = NULL) {
  data_expr <- substitute(.data)

  # Capture quosures before force() for consistency with other taps
  dots <- rlang::enquos(...)
  wt_quo <- rlang::enquo(.wt)

  # Validate arguments that don't require .data evaluation
  if (!inherits(.trail, "audit_trail")) {
    cli::cli_abort("{.arg .trail} must be an {.cls audit_trail} object.")
  }
  if (!is.null(.label)) {
    if (!is.character(.label) || length(.label) != 1L || is.na(.label)) {
      cli::cli_abort("{.arg .label} must be a single character string or NULL.")
    }
  }

  .sort <- match.arg(.sort)
  .na <- match.arg(.na)
  .display <- match.arg(.display)

  force(.data)

  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble, not {.cls {class(.data)}}.")
  }

  # Build tab result as a custom diagnostic
  tab_result <- tab(.data, !!!dots, .wt = !!wt_quo,
                    .sort = .sort, .cutoff = .cutoff, .na = .na,
                    .display = .display)

  # Build a descriptive name for the tab diagnostic
  var_names <- vapply(dots, rlang::as_label, character(1))
  tab_name <- paste0("tab(", paste(var_names, collapse = ", "), ")")

  # Delegate to audit_tap with the tab result as a custom function
  audit_tap(.data, .trail = .trail, .label = .label,
            .fns = stats::setNames(list(function(.d) tab_result), tab_name),
            .numeric_summary = .numeric_summary,
            .cols_include = .cols_include,
            .cols_exclude = .cols_exclude)

  invisible(.data)
}

#' Validate Primary Keys
#'
#' Tests whether a set of columns constitute primary keys of a data.frame,
#' i.e., whether they uniquely identify every row in the table.
#'
#' @param .data A data.frame or tibble.
#' @param keys Character vector of column names to test as primary keys.
#'
#' @returns An S3 object of class `validate_pk` containing:
#' \describe{
#'   \item{table_name}{Name of the input table from the original call}
#'   \item{keys}{Character vector of column names tested}
#'   \item{is_primary_key}{Logical: TRUE if keys uniquely identify all rows AND
#'     no key column contains NA values}
#'   \item{n_rows}{Total number of rows in the table}
#'   \item{n_unique_keys}{Number of distinct key combinations}
#'   \item{n_duplicate_keys}{Number of key combinations that appear more than once}
#'   \item{duplicate_keys}{A data.frame of duplicated key values with their counts}
#'   \item{has_numeric_keys}{Logical: TRUE if any key column is of type double}
#'   \item{has_na_keys}{Logical: TRUE if any key column contains NA values}
#'   \item{na_in_keys}{Named logical vector indicating which key columns contain NAs}
#' }
#'
#' @examples
#' df <- data.frame(
#'   id = c(1L, 2L, 3L, 4L),
#'   group = c("A", "A", "B", "B"),
#'   value = c(10, 20, 30, 40)
#' )
#' validate_primary_keys(df, "id")
#' validate_primary_keys(df, "group")
#'
#' @family join validation
#' @export
validate_primary_keys <- function(.data, keys) {
  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble.")
  }

  table_name <- deparse(substitute(.data))

  if (!is.character(keys) || length(keys) == 0L) {
    cli::cli_abort("{.arg keys} must be a non-empty character vector.")
  }

  miss <- setdiff(keys, names(.data))
  if (length(miss) > 0L) {
    cli::cli_abort("Column{?s} not found in table: {.field {miss}}")
  }

  # Warn about numeric (double) keys
  numeric_keys <- keys[vapply(.data[keys], is.double, logical(1))]
  has_numeric_keys <- length(numeric_keys) > 0L
  if (has_numeric_keys) {
    cli::cli_warn(
      "Key column{?s} {.field {numeric_keys}} {?is/are} numeric (double). Consider using integer or character for exact matching."
    )
  }

  # Check for NA values in key columns (NA keys violate PK semantics)
  na_in_keys <- vapply(keys, function(k) anyNA(.data[[k]]), logical(1L))
  has_na_keys <- any(na_in_keys)

  # Count occurrences of each key combination
  key_counts <- dplyr::count(.data, dplyr::across(dplyr::all_of(keys)), name = "n")
  n_rows <- nrow(.data)
  n_unique_keys <- nrow(key_counts)

  # Identify duplicates
  duplicate_keys <- dplyr::filter(key_counts, .data$n > 1L)
  n_duplicate_keys <- nrow(duplicate_keys)

  is_primary_key <- (n_unique_keys == n_rows) && !has_na_keys

  out <- list(
    table_name = table_name,
    keys = keys,
    is_primary_key = is_primary_key,
    n_rows = n_rows,
    n_unique_keys = n_unique_keys,
    n_duplicate_keys = n_duplicate_keys,
    duplicate_keys = duplicate_keys,
    has_numeric_keys = has_numeric_keys,
    has_na_keys = has_na_keys,
    na_in_keys = na_in_keys
  )
  structure(out, class = c("validate_pk", "list"))
}

#' @rdname validate_primary_keys
#' @param x An object to print.
#' @param ... Additional arguments (currently unused).
#' @export
print.validate_pk <- function(x, ...) {
  cli::cli_h1("Primary Key Validation")
  cli::cli_text("Table: {x$table_name}")
  cli::cli_text("Key column{?s}: {.field {x$keys}}")
  cli::cli_text("")

  tbl <- data.frame(
    Metric = c("Total rows", "Unique key combinations", "Duplicate key combos"),
    Value = c(
      format(x$n_rows, big.mark = ",", trim = TRUE),
      format(x$n_unique_keys, big.mark = ",", trim = TRUE),
      format(x$n_duplicate_keys, big.mark = ",", trim = TRUE)
    ),
    stringsAsFactors = FALSE
  )
  .cli_table(tbl, right_align = "Value")

  cli::cli_text("")
  if (x$is_primary_key) {
    cli::cli_alert_success("YES - Keys uniquely identify all rows.")
  } else {
    cli::cli_alert_danger("NO - Keys do NOT uniquely identify all rows.")
    if (x$n_duplicate_keys > 0L) {
      cli::cli_text("")
      cli::cli_text("{.strong Duplicate keys (showing up to 10):}")
      print(utils::head(x$duplicate_keys, 10L))
    }
  }

  if (x$has_na_keys) {
    na_cols <- names(x$na_in_keys)[x$na_in_keys]
    cli::cli_text("")
    cli::cli_alert_warning("NA values found in key column{?s}: {.field {na_cols}}. Primary keys must not contain NAs.")
  }

  if (x$has_numeric_keys) {
    cli::cli_text("")
    cli::cli_alert_warning("Numeric (double) key column(s) detected.")
  }

  invisible(x)
}

#' Validate Variable Relationship
#'
#' Determines the relationship between two variables in a data.frame:
#' one-to-one, one-to-many, many-to-one, or many-to-many.
#'
#' @param .data A data.frame or tibble.
#' @param var1 Character string: name of the first variable.
#' @param var2 Character string: name of the second variable.
#'
#' @returns An S3 object of class `validate_var_rel` containing:
#' \describe{
#'   \item{table_name}{Name of the input table}
#'   \item{var1, var2}{Names of the variables analyzed}
#'   \item{relation}{Character string: "one-to-one", "one-to-many",
#'     "many-to-one", or "many-to-many"}
#'   \item{var1_unique}{Number of distinct values in var1}
#'   \item{var2_unique}{Number of distinct values in var2}
#'   \item{n_combinations}{Number of unique (var1, var2) pairs}
#'   \item{var1_has_dups}{Does any var1 value map to multiple var2 values?}
#'   \item{var2_has_dups}{Does any var2 value map to multiple var1 values?}
#' }
#'
#' @details
#' Only accepts variables of type character, integer, or factor. Numeric
#' (double) variables are not allowed due to floating-point comparison issues.
#'
#' @examples
#' df <- data.frame(
#'   person_id = c(1L, 2L, 3L, 4L),
#'   department = c("Sales", "Sales", "Engineering", "Engineering"),
#'   country = c("US", "US", "US", "UK")
#' )
#' validate_var_relationship(df, "person_id", "department")
#'
#' @family join validation
#' @export
validate_var_relationship <- function(.data, var1, var2) {
  if (!is.data.frame(.data)) {
    cli::cli_abort("{.arg .data} must be a data.frame or tibble.")
  }

  table_name <- deparse(substitute(.data))

  if (!is.character(var1) || length(var1) != 1L) {
    cli::cli_abort("{.arg var1} must be a single character string.")
  }
  if (!is.character(var2) || length(var2) != 1L) {
    cli::cli_abort("{.arg var2} must be a single character string.")
  }

  if (!var1 %in% names(.data)) {
    cli::cli_abort("Variable {.field {var1}} not found in table.")
  }
  if (!var2 %in% names(.data)) {
    cli::cli_abort("Variable {.field {var2}} not found in table.")
  }

  # Check allowed types
  allowed_types <- c("character", "integer", "factor")
  var1_class <- class(.data[[var1]])[1L]
  var2_class <- class(.data[[var2]])[1L]

  if (!var1_class %in% allowed_types) {
    cli::cli_abort("Variable {.field {var1}} must be character, integer, or factor. Got: {.cls {var1_class}}")
  }
  if (!var2_class %in% allowed_types) {
    cli::cli_abort("Variable {.field {var2}} must be character, integer, or factor. Got: {.cls {var2_class}}")
  }

  # Unique combinations
  combos <- dplyr::distinct(.data, dplyr::across(dplyr::all_of(c(var1, var2))))
  n_combinations <- nrow(combos)

  var1_unique <- dplyr::n_distinct(.data[[var1]])
  var2_unique <- dplyr::n_distinct(.data[[var2]])

  # Check if var1 has duplicates (same var1 -> different var2)
  var1_mapping <- dplyr::count(combos, dplyr::across(dplyr::all_of(var1)), name = "n")
  var1_has_dups <- any(var1_mapping[["n"]] > 1L)

  # Check if var2 has duplicates (same var2 -> different var1)
  var2_mapping <- dplyr::count(combos, dplyr::across(dplyr::all_of(var2)), name = "n")
  var2_has_dups <- any(var2_mapping[["n"]] > 1L)

  # Determine relationship
  relation <- if (!var1_has_dups && !var2_has_dups) {
    "one-to-one"
  } else if (!var1_has_dups && var2_has_dups) {
    "many-to-one"
  } else if (var1_has_dups && !var2_has_dups) {
    "one-to-many"
  } else {
    "many-to-many"
  }

  out <- list(
    table_name = table_name,
    var1 = var1,
    var2 = var2,
    relation = relation,
    var1_unique = var1_unique,
    var2_unique = var2_unique,
    n_combinations = n_combinations,
    var1_has_dups = var1_has_dups,
    var2_has_dups = var2_has_dups
  )
  structure(out, class = c("validate_var_rel", "list"))
}

#' @rdname validate_var_relationship
#' @param x An object to print.
#' @param ... Additional arguments (currently unused).
#' @export
print.validate_var_rel <- function(x, ...) {
  cli::cli_h1("Variable Relationship Validation")
  cli::cli_text("Table: {x$table_name}")
  cli::cli_text("Variables: {.field {x$var1}} \u2194 {.field {x$var2}}")
  cli::cli_text("")

  tbl <- data.frame(
    Metric = c(
      paste("Unique values in", x$var1),
      paste("Unique values in", x$var2),
      paste0("Unique (", x$var1, ", ", x$var2, ") pairs")
    ),
    Value = c(
      format(x$var1_unique, big.mark = ",", trim = TRUE),
      format(x$var2_unique, big.mark = ",", trim = TRUE),
      format(x$n_combinations, big.mark = ",", trim = TRUE)
    ),
    stringsAsFactors = FALSE
  )
  .cli_table(tbl, right_align = "Value")

  cli::cli_text("")
  v1_dir <- if (x$var1_has_dups) "one-to-many" else "one-to-one"
  v2_dir <- if (x$var2_has_dups) "one-to-many" else "one-to-one"
  cli::cli_text("  {x$var1} \u2192 {x$var2}: {v1_dir}")
  cli::cli_text("  {x$var2} \u2192 {x$var1}: {v2_dir}")
  cli::cli_text("")
  cli::cli_text("Relationship: {.strong {toupper(x$relation)}}")

  invisible(x)
}

test_that("validate_primary_keys detects valid primary key", {
  df <- data.frame(id = 1:4, group = c("A", "A", "B", "B"),
                   stringsAsFactors = FALSE)
  result <- validate_primary_keys(df, "id")

  expect_s3_class(result, "validate_pk")
  expect_true(result$is_primary_key)
  expect_equal(result$n_rows, 4L)
  expect_equal(result$n_unique_keys, 4L)
  expect_equal(result$n_duplicate_keys, 0L)
})

test_that("validate_primary_keys detects non-primary key", {
  df <- data.frame(id = 1:4, group = c("A", "A", "B", "B"),
                   stringsAsFactors = FALSE)
  result <- validate_primary_keys(df, "group")

  expect_false(result$is_primary_key)
  expect_equal(result$n_unique_keys, 2L)
  expect_equal(result$n_duplicate_keys, 2L)
  expect_true(nrow(result$duplicate_keys) > 0L)
})

test_that("validate_primary_keys works with composite keys", {
  df <- data.frame(a = c(1L, 1L, 2L, 2L), b = c("x", "y", "x", "y"),
                   stringsAsFactors = FALSE)
  result <- validate_primary_keys(df, c("a", "b"))

  expect_true(result$is_primary_key)
  expect_equal(result$n_unique_keys, 4L)
})

test_that("validate_primary_keys warns about numeric double keys", {
  df <- data.frame(id = c(1.1, 2.2, 3.3))
  expect_warning(result <- validate_primary_keys(df, "id"), "numeric")
  expect_true(result$has_numeric_keys)
})

test_that("validate_primary_keys errors on missing columns", {
  df <- data.frame(id = 1:3)
  expect_error(validate_primary_keys(df, "nonexistent"), "not found")
})

test_that("validate_primary_keys errors on non-character keys", {
  df <- data.frame(id = 1:3)
  expect_error(validate_primary_keys(df, 1), "character vector")
})

test_that("print.validate_pk produces output for valid PK", {
  df <- data.frame(id = 1:3)
  result <- validate_primary_keys(df, "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Primary Key", combined))
  expect_true(grepl("YES", combined))
})

test_that("print.validate_pk shows duplicates for invalid PK", {
  df <- data.frame(id = c(1L, 1L, 2L))
  result <- validate_primary_keys(df, "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("NO", combined))
})

test_that("validate_primary_keys works with tibbles", {
  skip_if_not_installed("dplyr")
  df <- dplyr::tibble(id = 1:5, val = letters[1:5])
  result <- validate_primary_keys(df, "id")
  expect_true(result$is_primary_key)
})

# validate_var_relationship tests

test_that("validate_var_relationship detects one-to-one", {
  df <- data.frame(a = c(1L, 2L, 3L), b = c("x", "y", "z"),
                   stringsAsFactors = FALSE)
  result <- validate_var_relationship(df, "a", "b")

  expect_s3_class(result, "validate_var_rel")
  expect_equal(result$relation, "one-to-one")
  expect_false(result$var1_has_dups)
  expect_false(result$var2_has_dups)
})

test_that("validate_var_relationship detects one-to-many", {
  # a=1 maps to {x, y}, a=2 maps to {z}: var1_has_dups = TRUE
  # b=x maps to {1}, b=y maps to {1}, b=z maps to {2}: var2_has_dups = FALSE
  df <- data.frame(a = c(1L, 1L, 2L), b = c("x", "y", "z"),
                   stringsAsFactors = FALSE)
  result <- validate_var_relationship(df, "a", "b")

  expect_equal(result$relation, "one-to-many")
  expect_true(result$var1_has_dups)
  expect_false(result$var2_has_dups)
})

test_that("validate_var_relationship detects many-to-one", {
  df <- data.frame(dept = c("Sales", "Sales", "Eng", "Eng"),
                   person = c(1L, 2L, 3L, 4L),
                   stringsAsFactors = FALSE)
  result <- validate_var_relationship(df, "person", "dept")

  expect_equal(result$relation, "many-to-one")
})

test_that("validate_var_relationship detects many-to-many", {
  df <- data.frame(a = c(1L, 1L, 2L, 2L), b = c("x", "y", "x", "y"),
                   stringsAsFactors = FALSE)
  result <- validate_var_relationship(df, "a", "b")

  expect_equal(result$relation, "many-to-many")
  expect_true(result$var1_has_dups)
  expect_true(result$var2_has_dups)
})

test_that("validate_var_relationship rejects numeric double columns", {
  df <- data.frame(a = c(1.1, 2.2), b = c("x", "y"), stringsAsFactors = FALSE)
  expect_error(validate_var_relationship(df, "a", "b"), "character, integer, or factor")
})

test_that("validate_var_relationship errors on missing variable", {
  df <- data.frame(a = 1:3)
  expect_error(validate_var_relationship(df, "a", "nonexistent"), "not found")
})

test_that("print.validate_var_rel produces output", {
  df <- data.frame(a = c(1L, 2L, 3L), b = c("x", "y", "z"),
                   stringsAsFactors = FALSE)
  result <- validate_var_relationship(df, "a", "b")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Variable Relationship", combined))
  expect_true(grepl("ONE-TO-ONE", combined))
})

test_that("validate_var_relationship works with factors", {
  df <- data.frame(a = factor(c("x", "y", "z")), b = c(1L, 2L, 3L))
  result <- validate_var_relationship(df, "a", "b")

  expect_equal(result$relation, "one-to-one")
})

# ---------------------------------------------------------------------------
# validate_primary_keys: NA key values (Fix 2)
# ---------------------------------------------------------------------------

test_that("validate_primary_keys returns FALSE when single NA in integer key", {
  df <- data.frame(id = c(1L, 2L, NA_integer_))
  result <- validate_primary_keys(df, "id")

  expect_false(result$is_primary_key)
  expect_true(result$has_na_keys)
  expect_true(result$na_in_keys[["id"]])
})

test_that("validate_primary_keys returns FALSE when single NA in character key", {
  df <- data.frame(id = c("a", "b", NA_character_), stringsAsFactors = FALSE)
  result <- validate_primary_keys(df, "id")

  expect_false(result$is_primary_key)
  expect_true(result$has_na_keys)
})

test_that("validate_primary_keys has_na_keys=FALSE when no NAs present", {
  df <- data.frame(id = 1:3)
  result <- validate_primary_keys(df, "id")

  expect_true(result$is_primary_key)
  expect_false(result$has_na_keys)
  expect_false(any(result$na_in_keys))
})

test_that("validate_primary_keys returns FALSE with multiple NAs (duplicate NA keys)", {
  df <- data.frame(id = c(1L, NA_integer_, NA_integer_))
  result <- validate_primary_keys(df, "id")

  expect_false(result$is_primary_key)
  expect_true(result$has_na_keys)
})

test_that("print.validate_pk warns about NA key columns", {
  df <- data.frame(id = c(1L, 2L, NA_integer_))
  result <- validate_primary_keys(df, "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("NA values", combined, ignore.case = TRUE))
})

# Bind a list of data.frames, dropping NULLs. Returns NULL if nothing remains.
.rbind_compact <- function(xs) {
  xs <- Filter(Negate(is.null), xs)
  if (!length(xs)) return(NULL)
  out <- do.call(rbind, xs)
  row.names(out) <- NULL
  out
}

# Shared helper: compute numeric summary, discrepancies, and row-level disc count
# from pre-computed col_diffs. key_fn(keep_indices) returns a key data.frame.
.analyze_numeric_diffs <- function(col_diffs, num_cols, tol, top_n, key_fn) {
  # Numeric summary (quantiles + n_over_tol including NA-vs-value pairs)
  num_list <- lapply(num_cols, function(nm) {
    cd <- col_diffs[[nm]]
    d_clean <- cd$d[!is.na(cd$d)]
    if (!length(d_clean) && !any(cd$na_mismatch)) return(NULL)
    n_over <- sum(d_clean > tol) + sum(cd$na_mismatch)
    if (length(d_clean)) {
      qs <- quantile(d_clean, probs = c(0, 0.25, 0.5, 0.75, 1),
                     na.rm = TRUE, names = FALSE)
    } else {
      qs <- rep(NA_real_, 5L)
    }
    data.frame(column = nm, n = length(d_clean),
               min = qs[1], q25 = qs[2], median = qs[3], q75 = qs[4], max = qs[5],
               n_over_tol = n_over,
               stringsAsFactors = FALSE)
  })
  num_summary <- .rbind_compact(num_list)

  # Row-level discrepancies (numeric diffs > tol OR NA-vs-value)
  disc_list <- lapply(num_cols, function(nm) {
    cd <- col_diffs[[nm]]
    keep <- which((!is.na(cd$d) & cd$d > tol) | cd$na_mismatch)
    if (!length(keep)) return(NULL)
    # Relative difference: abs_diff / max(|value_x|, |value_y|), as proportion
    pct <- ifelse(
      is.na(cd$d[keep]), NA_real_,
      ifelse(cd$v1[keep] == 0 & cd$v2[keep] == 0, 0,
             cd$d[keep] / pmax(abs(cd$v1[keep]), abs(cd$v2[keep])))
    )
    df <- cbind(
      key_fn(keep),
      data.frame(column = nm,
                 value_x = cd$v1[keep], value_y = cd$v2[keep],
                 abs_diff = cd$d[keep], pct_diff = pct,
                 stringsAsFactors = FALSE)
    )
    # Sort: finite diffs descending, then NA diffs at the end
    df <- df[order(is.na(df$abs_diff), -replace(df$abs_diff, is.na(df$abs_diff), 0)), ]
    head(df, top_n)
  })
  discrepancies <- .rbind_compact(disc_list)

  list(
    num_summary = num_summary,
    discrepancies = discrepancies
  )
}

# Build a match_summary list from disc counts and unmatched counts
.build_match_summary <- function(n_with_disc, n_matched, n_only_x, n_only_y) {
  n_no_disc <- n_matched - n_with_disc
  pct_no <- if (n_matched > 0L) n_no_disc / n_matched else NA_real_
  pct_with <- if (n_matched > 0L) n_with_disc / n_matched else NA_real_
  list(
    only_x = n_only_x,
    only_y = n_only_y,
    matched_no_disc = n_no_disc,
    matched_with_disc = n_with_disc,
    pct_no_disc = pct_no,
    pct_with_disc = pct_with
  )
}

# Shared helper: compute categorical summary, discrepancies, and row-level disc count
# from pre-computed col_diffs_cat. key_fn(keep_indices) returns a key data.frame.
.analyze_categorical_diffs <- function(col_diffs_cat, cat_cols, top_n, key_fn) {
  # Categorical summary (n_compared, n_mismatched, pct, NA mismatches)
  cat_list <- lapply(cat_cols, function(nm) {
    cd <- col_diffs_cat[[nm]]
    n_compared <- sum(!is.na(cd$v1) | !is.na(cd$v2))
    n_mis <- sum(cd$mismatch)
    n_na <- sum(cd$na_mismatch)
    if (n_compared == 0L && n_na == 0L) return(NULL)
    data.frame(column = nm, n_compared = n_compared,
               n_mismatched = n_mis,
               pct_mismatched = if (n_compared > 0L) n_mis / n_compared else NA_real_,
               n_na_mismatch = n_na,
               stringsAsFactors = FALSE)
  })
  cat_summary <- .rbind_compact(cat_list)

  # Row-level categorical discrepancies
  disc_list <- lapply(cat_cols, function(nm) {
    cd <- col_diffs_cat[[nm]]
    keep <- which(cd$mismatch)
    if (!length(keep)) return(NULL)
    df <- cbind(
      key_fn(keep),
      data.frame(column = nm,
                 value_x = cd$v1[keep], value_y = cd$v2[keep],
                 stringsAsFactors = FALSE)
    )
    head(df, top_n)
  })
  cat_discrepancies <- .rbind_compact(disc_list)

  list(
    cat_summary = cat_summary,
    cat_discrepancies = cat_discrepancies
  )
}

#' Compare Two Tables
#'
#' Compares two data.frames or tibbles by examining column names, row counts,
#' key overlap, numeric discrepancies, and categorical discrepancies. Useful for
#' validating data processing pipelines.
#'
#' @param x First data.frame or tibble to compare.
#' @param y Second data.frame or tibble to compare.
#' @param key_cols Character vector of column names to use as keys for matching
#'   rows. If `NULL` (default), automatically detects character, factor, and
#'   integer columns as keys.
#' @param tol Numeric tolerance for comparing numeric columns. Differences
#'   less than or equal to `tol` are considered equal. Defaults to
#'   [`.Machine$double.eps`][.Machine] (machine double-precision).
#' @param top_n Maximum number of row-level discrepancies to store **per
#'   column** (numeric and categorical), and maximum unmatched keys to store.
#'   Defaults to `Inf` (all). Unmatched keys are stored in arbitrary order.
#' @param compare_cols Character vector of column names to compare. If `NULL`
#'   (default), all common non-key columns are compared. Mutually exclusive
#'   with `exclude_cols`.
#' @param exclude_cols Character vector of column names to exclude from
#'   comparison. If `NULL` (default), no columns are excluded. Mutually
#'   exclusive with `compare_cols`.
#'
#' @returns An S3 object of class `compare_tbl` containing:
#' \describe{
#'   \item{name_x, name_y}{Names of the compared objects}
#'   \item{common_columns}{Column names present in both tables}
#'   \item{only_x}{Column names only in x}
#'   \item{only_y}{Column names only in y}
#'   \item{type_mismatches}{Data.frame of columns with different types, or NULL}
#'   \item{nrow_x}{Number of rows in x}
#'   \item{nrow_y}{Number of rows in y}
#'   \item{key_summary}{Summary of key overlap, or NULL}
#'   \item{numeric_summary}{Data.frame of numeric discrepancy quantiles (with
#'     `n_over_tol` count), or NULL}
#'   \item{comparison_method}{How columns were compared (`"keys"`,
#'     `"row_index"`, or `NA`)}
#'   \item{rows_matched}{Number of rows matched on keys}
#'   \item{tol}{The tolerance used}
#'   \item{top_n}{The top_n used}
#'   \item{discrepancies}{Data.frame of row-level numeric discrepancies
#'     exceeding `tol` (or where one side is `NA`), with key columns (or
#'     `row_index`), `column`, `value_x`, `value_y`, `abs_diff`, and
#'     `pct_diff` (relative difference as a proportion). NULL if none.}
#'   \item{categorical_summary}{Data.frame with `column`, `n_compared`,
#'     `n_mismatched`, `pct_mismatched` (proportion, 0--1), `n_na_mismatch`,
#'     or NULL}
#'   \item{categorical_discrepancies}{Data.frame of row-level categorical
#'     discrepancies with key columns (or `row_index`), `column`, `value_x`,
#'     `value_y`. NULL if none.}
#'   \item{total_discrepancies}{Total number of cell-level discrepancies
#'     across all column types (not limited by `top_n`)}
#'   \item{only_x_keys}{Data.frame of key combinations only in x (up to
#'     `top_n` rows), or NULL}
#'   \item{only_y_keys}{Data.frame of key combinations only in y (up to
#'     `top_n` rows), or NULL}
#'   \item{match_summary}{List with `only_x`, `only_y`, `matched_no_disc`,
#'     `matched_with_disc`, `pct_no_disc` (proportion, 0--1),
#'     `pct_with_disc` (proportion, 0--1)}
#' }
#'
#' Use [as.data.frame()] to extract all discrepancies (numeric and categorical)
#' as a single tidy data.frame.
#'
#' @examples
#' x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
#' y <- data.frame(id = 1:3, value = c(10.1, 20.0, 30.5))
#' compare_tables(x, y)
#'
#' # With tolerance — differences <= 0.15 are considered equal
#' compare_tables(x, y, tol = 0.15)
#'
#' # Categorical columns are also compared
#' a <- data.frame(id = 1:3, status = c("ok", "warn", "fail"),
#'                  stringsAsFactors = FALSE)
#' b <- data.frame(id = 1:3, status = c("ok", "warn", "error"),
#'                  stringsAsFactors = FALSE)
#' compare_tables(a, b)
#'
#' @family join validation
#' @export
compare_tables <- function(x, y, key_cols = NULL, tol = .Machine$double.eps,
                           top_n = Inf, compare_cols = NULL,
                           exclude_cols = NULL) {
  name_x <- deparse(substitute(x))
  name_y <- deparse(substitute(y))

  if (!is.data.frame(x)) {
    cli::cli_abort("{.arg x} must be a data.frame or tibble.")
  }
  if (!is.data.frame(y)) {
    cli::cli_abort("{.arg y} must be a data.frame or tibble.")
  }
  if (!is.numeric(tol) || length(tol) != 1L || is.na(tol) || tol < 0) {
    cli::cli_abort("{.arg tol} must be a single non-negative number.")
  }
  if (!is.numeric(top_n) || length(top_n) != 1L || is.na(top_n) || top_n <= 0) {
    cli::cli_abort("{.arg top_n} must be a single positive number or {.code Inf}.")
  }
  if (!is.null(compare_cols) && !is.null(exclude_cols)) {
    cli::cli_abort("{.arg compare_cols} and {.arg exclude_cols} cannot both be specified.")
  }
  if (!is.null(compare_cols) && (!is.character(compare_cols) || length(compare_cols) == 0L)) {
    cli::cli_abort("{.arg compare_cols} must be a character vector.")
  }
  if (!is.null(exclude_cols) && (!is.character(exclude_cols) || length(exclude_cols) == 0L)) {
    cli::cli_abort("{.arg exclude_cols} must be a character vector.")
  }

  n1 <- nrow(x)
  n2 <- nrow(y)

  names1 <- names(x)
  names2 <- names(y)

  common_cols <- intersect(names1, names2)
  only_x <- setdiff(names1, names2)
  only_y <- setdiff(names2, names1)

  # Type mismatches in common columns
  type_mismatches <- NULL
  if (length(common_cols) > 0L) {
    type_list <- lapply(common_cols, function(nm) {
      type1 <- class(x[[nm]])[1L]
      type2 <- class(y[[nm]])[1L]
      if (type1 != type2) {
        data.frame(column = nm, type_x = type1, type_y = type2,
                   stringsAsFactors = FALSE)
      } else {
        NULL
      }
    })
    type_mismatches <- .rbind_compact(type_list)
  }

  if (length(common_cols) == 0L) {
    cli::cli_abort("No matching column names between {.arg x} and {.arg y}.")
  }

  # Determine keys
  mismatched_cols <- if (!is.null(type_mismatches)) type_mismatches$column else character(0L)
  auto_keys <- FALSE

  if (is.null(key_cols)) {
    key_cols <- common_cols[vapply(common_cols, function(nm) {
      if (nm %in% mismatched_cols) return(FALSE)
      is_key_type <- function(col) is.integer(col) || is.factor(col) || is.character(col)
      is_key_type(x[[nm]]) && is_key_type(y[[nm]])
    }, logical(1L))]
    auto_keys <- TRUE
  } else {
    miss1 <- setdiff(key_cols, names1)
    miss2 <- setdiff(key_cols, names2)
    if (length(miss1) > 0L || length(miss2) > 0L) {
      cli::cli_abort(c(
        "Some {.arg key_cols} not present in both tables.",
        if (length(miss1)) paste0("Missing from x: ", paste(miss1, collapse = ", ")),
        if (length(miss2)) paste0("Missing from y: ", paste(miss2, collapse = ", "))
      ))
    }
  }

  # Key overlap
  key_summary <- NULL
  only_x_keys_df <- NULL
  only_y_keys_df <- NULL
  if (length(key_cols) > 0L) {
    u1 <- dplyr::distinct(x[key_cols])
    u2 <- dplyr::distinct(y[key_cols])
    match_keys <- dplyr::intersect(u1, u2)
    only_x_keys_full <- dplyr::setdiff(u1, u2)
    only_y_keys_full <- dplyr::setdiff(u2, u1)

    key_summary <- list(
      keys = key_cols,
      auto = auto_keys,
      x_unique = nrow(u1),
      y_unique = nrow(u2),
      matches = nrow(match_keys),
      only_x = nrow(only_x_keys_full),
      only_y = nrow(only_y_keys_full)
    )

    if (nrow(only_x_keys_full) > 0L) {
      only_x_keys_df <- as.data.frame(head(only_x_keys_full, top_n),
                                       stringsAsFactors = FALSE)
    }
    if (nrow(only_y_keys_full) > 0L) {
      only_y_keys_df <- as.data.frame(head(only_y_keys_full, top_n),
                                       stringsAsFactors = FALSE)
    }
  }

  # Determine value columns to compare (non-key common columns)
  value_cols <- setdiff(common_cols, key_cols)
  if (!is.null(compare_cols)) {
    bad <- setdiff(compare_cols, value_cols)
    if (length(bad) > 0L) {
      cli::cli_abort(c(
        "Some {.arg compare_cols} not found among common non-key columns.",
        paste0("Not found: ", paste(bad, collapse = ", "))
      ))
    }
    value_cols <- intersect(value_cols, compare_cols)
  }
  if (!is.null(exclude_cols)) {
    unknown <- setdiff(exclude_cols, value_cols)
    if (length(unknown) > 0L) {
      cli::cli_warn(
        "Some {.arg exclude_cols} not found among common non-key columns and will be ignored: {.field {unknown}}."
      )
    }
    value_cols <- setdiff(value_cols, exclude_cols)
  }

  # Split value columns into numeric and categorical
  num_cols <- value_cols[vapply(value_cols, function(nm) {
    if (nm %in% mismatched_cols) return(FALSE)
    is.numeric(x[[nm]]) && is.numeric(y[[nm]])
  }, logical(1L))]

  cat_cols <- value_cols[vapply(value_cols, function(nm) {
    if (nm %in% mismatched_cols) return(FALSE)
    if (nm %in% num_cols) return(FALSE)
    TRUE
  }, logical(1L))]

  num_summary <- NULL
  cat_summary <- NULL
  comparison_method <- NA_character_
  rows_matched <- NA_integer_
  discrepancies <- NULL
  cat_discrepancies <- NULL
  match_summary <- NULL

  has_num <- length(num_cols) > 0L
  has_cat <- length(cat_cols) > 0L

  if (has_num || has_cat) {
    if (length(key_cols) == 0L) {
      comparison_method <- "row_index"
      maxn <- min(n1, n2)
      idx <- seq_len(maxn)

      key_fn <- function(keep) {
        data.frame(row_index = idx[keep], stringsAsFactors = FALSE)
      }

      # Numeric analysis
      if (has_num) {
        col_diffs <- lapply(num_cols, function(nm) {
          v1 <- x[[nm]][idx]
          v2 <- y[[nm]][idx]
          d <- abs(v1 - v2)
          na_mismatch <- is.na(v1) != is.na(v2)
          list(v1 = v1, v2 = v2, d = d, na_mismatch = na_mismatch)
        })
        names(col_diffs) <- num_cols
        analysis <- .analyze_numeric_diffs(col_diffs, num_cols, tol, top_n, key_fn)
        num_summary <- analysis$num_summary
        discrepancies <- analysis$discrepancies
      }

      # Categorical analysis
      if (has_cat) {
        col_diffs_cat <- lapply(cat_cols, function(nm) {
          v1 <- as.character(x[[nm]][idx])
          v2 <- as.character(y[[nm]][idx])
          na_mismatch <- is.na(v1) != is.na(v2)
          mismatch <- na_mismatch | (!is.na(v1) & !is.na(v2) & v1 != v2)
          list(v1 = v1, v2 = v2, mismatch = mismatch, na_mismatch = na_mismatch)
        })
        names(col_diffs_cat) <- cat_cols
        cat_analysis <- .analyze_categorical_diffs(col_diffs_cat, cat_cols, top_n, key_fn)
        cat_summary <- cat_analysis$cat_summary
        cat_discrepancies <- cat_analysis$cat_discrepancies
      }

      # Combined row-level disc count (reuse col_diffs / col_diffs_cat)
      row_has_any <- rep(FALSE, maxn)
      if (has_num) {
        for (nm in num_cols) {
          cd <- col_diffs[[nm]]
          row_has_any <- row_has_any |
            (!is.na(cd$d) & cd$d > tol) | cd$na_mismatch
        }
      }
      if (has_cat) {
        for (nm in cat_cols) {
          row_has_any <- row_has_any | col_diffs_cat[[nm]]$mismatch
        }
      }

      n_only_x <- n1 - maxn
      n_only_y <- n2 - maxn
      match_summary <- .build_match_summary(
        sum(row_has_any), maxn, n_only_x, n_only_y
      )

      # Unmatched row indices for row-index mode
      if (n_only_x > 0L) {
        extra_idx <- seq.int(maxn + 1L, n1)
        only_x_keys_df <- data.frame(
          row_index = head(extra_idx, top_n),
          stringsAsFactors = FALSE
        )
      }
      if (n_only_y > 0L) {
        extra_idx <- seq.int(maxn + 1L, n2)
        only_y_keys_df <- data.frame(
          row_index = head(extra_idx, top_n),
          stringsAsFactors = FALSE
        )
      }
    } else {
      comparison_method <- "keys"
      # Warn if keys are non-unique — inner_join can cause cartesian expansion
      x_dup <- nrow(x) != nrow(u1)
      y_dup <- nrow(y) != nrow(u2)
      if (x_dup || y_dup) {
        dup_tbls <- c(if (x_dup) name_x, if (y_dup) name_y)
        cli::cli_warn(
          "Key columns are not unique in {dup_tbls}; comparison may be inflated by cartesian row expansion."
        )
      }
      all_value_cols <- union(num_cols, cat_cols)
      x_sub <- x[c(key_cols, all_value_cols)]
      y_sub <- y[c(key_cols, all_value_cols)]
      merged <- dplyr::inner_join(x_sub, y_sub, by = key_cols, suffix = c(".x", ".y"))
      rows_matched <- nrow(merged)

      key_fn <- function(keep) merged[keep, key_cols, drop = FALSE]

      # Numeric analysis
      if (has_num) {
        col_diffs <- lapply(num_cols, function(nm) {
          v1 <- merged[[paste0(nm, ".x")]]
          v2 <- merged[[paste0(nm, ".y")]]
          d <- abs(v1 - v2)
          na_mismatch <- is.na(v1) != is.na(v2)
          list(v1 = v1, v2 = v2, d = d, na_mismatch = na_mismatch)
        })
        names(col_diffs) <- num_cols
        analysis <- .analyze_numeric_diffs(col_diffs, num_cols, tol, top_n, key_fn)
        num_summary <- analysis$num_summary
        discrepancies <- analysis$discrepancies
      }

      # Categorical analysis
      if (has_cat) {
        col_diffs_cat <- lapply(cat_cols, function(nm) {
          v1 <- as.character(merged[[paste0(nm, ".x")]])
          v2 <- as.character(merged[[paste0(nm, ".y")]])
          na_mismatch <- is.na(v1) != is.na(v2)
          mismatch <- na_mismatch | (!is.na(v1) & !is.na(v2) & v1 != v2)
          list(v1 = v1, v2 = v2, mismatch = mismatch, na_mismatch = na_mismatch)
        })
        names(col_diffs_cat) <- cat_cols
        cat_analysis <- .analyze_categorical_diffs(col_diffs_cat, cat_cols, top_n, key_fn)
        cat_summary <- cat_analysis$cat_summary
        cat_discrepancies <- cat_analysis$cat_discrepancies
      }

      # Combined row-level disc count
      row_has_any <- rep(FALSE, rows_matched)
      if (has_num) {
        for (nm in num_cols) {
          cd <- col_diffs[[nm]]
          row_has_any <- row_has_any |
            (!is.na(cd$d) & cd$d > tol) | cd$na_mismatch
        }
      }
      if (has_cat) {
        for (nm in cat_cols) {
          row_has_any <- row_has_any | col_diffs_cat[[nm]]$mismatch
        }
      }

      match_summary <- .build_match_summary(
        sum(row_has_any), rows_matched,
        key_summary$only_x, key_summary$only_y
      )
    }
  } else if (length(key_cols) > 0L) {
    # Keys exist but no value columns to compare
    match_summary <- list(
      only_x = key_summary$only_x,
      only_y = key_summary$only_y,
      matched_no_disc = key_summary$matches,
      matched_with_disc = 0L,
      pct_no_disc = if (key_summary$matches > 0L) 1 else NA_real_,
      pct_with_disc = if (key_summary$matches > 0L) 0 else NA_real_
    )
  }

  # Total discrepancies (true count from summaries, not limited by top_n)
  n_num_total <- if (!is.null(num_summary)) sum(num_summary$n_over_tol) else 0L
  n_cat_total <- if (!is.null(cat_summary)) sum(cat_summary$n_mismatched) else 0L
  total_discrepancies <- n_num_total + n_cat_total

  result <- list(
    name_x = name_x,
    name_y = name_y,
    common_columns = common_cols,
    only_x = only_x,
    only_y = only_y,
    type_mismatches = type_mismatches,
    nrow_x = n1,
    nrow_y = n2,
    key_summary = key_summary,
    numeric_summary = num_summary,
    comparison_method = comparison_method,
    rows_matched = rows_matched,
    tol = tol,
    top_n = top_n,
    discrepancies = discrepancies,
    categorical_summary = cat_summary,
    categorical_discrepancies = cat_discrepancies,
    total_discrepancies = total_discrepancies,
    only_x_keys = only_x_keys_df,
    only_y_keys = only_y_keys_df,
    match_summary = match_summary
  )
  structure(result, class = c("compare_tbl", "list"))
}

# Print "... and N more" with an optional note when top_n truncated storage
.print_remaining <- function(total, n_displayed, top_n) {
  remaining <- total - n_displayed
  if (remaining <= 0L) return(invisible())
  msg <- format(remaining, big.mark = ",")
  if (is.finite(top_n) && top_n < total) {
    cli::cli_text("    {.emph ... and {msg} more ({format(top_n, big.mark = ',')} stored)}")
  } else {
    cli::cli_text("    {.emph ... and {msg} more}")
  }
}

#' @rdname compare_tables
#' @param show_n Maximum number of rows to display for discrepancies and
#'   unmatched keys in the printed output. Defaults to `5L`.
#' @param ... Additional arguments (currently unused).
#' @export
print.compare_tbl <- function(x, show_n = 5L, ...) {
  if (!is.numeric(show_n) || length(show_n) != 1L || is.na(show_n) || show_n < 0) {
    cli::cli_abort("{.arg show_n} must be a single non-negative integer.")
  }
  show_n <- as.integer(show_n)

  cli::cli_h1("Table Comparison: {x$name_x} vs {x$name_y}")
  sec <- 0L

  # Row counts
  sec <- sec + 1L
  cli::cli_text("{.strong {sec}. Row counts}")
  row_diff <- x$nrow_x - x$nrow_y
  cli::cli_text("  {x$name_x}: {format(x$nrow_x, big.mark = ',')} rows")
  cli::cli_text("  {x$name_y}: {format(x$nrow_y, big.mark = ',')} rows")
  diff_str <- .format_delta(row_diff)
  cli::cli_text("  Difference: {diff_str}")

  # Column names
  cli::cli_text("")
  sec <- sec + 1L
  cli::cli_text("{.strong {sec}. Column names}")
  cli::cli_text("  Matching columns: {length(x$common_columns)}")
  if (length(x$only_x) > 0L) {
    cli::cli_text("  Only in {x$name_x}: {length(x$only_x)} ({.field {x$only_x}})")
  } else {
    cli::cli_text("  Only in {x$name_x}: 0")
  }
  if (length(x$only_y) > 0L) {
    cli::cli_text("  Only in {x$name_y}: {length(x$only_y)} ({.field {x$only_y}})")
  } else {
    cli::cli_text("  Only in {x$name_y}: 0")
  }

  n_mismatches <- if (is.null(x$type_mismatches)) 0L else nrow(x$type_mismatches)
  cli::cli_text("  Type mismatches: {n_mismatches}")
  if (n_mismatches > 0L) {
    print(x$type_mismatches)
  }

  # Key summary
  cli::cli_text("")
  sec <- sec + 1L
  cli::cli_text("{.strong {sec}. Key columns}")
  ks <- x$key_summary
  if (is.null(ks)) {
    if (is.na(x$comparison_method) || x$comparison_method != "row_index") {
      cli::cli_text("  No common character/factor/integer columns found.")
    } else {
      cli::cli_text("  No key columns; comparing by row index.")
    }
  } else {
    auto_lbl <- if (ks$auto) " (auto-detected)" else ""
    cli::cli_text("  Key columns: {.field {ks$keys}}{auto_lbl}")
    cli::cli_text("  Distinct combos in {x$name_x}: {format(ks$x_unique, big.mark = ',')}")
    cli::cli_text("  Distinct combos in {x$name_y}: {format(ks$y_unique, big.mark = ',')}")
  }

  # Row matching (only shown when match_summary exists)
  ms <- x$match_summary
  if (!is.null(ms)) {
    cli::cli_text("")
    sec <- sec + 1L
    tol_val <- x$tol
    tol_lbl <- if (!is.null(tol_val) && tol_val > .Machine$double.eps) {
      paste0(" (tol = ", format(tol_val, scientific = FALSE), ")")
    } else {
      ""
    }
    cli::cli_text("{.strong {sec}. Row matching}{tol_lbl}")
    cli::cli_text("  Only in {x$name_x}: {format(ms$only_x, big.mark = ',')}")
    cli::cli_text("  Only in {x$name_y}: {format(ms$only_y, big.mark = ',')}")
    if (!is.na(ms$pct_no_disc)) {
      pct_no_lbl <- round(ms$pct_no_disc * 100, 1)
      pct_with_lbl <- round(ms$pct_with_disc * 100, 1)
      cli::cli_text("  Matched, no discrepancies: {format(ms$matched_no_disc, big.mark = ',')} ({pct_no_lbl}%)")
      cli::cli_text("  Matched, with discrepancies: {format(ms$matched_with_disc, big.mark = ',')} ({pct_with_lbl}%)")
    } else {
      cli::cli_text("  Matched, no discrepancies: {format(ms$matched_no_disc, big.mark = ',')}")
      cli::cli_text("  Matched, with discrepancies: {format(ms$matched_with_disc, big.mark = ',')}")
    }

    # Total cell discrepancies
    td <- x$total_discrepancies
    if (!is.null(td) && td > 0L) {
      n_num_d <- if (!is.null(x$numeric_summary)) sum(x$numeric_summary$n_over_tol) else 0L
      n_cat_d <- if (!is.null(x$categorical_summary)) sum(x$categorical_summary$n_mismatched) else 0L
      cli::cli_text("  Total cell discrepancies: {format(td, big.mark = ',')} ({format(n_num_d, big.mark = ',')} numeric, {format(n_cat_d, big.mark = ',')} categorical)")
    }

    # Show unmatched keys
    if (!is.null(x$only_x_keys) && nrow(x$only_x_keys) > 0L) {
      cli::cli_text("")
      cli::cli_text("  Unmatched keys in {x$name_x}:")
      display_keys <- head(x$only_x_keys, show_n)
      .cli_table(display_keys, indent = 4L)
      .print_remaining(ms$only_x, nrow(display_keys), x$top_n)
    }
    if (!is.null(x$only_y_keys) && nrow(x$only_y_keys) > 0L) {
      cli::cli_text("")
      cli::cli_text("  Unmatched keys in {x$name_y}:")
      display_keys <- head(x$only_y_keys, show_n)
      .cli_table(display_keys, indent = 4L)
      .print_remaining(ms$only_y, nrow(display_keys), x$top_n)
    }
  }

  # Numeric discrepancies
  cli::cli_text("")
  sec <- sec + 1L
  cli::cli_text("{.strong {sec}. Numeric discrepancies (absolute differences)}")
  if (is.na(x$comparison_method)) {
    cli::cli_text("  No common numeric columns found.")
  } else {
    if (x$comparison_method == "row_index") {
      cli::cli_text("  Compared by row index.")
    } else {
      cli::cli_text("  Compared after merging on keys.")
      cli::cli_text("  Rows matched: {format(x$rows_matched, big.mark = ',')}")
    }

    if (!is.null(x$numeric_summary) && nrow(x$numeric_summary) > 0L) {
      ns <- x$numeric_summary
      display_tbl <- data.frame(
        Column     = ns$column,
        N          = format(ns$n, big.mark = ",", trim = TRUE),
        Min        = format(round(ns$min, 4), trim = TRUE),
        Q25        = format(round(ns$q25, 4), trim = TRUE),
        Median     = format(round(ns$median, 4), trim = TRUE),
        Q75        = format(round(ns$q75, 4), trim = TRUE),
        Max        = format(round(ns$max, 4), trim = TRUE),
        `>tol`     = format(ns$n_over_tol, big.mark = ",", trim = TRUE),
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
      cli::cli_text("")
      .cli_table(display_tbl, right_align = c("N", "Min", "Q25", "Median", "Q75", "Max", ">tol"), indent = 4L)
    }

    # Show top discrepancies
    if (!is.null(x$discrepancies) && nrow(x$discrepancies) > 0L) {
      cli::cli_text("")
      cli::cli_text("  Top discrepancies:")
      disc_show <- head(x$discrepancies, show_n)
      # Format numeric columns for display (handle NA values)
      disc_display <- disc_show
      disc_display$value_x <- ifelse(is.na(disc_display$value_x), "NA",
        format(round(disc_display$value_x, 4), trim = TRUE))
      disc_display$value_y <- ifelse(is.na(disc_display$value_y), "NA",
        format(round(disc_display$value_y, 4), trim = TRUE))
      disc_display$abs_diff <- ifelse(is.na(disc_display$abs_diff), "NA",
        format(round(disc_display$abs_diff, 4), trim = TRUE))
      disc_display$pct_diff <- ifelse(is.na(disc_display$pct_diff), "NA",
        paste0(format(round(disc_display$pct_diff * 100, 1), trim = TRUE), "%"))
      # Determine which columns to right-align
      ralign <- intersect(c("row_index", "value_x", "value_y", "abs_diff", "pct_diff"), names(disc_display))
      .cli_table(disc_display, right_align = ralign, indent = 4L)
      remaining <- nrow(x$discrepancies) - nrow(disc_show)
      if (remaining > 0L) {
        cli::cli_text("    {.emph ... and {format(remaining, big.mark = ',')} more}")
      }
    }
  }

  # Categorical discrepancies
  cli::cli_text("")
  sec <- sec + 1L
  cli::cli_text("{.strong {sec}. Categorical discrepancies}")
  has_comparison <- !is.na(x$comparison_method)
  if (!has_comparison) {
    cli::cli_text("  No value columns compared.")
  } else if (is.null(x$categorical_summary) || nrow(x$categorical_summary) == 0L) {
    cli::cli_text("  No categorical discrepancies found.")
  } else {
    cs <- x$categorical_summary
    display_tbl <- data.frame(
      Column     = cs$column,
      N          = format(cs$n_compared, big.mark = ",", trim = TRUE),
      Mismatched = format(cs$n_mismatched, big.mark = ",", trim = TRUE),
      `%Mismatch` = ifelse(is.na(cs$pct_mismatched), "NA",
        paste0(format(round(cs$pct_mismatched * 100, 1), trim = TRUE), "%")),
      `NA mismatch` = format(cs$n_na_mismatch, big.mark = ",", trim = TRUE),
      stringsAsFactors = FALSE,
      check.names = FALSE
    )
    cli::cli_text("")
    .cli_table(display_tbl, right_align = c("N", "Mismatched", "%Mismatch", "NA mismatch"), indent = 4L)

    n_total_na <- sum(cs$n_na_mismatch)
    if (n_total_na > 0L) {
      cli::cli_text("  {.emph Note: {format(n_total_na, big.mark = ',')} NA-vs-value mismatch(es) are included in the mismatch counts.}")
    }

    if (!is.null(x$categorical_discrepancies) && nrow(x$categorical_discrepancies) > 0L) {
      cli::cli_text("")
      cli::cli_text("  Top discrepancies:")
      disc_show <- head(x$categorical_discrepancies, show_n)
      disc_display <- disc_show
      disc_display$value_x <- ifelse(is.na(disc_display$value_x), "NA", disc_display$value_x)
      disc_display$value_y <- ifelse(is.na(disc_display$value_y), "NA", disc_display$value_y)
      ralign <- intersect(c("row_index"), names(disc_display))
      .cli_table(disc_display, right_align = ralign, indent = 4L)
      remaining <- nrow(x$categorical_discrepancies) - nrow(disc_show)
      if (remaining > 0L) {
        cli::cli_text("    {.emph ... and {format(remaining, big.mark = ',')} more}")
      }
    }
  }

  invisible(x)
}

# -- as.data.frame method ----------------------------------------------------

#' @rdname compare_tables
#' @param row.names Passed to [as.data.frame()]. Default `NULL`.
#' @param optional Passed to [as.data.frame()]. Default `FALSE`.
#' @export
as.data.frame.compare_tbl <- function(x, row.names = NULL, optional = FALSE, ...) {
  # Determine key column names from either source
  key_nms <- character(0L)
  if (!is.null(x$discrepancies) && nrow(x$discrepancies) > 0L) {
    key_nms <- setdiff(names(x$discrepancies),
                       c("column", "value_x", "value_y", "abs_diff", "pct_diff"))
  } else if (!is.null(x$categorical_discrepancies) && nrow(x$categorical_discrepancies) > 0L) {
    key_nms <- setdiff(names(x$categorical_discrepancies),
                       c("column", "value_x", "value_y"))
  }

  num_part <- NULL
  if (!is.null(x$discrepancies) && nrow(x$discrepancies) > 0L) {
    d <- x$discrepancies
    d$value_x <- as.character(d$value_x)
    d$value_y <- as.character(d$value_y)
    d$type <- "numeric"
    num_part <- d[c(key_nms, "column", "type", "value_x", "value_y", "abs_diff", "pct_diff")]
  }

  cat_part <- NULL
  if (!is.null(x$categorical_discrepancies) && nrow(x$categorical_discrepancies) > 0L) {
    d <- x$categorical_discrepancies
    d$type <- "categorical"
    d$abs_diff <- NA_real_
    d$pct_diff <- NA_real_
    cat_part <- d[c(key_nms, "column", "type", "value_x", "value_y", "abs_diff", "pct_diff")]
  }

  out <- .rbind_compact(list(num_part, cat_part))
  if (is.null(out)) {
    # Return empty data.frame with correctly typed columns
    out <- data.frame(
      column = character(0L), type = character(0L),
      value_x = character(0L), value_y = character(0L),
      abs_diff = numeric(0L), pct_diff = numeric(0L),
      stringsAsFactors = FALSE
    )
    if (length(key_nms)) {
      key_df <- as.data.frame(
        setNames(rep(list(character(0L)), length(key_nms)), key_nms),
        stringsAsFactors = FALSE
      )
      out <- cbind(key_df, out)
    }
  }
  row.names(out) <- row.names
  out
}

test_that("compare_tables detects identical tables", {
  x <- data.frame(id = 1:3, value = c(10, 20, 30))
  result <- compare_tables(x, x)

  expect_s3_class(result, "compare_tbl")
  expect_equal(result$nrow_x, 3L)
  expect_equal(result$nrow_y, 3L)
  expect_equal(length(result$only_x), 0L)
  expect_equal(length(result$only_y), 0L)
})

test_that("compare_tables detects column differences", {
  x <- data.frame(id = 1:3, value = 1:3, extra_x = 4:6)
  y <- data.frame(id = 1:3, value = 1:3, extra_y = 7:9)
  result <- compare_tables(x, y)

  expect_equal(result$only_x, "extra_x")
  expect_equal(result$only_y, "extra_y")
  expect_true("id" %in% result$common_columns)
  expect_true("value" %in% result$common_columns)
})

test_that("compare_tables detects row count differences", {
  x <- data.frame(id = 1:5, value = 1:5)
  y <- data.frame(id = 1:3, value = 1:3)
  result <- compare_tables(x, y)

  expect_equal(result$nrow_x, 5L)
  expect_equal(result$nrow_y, 3L)
})

test_that("compare_tables detects type mismatches", {
  x <- data.frame(id = 1:3, mixed = c(1L, 2L, 3L))
  y <- data.frame(id = 1:3, mixed = c("a", "b", "c"), stringsAsFactors = FALSE)
  result <- compare_tables(x, y)

  expect_false(is.null(result$type_mismatches))
  expect_equal(nrow(result$type_mismatches), 1L)
  expect_equal(result$type_mismatches$column, "mixed")
})

test_that("compare_tables auto-detects keys", {
  x <- data.frame(id = c(1L, 2L, 3L), name = c("a", "b", "c"),
                   value = c(10.0, 20.0, 30.0), stringsAsFactors = FALSE)
  y <- data.frame(id = c(1L, 2L, 3L), name = c("a", "b", "c"),
                   value = c(10.1, 20.0, 30.5), stringsAsFactors = FALSE)
  result <- compare_tables(x, y)

  expect_false(is.null(result$key_summary))
  expect_true(result$key_summary$auto)
  expect_true("id" %in% result$key_summary$keys)
  expect_true("name" %in% result$key_summary$keys)
})

test_that("compare_tables works with explicit key_cols", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.1, 20.0, 30.5))
  result <- compare_tables(x, y, key_cols = "id")

  expect_false(is.null(result$key_summary))
  expect_false(result$key_summary$auto)
  expect_equal(result$key_summary$keys, "id")
})

test_that("compare_tables computes numeric discrepancies with keys", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.5, 20.0, 30.5))
  result <- compare_tables(x, y, key_cols = "id")

  expect_false(is.null(result$numeric_summary))
  expect_equal(result$comparison_method, "keys")
  expect_equal(result$rows_matched, 3L)
  expect_equal(result$numeric_summary$column, "value")
})

test_that("compare_tables computes numeric discrepancies by row index", {
  x <- data.frame(value = c(10.0, 20.0, 30.0))
  y <- data.frame(value = c(10.5, 20.0, 30.5))
  result <- compare_tables(x, y)

  expect_equal(result$comparison_method, "row_index")
  expect_false(is.null(result$numeric_summary))
})

test_that("compare_tables errors with no common columns", {
  x <- data.frame(a = 1:3)
  y <- data.frame(b = 1:3)
  expect_error(compare_tables(x, y), "No matching column names")
})

test_that("compare_tables errors with invalid key_cols", {
  x <- data.frame(id = 1:3)
  y <- data.frame(id = 1:3)
  expect_error(compare_tables(x, y, key_cols = "nonexistent"), "not present")
})

test_that("compare_tables key overlap summary is correct", {
  x <- data.frame(id = c(1L, 2L, 3L), value = c(10, 20, 30))
  y <- data.frame(id = c(2L, 3L, 4L), value = c(40, 50, 60))
  result <- compare_tables(x, y, key_cols = "id")

  expect_equal(result$key_summary$matches, 2L)
  expect_equal(result$key_summary$only_x, 1L)
  expect_equal(result$key_summary$only_y, 1L)
})

test_that("print.compare_tbl produces output", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.1, 20.0, 30.5))
  result <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Table Comparison", combined))
})

test_that("compare_tables works with tibbles", {
  skip_if_not_installed("dplyr")
  x <- dplyr::tibble(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- dplyr::tibble(id = 1:3, value = c(10.1, 20.0, 30.5))
  result <- compare_tables(x, y, key_cols = "id")

  expect_s3_class(result, "compare_tbl")
  expect_false(is.null(result$numeric_summary))
})

test_that("compare_tables handles no numeric columns", {
  x <- data.frame(id = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = c("a", "b", "d"), stringsAsFactors = FALSE)
  result <- compare_tables(x, y)

  expect_true(is.na(result$comparison_method))
  expect_null(result$numeric_summary)
})

test_that("compare_tables warns when key columns are non-unique", {
  x <- data.frame(id = c(1L, 1L, 2L), value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = c(1L, 2L), value = c(10.1, 30.5))

  expect_warning(
    compare_tables(x, y, key_cols = "id"),
    "not unique"
  )
})

# --- tol, top_n, discrepancies, match_summary tests ---

test_that("tol = 0 is backward compatible", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.5, 20.0, 30.5))
  result <- compare_tables(x, y, key_cols = "id")

  expect_equal(result$tol, .Machine$double.eps)
  expect_equal(result$top_n, Inf)
  expect_false(is.null(result$match_summary))
  expect_false(is.null(result$discrepancies))
  expect_equal(result$numeric_summary$n_over_tol, 2L)
})

test_that("tol filters discrepancies correctly", {
  x <- data.frame(id = 1:4, value = c(10.0, 20.0, 30.0, 40.0))
  y <- data.frame(id = 1:4, value = c(10.1, 20.0, 30.5, 40.2))
  result <- compare_tables(x, y, key_cols = "id", tol = 0.15)

  # Only id=3 (diff=0.5) and id=4 (diff=0.2) exceed tol=0.15
  expect_equal(nrow(result$discrepancies), 2L)
  expect_true(all(result$discrepancies$abs_diff > 0.15))
  expect_equal(result$numeric_summary$n_over_tol, 2L)
})

test_that("tol classifies matched rows correctly", {
  x <- data.frame(id = 1:4, value = c(10.0, 20.0, 30.0, 40.0))
  y <- data.frame(id = 1:4, value = c(10.1, 20.0, 30.5, 40.0))
  result <- compare_tables(x, y, key_cols = "id", tol = 0.15)

  ms <- result$match_summary
  # id=1 (diff=0.1 <= 0.15): no disc; id=2,4 (diff=0): no disc; id=3 (diff=0.5): disc
  expect_equal(ms$matched_no_disc, 3L)
  expect_equal(ms$matched_with_disc, 1L)
  expect_equal(ms$pct_no_disc, 0.75)
  expect_equal(ms$pct_with_disc, 0.25)
})

test_that("top_n truncates discrepancies per column", {
  x <- data.frame(id = 1:5, value = c(10, 20, 30, 40, 50))
  y <- data.frame(id = 1:5, value = c(11, 22, 33, 44, 55))
  result <- compare_tables(x, y, key_cols = "id", top_n = 2)

  # All 5 rows differ, but only top 2 per column stored
  expect_equal(nrow(result$discrepancies), 2L)
  # Should be the largest diffs (id=5: diff=5, id=4: diff=4)
  expect_equal(result$discrepancies$abs_diff[1], 5)
  expect_equal(result$discrepancies$abs_diff[2], 4)
})

test_that("top_n = Inf returns all discrepancies", {
  x <- data.frame(id = 1:5, value = c(10, 20, 30, 40, 50))
  y <- data.frame(id = 1:5, value = c(11, 22, 33, 44, 55))
  result <- compare_tables(x, y, key_cols = "id", top_n = Inf)

  expect_equal(nrow(result$discrepancies), 5L)
})

test_that("discrepancies data.frame has correct structure with keys", {
  x <- data.frame(id = c("a", "b", "c"), date = c("2024-01-01", "2024-01-02", "2024-01-03"),
                   value = c(10.0, 20.0, 30.0), stringsAsFactors = FALSE)
  y <- data.frame(id = c("a", "b", "c"), date = c("2024-01-01", "2024-01-02", "2024-01-03"),
                   value = c(10.5, 20.0, 30.5), stringsAsFactors = FALSE)
  result <- compare_tables(x, y, key_cols = c("id", "date"))

  disc <- result$discrepancies
  expect_true("id" %in% names(disc))
  expect_true("date" %in% names(disc))
  expect_true("column" %in% names(disc))
  expect_true("value_x" %in% names(disc))
  expect_true("value_y" %in% names(disc))
  expect_true("abs_diff" %in% names(disc))
})

test_that("discrepancies use row_index in row-index mode", {
  x <- data.frame(value = c(10.0, 20.0, 30.0))
  y <- data.frame(value = c(10.5, 20.0, 30.5))
  result <- compare_tables(x, y)

  disc <- result$discrepancies
  expect_true("row_index" %in% names(disc))
  expect_false("id" %in% names(disc))
})

test_that("only_x_keys and only_y_keys data.frames are stored", {
  x <- data.frame(id = c(1L, 2L, 3L), value = c(10, 20, 30))
  y <- data.frame(id = c(2L, 3L, 4L), value = c(40, 50, 60))
  result <- compare_tables(x, y, key_cols = "id")

  expect_false(is.null(result$only_x_keys))
  expect_equal(nrow(result$only_x_keys), 1L)
  expect_equal(result$only_x_keys$id, 1L)

  expect_false(is.null(result$only_y_keys))
  expect_equal(nrow(result$only_y_keys), 1L)
  expect_equal(result$only_y_keys$id, 4L)
})

test_that("top_n truncates unmatched key data.frames", {
  x <- data.frame(id = 1:10, value = 1:10)
  y <- data.frame(id = 6:15, value = 6:15)
  result <- compare_tables(x, y, key_cols = "id", top_n = 3)

  # x has ids 1-5 unmatched, but top_n = 3
  expect_equal(nrow(result$only_x_keys), 3L)
  # y has ids 11-15 unmatched, but top_n = 3
  expect_equal(nrow(result$only_y_keys), 3L)
  # key_summary still has full counts
  expect_equal(result$key_summary$only_x, 5L)
  expect_equal(result$key_summary$only_y, 5L)
})

test_that("match_summary counts are correct with key overlap", {
  x <- data.frame(id = c(1L, 2L, 3L, 4L), value = c(10, 20, 30, 40))
  y <- data.frame(id = c(2L, 3L, 5L), value = c(25, 30, 50))
  result <- compare_tables(x, y, key_cols = "id")

  ms <- result$match_summary
  expect_equal(ms$only_x, 2L)    # ids 1, 4

  expect_equal(ms$only_y, 1L)    # id 5
  # ids 2,3 match; id=2 differs (20 vs 25), id=3 same (30 vs 30)
  expect_equal(ms$matched_no_disc, 1L)
  expect_equal(ms$matched_with_disc, 1L)
  expect_equal(ms$pct_no_disc, 0.5)
  expect_equal(ms$pct_with_disc, 0.5)
})

test_that("match_summary in row-index mode", {
  x <- data.frame(value = c(10.0, 20.0, 30.0, 40.0, 50.0))
  y <- data.frame(value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y)

  ms <- result$match_summary
  expect_equal(ms$only_x, 2L)
  expect_equal(ms$only_y, 0L)
  expect_equal(ms$matched_no_disc, 3L)
  expect_equal(ms$matched_with_disc, 0L)
})

test_that("no discrepancies returns NULL discrepancies", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y, key_cols = "id")

  expect_null(result$discrepancies)
  expect_equal(result$match_summary$matched_no_disc, 3L)
  expect_equal(result$match_summary$matched_with_disc, 0L)
})

test_that("match_summary works with no numeric columns", {
  x <- data.frame(id = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = c("a", "b", "d"), stringsAsFactors = FALSE)
  result <- compare_tables(x, y)

  ms <- result$match_summary
  expect_equal(ms$only_x, 1L)
  expect_equal(ms$only_y, 1L)
  expect_equal(ms$matched_no_disc, 2L)
  expect_equal(ms$matched_with_disc, 0L)
})

test_that("print.compare_tbl shows new sections", {
  x <- data.frame(id = c(1L, 2L, 3L, 4L), value = c(10.0, 20.0, 30.0, 40.0))
  y <- data.frame(id = c(2L, 3L, 5L), value = c(25.0, 30.0, 50.0))
  result <- compare_tables(x, y, key_cols = "id", tol = 0.5)

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Row matching", combined))
  expect_true(grepl("tol = 0.5", combined))
  expect_true(grepl("Unmatched keys", combined))
  expect_true(grepl("Top discrepancies", combined))
  expect_true(grepl(">tol", combined))
})

test_that("print.compare_tbl respects show_n", {
  x <- data.frame(id = 1:10, value = as.numeric(1:10))
  y <- data.frame(id = 1:10, value = as.numeric(1:10) + 1)
  result <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(result, show_n = 2), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("and 8 more", combined))
})

test_that("tol validation rejects invalid values", {
  x <- data.frame(id = 1:3, value = 1:3)
  y <- data.frame(id = 1:3, value = 1:3)
  expect_error(compare_tables(x, y, tol = -1), "non-negative")
  expect_error(compare_tables(x, y, tol = "abc"), "non-negative")
  expect_error(compare_tables(x, y, tol = c(0, 1)), "non-negative")
})

test_that("top_n validation rejects invalid values", {
  x <- data.frame(id = 1:3, value = 1:3)
  y <- data.frame(id = 1:3, value = 1:3)
  expect_error(compare_tables(x, y, top_n = 0), "positive")
  expect_error(compare_tables(x, y, top_n = -1), "positive")
  expect_error(compare_tables(x, y, top_n = "abc"), "positive")
})

test_that("NA-vs-value is treated as discrepancy with abs_diff = NA (keys mode)", {
  x <- data.frame(id = 1:3, value = c(10.0, NA, 30.0))
  y <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y, key_cols = "id")

  # id=2: NA vs 20.0 should be a discrepancy

  expect_false(is.null(result$discrepancies))
  expect_equal(nrow(result$discrepancies), 1L)
  expect_equal(result$discrepancies$id, 2L)
  expect_true(is.na(result$discrepancies$abs_diff))
  expect_true(is.na(result$discrepancies$value_x))
  expect_equal(result$discrepancies$value_y, 20.0)

  # match_summary: 2 no-disc (id=1,3), 1 with-disc (id=2)
  expect_equal(result$match_summary$matched_no_disc, 2L)
  expect_equal(result$match_summary$matched_with_disc, 1L)
})

test_that("NA-vs-value is treated as discrepancy with abs_diff = NA (row-index mode)", {
  x <- data.frame(value = c(10.0, NA, 30.0))
  y <- data.frame(value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y)

  expect_false(is.null(result$discrepancies))
  expect_equal(nrow(result$discrepancies), 1L)
  expect_equal(result$discrepancies$row_index, 2L)
  expect_true(is.na(result$discrepancies$abs_diff))
  expect_equal(result$match_summary$matched_with_disc, 1L)
})

test_that("both-NA is NOT treated as discrepancy", {
  x <- data.frame(id = 1:3, value = c(10.0, NA, 30.0))
  y <- data.frame(id = 1:3, value = c(10.0, NA, 30.0))
  result <- compare_tables(x, y, key_cols = "id")

  expect_null(result$discrepancies)
  expect_equal(result$match_summary$matched_no_disc, 3L)
  expect_equal(result$match_summary$matched_with_disc, 0L)
})

test_that("n_over_tol includes NA-vs-value pairs", {
  x <- data.frame(id = 1:4, value = c(10.0, NA, 30.0, 40.0))
  y <- data.frame(id = 1:4, value = c(10.5, 20.0, 30.0, 40.0))
  result <- compare_tables(x, y, key_cols = "id")

  # id=1 (diff=0.5 > eps): over tol; id=2 (NA vs 20): NA mismatch
  # n_over_tol should count both
  expect_equal(result$numeric_summary$n_over_tol, 2L)
})

test_that("NA discrepancies are sorted to the end", {
  x <- data.frame(id = 1:4, value = c(10.0, NA, 30.0, 40.0))
  y <- data.frame(id = 1:4, value = c(15.0, 20.0, 30.0, 42.0))
  result <- compare_tables(x, y, key_cols = "id")

  disc <- result$discrepancies
  # id=1 (diff=5), id=4 (diff=2) should come before id=2 (NA), id=3 matches
  expect_equal(nrow(disc), 3L)
  expect_false(is.na(disc$abs_diff[1]))  # largest finite diff first
  expect_false(is.na(disc$abs_diff[2]))  # second finite diff
  expect_true(is.na(disc$abs_diff[3]))   # NA at end
})

test_that("multiple numeric columns produce correct discrepancies", {
  x <- data.frame(id = 1:3, val_a = c(10.0, 20.0, 30.0), val_b = c(1.0, 2.0, 3.0))
  y <- data.frame(id = 1:3, val_a = c(10.5, 20.0, 30.5), val_b = c(1.0, 2.5, 3.0))
  result <- compare_tables(x, y, key_cols = "id")

  disc <- result$discrepancies
  # val_a: id=1 (0.5), id=3 (0.5); val_b: id=2 (0.5)
  expect_equal(nrow(disc), 3L)
  expect_true("val_a" %in% disc$column)
  expect_true("val_b" %in% disc$column)
  expect_equal(sum(disc$column == "val_a"), 2L)
  expect_equal(sum(disc$column == "val_b"), 1L)
})

test_that("top_n applies per-column with multiple numeric columns", {
  x <- data.frame(id = 1:5, val_a = as.numeric(1:5), val_b = as.numeric(1:5))
  y <- data.frame(id = 1:5, val_a = as.numeric(1:5) + 1, val_b = as.numeric(1:5) + 1)
  result <- compare_tables(x, y, key_cols = "id", top_n = 2)

  disc <- result$discrepancies
  # top_n = 2 per column, 2 columns => up to 4 rows
  expect_equal(sum(disc$column == "val_a"), 2L)
  expect_equal(sum(disc$column == "val_b"), 2L)
})

test_that("tol works correctly in row-index mode", {
  x <- data.frame(value = c(10.0, 20.0, 30.0, 40.0))
  y <- data.frame(value = c(10.1, 20.0, 30.5, 40.2))
  result <- compare_tables(x, y, tol = 0.15)

  # Only row 3 (diff=0.5) and row 4 (diff=0.2) exceed tol=0.15
  expect_equal(nrow(result$discrepancies), 2L)
  expect_true(all(result$discrepancies$abs_diff > 0.15))
  expect_equal(result$match_summary$matched_no_disc, 2L)
  expect_equal(result$match_summary$matched_with_disc, 2L)
  expect_equal(result$numeric_summary$n_over_tol, 2L)
})

test_that("only_x_keys and only_y_keys in row-index mode", {
  x <- data.frame(value = c(10.0, 20.0, 30.0, 40.0, 50.0))
  y <- data.frame(value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y)

  expect_false(is.null(result$only_x_keys))
  expect_equal(nrow(result$only_x_keys), 2L)
  expect_equal(result$only_x_keys$row_index, c(4L, 5L))
  expect_null(result$only_y_keys)
})

test_that("print with no discrepancies omits Top discrepancies section", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_false(grepl("Top discrepancies", combined))
  expect_true(grepl("Row matching", combined))
})

test_that("print section numbering is sequential even without match_summary", {
  x <- data.frame(value = c(10.0, 20.0, 30.0))
  y <- data.frame(value = c(10.0, 20.0, 30.0))
  result <- compare_tables(x, y)

  # No match_summary (all identical, row-index mode has match_summary)
  # but let's test a case with no match_summary: identical tables with no keys
  # Actually row-index mode now always has match_summary when num_cols > 0.
  # Test the no-numeric-no-key case:
  x2 <- data.frame(a = c(1L, 2L, 3L))
  y2 <- data.frame(a = c(1L, 2L, 3L))
  result2 <- compare_tables(x2, y2)

  output <- capture.output(print(result2), type = "message")
  combined <- paste(output, collapse = "\n")
  # Should have sections 1, 2, 3, 4 (row matching), 5 (numeric) — or skip 4 if no match_summary
  # With keys auto-detected (integer col "a"), match_summary exists
  expect_true(grepl("1\\. Row counts", combined))
  expect_true(grepl("2\\. Column names", combined))
  expect_true(grepl("3\\. Key columns", combined))
})

test_that("print.compare_tbl suppresses tol label at default eps", {
  x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
  y <- data.frame(id = 1:3, value = c(10.5, 20.0, 30.5))
  result <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  # Default eps should NOT show tol label
  expect_false(grepl("tol =", combined))
})

test_that("print.compare_tbl shows top_n cap in remaining message", {
  x <- data.frame(id = 1:10, value = 1:10)
  y <- data.frame(id = 6:15, value = 6:15)
  result <- compare_tables(x, y, key_cols = "id", top_n = 3)

  output <- capture.output(print(result, show_n = 2), type = "message")
  combined <- paste(output, collapse = "\n")
  # top_n = 3 < total unmatched (5), so should mention "stored"
  expect_true(grepl("3 stored", combined))
})

test_that("print.compare_tbl validates show_n", {
  x <- data.frame(id = 1:3, value = 1:3)
  y <- data.frame(id = 1:3, value = 1:3)
  result <- compare_tables(x, y, key_cols = "id")
  expect_error(print(result, show_n = -1), "non-negative")
  expect_error(print(result, show_n = "abc"), "non-negative")
})

test_that("zero-row tables are handled gracefully", {
  x <- data.frame(id = integer(0), value = numeric(0))
  y <- data.frame(id = integer(0), value = numeric(0))
  result <- compare_tables(x, y, key_cols = "id")

  expect_equal(result$nrow_x, 0L)
  expect_equal(result$nrow_y, 0L)
  expect_equal(result$key_summary$matches, 0L)
  expect_null(result$discrepancies)
  expect_null(result$numeric_summary)
})

test_that("print handles NA values in discrepancies display", {
  x <- data.frame(id = 1:3, value = c(10.0, NA, 30.0))
  y <- data.frame(id = 1:3, value = c(15.0, 20.0, 30.0))
  result <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(result), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("NA", combined))
  expect_true(grepl("Top discrepancies", combined))
})

# -- compare_cols / exclude_cols ----------------------------------------------

test_that("compare_cols selects only specified columns", {
  x <- data.frame(id = 1:3, a = c(1, 2, 3), b = c(4, 5, 6))
  y <- data.frame(id = 1:3, a = c(1, 2, 4), b = c(4, 5, 7))
  r <- compare_tables(x, y, key_cols = "id", compare_cols = "a")
  expect_equal(r$discrepancies$column, "a")
})

test_that("exclude_cols removes specified columns from comparison", {
  x <- data.frame(id = 1:3, a = c(1, 2, 3), b = c(4, 5, 6))
  y <- data.frame(id = 1:3, a = c(1, 2, 4), b = c(4, 5, 7))
  r <- compare_tables(x, y, key_cols = "id", exclude_cols = "a")
  expect_equal(r$discrepancies$column, "b")
})

test_that("compare_cols and exclude_cols are mutually exclusive", {
  x <- data.frame(id = 1:3, a = 1:3)
  y <- data.frame(id = 1:3, a = 1:3)
  expect_error(
    compare_tables(x, y, compare_cols = "a", exclude_cols = "a"),
    "cannot both be specified"
  )
})

test_that("compare_cols errors on unknown column names", {
  x <- data.frame(id = 1:3, a = 1:3)
  y <- data.frame(id = 1:3, a = 1:3)
  expect_error(
    compare_tables(x, y, key_cols = "id", compare_cols = "nonexistent"),
    "Not found"
  )
})

test_that("exclude_cols warns on nonexistent columns", {
  x <- data.frame(id = 1:3, a = 1:3)
  y <- data.frame(id = 1:3, a = 1:3)
  expect_warning(
    compare_tables(x, y, key_cols = "id", exclude_cols = "nonexistent"),
    "not found among common non-key columns"
  )
})

test_that("compare_cols silently excludes key columns", {
  x <- data.frame(id = 1:3, a = c(1, 2, 3))
  y <- data.frame(id = 1:3, a = c(1, 2, 4))
  # "id" is a key so asking for it in compare_cols is a non-key column error
  expect_error(
    compare_tables(x, y, key_cols = "id", compare_cols = c("id", "a")),
    "Not found"
  )
})

# -- Categorical discrepancy detection ---------------------------------------

test_that("character column discrepancies detected in keys mode", {
  x <- data.frame(id = 1:4, cat = c("a", "b", "c", "d"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:4, cat = c("a", "x", "c", "z"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  expect_false(is.null(r$categorical_summary))
  expect_equal(r$categorical_summary$column, "cat")
  expect_equal(r$categorical_summary$n_compared, 4L)
  expect_equal(r$categorical_summary$n_mismatched, 2L)
  expect_equal(r$categorical_summary$pct_mismatched, 0.5)
  expect_equal(r$categorical_summary$n_na_mismatch, 0L)

  expect_false(is.null(r$categorical_discrepancies))
  expect_equal(nrow(r$categorical_discrepancies), 2L)
  expect_true(all(c("id", "column", "value_x", "value_y") %in% names(r$categorical_discrepancies)))
})

test_that("character column discrepancies detected in row-index mode", {
  x <- data.frame(val = c(1, 2), cat = c("a", "b"), stringsAsFactors = FALSE)
  y <- data.frame(val = c(1, 3), cat = c("a", "x"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = character(0))

  expect_false(is.null(r$categorical_discrepancies))
  expect_equal(nrow(r$categorical_discrepancies), 1L)
  expect_equal(r$categorical_discrepancies$row_index, 2L)
  expect_equal(r$categorical_discrepancies$value_x, "b")
  expect_equal(r$categorical_discrepancies$value_y, "x")
})

test_that("factor columns compared as character values", {
  x <- data.frame(id = 1:3, cat = factor(c("a", "b", "c")))
  y <- data.frame(id = 1:3, cat = factor(c("a", "x", "c")))
  r <- compare_tables(x, y, key_cols = "id")

  expect_false(is.null(r$categorical_discrepancies))
  expect_equal(nrow(r$categorical_discrepancies), 1L)
  expect_equal(r$categorical_discrepancies$value_x, "b")
  expect_equal(r$categorical_discrepancies$value_y, "x")
})

test_that("logical column discrepancies detected", {
  x <- data.frame(id = 1:3, flag = c(TRUE, FALSE, TRUE))
  y <- data.frame(id = 1:3, flag = c(TRUE, TRUE, TRUE))
  r <- compare_tables(x, y, key_cols = "id")

  expect_equal(r$categorical_summary$n_mismatched, 1L)
  expect_equal(nrow(r$categorical_discrepancies), 1L)
})

test_that("Date column discrepancies detected", {
  x <- data.frame(id = 1:3, dt = as.Date(c("2024-01-01", "2024-02-01", "2024-03-01")))
  y <- data.frame(id = 1:3, dt = as.Date(c("2024-01-01", "2024-02-15", "2024-03-01")))
  r <- compare_tables(x, y, key_cols = "id")

  expect_equal(r$categorical_summary$n_mismatched, 1L)
  expect_equal(nrow(r$categorical_discrepancies), 1L)
})

test_that("NA-vs-value is categorical discrepancy; both-NA is not", {
  x <- data.frame(id = 1:4, cat = c("a", NA, "c", NA), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:4, cat = c("a", "b", NA, NA), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  expect_equal(r$categorical_summary$n_mismatched, 2L)
  expect_equal(r$categorical_summary$n_na_mismatch, 2L)
  expect_equal(nrow(r$categorical_discrepancies), 2L)
})

test_that("top_n applies per-column for categorical discrepancies", {
  x <- data.frame(id = 1:5,
                   a = c("a", "b", "c", "d", "e"),
                   b = c("x", "y", "z", "w", "v"),
                   stringsAsFactors = FALSE)
  y <- data.frame(id = 1:5,
                   a = c("A", "B", "C", "D", "E"),
                   b = c("X", "Y", "Z", "W", "V"),
                   stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id", top_n = 2)

  # 2 per column = 4 total stored
  expect_equal(nrow(r$categorical_discrepancies), 4L)
  expect_equal(sum(r$categorical_discrepancies$column == "a"), 2L)
  expect_equal(sum(r$categorical_discrepancies$column == "b"), 2L)
})

test_that("match_summary counts rows with either numeric or categorical discrepancies", {
  x <- data.frame(id = 1:4, val = c(1, 2, 3, 4),
                   cat = c("a", "b", "c", "d"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:4, val = c(1, 2.5, 3, 4),
                   cat = c("a", "b", "x", "d"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  # Row 2: numeric disc only, Row 3: categorical disc only
  expect_equal(r$match_summary$matched_with_disc, 2L)
  expect_equal(r$match_summary$matched_no_disc, 2L)
})

test_that("type-mismatched columns excluded from categorical comparison", {
  x <- data.frame(id = 1:3, cat = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3, cat = c(1L, 2L, 3L))
  r <- compare_tables(x, y, key_cols = "id")

  expect_null(r$categorical_summary)
  expect_null(r$categorical_discrepancies)
  expect_false(is.null(r$type_mismatches))
})

# -- pct_diff -----------------------------------------------------------------

test_that("pct_diff column present in numeric discrepancies", {
  x <- data.frame(id = 1:3, value = c(10, 20, 30))
  y <- data.frame(id = 1:3, value = c(10, 25, 30))
  r <- compare_tables(x, y, key_cols = "id")

  expect_true("pct_diff" %in% names(r$discrepancies))
  # 5/25 = 0.2
  expect_equal(r$discrepancies$pct_diff, 0.2)
})

test_that("pct_diff handles both-zero case", {
  x <- data.frame(id = 1:2, value = c(0, 10))
  y <- data.frame(id = 1:2, value = c(0, 15))
  r <- compare_tables(x, y, key_cols = "id")

  # Only row 2 has a discrepancy (both-zero diff = 0, not > tol)
  expect_equal(nrow(r$discrepancies), 1L)
  # 5/15 = 0.333...
  expect_equal(r$discrepancies$pct_diff, 5 / 15, tolerance = 1e-10)
})

test_that("pct_diff handles one-zero case", {
  x <- data.frame(id = 1L, value = 0)
  y <- data.frame(id = 1L, value = 5)
  r <- compare_tables(x, y, key_cols = "id")

  # 5/max(0,5) = 1.0
  expect_equal(r$discrepancies$pct_diff, 1.0)
})

test_that("pct_diff is NA when value is NA", {
  x <- data.frame(id = 1L, value = NA_real_)
  y <- data.frame(id = 1L, value = 5)
  r <- compare_tables(x, y, key_cols = "id")

  expect_true(is.na(r$discrepancies$pct_diff))
})

# -- total_discrepancies ------------------------------------------------------

test_that("total_discrepancies counts numeric discrepancies only", {
  x <- data.frame(id = 1:3, value = c(10, 20, 30))
  y <- data.frame(id = 1:3, value = c(10, 25, 35))
  r <- compare_tables(x, y, key_cols = "id")

  expect_equal(r$total_discrepancies, 2L)
})

test_that("total_discrepancies counts both numeric and categorical", {
  x <- data.frame(id = 1:3, val = c(1, 2, 3),
                   cat = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3, val = c(1, 2.5, 3),
                   cat = c("a", "x", "c"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  # 1 numeric + 1 categorical = 2

  expect_equal(r$total_discrepancies, 2L)
})

test_that("total_discrepancies is not limited by top_n", {
  x <- data.frame(id = 1:5, val = c(1, 2, 3, 4, 5))
  y <- data.frame(id = 1:5, val = c(2, 3, 4, 5, 6))
  r <- compare_tables(x, y, key_cols = "id", top_n = 2)

  # All 5 rows differ but only 2 stored
  expect_equal(nrow(r$discrepancies), 2L)
  expect_equal(r$total_discrepancies, 5L)
})

# -- as.data.frame.compare_tbl -----------------------------------------------

test_that("as.data.frame returns correct columns for mixed discrepancies", {
  x <- data.frame(id = 1:3, val = c(1, 2, 3),
                   cat = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3, val = c(1, 2.5, 3),
                   cat = c("a", "x", "c"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")
  df <- as.data.frame(r)

  expect_s3_class(df, "data.frame")
  expect_true(all(c("id", "column", "type", "value_x", "value_y",
                     "abs_diff", "pct_diff") %in% names(df)))
  expect_equal(nrow(df), 2L)
  expect_equal(sum(df$type == "numeric"), 1L)
  expect_equal(sum(df$type == "categorical"), 1L)
})

test_that("as.data.frame works with numeric-only discrepancies", {
  x <- data.frame(id = 1:3, val = c(1, 2, 3))
  y <- data.frame(id = 1:3, val = c(1, 2.5, 3))
  r <- compare_tables(x, y, key_cols = "id")
  df <- as.data.frame(r)

  expect_equal(nrow(df), 1L)
  expect_equal(df$type, "numeric")
})

test_that("as.data.frame works with categorical-only discrepancies", {
  x <- data.frame(id = 1:3, cat = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3, cat = c("a", "x", "c"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")
  df <- as.data.frame(r)

  expect_equal(nrow(df), 1L)
  expect_equal(df$type, "categorical")
  expect_true(is.na(df$abs_diff))
  expect_true(is.na(df$pct_diff))
})

test_that("as.data.frame returns 0-row data.frame with no discrepancies", {
  x <- data.frame(id = 1:3, val = c(1, 2, 3))
  y <- data.frame(id = 1:3, val = c(1, 2, 3))
  r <- compare_tables(x, y, key_cols = "id")
  df <- as.data.frame(r)

  expect_equal(nrow(df), 0L)
  expect_true("column" %in% names(df))
  expect_true("type" %in% names(df))
})

test_that("as.data.frame converts numeric values to character", {
  x <- data.frame(id = 1:2, val = c(1, 2))
  y <- data.frame(id = 1:2, val = c(1, 3))
  r <- compare_tables(x, y, key_cols = "id")
  df <- as.data.frame(r)

  expect_type(df$value_x, "character")
  expect_type(df$value_y, "character")
})

# -- Print method updates ----------------------------------------------------

test_that("print shows categorical discrepancies section", {
  x <- data.frame(id = 1:3, cat = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3, cat = c("a", "x", "c"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(r), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Categorical discrepancies", combined))
  expect_true(grepl("Mismatched", combined))
})

test_that("print shows pct_diff column in numeric discrepancies", {
  x <- data.frame(id = 1:3, value = c(10, 20, 30))
  y <- data.frame(id = 1:3, value = c(10, 25, 30))
  r <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(r), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("pct_diff", combined))
})

test_that("print shows total cell discrepancies line", {
  x <- data.frame(id = 1:3, val = c(1, 2, 3),
                   cat = c("a", "b", "c"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:3, val = c(1, 2.5, 3),
                   cat = c("a", "x", "c"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(r), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("Total cell discrepancies", combined))
  expect_true(grepl("numeric", combined))
  expect_true(grepl("categorical", combined))
})

test_that("print handles show_n for categorical discrepancies", {
  x <- data.frame(id = 1:5, cat = c("a", "b", "c", "d", "e"), stringsAsFactors = FALSE)
  y <- data.frame(id = 1:5, cat = c("A", "B", "C", "D", "E"), stringsAsFactors = FALSE)
  r <- compare_tables(x, y, key_cols = "id")

  output <- capture.output(print(r, show_n = 2L), type = "message")
  combined <- paste(output, collapse = "\n")
  expect_true(grepl("and 3 more", combined))
})

---
title: "Diagnostic Functions Guide"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Diagnostic Functions Guide}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message=FALSE}
library(tidyaudit)
library(dplyr)
```

Before you build a pipeline, you need to understand your data. Before you trust
a join, you need to verify the keys. Before you filter, you want to know what
you'll lose. tidyaudit's diagnostic functions help with the detective work that
comes before, during, and after every analysis.

These functions are designed for **interactive use** — the questions you ask in
the console while exploring a dataset or debugging a transformation. They
complement the [audit trail system](tidyaudit.html), which instruments
production pipelines.

## Before you join

You're about to left-join orders to customers. Will every order match? Will the
join create duplicates? How many rows will end up with NAs?

`validate_join()` analyzes a potential join **without performing it**, reporting
match rates, relationship type, duplicate keys, and unmatched rows:

```{r validate-join}
orders <- data.frame(
  id     = c(1L, 2L, 3L, 3L, 4L, 5L),
  amount = c(100, 200, 150, 175, 300, 50)
)
customers <- data.frame(
  id   = c(2L, 3L, 6L),
  name = c("Alice", "Bob", "Carol")
)

validate_join(orders, customers, by = "id")
```

### Different key names

When the key columns have different names, use a named vector:

```{r validate-join-named}
products <- data.frame(prod_id = 1:3, price = c(10, 20, 30))
sales    <- data.frame(item_id = c(1L, 1L, 2L), qty = c(5, 3, 7))

validate_join(products, sales, by = c("prod_id" = "item_id"))
```

### Tracking the impact on a metric

Want to know the revenue at risk from unmatched keys? Use `stat` (same column
name in both tables) or `stat_x`/`stat_y` (different column names):

```{r validate-join-stat}
x <- data.frame(id = 1:4, revenue = c(100, 200, 300, 400))
y <- data.frame(id = c(2L, 3L, 5L), cost = c(10, 20, 30))

validate_join(x, y, by = "id", stat_x = "revenue", stat_y = "cost")
```

## Are your keys unique?

You assume `id` uniquely identifies every row. Does it?

`validate_primary_keys()` tests whether a set of columns forms a valid primary
key:

```{r validate-pk}
df <- data.frame(
  id    = c(1L, 2L, 3L, 3L, 4L),
  group = c("A", "A", "B", "C", "A"),
  value = c(10, 20, 30, 40, 50)
)

# Single column -- not unique
validate_primary_keys(df, "id")

# Composite key -- unique
validate_primary_keys(df, c("id", "group"))
```

### Beyond uniqueness: what's the relationship?

`validate_var_relationship()` determines the functional relationship between two
columns — one-to-one, one-to-many, many-to-one, or many-to-many:

```{r validate-var-rel}
df2 <- data.frame(
  dept    = c("Sales", "Sales", "Engineering", "Engineering"),
  manager = c("Ann", "Ann", "Bob", "Bob")
)
validate_var_relationship(df2, "dept", "manager")
```

## What changed between versions?

You have yesterday's extract and today's extract. Or the data before your
transformation and after. What's different?

`compare_tables()` compares two data frames by examining columns, row counts,
key overlap, and numeric discrepancies:

```{r compare-tables}
before <- data.frame(id = 1:5, value = c(10.0, 20.0, 30.0, 40.0, 50.0))
after  <- data.frame(id = 1:5, value = c(10.0, 22.5, 30.0, 40.0, 55.0))

compare_tables(before, after)
```

## Filter with visibility

`dplyr::filter()` silently removes rows. You apply the condition, the rows
vanish, and you move on. `filter_keep()` and `filter_drop()` do the same
filtering but tell you exactly what was removed — and optionally warn you when
too much is gone.

### filter_keep

Keeps rows where the condition is `TRUE` (same as `dplyr::filter()`), with
diagnostic output:

```{r filter-keep}
sales <- data.frame(
  id     = 1:10,
  amount = c(500, 25, 1200, 80, 3000, 15, 750, 40, 2000, 60),
  status = rep(c("valid", "suspect"), 5)
)

result <- filter_keep(sales, amount > 100, .stat = amount)
```

### filter_drop

Drops rows where the condition is `TRUE` (the inverse):

```{r filter-drop}
result2 <- filter_drop(sales, status == "suspect", .stat = amount)
```

### Warning thresholds

Set `.warn_threshold` to get a warning when too many rows are dropped — a
safety net for production pipelines:

```{r filter-warn, warning=TRUE}
filter_keep(sales, amount > 1000, .stat = amount, .warn_threshold = 0.5)
```

## Diagnose data quality

You just received a dataset. Where are the gaps?

### Missing values

`diagnose_nas()` reports NA counts and percentages for every column:

```{r diagnose-nas}
messy <- data.frame(
  id    = 1:6,
  name  = c("A", NA, "C", "D", NA, "F"),
  score = c(10, 20, NA, NA, 50, NA),
  grade = c("A", "B", "C", NA, "A", "B")
)

diagnose_nas(messy)
```

### Column summaries

`summarize_column()` gives type-appropriate statistics for a single vector —
quantiles for numeric, value counts for character, balance for logical:

```{r summarize-column}
summarize_column(c(1, 2, 3, NA, 5, 10, 100))
summarize_column(c("apple", "banana", "apple", "cherry", NA))
```

`get_summary_table()` applies this across all columns (or a selection):

```{r get-summary-table}
get_summary_table(messy)
```

## Frequency tables

Base R's `table()` gives you raw counts. `tab()` gives you percentages,
cumulative totals, sorting, cutoffs, weighting, and two-way crosstabulations.

### One-way tables

```{r tab-oneway}
tab(mtcars, cyl)
```

### Sorting and cutoffs

```{r tab-sort}
# Sort by frequency
tab(mtcars, carb, .sort = "freq_desc")

# Keep only top-2 values, collapse rest into (Other)
tab(mtcars, carb, .cutoff = 2)
```

### Two-way crosstabulations

```{r tab-twoway}
tab(mtcars, cyl, gear)

# Show row percentages instead of counts
tab(mtcars, cyl, gear, .display = "row_pct")
```

### Weighted tabulation

```{r tab-weighted}
tab(mtcars, cyl, .wt = mpg)
```

## String quality

Duplicates hiding behind case differences, leading spaces, and encoding issues
are among the most common causes of failed joins and inflated group counts.

`diagnose_strings()` audits a character vector for these problems:

```{r diagnose-strings}
firms <- c("Apple", "APPLE", "apple", "  Microsoft ", "Google", NA, "")
diagnose_strings(firms)
```

## Audit your transformations

You're about to apply `trimws()` to a column, or `round()` to a price vector.
How many values will actually change? What will they look like afterward?

`audit_transform()` shows exactly what a transformation does. It automatically
detects the vector type and computes type-appropriate diagnostics:

- **Numeric**: mean/min/max shift, proportion changed
- **Date/POSIXct**: range change, proportion shifted
- **Factor**: level remapping (gained/lost levels)
- **Logical**: TRUE/FALSE/NA balance shift
- **Character**: string-level before/after comparison

### Character vectors

```{r audit-transform}
audit_transform(firms, trimws)
audit_transform(firms, tolower)
```

### Numeric vectors

```{r audit-transform-numeric}
prices <- c(10.456, 20.789, 30.123, NA, 50.999)
audit_transform(prices, round)
```

### Date vectors

```{r audit-transform-date}
dates <- as.Date(c("2024-01-15", "2024-06-30", "2024-12-01", NA))
audit_transform(dates, function(d) d + 30)
```

### Factor vectors

```{r audit-transform-factor}
sizes <- factor(c("S", "M", "L", "XL", "XXL", "S", "M"))
audit_transform(sizes, function(f) {
  levels(f)[levels(f) %in% c("XL", "XXL")] <- "XL+"
  f
})
```

### Logical vectors

```{r audit-transform-logical}
flags <- c(TRUE, FALSE, TRUE, NA, FALSE)
audit_transform(flags, function(x) !x)
```

---
title: "Audit Trail Walkthrough"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Audit Trail Walkthrough}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message=FALSE}
library(tidyaudit)
library(dplyr)
```

## Tracking what happens in your pipeline

Every filter, join, and transformation in a data pipeline encodes an
assumption about the data. How many rows did that filter actually drop? Did the
join introduce NAs or duplicates? Which records did not match?
How much revenue was lost at each step?

These questions matter. Even when these answers are clear in the moment, they become hard to recall
during code review, debugging, or when someone inherits your pipeline months
later. An audit trail keeps that record for you.

tidyaudit captures **metadata-only snapshots** at each step of a pipe — row
counts, column counts, NA totals, numeric summaries — without storing the data
itself. The trail is a lightweight, structured record of your pipeline's
behavior that you can print, diff, export, and share.

## Your first trail

Create a trail object and insert `audit_tap()` calls into your pipeline. Each
tap records a snapshot and passes the data through unchanged — three taps,
three snapshots, one timeline.

```{r basic-trail}
# Sample data
orders <- data.frame(
  id       = 1:20,
  customer = rep(c("Alice", "Bob", "Carol", "Dan", "Eve"), 4),
  amount   = c(150, 200, 50, 300, 75, 120, 400, 90, 250, 60,
               180, 210, 45, 320, 85, 130, 380, 95, 270, 55),
  status   = rep(c("complete", "pending", "complete", "cancelled", "complete"), 4)
)

trail <- audit_trail("order_pipeline")

result <- orders |>
  audit_tap(trail, "raw") |>
  filter(status == "complete") |>
  audit_tap(trail, "complete_only") |>
  mutate(tax = amount * 0.1) |>
  audit_tap(trail, "with_tax")
```

Print the trail to see the full timeline:

```{r print-trail}
print(trail)
```

The timeline shows row counts, column counts, NA totals, and change summaries
between consecutive steps. Notice how the filter reduced 20 rows to 12, and the
mutate added a column — all captured automatically.

## Operation-aware taps

Plain `audit_tap()` records what the data looks like at a given point, but it
can't tell you *why* it changed. Operation-aware taps solve this — they
perform the dplyr operation AND record enriched diagnostics in a single step.

### Join taps

Joins are where data quality problems hide. A left join can silently introduce
NAs, inflate row counts, or produce unexpected many-to-many relationships.
Replace `dplyr::left_join()` with `left_join_tap()` to capture match rates,
relationship type, and duplicate key information automatically:

```{r join-tap}
customers <- data.frame(
  customer = c("Alice", "Bob", "Carol", "Dan"),
  region   = c("East", "West", "East", "North")
)

trail2 <- audit_trail("join_pipeline")

result2 <- orders |>
  audit_tap(trail2, "raw") |>
  left_join_tap(customers, by = "customer",
                .trail = trail2, .label = "with_region")

print(trail2)
```

The `Type` column now shows the join type, relationship, and match rate — all
without leaving the pipe. All six dplyr join types are supported:
`left_join_tap()`, `right_join_tap()`, `inner_join_tap()`, `full_join_tap()`,
`anti_join_tap()`, `semi_join_tap()`.

### Filter taps

Filters are invisible by default — you remove rows and never know what you
lost. `filter_tap()` keeps matching rows (like `dplyr::filter()`) while
recording exactly how many rows were dropped:

```{r filter-tap}
trail3 <- audit_trail("filter_pipeline")

result3 <- orders |>
  audit_tap(trail3, "raw") |>
  filter_tap(status == "complete",
             .trail = trail3, .label = "complete_only") |>
  filter_tap(amount > 100,
             .trail = trail3, .label = "high_value",
             .stat = amount)

print(trail3)
```

The `.stat` argument is the feature that makes filter taps indispensable for
financial and business pipelines: it tracks how much of a numeric column was
removed at each step. You can see not just that you dropped 4 rows, but that
those rows represented a specific dollar amount.

`filter_out_tap()` works the same way but drops matching rows (the inverse).

## Comparing snapshots

`audit_diff()` gives you a detailed before/after comparison between any two
snapshots in the trail — not just adjacent ones:

```{r audit-diff}
audit_diff(trail3, "raw", "high_value")
```

This shows row/column/NA deltas, columns added or removed, and numeric
distribution shifts across the common columns.

## Full audit report

`audit_report()` prints the complete trail summary plus all consecutive diffs
in one call — the full story of what your pipeline did:

```{r audit-report}
audit_report(trail3)
```

## Domain-specific diagnostics

The built-in metrics cover structure and shape (rows, columns, NAs, numeric
summaries). But your domain has its own questions: how many valid records
remain? What's the total revenue at this stage? Is a business rule still
satisfied?

Pass a named list of functions via `.fns` to compute anything domain-specific
at any tap. Each function receives the data and its return value is stored in
the snapshot:

```{r custom-fns}
trail4 <- audit_trail("custom_example")

result4 <- orders |>
  audit_tap(trail4, "raw", .fns = list(
    n_complete   = ~sum(.$status == "complete"),          # scalar
    amount_stats = ~c(mean = mean(.$amount),              # named vector
                      max  = max(.$amount))
  )) |>
  filter(status == "complete") |>
  audit_tap(trail4, "complete_only", .fns = list(
    n_complete   = ~sum(.$status == "complete"),
    amount_stats = ~c(mean = mean(.$amount),
                      max  = max(.$amount))
  ))
```

Custom results appear as inline annotations directly below each snapshot row:

```{r print-custom}
print(trail4)
```

The rendering rules are:

- **Scalar** return value: `fn_name: value`
- **Named vector or named list of scalars**: `fn_name: key=val, key=val`
  (truncated at 60 characters)
- **Complex object** (data frame, unnamed vector, nested list):
  `fn_name: [complex -- see audit_report()]`

To suppress annotations and display only the main table:

```{r show-custom-false}
print(trail4, show_custom = FALSE)
```

## Snapshot controls

On wide datasets, computing numeric summaries for every column is unnecessary
and slows the pipeline down. Three parameters on all tap functions let you
control what gets captured:

- `.numeric_summary = FALSE` — skip quantile computation entirely
- `.cols_include` — character vector of column names to include in the
  snapshot's schema (mutually exclusive with `.cols_exclude`)
- `.cols_exclude` — character vector of column names to exclude

Core invariants — `nrow`, `ncol`, and `total_nas` — are always recorded
regardless of these settings.

```{r snapshot-controls}
wide_data <- cbind(orders, matrix(rnorm(20 * 50), nrow = 20))

trail_ctrl <- audit_trail("snapshot_controls")

wide_data |>
  audit_tap(trail_ctrl, "full_snapshot") |>
  audit_tap(trail_ctrl, "minimal",
            .numeric_summary = FALSE,
            .cols_include = c("id", "amount", "status"))

print(trail_ctrl)
```

The "minimal" snapshot still knows the data has `r ncol(wide_data)` columns and
20 rows, but its schema only describes the three columns you asked for and
contains no numeric summaries.

## Tabulation in pipelines

`tab()` produces one-way frequency tables or two-way crosstabulations:

```{r tab-standalone}
tab(orders, status)
tab(orders, status, customer)
```

`tab_tap()` embeds a tabulation inside a pipeline as a custom diagnostic
annotation. It runs `tab()` on the data, stores the result in the snapshot, and
returns the data unchanged — useful for tracking how categorical distributions
shift across pipeline steps:

```{r tab-tap}
trail_tab <- audit_trail("tab_pipeline")

result_tab <- orders |>
  tab_tap(status, .trail = trail_tab, .label = "status_dist") |>
  filter(status == "complete") |>
  tab_tap(customer, .trail = trail_tab, .label = "customer_dist",
          .sort = "freq_desc")

print(trail_tab)
```

## Standalone mode

All tap functions work without a trail. When `.trail = NULL` (the default):

- **No diagnostic args**: behaves like the plain dplyr function
- **With `.stat` or `.warn_threshold`**: runs diagnostics and prints results
  without recording to a trail

This makes it easy to add quick diagnostics to any pipeline without setting up a
full trail:

```{r null-trail}
# Plain filter -- no diagnostics
orders |> filter_tap(amount > 100) |> nrow()

# Diagnostics without a trail
orders |> filter_tap(amount > 100, .stat = amount) |> invisible()
```

## Exporting and sharing trails

Trails live in-memory as environment-based S3 objects. To share them with
colleagues, CI systems, or documentation, convert them to portable formats.

### Converting to R objects

```{r trail-to-objects}
# As a plain R list (suitable for jsonlite::toJSON())
trail_list <- trail_to_list(trail3)
str(trail_list, max.level = 2)

# As a data.frame (one row per snapshot)
trail_df <- trail_to_df(trail3)
print(trail_df)
```

### Saving and loading trails

RDS format preserves all R types and round-trips perfectly:

```{r trail-rds}
tmp_rds <- tempfile(fileext = ".rds")
write_trail(trail3, tmp_rds)
restored <- read_trail(tmp_rds)
print(restored)
```

JSON format is available for interoperability with other tools (requires
jsonlite):

```{r trail-json}
tmp_json <- tempfile(fileext = ".json")
write_trail(trail3, tmp_json, format = "json")
```

### HTML visualization

`audit_export()` produces a self-contained HTML file — one file you can email,
embed in documentation, or drop into a compliance folder:

```{r trail-html}
audit_export(trail3, tempfile(fileext = ".html"))
```

The HTML page renders an interactive pipeline flow diagram with light/dark theme
toggle. Nodes are clickable and expand to show column schema and diagnostics.
Edges show the full diff between adjacent snapshots. No server or internet
connection required.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate-join.R
\name{validate_join}
\alias{validate_join}
\alias{print.validate_join}
\alias{summary.validate_join}
\title{Validate Join Operations Between Two Tables}
\usage{
validate_join(x, y, by = NULL, stat = NULL, stat_x = NULL, stat_y = NULL)

\method{print}{validate_join}(x, ...)

\method{summary}{validate_join}(object, ...)
}
\arguments{
\item{x}{A data.frame or tibble (left table).}

\item{y}{A data.frame or tibble (right table).}

\item{by}{A character vector of column names to join on. Use a named vector
\code{c("key_x" = "key_y")} when column names differ between tables. Unnamed
elements are used for both tables.}

\item{stat}{Optional single column name (string) to track in both tables when
the column name is the same. Ignored if \code{stat_x} or \code{stat_y} is provided.}

\item{stat_x}{Optional column name (string) for a numeric statistic in \code{x}.}

\item{stat_y}{Optional column name (string) for a numeric statistic in \code{y}.}

\item{...}{Additional arguments (currently unused).}

\item{object}{A \code{validate_join} object to summarize.}
}
\value{
An S3 object of class \code{validate_join} containing:
\describe{
\item{x_name, y_name}{Names of the input tables from the original call}
\item{by_x, by_y}{Key columns used for the join}
\item{counts}{List with row counts, match rates, and overlap statistics}
\item{stat}{When \code{stat}, \code{stat_x}, or \code{stat_y} is provided, a list with
stat diagnostics per table. \code{NULL} when no stat is provided.}
\item{duplicates}{List with duplicate key information for each table}
\item{summary_table}{A data.frame summarizing the join diagnostics}
\item{relation}{Character string describing the relationship}
\item{keys_only_in_x}{Unmatched keys from x}
\item{keys_only_in_y}{Unmatched keys from y}
}
}
\description{
Analyzes a potential join between two data.frames or tibbles without
performing the full join. Reports relationship type (one-to-one,
one-to-many, etc.), match rates, duplicate keys, and unmatched rows.
Optionally tracks a numeric statistic column through the join to quantify
impact.
}
\examples{
x <- data.frame(id = c(1L, 2L, 3L, 3L), value = c("a", "b", "c", "d"))
y <- data.frame(id = c(2L, 3L, 4L), score = c(10, 20, 30))
result <- validate_join(x, y, by = "id")
print(result)

# Track a stat column with different names in each table
x2 <- data.frame(id = 1:3, sales = c(100, 200, 300))
y2 <- data.frame(id = 2:4, cost = c(10, 20, 30))
validate_join(x2, y2, by = "id", stat_x = "sales", stat_y = "cost")

}
\seealso{
Other join validation:
\code{\link[=compare_tables]{compare_tables()}},
\code{\link[=validate_primary_keys]{validate_primary_keys()}},
\code{\link[=validate_var_relationship]{validate_var_relationship()}}
}
\concept{join validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-snap.R, R/audit-trail.R
\name{print.audit_snap}
\alias{print.audit_snap}
\alias{audit_trail}
\alias{print.audit_trail}
\title{Create an Audit Trail}
\usage{
\method{print}{audit_snap}(x, ...)

audit_trail(name = NULL)

\method{print}{audit_trail}(x, show_custom = TRUE, ...)
}
\arguments{
\item{x}{An object to print.}

\item{...}{Additional arguments (currently unused).}

\item{name}{Optional name for the trail. If \code{NULL}, a timestamped name is
generated automatically.}

\item{show_custom}{Logical. If \code{TRUE} (default), inline annotations (one
indented line per custom function) are printed below each snapshot that has
custom diagnostics. Set to \code{FALSE} to suppress them and display only the
main timeline table.}
}
\value{
An \code{audit_trail} object (S3 class wrapping an environment).
}
\description{
Creates an audit trail object that captures metadata snapshots at each step
of a data pipeline. The trail uses environment-based reference semantics so
it can be modified in place inside pipes via \code{\link[=audit_tap]{audit_tap()}}.
}
\examples{
trail <- audit_trail("my_analysis")
print(trail)

}
\seealso{
Other audit trail:
\code{\link[=audit_diff]{audit_diff()}},
\code{\link[=audit_report]{audit_report()}},
\code{\link[=audit_tap]{audit_tap()}},
\code{\link[=tab_tap]{tab_tap()}}
}
\concept{audit trail}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabulation.R
\name{tab}
\alias{tab}
\alias{print.tidyaudit_tab}
\alias{as.data.frame.tidyaudit_tab}
\title{Tabulate Variables}
\usage{
tab(
  .data,
  ...,
  .wt = NULL,
  .sort = c("value_asc", "value_desc", "freq_desc", "freq_asc"),
  .cutoff = NULL,
  .na = c("include", "exclude", "only"),
  .display = c("count", "row_pct", "col_pct", "total_pct")
)

\method{print}{tidyaudit_tab}(x, ...)

\method{as.data.frame}{tidyaudit_tab}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{.data}{A data.frame or tibble.}

\item{...}{Additional arguments (currently unused).}

\item{.wt}{Optional unquoted column to use as frequency weights.
When supplied, frequencies are weighted sums instead of row counts.}

\item{.sort}{How to order the rows (and columns in two-way tables).
\code{"value_asc"} (default) sorts alphabetically (or by factor levels),
\code{"value_desc"} sorts in reverse, \code{"freq_desc"} sorts by frequency
descending, \code{"freq_asc"} sorts by frequency ascending.}

\item{.cutoff}{Controls how many values to display.
An integer >= 1 keeps the top-N values by frequency.
A number in (0, 1) keeps values that cumulatively account for that
proportion of the total. Remaining values are grouped under \code{"(Other)"}.
For two-way tables, the cutoff applies to the row variable only.}

\item{.na}{How to handle \code{NA} values. \code{"include"} (default) treats \code{NA} as a
category, \code{"exclude"} drops \code{NA} rows before tabulation, \code{"only"} shows
only \code{NA} rows.}

\item{.display}{Cell contents for two-way crosstabulations. One of
\code{"count"} (default), \code{"row_pct"}, \code{"col_pct"}, or \code{"total_pct"}.
Ignored for one-way tables.}

\item{x}{A \code{tidyaudit_tab} object.}

\item{row.names}{Passed to \code{\link[=as.data.frame]{as.data.frame()}}. Default \code{NULL}.}

\item{optional}{Passed to \code{\link[=as.data.frame]{as.data.frame()}}. Default \code{FALSE}.}
}
\value{
An S3 object of class \code{tidyaudit_tab}. Use \code{\link[=as.data.frame]{as.data.frame()}} to
extract the underlying table.
}
\description{
Produces one-way frequency tables or two-way crosstabulations. One variable
gives counts, percentages, and cumulative percentages; two variables give a
crosstabulation matrix with row/column totals.
}
\examples{
tab(mtcars, cyl)
tab(mtcars, cyl, .sort = "freq_desc")
tab(mtcars, cyl, gear)
tab(mtcars, cyl, gear, .display = "row_pct")
tab(mtcars, cyl, .wt = mpg)
tab(mtcars, cyl, .cutoff = 2)

}
\seealso{
Other data quality:
\code{\link[=audit_transform]{audit_transform()}},
\code{\link[=diagnose_nas]{diagnose_nas()}},
\code{\link[=diagnose_strings]{diagnose_strings()}},
\code{\link[=get_summary_table]{get_summary_table()}},
\code{\link[=summarize_column]{summarize_column()}}
}
\concept{data quality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-export.R
\name{write_trail}
\alias{write_trail}
\title{Write an Audit Trail to a File}
\usage{
write_trail(.trail, file, format = c("rds", "json"))
}
\arguments{
\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}

\item{file}{Path to the output file. A \code{.rds} extension is conventional
for \code{format = "rds"}; \code{.json} for \code{format = "json"}.}

\item{format}{One of \code{"rds"} (default) or \code{"json"}. The JSON format
requires the \pkg{jsonlite} package to be installed.}
}
\value{
\code{.trail}, invisibly.
}
\description{
Saves an \code{\link[=audit_trail]{audit_trail()}} to disk as either an RDS file (default) or a
JSON file. The RDS format preserves all R types and can be restored
perfectly with \code{\link[=read_trail]{read_trail()}}. The JSON format produces a human-readable
representation suitable for archiving or interoperability with other tools.
}
\note{
Custom diagnostic results (the \code{custom} field, populated via \code{.fns}
in \code{\link[=audit_tap]{audit_tap()}}) are serialised on a best-effort basis for JSON output.
Complex R objects such as environments or functions cannot be represented
in JSON and will cause an error.
}
\examples{
trail <- audit_trail("example")
mtcars |> audit_tap(trail, "raw")
tmp <- tempfile(fileext = ".rds")
write_trail(trail, tmp)
restored <- read_trail(tmp)

}
\seealso{
\code{\link[=read_trail]{read_trail()}}, \code{\link[=trail_to_list]{trail_to_list()}}

Other audit export:
\code{\link[=audit_export]{audit_export()}},
\code{\link[=read_trail]{read_trail()}},
\code{\link[=trail_to_df]{trail_to_df()}},
\code{\link[=trail_to_list]{trail_to_list()}}
}
\concept{audit export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-report.R
\name{audit_report}
\alias{audit_report}
\title{Generate an Audit Report}
\usage{
audit_report(.trail, format = "console")
}
\arguments{
\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}

\item{format}{Report format. Currently only \code{"console"} is supported.}
}
\value{
\code{.trail}, invisibly.
}
\description{
Prints a full audit report for a trail, including the trail summary, all
diffs between consecutive snapshots, custom diagnostic results, and a final
data profile.
}
\examples{
trail <- audit_trail("example")
mtcars |>
  audit_tap(trail, "raw") |>
  dplyr::filter(mpg > 20) |>
  audit_tap(trail, "filtered")
audit_report(trail)

}
\seealso{
Other audit trail:
\code{\link[=audit_diff]{audit_diff()}},
\code{\link[=audit_tap]{audit_tap()}},
\code{\link[=print.audit_snap]{print.audit_snap()}},
\code{\link[=tab_tap]{tab_tap()}}
}
\concept{audit trail}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-tap.R
\name{audit_tap}
\alias{audit_tap}
\title{Record a Pipeline Snapshot}
\usage{
audit_tap(
  .data,
  .trail,
  .label = NULL,
  .fns = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)
}
\arguments{
\item{.data}{A data.frame or tibble flowing through the pipe.}

\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}

\item{.label}{Optional character label for this snapshot. If \code{NULL}, an
auto-generated label like \code{"step_1"} is used.}

\item{.fns}{Optional named list of diagnostic functions (or formula lambdas)
to run on \code{.data}. Results are stored in the snapshot.}

\item{.numeric_summary}{Logical. If \code{FALSE}, skip numeric summary
computation in the snapshot (default \code{TRUE}).}

\item{.cols_include}{Character vector of column names to include in the
snapshot schema, or \code{NULL} (the default) to include all columns. Mutually
exclusive with \code{.cols_exclude}.}

\item{.cols_exclude}{Character vector of column names to exclude from the
snapshot schema, or \code{NULL} (the default). Mutually exclusive with
\code{.cols_include}.}
}
\value{
\code{.data}, unchanged, returned invisibly. The function is a
transparent pass-through; its only effect is the side effect on \code{.trail}.
}
\description{
Transparent pipe pass-through that captures a metadata snapshot and appends
it to an audit trail. Returns \code{.data} unchanged — the function's only purpose
is its side effect on \code{.trail}.
}
\examples{
trail <- audit_trail("example")
result <- mtcars |>
  audit_tap(trail, "raw") |>
  dplyr::filter(mpg > 20) |>
  audit_tap(trail, "filtered")
print(trail)

}
\seealso{
Other audit trail:
\code{\link[=audit_diff]{audit_diff()}},
\code{\link[=audit_report]{audit_report()}},
\code{\link[=print.audit_snap]{print.audit_snap()}},
\code{\link[=tab_tap]{tab_tap()}}
}
\concept{audit trail}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare-tables.R
\name{compare_tables}
\alias{compare_tables}
\alias{print.compare_tbl}
\alias{as.data.frame.compare_tbl}
\title{Compare Two Tables}
\usage{
compare_tables(
  x,
  y,
  key_cols = NULL,
  tol = .Machine$double.eps,
  top_n = Inf,
  compare_cols = NULL,
  exclude_cols = NULL
)

\method{print}{compare_tbl}(x, show_n = 5L, ...)

\method{as.data.frame}{compare_tbl}(x, row.names = NULL, optional = FALSE, ...)
}
\arguments{
\item{x}{First data.frame or tibble to compare.}

\item{y}{Second data.frame or tibble to compare.}

\item{key_cols}{Character vector of column names to use as keys for matching
rows. If \code{NULL} (default), automatically detects character, factor, and
integer columns as keys.}

\item{tol}{Numeric tolerance for comparing numeric columns. Differences
less than or equal to \code{tol} are considered equal. Defaults to
\code{\link[=.Machine]{.Machine$double.eps}} (machine double-precision).}

\item{top_n}{Maximum number of row-level discrepancies to store \strong{per
column} (numeric and categorical), and maximum unmatched keys to store.
Defaults to \code{Inf} (all). Unmatched keys are stored in arbitrary order.}

\item{compare_cols}{Character vector of column names to compare. If \code{NULL}
(default), all common non-key columns are compared. Mutually exclusive
with \code{exclude_cols}.}

\item{exclude_cols}{Character vector of column names to exclude from
comparison. If \code{NULL} (default), no columns are excluded. Mutually
exclusive with \code{compare_cols}.}

\item{show_n}{Maximum number of rows to display for discrepancies and
unmatched keys in the printed output. Defaults to \code{5L}.}

\item{...}{Additional arguments (currently unused).}

\item{row.names}{Passed to \code{\link[=as.data.frame]{as.data.frame()}}. Default \code{NULL}.}

\item{optional}{Passed to \code{\link[=as.data.frame]{as.data.frame()}}. Default \code{FALSE}.}
}
\value{
An S3 object of class \code{compare_tbl} containing:
\describe{
\item{name_x, name_y}{Names of the compared objects}
\item{common_columns}{Column names present in both tables}
\item{only_x}{Column names only in x}
\item{only_y}{Column names only in y}
\item{type_mismatches}{Data.frame of columns with different types, or NULL}
\item{nrow_x}{Number of rows in x}
\item{nrow_y}{Number of rows in y}
\item{key_summary}{Summary of key overlap, or NULL}
\item{numeric_summary}{Data.frame of numeric discrepancy quantiles (with
\code{n_over_tol} count), or NULL}
\item{comparison_method}{How columns were compared (\code{"keys"},
\code{"row_index"}, or \code{NA})}
\item{rows_matched}{Number of rows matched on keys}
\item{tol}{The tolerance used}
\item{top_n}{The top_n used}
\item{discrepancies}{Data.frame of row-level numeric discrepancies
exceeding \code{tol} (or where one side is \code{NA}), with key columns (or
\code{row_index}), \code{column}, \code{value_x}, \code{value_y}, \code{abs_diff}, and
\code{pct_diff} (relative difference as a proportion). NULL if none.}
\item{categorical_summary}{Data.frame with \code{column}, \code{n_compared},
\code{n_mismatched}, \code{pct_mismatched} (proportion, 0--1), \code{n_na_mismatch},
or NULL}
\item{categorical_discrepancies}{Data.frame of row-level categorical
discrepancies with key columns (or \code{row_index}), \code{column}, \code{value_x},
\code{value_y}. NULL if none.}
\item{total_discrepancies}{Total number of cell-level discrepancies
across all column types (not limited by \code{top_n})}
\item{only_x_keys}{Data.frame of key combinations only in x (up to
\code{top_n} rows), or NULL}
\item{only_y_keys}{Data.frame of key combinations only in y (up to
\code{top_n} rows), or NULL}
\item{match_summary}{List with \code{only_x}, \code{only_y}, \code{matched_no_disc},
\code{matched_with_disc}, \code{pct_no_disc} (proportion, 0--1),
\code{pct_with_disc} (proportion, 0--1)}
}

Use \code{\link[=as.data.frame]{as.data.frame()}} to extract all discrepancies (numeric and categorical)
as a single tidy data.frame.
}
\description{
Compares two data.frames or tibbles by examining column names, row counts,
key overlap, numeric discrepancies, and categorical discrepancies. Useful for
validating data processing pipelines.
}
\examples{
x <- data.frame(id = 1:3, value = c(10.0, 20.0, 30.0))
y <- data.frame(id = 1:3, value = c(10.1, 20.0, 30.5))
compare_tables(x, y)

# With tolerance — differences <= 0.15 are considered equal
compare_tables(x, y, tol = 0.15)

# Categorical columns are also compared
a <- data.frame(id = 1:3, status = c("ok", "warn", "fail"),
                 stringsAsFactors = FALSE)
b <- data.frame(id = 1:3, status = c("ok", "warn", "error"),
                 stringsAsFactors = FALSE)
compare_tables(a, b)

}
\seealso{
Other join validation:
\code{\link[=validate_join]{validate_join()}},
\code{\link[=validate_primary_keys]{validate_primary_keys()}},
\code{\link[=validate_var_relationship]{validate_var_relationship()}}
}
\concept{join validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/join-tap.R
\name{join_tap}
\alias{join_tap}
\alias{left_join_tap}
\alias{right_join_tap}
\alias{inner_join_tap}
\alias{full_join_tap}
\alias{anti_join_tap}
\alias{semi_join_tap}
\title{Operation-Aware Join Taps}
\usage{
left_join_tap(
  .data,
  y,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)

right_join_tap(
  .data,
  y,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)

inner_join_tap(
  .data,
  y,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)

full_join_tap(
  .data,
  y,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)

anti_join_tap(
  .data,
  y,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)

semi_join_tap(
  .data,
  y,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)
}
\arguments{
\item{.data}{A data.frame or tibble (left table in the join).}

\item{y}{A data.frame or tibble (right table in the join).}

\item{...}{Arguments passed to the corresponding \verb{dplyr::*_join()} function,
including \code{by}, \code{suffix}, \code{keep}, \code{multiple}, \code{unmatched}, etc. The \code{by}
argument should be passed by name for enriched diagnostics.}

\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object, or \code{NULL} (the default). When
\code{NULL}, behavior depends on \code{.stat}: if \code{.stat} is also \code{NULL}, a plain
dplyr join is performed; if \code{.stat} is provided, \code{\link[=validate_join]{validate_join()}}
diagnostics are printed before the join.}

\item{.label}{Optional character label for this snapshot. If \code{NULL},
auto-generated as \code{"left_join_1"} etc.}

\item{.stat}{An unquoted column name for stat tracking, e.g., \code{amount}.
Passed to \code{\link[=validate_join]{validate_join()}}.}

\item{.numeric_summary}{Logical. If \code{FALSE}, skip numeric summary
computation in the snapshot (default \code{TRUE}).}

\item{.cols_include}{Character vector of column names to include in the
snapshot schema, or \code{NULL} (the default) to include all columns. Mutually
exclusive with \code{.cols_exclude}.}

\item{.cols_exclude}{Character vector of column names to exclude from the
snapshot schema, or \code{NULL} (the default). Mutually exclusive with
\code{.cols_include}.}
}
\value{
The joined data.frame or tibble (same as the corresponding
\verb{dplyr::*_join()}).
}
\description{
Performs a dplyr join AND records enriched diagnostics in an audit trail.
These functions replace the pattern of wrapping a join with two
\code{\link[=audit_tap]{audit_tap()}} calls, capturing information that plain taps cannot:
match rates, relationship type, duplicate keys, and unmatched row counts.
}
\details{
Enriched diagnostics (match rates, relationship type, duplicate keys) require
equality joins — \code{by} as a character vector, named character vector, or
simple equality \code{join_by()} expression (e.g., \code{join_by(id)},
\code{join_by(a == b)}). For non-equi \code{join_by()} expressions, the tap records
a basic snapshot without match-rate diagnostics.

All dplyr join features (\code{join_by}, \code{multiple}, \code{unmatched}, \code{suffix}, etc.)
work unchanged via \code{...}.

When \code{.trail} is \code{NULL}:
\itemize{
\item \code{.stat} also \code{NULL}: plain dplyr join
\item \code{.stat} provided: prints \code{\link[=validate_join]{validate_join()}} diagnostics, then joins
\item \code{.label} provided: warns that label is ignored
}
}
\examples{
orders <- data.frame(id = 1:4, amount = c(100, 200, 300, 400))
customers <- data.frame(id = c(2, 3, 5), name = c("A", "B", "C"))

# With trail
trail <- audit_trail("join_example")
result <- orders |>
  audit_tap(trail, "raw") |>
  left_join_tap(customers, by = "id", .trail = trail, .label = "joined")
print(trail)

# Without trail (plain join)
result2 <- left_join_tap(orders, customers, by = "id")

}
\seealso{
Other operation taps:
\code{\link[=filter_tap]{filter_tap()}}
}
\concept{operation taps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-diff.R
\name{audit_diff}
\alias{audit_diff}
\alias{print.audit_diff}
\title{Compare Two Audit Trail Snapshots}
\usage{
audit_diff(.trail, from, to)

\method{print}{audit_diff}(x, ...)
}
\arguments{
\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}

\item{from}{Label (character) or index (integer) of the first snapshot.}

\item{to}{Label (character) or index (integer) of the second snapshot.}

\item{x}{An \code{audit_diff} object to print.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An \code{audit_diff} object (S3 list).
}
\description{
Computes detailed differences between any two snapshots in an audit trail,
including row/column/NA deltas, columns added/removed, type changes,
per-column NA changes, and numeric distribution shifts.
}
\examples{
trail <- audit_trail("example")
mtcars |>
  audit_tap(trail, "raw") |>
  dplyr::filter(mpg > 20) |>
  audit_tap(trail, "filtered")
audit_diff(trail, "raw", "filtered")

}
\seealso{
Other audit trail:
\code{\link[=audit_report]{audit_report()}},
\code{\link[=audit_tap]{audit_tap()}},
\code{\link[=print.audit_snap]{print.audit_snap()}},
\code{\link[=tab_tap]{tab_tap()}}
}
\concept{audit trail}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-tap.R
\name{filter_tap}
\alias{filter_tap}
\alias{filter_out_tap}
\title{Operation-Aware Filter Taps}
\usage{
filter_tap(
  .data,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .quiet = FALSE,
  .warn_threshold = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)

filter_out_tap(
  .data,
  ...,
  .trail = NULL,
  .label = NULL,
  .stat = NULL,
  .quiet = FALSE,
  .warn_threshold = NULL,
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)
}
\arguments{
\item{.data}{A data.frame or tibble.}

\item{...}{Filter conditions, evaluated in the context of \code{.data} using
tidy evaluation (same as \code{\link[dplyr:filter]{dplyr::filter()}}).}

\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object, or \code{NULL} (the default). When
\code{NULL}, behavior depends on diagnostic arguments: if none are provided,
a plain \code{dplyr::filter()} is performed; if \code{.stat}, \code{.warn_threshold}, or
\code{.quiet = TRUE} is provided, delegates to \code{\link[=filter_keep]{filter_keep()}} or
\code{\link[=filter_drop]{filter_drop()}}.}

\item{.label}{Optional character label for this snapshot. If \code{NULL},
auto-generated as \code{"filter_1"} etc.}

\item{.stat}{An unquoted column or expression to total, e.g., \code{amount},
\code{price * qty}. Reports the stat amount dropped and its share of the total.}

\item{.quiet}{Logical. If \code{TRUE}, suppress printing diagnostics (default
\code{FALSE}).}

\item{.warn_threshold}{Numeric between 0 and 1. If set and the proportion of
dropped rows exceeds this threshold, a warning is issued.}

\item{.numeric_summary}{Logical. If \code{FALSE}, skip numeric summary
computation in the snapshot (default \code{TRUE}).}

\item{.cols_include}{Character vector of column names to include in the
snapshot schema, or \code{NULL} (the default) to include all columns. Mutually
exclusive with \code{.cols_exclude}.}

\item{.cols_exclude}{Character vector of column names to exclude from the
snapshot schema, or \code{NULL} (the default). Mutually exclusive with
\code{.cols_include}.}
}
\value{
The filtered data.frame or tibble.
}
\description{
Performs a diagnostic filter AND records filter diagnostics in an audit trail.
\code{filter_tap()} keeps matching rows (like \code{\link[dplyr:filter]{dplyr::filter()}}),
\code{filter_out_tap()} drops matching rows (the inverse).
}
\details{
When \code{.trail} is \code{NULL}:
\itemize{
\item No diagnostic args: plain \code{dplyr::filter()} / \code{dplyr::filter_out()}
\item Diagnostic args provided: delegates to \code{\link[=filter_keep]{filter_keep()}} /
\code{\link[=filter_drop]{filter_drop()}} (prints diagnostics but no trail recording)
\item \code{.label} provided: warns that label is ignored
}
}
\examples{
df <- data.frame(id = 1:10, amount = 1:10 * 100, flag = rep(c(TRUE, FALSE), 5))

# With trail
trail <- audit_trail("filter_example")
result <- df |>
  audit_tap(trail, "raw") |>
  filter_tap(amount > 300, .trail = trail, .label = "big_only")
print(trail)

# Inverse: drop matching rows
trail2 <- audit_trail("filter_out_example")
result2 <- df |>
  audit_tap(trail2, "raw") |>
  filter_out_tap(flag == FALSE, .trail = trail2, .label = "flagged_only")
print(trail2)

# Without trail (plain filter)
result3 <- filter_tap(df, amount > 300)

}
\seealso{
Other operation taps:
\code{\link{join_tap}}
}
\concept{operation taps}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate-keys.R
\name{validate_var_relationship}
\alias{validate_var_relationship}
\alias{print.validate_var_rel}
\title{Validate Variable Relationship}
\usage{
validate_var_relationship(.data, var1, var2)

\method{print}{validate_var_rel}(x, ...)
}
\arguments{
\item{.data}{A data.frame or tibble.}

\item{var1}{Character string: name of the first variable.}

\item{var2}{Character string: name of the second variable.}

\item{x}{An object to print.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An S3 object of class \code{validate_var_rel} containing:
\describe{
\item{table_name}{Name of the input table}
\item{var1, var2}{Names of the variables analyzed}
\item{relation}{Character string: "one-to-one", "one-to-many",
"many-to-one", or "many-to-many"}
\item{var1_unique}{Number of distinct values in var1}
\item{var2_unique}{Number of distinct values in var2}
\item{n_combinations}{Number of unique (var1, var2) pairs}
\item{var1_has_dups}{Does any var1 value map to multiple var2 values?}
\item{var2_has_dups}{Does any var2 value map to multiple var1 values?}
}
}
\description{
Determines the relationship between two variables in a data.frame:
one-to-one, one-to-many, many-to-one, or many-to-many.
}
\details{
Only accepts variables of type character, integer, or factor. Numeric
(double) variables are not allowed due to floating-point comparison issues.
}
\examples{
df <- data.frame(
  person_id = c(1L, 2L, 3L, 4L),
  department = c("Sales", "Sales", "Engineering", "Engineering"),
  country = c("US", "US", "US", "UK")
)
validate_var_relationship(df, "person_id", "department")

}
\seealso{
Other join validation:
\code{\link[=compare_tables]{compare_tables()}},
\code{\link[=validate_join]{validate_join()}},
\code{\link[=validate_primary_keys]{validate_primary_keys()}}
}
\concept{join validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-export.R
\name{trail_to_list}
\alias{trail_to_list}
\title{Convert an Audit Trail to a Plain List}
\usage{
trail_to_list(.trail)
}
\arguments{
\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}
}
\value{
A named list with elements \code{name}, \code{created_at} (ISO 8601 string),
\code{n_snapshots}, and \code{snapshots} (a named list keyed by snapshot label).
}
\description{
Converts an \code{\link[=audit_trail]{audit_trail()}} object to a plain R list suitable for
serialisation with \code{\link[jsonlite:toJSON]{jsonlite::toJSON()}}. All POSIXct timestamps are
converted to ISO 8601 character strings and data.frames are converted
to lists of named rows for JSON compatibility.
}
\examples{
trail <- audit_trail("example")
mtcars |> audit_tap(trail, "raw")
lst <- trail_to_list(trail)
str(lst, max.level = 2)

}
\seealso{
Other audit export:
\code{\link[=audit_export]{audit_export()}},
\code{\link[=read_trail]{read_trail()}},
\code{\link[=trail_to_df]{trail_to_df()}},
\code{\link[=write_trail]{write_trail()}}
}
\concept{audit export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-export.R
\name{read_trail}
\alias{read_trail}
\title{Read an Audit Trail from a File}
\usage{
read_trail(file, format = NULL)
}
\arguments{
\item{file}{Path to an RDS or JSON file created by \code{\link[=write_trail]{write_trail()}}.}

\item{format}{One of \code{"rds"}, \code{"json"}, or \code{NULL} (default). When \code{NULL},
the format is inferred from the file extension.}
}
\value{
A reconstructed \code{\link[=audit_trail]{audit_trail()}} object with all S3 classes
restored.
}
\description{
Restores an \code{\link[=audit_trail]{audit_trail()}} previously saved with \code{\link[=write_trail]{write_trail()}}. The
file format is detected automatically from the file extension (\code{.rds} for
RDS, \code{.json} for JSON), or can be specified explicitly via \code{format}.
}
\examples{
trail <- audit_trail("example")
mtcars |> audit_tap(trail, "raw")
tmp <- tempfile(fileext = ".rds")
write_trail(trail, tmp)
restored <- read_trail(tmp)
print(restored)

}
\seealso{
\code{\link[=write_trail]{write_trail()}}

Other audit export:
\code{\link[=audit_export]{audit_export()}},
\code{\link[=trail_to_df]{trail_to_df()}},
\code{\link[=trail_to_list]{trail_to_list()}},
\code{\link[=write_trail]{write_trail()}}
}
\concept{audit export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-summary.R
\name{summarize_column}
\alias{summarize_column}
\title{Summarize a Single Column}
\usage{
summarize_column(x)
}
\arguments{
\item{x}{A vector to summarize.}
}
\value{
A named character vector with summary statistics including:
type, unique count, missing count, missing share (proportion from 0 to 1),
most frequent value (for non-numeric), mean, sd, min, quartiles (q25, q50,
q75), max, and three example values.
}
\description{
Computes summary statistics for a vector. Handles numeric, character,
factor, logical, Date, and other types with appropriate statistics for each.
}
\examples{
summarize_column(c(1, 2, 3, NA, 5))
summarize_column(c("a", "b", "a", "c"))

}
\seealso{
Other data quality:
\code{\link[=audit_transform]{audit_transform()}},
\code{\link[=diagnose_nas]{diagnose_nas()}},
\code{\link[=diagnose_strings]{diagnose_strings()}},
\code{\link[=get_summary_table]{get_summary_table()}},
\code{\link[=tab]{tab()}}
}
\concept{data quality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tab-tap.R
\name{tab_tap}
\alias{tab_tap}
\title{Record a Tabulation Snapshot in a Pipeline}
\usage{
tab_tap(
  .data,
  ...,
  .trail,
  .label = NULL,
  .wt = NULL,
  .sort = c("value_asc", "value_desc", "freq_desc", "freq_asc"),
  .cutoff = NULL,
  .na = c("include", "exclude", "only"),
  .display = c("count", "row_pct", "col_pct", "total_pct"),
  .numeric_summary = TRUE,
  .cols_include = NULL,
  .cols_exclude = NULL
)
}
\arguments{
\item{.data}{A data.frame or tibble.}

\item{...}{Additional arguments (currently unused).}

\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}

\item{.label}{Character label for this snapshot.}

\item{.wt}{Optional unquoted column to use as frequency weights.
When supplied, frequencies are weighted sums instead of row counts.}

\item{.sort}{How to order the rows (and columns in two-way tables).
\code{"value_asc"} (default) sorts alphabetically (or by factor levels),
\code{"value_desc"} sorts in reverse, \code{"freq_desc"} sorts by frequency
descending, \code{"freq_asc"} sorts by frequency ascending.}

\item{.cutoff}{Controls how many values to display.
An integer >= 1 keeps the top-N values by frequency.
A number in (0, 1) keeps values that cumulatively account for that
proportion of the total. Remaining values are grouped under \code{"(Other)"}.
For two-way tables, the cutoff applies to the row variable only.}

\item{.na}{How to handle \code{NA} values. \code{"include"} (default) treats \code{NA} as a
category, \code{"exclude"} drops \code{NA} rows before tabulation, \code{"only"} shows
only \code{NA} rows.}

\item{.display}{Cell contents for two-way crosstabulations. One of
\code{"count"} (default), \code{"row_pct"}, \code{"col_pct"}, or \code{"total_pct"}.
Ignored for one-way tables.}

\item{.numeric_summary}{Logical. If \code{FALSE}, skip numeric summary
computation in the snapshot (default \code{TRUE}).}

\item{.cols_include}{Character vector of column names to include in the
snapshot schema, or \code{NULL} (the default) to include all columns. Mutually
exclusive with \code{.cols_exclude}.}

\item{.cols_exclude}{Character vector of column names to exclude from the
snapshot schema, or \code{NULL} (the default). Mutually exclusive with
\code{.cols_include}.}
}
\value{
\code{.data}, unchanged, returned invisibly.
}
\description{
Transparent pipe pass-through that runs \code{\link[=tab]{tab()}} on the data and stores the
result as a custom diagnostic annotation in the audit trail snapshot.
Returns \code{.data} unchanged.
}
\examples{
trail <- audit_trail("example")
result <- mtcars |>
  tab_tap(cyl, .trail = trail, .label = "by_cyl") |>
  dplyr::filter(mpg > 20) |>
  tab_tap(cyl, .trail = trail, .label = "by_cyl_filtered")
print(trail)

}
\seealso{
Other audit trail:
\code{\link[=audit_diff]{audit_diff()}},
\code{\link[=audit_report]{audit_report()}},
\code{\link[=audit_tap]{audit_tap()}},
\code{\link[=print.audit_snap]{print.audit_snap()}}
}
\concept{audit trail}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-export.R
\name{audit_export}
\alias{audit_export}
\title{Export an Audit Trail as a Self-Contained HTML File}
\usage{
audit_export(.trail, file = NULL)
}
\arguments{
\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}

\item{file}{Path to the output \code{.html} file. If \code{NULL}
(the default), writes to a temporary file and opens it in the default
browser via \code{\link[utils:browseURL]{utils::browseURL()}}.}
}
\value{
The file path (character), invisibly.
}
\description{
Produces a standalone HTML file that visualises the audit trail as an
interactive pipeline flow diagram. The file is completely self-contained
— no server, internet connection, or R installation is required to view
it. Open it in any browser.
}
\details{
The trail is serialised via \code{\link[=trail_to_list]{trail_to_list()}} and embedded as JSON inside
an HTML template with inline CSS and vanilla JavaScript. The visualisation
features:
\itemize{
\item Horizontal pipeline flow diagram with colour-coded nodes per
operation type (snapshot, join, filter).
\item Edges annotated with key deltas (match rate, drop \\%, columns
added).
\item Clickable nodes expanding to show column schema, operation
diagnostics, and custom \code{.fns} results.
\item Clickable edges showing the full diff between adjacent snapshots.
\item Light / dark theme toggle.
\item Collapsible JSON export panel.
}
}
\examples{
\donttest{
trail <- audit_trail("demo")
mtcars |> audit_tap(trail, "raw")
dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")
audit_export(trail, tempfile(fileext = ".html"))
}

}
\seealso{
\code{\link[=trail_to_list]{trail_to_list()}}, \code{\link[=write_trail]{write_trail()}}

Other audit export:
\code{\link[=read_trail]{read_trail()}},
\code{\link[=trail_to_df]{trail_to_df()}},
\code{\link[=trail_to_list]{trail_to_list()}},
\code{\link[=write_trail]{write_trail()}}
}
\concept{audit export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/validate-keys.R
\name{validate_primary_keys}
\alias{validate_primary_keys}
\alias{print.validate_pk}
\title{Validate Primary Keys}
\usage{
validate_primary_keys(.data, keys)

\method{print}{validate_pk}(x, ...)
}
\arguments{
\item{.data}{A data.frame or tibble.}

\item{keys}{Character vector of column names to test as primary keys.}

\item{x}{An object to print.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An S3 object of class \code{validate_pk} containing:
\describe{
\item{table_name}{Name of the input table from the original call}
\item{keys}{Character vector of column names tested}
\item{is_primary_key}{Logical: TRUE if keys uniquely identify all rows AND
no key column contains NA values}
\item{n_rows}{Total number of rows in the table}
\item{n_unique_keys}{Number of distinct key combinations}
\item{n_duplicate_keys}{Number of key combinations that appear more than once}
\item{duplicate_keys}{A data.frame of duplicated key values with their counts}
\item{has_numeric_keys}{Logical: TRUE if any key column is of type double}
\item{has_na_keys}{Logical: TRUE if any key column contains NA values}
\item{na_in_keys}{Named logical vector indicating which key columns contain NAs}
}
}
\description{
Tests whether a set of columns constitute primary keys of a data.frame,
i.e., whether they uniquely identify every row in the table.
}
\examples{
df <- data.frame(
  id = c(1L, 2L, 3L, 4L),
  group = c("A", "A", "B", "B"),
  value = c(10, 20, 30, 40)
)
validate_primary_keys(df, "id")
validate_primary_keys(df, "group")

}
\seealso{
Other join validation:
\code{\link[=compare_tables]{compare_tables()}},
\code{\link[=validate_join]{validate_join()}},
\code{\link[=validate_var_relationship]{validate_var_relationship()}}
}
\concept{join validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-summary.R
\name{get_summary_table}
\alias{get_summary_table}
\title{Generate Summary Table for a Data Frame}
\usage{
get_summary_table(.data, cols = NULL)
}
\arguments{
\item{.data}{A data.frame or tibble to summarize.}

\item{cols}{Optional character vector of column names to summarize. If
\code{NULL} (the default), all columns are summarized.}
}
\value{
A data.frame with one row per column containing summary statistics.
}
\description{
Creates a comprehensive summary of all columns in a data.frame, including
type, missing values, descriptive statistics, and example values.
}
\examples{
df <- data.frame(
  id = 1:100,
  value = rnorm(100),
  category = sample(letters[1:5], 100, replace = TRUE)
)
get_summary_table(df)

}
\seealso{
Other data quality:
\code{\link[=audit_transform]{audit_transform()}},
\code{\link[=diagnose_nas]{diagnose_nas()}},
\code{\link[=diagnose_strings]{diagnose_strings()}},
\code{\link[=summarize_column]{summarize_column()}},
\code{\link[=tab]{tab()}}
}
\concept{data quality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-export.R
\name{trail_to_df}
\alias{trail_to_df}
\title{Convert an Audit Trail to a Data Frame}
\usage{
trail_to_df(.trail)
}
\arguments{
\item{.trail}{An \code{\link[=audit_trail]{audit_trail()}} object.}
}
\value{
A \code{data.frame} with columns \code{index}, \code{label}, \code{type}, \code{timestamp},
\code{nrow}, \code{ncol}, \code{total_nas}, \code{all_columns}, \code{schema}, \code{numeric_summary},
\code{changes}, \code{diagnostics}, \code{custom}, \code{pipeline}, and \code{controls}. Trail
\code{name} and \code{created_at} are stored as attributes \code{"trail_name"} and
\code{"created_at"}.
}
\description{
Returns a plain \code{data.frame} with one row per snapshot. Nested fields
(\code{all_columns}, \code{schema}, \code{numeric_summary}, \code{changes}, \code{diagnostics},
\code{custom}, \code{pipeline}, \code{controls}) become list-columns. Trail metadata is
stored as attributes on the result.
}
\examples{
trail <- audit_trail("example")
mtcars |> audit_tap(trail, "raw")
dplyr::filter(mtcars, mpg > 20) |> audit_tap(trail, "filtered")
df <- trail_to_df(trail)
print(df)
attr(df, "trail_name")

}
\seealso{
Other audit export:
\code{\link[=audit_export]{audit_export()}},
\code{\link[=read_trail]{read_trail()}},
\code{\link[=trail_to_list]{trail_to_list()}},
\code{\link[=write_trail]{write_trail()}}
}
\concept{audit export}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-summary.R
\name{diagnose_nas}
\alias{diagnose_nas}
\alias{print.diagnose_na}
\title{Diagnose Missing Values}
\usage{
diagnose_nas(.data)

\method{print}{diagnose_na}(x, ...)
}
\arguments{
\item{.data}{A data.frame or tibble to diagnose.}

\item{x}{An object to print.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An S3 object of class \code{diagnose_na} containing:
\describe{
\item{table}{A data.frame with columns \code{variable}, \code{n_na}, \code{pct_na}, and
\code{n_valid}, sorted by \code{pct_na} descending.}
\item{n_cols}{Total number of columns in the input.}
\item{n_with_na}{Number of columns that have at least one NA.}
}
}
\description{
Reports NA counts and percentages for each column in a data.frame, sorted
by missing percentage in descending order.
}
\examples{
df <- data.frame(
  a = c(1, NA, 3),
  b = c(NA, NA, "x"),
  c = c(TRUE, FALSE, TRUE)
)
diagnose_nas(df)

}
\seealso{
Other data quality:
\code{\link[=audit_transform]{audit_transform()}},
\code{\link[=diagnose_strings]{diagnose_strings()}},
\code{\link[=get_summary_table]{get_summary_table()}},
\code{\link[=summarize_column]{summarize_column()}},
\code{\link[=tab]{tab()}}
}
\concept{data quality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit-transform.R
\name{audit_transform}
\alias{audit_transform}
\alias{print.audit_transform}
\title{Audit a Vector Transformation}
\usage{
audit_transform(
  x,
  clean_fn,
  name = NULL,
  .tolerance = sqrt(.Machine$double.eps)
)

\method{print}{audit_transform}(x, ...)
}
\arguments{
\item{x}{Vector to transform. Accepted types: character, numeric,
Date, POSIXct, factor, or logical.}

\item{clean_fn}{A function applied to \code{x} that returns a vector of the
same length, \strong{or} a pre-computed vector of the same length (used
directly as the transformation result).}

\item{name}{Optional name for the variable (used in output). If \code{NULL},
captures the variable name from the call.}

\item{.tolerance}{Numeric tolerance used for the "changed beyond tolerance"
diagnostic (numeric type only). Defaults to \code{sqrt(.Machine$double.eps)}.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An S3 object of class \code{audit_transform} containing:
\describe{
\item{name}{Name of the variable}
\item{clean_fn_name}{Name of the transformation function, or
\code{"<pre-computed>"} when a vector was supplied directly}
\item{type_class}{Detected type: \code{"character"}, \code{"numeric"},
\code{"Date"}, \code{"POSIXct"}, \code{"factor"}, or \code{"logical"}}
\item{n_total}{Total number of elements}
\item{n_changed}{Count of values that changed (including NA status changes)}
\item{n_unchanged}{Count of values that stayed the same}
\item{n_na_before}{Count of NA values before transformation}
\item{n_na_after}{Count of NA values after transformation}
\item{pct_changed}{Percentage of total elements that changed}
\item{change_examples}{Data frame with before/after pairs (up to 10)}
\item{diagnostics}{Type-specific diagnostic list, or \code{NULL} for character}
\item{cleaned}{The transformed vector, retaining its type}
}
}
\description{
Applies a transformation function to a vector and reports what changed.
Works with any vector type: character, numeric, Date/POSIXct, factor, or
logical. Diagnostics are adapted to the detected input type.
}
\examples{
# Character
x <- c("  hello ", "WORLD", "  foo  ", NA)
result <- audit_transform(x, trimws)
result$cleaned

# Numeric
prices <- c(10.5, 20.0, NA, 30.0)
audit_transform(prices, function(v) round(v))

# Pre-computed result
audit_transform(prices, round(prices))

}
\seealso{
\code{\link[=diagnose_strings]{diagnose_strings()}}

Other data quality:
\code{\link[=diagnose_nas]{diagnose_nas()}},
\code{\link[=diagnose_strings]{diagnose_strings()}},
\code{\link[=get_summary_table]{get_summary_table()}},
\code{\link[=summarize_column]{summarize_column()}},
\code{\link[=tab]{tab()}}
}
\concept{data quality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tidyaudit-package.R
\docType{package}
\name{tidyaudit-package}
\alias{tidyaudit}
\alias{tidyaudit-package}
\title{tidyaudit: Pipeline Audit Trails and Data Diagnostics for 'tidyverse' Workflows}
\description{
Provides pipeline audit trails and data diagnostics for 'tidyverse' workflows. The audit trail system captures lightweight metadata snapshots at each step of a pipeline, building a structured record without storing the data itself. Operation-aware taps enrich snapshots with join match rates and filter drop statistics. Trails can be serialized to 'JSON' or 'RDS' and exported as self-contained 'HTML' visualizations. Also includes diagnostic functions for interactive data analysis including frequency tables, string quality auditing, and data comparison.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fpcordeiro.github.io/tidyaudit/}
  \item \url{https://github.com/fpcordeiro/tidyaudit}
  \item Report bugs at \url{https://github.com/fpcordeiro/tidyaudit/issues}
}

}
\author{
\strong{Maintainer}: Fernando Cordeiro \email{fernandolpcordeiro@gmail.com} [copyright holder]

Authors:
\itemize{
  \item Fernando Cordeiro \email{fernandolpcordeiro@gmail.com} [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/string-cleaning.R
\name{diagnose_strings}
\alias{diagnose_strings}
\alias{print.diagnose_strings}
\title{Diagnose String Column Quality}
\usage{
diagnose_strings(x, name = NULL)

\method{print}{diagnose_strings}(x, ...)
}
\arguments{
\item{x}{Character vector to diagnose.}

\item{name}{Optional name for the variable (used in output). If \code{NULL},
captures the variable name from the call.}

\item{...}{Additional arguments (currently unused).}
}
\value{
An S3 object of class \code{diagnose_strings} containing:
\describe{
\item{name}{Name of the variable}
\item{n_total}{Total number of elements}
\item{n_na}{Count of NA values}
\item{n_empty}{Count of empty strings}
\item{n_whitespace_only}{Count of whitespace-only strings}
\item{n_leading_ws}{Count of strings with leading whitespace}
\item{n_trailing_ws}{Count of strings with trailing whitespace}
\item{n_non_ascii}{Count of strings with non-ASCII characters}
\item{n_case_variants}{Number of unique values with case variants}
\item{n_case_variant_groups}{Number of groups of case-insensitive duplicates}
\item{case_variant_examples}{Data.frame with examples of case variants}
}
}
\description{
Audits a character vector for common data quality issues including missing
values, empty strings, whitespace problems, non-ASCII characters, and case
inconsistencies. Requires the stringi package (in Suggests).
}
\examples{
firms <- c("Apple", "APPLE", "apple", "  Microsoft ", "Google", NA, "")
diagnose_strings(firms)

}
\seealso{
Other data quality:
\code{\link[=audit_transform]{audit_transform()}},
\code{\link[=diagnose_nas]{diagnose_nas()}},
\code{\link[=get_summary_table]{get_summary_table()}},
\code{\link[=summarize_column]{summarize_column()}},
\code{\link[=tab]{tab()}}
}
\concept{data quality}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter-diagnostics.R
\name{filter_drop}
\alias{filter_drop}
\alias{filter_drop.data.frame}
\title{Filter Data with Diagnostic Statistics (Drop)}
\usage{
filter_drop(.data, ...)

\method{filter_drop}{data.frame}(.data, ..., .stat = NULL, .quiet = FALSE, .warn_threshold = NULL)
}
\arguments{
\item{.data}{A data.frame, tibble, or other object.}

\item{...}{Filter conditions specifying rows to DROP, evaluated in the
context of \code{.data} using tidy evaluation.}

\item{.stat}{An unquoted column or expression to total, e.g., \code{amount},
\code{price * qty}. Reports the amount dropped and its share of the total.}

\item{.quiet}{Logical. If \code{TRUE}, suppress printing diagnostics.}

\item{.warn_threshold}{Numeric between 0 and 1. If set and the proportion of
dropped rows exceeds this threshold, a warning is issued.}
}
\value{
The filtered data.frame or tibble.
}
\description{
Filters a data.frame or tibble by DROPPING rows where the conditions are
TRUE, while reporting statistics about dropped rows and optionally the sum
of a statistic column that was dropped.
}
\section{Methods (by class)}{
\itemize{
\item \code{filter_drop(data.frame)}: Method for data.frame objects

}}
\examples{
df <- data.frame(
  id = 1:5,
  bad = c(FALSE, TRUE, FALSE, TRUE, FALSE),
  sales = 10:14
)
filter_drop(df, bad == TRUE)
filter_drop(df, bad == TRUE, .stat = sales)

}
\seealso{
Other filter diagnostics:
\code{\link[=filter_keep]{filter_keep()}}
}
\concept{filter diagnostics}
